Package: alakazam
Type: Package
Version: 1.4.3
Date: 2026-04-29
Authors@R: c(person("Susanna", "Marquez", role=c("cre", "aut"), 
                    email="susanna.marquez@yale.edu"),
             person("Namita", "Gupta", role=c("aut"), 
                    email="namita.gupta@yale.edu"),
             person("Nima", "Nouri", role=c("aut"), 
                    email="nima.nouri@yale.edu"),
             person("Ruoyi", "Jiang", role=c("aut"), 
                    email="ruoyi.jiang@yale.edu"),
             person("Julian", "Zhou", role=c("aut"), 
                    email="julian.zhou@bulldogs.yale.edu"),
             person("Kenneth", "Hoehn", role=c("aut"), 
                    email="kenneth.hoehn@yale.edu"),
             person("Daniel", "Gadala-Maria", role=c("ctb"), 
                    email="daniel.gadala-maria@yale.edu"),
             person("Edel", "Aron", role=c("ctb"), 
                    email="edel.aron@yale.edu"),
             person("Cole", "Jensen", role=c("aut"),
                    email="cole.jensen@yale.edu"),
             person("Gisela", "Gabernet", role=c("ctb"),
                    email="gisela.gabernet@yale.edu"),   
             person("Caroline", "Sullivan", role=c("ctb"),
                    email="caroline.sullivan@yale.edu"),
             person("Hailong", "Meng", role=c("ctb"),
                    email="hailong.meng@yale.edu"),
             person("Huimin", "Lyu", role=c("ctb"),
                    email="huimin.lyu@yale.edu"),
             person("Burhan", "Sabuwala", role=c("ctb"),
                    email="burhan.sabuwala@yale.edu"),
             person("Jason", "Vander Heiden", role=c("aut"), 
                    email="jason.vanderheiden@gmail.com"), 
             person("Steven", "Kleinstein", role=c("aut", "cph"), 
                    email="steven.kleinstein@yale.edu"))
Title: Immunoglobulin Clonal Lineage and Diversity Analysis
Description: Provides methods for high-throughput adaptive immune 
    receptor repertoire sequencing (AIRR-Seq; Rep-Seq) analysis. In 
    particular, immunoglobulin (Ig) sequence lineage reconstruction, 
    lineage topology analysis, diversity profiling, amino acid property 
    analysis and gene usage.
    Citations: 
    Gupta and Vander Heiden, et al (2017) <doi:10.1093/bioinformatics/btv359>,
    Stern, Yaari and Vander Heiden, et al (2014) <doi:10.1126/scitranslmed.3008879>.
License: AGPL-3
URL: https://alakazam.readthedocs.io/
BugReports: https://github.com/immcantation/alakazam/issues
LazyData: true
BuildVignettes: true
VignetteBuilder: knitr, rmarkdown
Encoding: UTF-8
LinkingTo: Rcpp
biocViews: Software, AnnotationData
Depends: R (>= 4.0), ggplot2 (>= 3.4.0)
Imports: airr (>= 1.4.1), ape, dplyr (>= 1.0), graphics, grid, igraph
        (>= 1.5.0), Matrix (>= 1.3-0), methods, progress, Rcpp (>=
        0.12.12), readr, rlang, scales, seqinr, stats, stringi, tibble,
        tidyr (>= 1.0), utils, Biostrings (>= 2.56.0),
        GenomicAlignments (>= 1.24.0), IRanges (>= 2.22.2)
Suggests: cigarillo, knitr, rmarkdown, testthat
RoxygenNote: 7.3.3
Collate: 'Alakazam.R' 'AminoAcids.R' 'Classes.R' 'Core.R' 'Data.R'
        'Diversity.R' 'Deprecated.R' 'Fastq.R' 'Gene.R' 'Lineage.R'
        'RcppExports.R' 'Sequence.R' 'Topology.R'
NeedsCompilation: yes
Packaged: 2026-04-29 13:38:10 UTC; root
Author: Susanna Marquez [cre, aut],
  Namita Gupta [aut],
  Nima Nouri [aut],
  Ruoyi Jiang [aut],
  Julian Zhou [aut],
  Kenneth Hoehn [aut],
  Daniel Gadala-Maria [ctb],
  Edel Aron [ctb],
  Cole Jensen [aut],
  Gisela Gabernet [ctb],
  Caroline Sullivan [ctb],
  Hailong Meng [ctb],
  Huimin Lyu [ctb],
  Burhan Sabuwala [ctb],
  Jason Vander Heiden [aut],
  Steven Kleinstein [aut, cph]
Maintainer: Susanna Marquez <susanna.marquez@yale.edu>
Repository: CRAN
Date/Publication: 2026-04-29 15:00:03 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-01 21:56:39 UTC; unix
Archs: alakazam.so.dSYM

Package: alakazam
Type: Package
Version: 1.4.3
Date: 2026-04-29
Authors@R: c(person("Susanna", "Marquez", role=c("cre", "aut"), 
                    email="susanna.marquez@yale.edu"),
             person("Namita", "Gupta", role=c("aut"), 
                    email="namita.gupta@yale.edu"),
             person("Nima", "Nouri", role=c("aut"), 
                    email="nima.nouri@yale.edu"),
             person("Ruoyi", "Jiang", role=c("aut"), 
                    email="ruoyi.jiang@yale.edu"),
             person("Julian", "Zhou", role=c("aut"), 
                    email="julian.zhou@bulldogs.yale.edu"),
             person("Kenneth", "Hoehn", role=c("aut"), 
                    email="kenneth.hoehn@yale.edu"),
             person("Daniel", "Gadala-Maria", role=c("ctb"), 
                    email="daniel.gadala-maria@yale.edu"),
             person("Edel", "Aron", role=c("ctb"), 
                    email="edel.aron@yale.edu"),
             person("Cole", "Jensen", role=c("aut"),
                    email="cole.jensen@yale.edu"),
             person("Gisela", "Gabernet", role=c("ctb"),
                    email="gisela.gabernet@yale.edu"),   
             person("Caroline", "Sullivan", role=c("ctb"),
                    email="caroline.sullivan@yale.edu"),
             person("Hailong", "Meng", role=c("ctb"),
                    email="hailong.meng@yale.edu"),
             person("Huimin", "Lyu", role=c("ctb"),
                    email="huimin.lyu@yale.edu"),
             person("Burhan", "Sabuwala", role=c("ctb"),
                    email="burhan.sabuwala@yale.edu"),
             person("Jason", "Vander Heiden", role=c("aut"), 
                    email="jason.vanderheiden@gmail.com"), 
             person("Steven", "Kleinstein", role=c("aut", "cph"), 
                    email="steven.kleinstein@yale.edu"))
Title: Immunoglobulin Clonal Lineage and Diversity Analysis
Description: Provides methods for high-throughput adaptive immune 
    receptor repertoire sequencing (AIRR-Seq; Rep-Seq) analysis. In 
    particular, immunoglobulin (Ig) sequence lineage reconstruction, 
    lineage topology analysis, diversity profiling, amino acid property 
    analysis and gene usage.
    Citations: 
    Gupta and Vander Heiden, et al (2017) <doi:10.1093/bioinformatics/btv359>,
    Stern, Yaari and Vander Heiden, et al (2014) <doi:10.1126/scitranslmed.3008879>.
License: AGPL-3
URL: https://alakazam.readthedocs.io/
BugReports: https://github.com/immcantation/alakazam/issues
LazyData: true
BuildVignettes: true
VignetteBuilder: knitr, rmarkdown
Encoding: UTF-8
LinkingTo: Rcpp
biocViews: Software, AnnotationData
Depends: R (>= 4.0), ggplot2 (>= 3.4.0)
Imports: airr (>= 1.4.1), ape, dplyr (>= 1.0), graphics, grid, igraph
        (>= 1.5.0), Matrix (>= 1.3-0), methods, progress, Rcpp (>=
        0.12.12), readr, rlang, scales, seqinr, stats, stringi, tibble,
        tidyr (>= 1.0), utils, Biostrings (>= 2.56.0),
        GenomicAlignments (>= 1.24.0), IRanges (>= 2.22.2)
Suggests: cigarillo, knitr, rmarkdown, testthat
RoxygenNote: 7.3.3
Collate: 'Alakazam.R' 'AminoAcids.R' 'Classes.R' 'Core.R' 'Data.R'
        'Diversity.R' 'Deprecated.R' 'Fastq.R' 'Gene.R' 'Lineage.R'
        'RcppExports.R' 'Sequence.R' 'Topology.R'
NeedsCompilation: yes
Packaged: 2026-04-29 13:38:10 UTC; root
Author: Susanna Marquez [cre, aut],
  Namita Gupta [aut],
  Nima Nouri [aut],
  Ruoyi Jiang [aut],
  Julian Zhou [aut],
  Kenneth Hoehn [aut],
  Daniel Gadala-Maria [ctb],
  Edel Aron [ctb],
  Cole Jensen [aut],
  Gisela Gabernet [ctb],
  Caroline Sullivan [ctb],
  Hailong Meng [ctb],
  Huimin Lyu [ctb],
  Burhan Sabuwala [ctb],
  Jason Vander Heiden [aut],
  Steven Kleinstein [aut, cph]
Maintainer: Susanna Marquez <susanna.marquez@yale.edu>
Repository: CRAN
Date/Publication: 2026-04-29 15:00:03 UTC

# Generated by roxygen2: do not edit by hand

export(ABBREV_AA)
export(DNA_COLORS)
export(DNA_IUPAC)
export(IG_COLORS)
export(IMGT_REGIONS)
export(IUPAC_AA)
export(IUPAC_DNA)
export(TR_COLORS)
export(aliphatic)
export(alphaDiversity)
export(aminoAcidProperties)
export(baseTheme)
export(buildPhylipLineage)
export(bulk)
export(calcCoverage)
export(calcDiversity)
export(charge)
export(checkColumns)
export(collapseDuplicates)
export(combineIgphyml)
export(countClones)
export(countGenes)
export(countPatterns)
export(cpuCount)
export(estimateAbundance)
export(extractVRegion)
export(getAAMatrix)
export(getAllele)
export(getChain)
export(getDNAMatrix)
export(getFamily)
export(getGene)
export(getLocus)
export(getMRCA)
export(getPathLengths)
export(getPositionQuality)
export(getSegment)
export(graphToPhylo)
export(gravy)
export(gridPlot)
export(groupGenes)
export(isValidAASeq)
export(junctionAlignment)
export(makeChangeoClone)
export(makeTempDir)
export(maskPositionsByQuality)
export(maskSeqEnds)
export(maskSeqGaps)
export(nonsquareDist)
export(padSeqEnds)
export(pairwiseDist)
export(pairwiseEqual)
export(permuteLabels)
export(phyloToGraph)
export(plotAbundanceCurve)
export(plotDiversityCurve)
export(plotDiversityTest)
export(plotEdgeTest)
export(plotMRCATest)
export(plotSubtrees)
export(polar)
export(progressBar)
export(rarefyDiversity)
export(readChangeoDb)
export(readFastqDb)
export(readIgphyml)
export(seqDist)
export(seqEqual)
export(sortGenes)
export(stoufferMeta)
export(summarizeSubtrees)
export(tableEdges)
export(testDiversity)
export(testEdges)
export(testMRCA)
export(translateDNA)
export(translateStrings)
export(writeChangeoDb)
exportClasses(AbundanceCurve)
exportClasses(ChangeoClone)
exportClasses(DiversityCurve)
exportClasses(EdgeTest)
exportClasses(MRCATest)
exportMethods(plot)
exportMethods(print)
import(ggplot2)
import(graphics)
import(methods)
import(utils)
importFrom(Biostrings,BString)
importFrom(Biostrings,extractAt)
importFrom(GenomicAlignments,explodeCigarOpLengths)
importFrom(GenomicAlignments,explodeCigarOps)
importFrom(IRanges,IRanges)
importFrom(Matrix,rowSums)
importFrom(Matrix,sparseMatrix)
importFrom(Rcpp,evalCpp)
importFrom(airr,read_rearrangement)
importFrom(airr,write_rearrangement)
importFrom(ape,di2multi)
importFrom(ape,ladderize)
importFrom(ape,read.fastq)
importFrom(ape,read.tree)
importFrom(ape,reorder.phylo)
importFrom(ape,root)
importFrom(dplyr,"%>%")
importFrom(dplyr,all_of)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,combine)
importFrom(dplyr,desc)
importFrom(dplyr,do)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,mutate_at)
importFrom(dplyr,n)
importFrom(dplyr,one_of)
importFrom(dplyr,rename)
importFrom(dplyr,right_join)
importFrom(dplyr,rowwise)
importFrom(dplyr,select)
importFrom(dplyr,slice)
importFrom(dplyr,summarize)
importFrom(dplyr,summarize_at)
importFrom(dplyr,transmute)
importFrom(dplyr,ungroup)
importFrom(igraph,E)
importFrom(igraph,V)
importFrom(igraph,all_shortest_paths)
importFrom(igraph,as_data_frame)
importFrom(igraph,as_edgelist)
importFrom(igraph,components)
importFrom(igraph,degree)
importFrom(igraph,distances)
importFrom(igraph,graph_from_adjacency_matrix)
importFrom(igraph,graph_from_data_frame)
importFrom(igraph,groups)
importFrom(igraph,make_directed_graph)
importFrom(igraph,make_graph)
importFrom(igraph,make_undirected_graph)
importFrom(igraph,set_vertex_attr)
importFrom(igraph,shortest_paths)
importFrom(igraph,vertex_attr)
importFrom(progress,progress_bar)
importFrom(readr,cols)
importFrom(readr,read_delim)
importFrom(readr,read_tsv)
importFrom(readr,write_delim)
importFrom(readr,write_tsv)
importFrom(rlang,":=")
importFrom(rlang,enquo)
importFrom(rlang,sym)
importFrom(rlang,syms)
importFrom(scales,log10_trans)
importFrom(scales,log2_trans)
importFrom(scales,math_format)
importFrom(scales,percent)
importFrom(scales,pretty_breaks)
importFrom(scales,scientific)
importFrom(scales,trans_breaks)
importFrom(scales,trans_format)
importFrom(seqinr,s2c)
importFrom(seqinr,translate)
importFrom(stats,cor)
importFrom(stats,cov)
importFrom(stats,dbinom)
importFrom(stats,dmultinom)
importFrom(stats,dnorm)
importFrom(stats,ecdf)
importFrom(stats,mad)
importFrom(stats,median)
importFrom(stats,na.omit)
importFrom(stats,pbinom)
importFrom(stats,pnorm)
importFrom(stats,qbinom)
importFrom(stats,qnorm)
importFrom(stats,rbinom)
importFrom(stats,rmultinom)
importFrom(stats,rnorm)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stringi,stri_count_boundaries)
importFrom(stringi,stri_count_fixed)
importFrom(stringi,stri_count_regex)
importFrom(stringi,stri_detect_fixed)
importFrom(stringi,stri_dup)
importFrom(stringi,stri_extract_all_regex)
importFrom(stringi,stri_extract_first_regex)
importFrom(stringi,stri_flatten)
importFrom(stringi,stri_join)
importFrom(stringi,stri_length)
importFrom(stringi,stri_pad_left)
importFrom(stringi,stri_pad_right)
importFrom(stringi,stri_paste)
importFrom(stringi,stri_replace_all_regex)
importFrom(stringi,stri_replace_first_regex)
importFrom(stringi,stri_split_fixed)
importFrom(tibble,tibble)
importFrom(tidyr,complete)
importFrom(tidyr,gather)
useDynLib(alakazam, .registration=TRUE)

[![](https://www.r-pkg.org/badges/version/alakazam)](https://cran.r-project.org/package=alakazam)
[![](http://cranlogs.r-pkg.org/badges/grand-total/alakazam)](https://www.r-pkg.org/pkg/alakazam)
[![](https://cranlogs.r-pkg.org/badges/alakazam)](https://www.r-pkg.org/pkg/alakazam)
[![](https://img.shields.io/static/v1?label=AIRR-C%20sw-tools%20v1&message=compliant&color=008AFF&labelColor=000000&style=plastic)](https://docs.airr-community.org/en/stable/swtools/airr_swtools_standard.html)

Alakazam
-------------------------------------------------------------------------------

Alakazam is part of the [Immcantation](http://immcantation.readthedocs.io) 
analysis framework for Adaptive Immune Receptor Repertoire sequencing 
(AIRR-seq) and provides a set of tools to investigate lymphocyte 
receptor clonal lineages, diversity, gene usage, and other repertoire level 
properties, with a focus on high-throughput immunoglobulin (Ig) sequencing.

Alakazam serves five main purposes:

1. Providing core functionality for other R packages in the Immcantation 
   framework. This includes common tasks such as file I/O, basic DNA sequence 
   manipulation, and interacting with V(D)J segment and gene annotations.
2. Providing an R interface for interacting with the output of the 
   [pRESTO](http://presto.readthedocs.io) and 
   [Change-O](http://changeo.readthedocs.io) tool suites.
3. Performing clonal abundance and diversity analysis on lymphocyte 
   repertoires.
4. Performing lineage reconstruction on clonal populations of Ig sequences 
   and analyzing the topology of the resultant lineage trees. 
5. Performing physicochemical property analyses of lymphocyte receptor 
   sequences.


Contact
-------------------------------------------------------------------------------

If you need help or have any questions, please contact the [Immcantation Group](mailto:immcantation@googlegroups.com).

If you have discovered a bug or have a feature request, you can open an issue using the [issue tracker](https://github.com/immcantation/alakazam/issues).

To receive alerts about Immcantation releases, news, events, and tutorials, join the [Immcantation News](https://groups.google.com/g/immcantation-news) Google Group. [Membership settings](https://groups.google.com/g/immcantation-news/membership) can be adjusted to change the frequency of email updates.

## ----eval=TRUE, warning=FALSE, message=FALSE----------------------------------
# Load required packages
library(alakazam)
library(dplyr)
library(scales)

# Subset example data
data(ExampleDb)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
# Quantify usage at the gene level
gene <- countGenes(ExampleDb, gene="v_call", groups="sample_id", mode="gene")
head(gene, n=4)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
# Assign sorted levels and subset to IGHV1
ighv1 <- gene %>%
    mutate(gene=factor(gene, levels=sortGenes(unique(gene), method="name"))) %>%
    filter(getFamily(gene) == "IGHV1")

# Plot V gene usage in the IGHV1 family by sample
g1 <- ggplot(ighv1, aes(x=gene, y=seq_freq)) +
    theme_bw() +
    ggtitle("IGHV1 Usage") +
    theme(axis.text.x=element_text(angle=45, hjust=1, vjust=1)) +
    ylab("Percent of repertoire") +
    xlab("") +
    scale_y_continuous(labels=percent) +
    scale_color_brewer(palette="Set1") +
    geom_point(aes(color=sample_id), size=5, alpha=0.8, position=position_dodge(width = 0.1))
plot(g1)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
# Quantify V family usage by sample
family <- countGenes(ExampleDb, gene="v_call", groups="sample_id", mode="family")

# Plot V family usage by sample
g2 <- ggplot(family, aes(x=gene, y=seq_freq)) +
    theme_bw() +
    ggtitle("Family Usage") +
    theme(axis.text.x=element_text(angle=45, hjust=1, vjust=1)) +
    ylab("Percent of repertoire") +
    xlab("") +
    scale_y_continuous(labels=percent) +
    scale_color_brewer(palette="Set1") +
    geom_point(aes(color=sample_id), size=5, alpha=0.8, position=position_dodge(width = 0.1))
plot(g2)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
# Quantify V family clonal usage by sample and isotype
family <- countGenes(ExampleDb, gene="v_call", groups=c("sample_id", "c_call"), 
                     clone="clone_id", mode="family")
head(family, n=4)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
# Subset to IGHM and IGHG for plotting
family <- filter(family, c_call %in% c("IGHM", "IGHG"))
# Plot V family clonal usage by sample and isotype
g3 <- ggplot(family, aes(x=gene, y=clone_freq)) +
    theme_bw() +
    ggtitle("Clonal Usage") +
    theme(axis.text.x=element_text(angle=45, hjust=1, vjust=1)) +
    ylab("Percent of repertoire") +
    xlab("") +
    scale_y_continuous(labels=percent) +
    scale_color_brewer(palette="Set1") +
    geom_point(aes(color=sample_id), size=5, alpha=0.8, position=position_dodge(width = 0.1)) +
    facet_grid(. ~ c_call)
plot(g3)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
# Calculate V family copy numbers by sample and isotype
family <- countGenes(ExampleDb, gene="v_call", groups=c("sample_id", "c_call"), 
                     mode="family", copy="duplicate_count")
head(family, n=4)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
# Subset to IGHM and IGHG for plotting
family <- filter(family, c_call %in% c("IGHM", "IGHG"))
# Plot V family copy abundance by sample and isotype
g4 <- ggplot(family, aes(x=gene, y=copy_freq)) +
    theme_bw() +
    ggtitle("Copy Number") +
    theme(axis.text.x=element_text(angle=45, hjust=1, vjust=1)) +
    ylab("Percent of repertoire") +
    xlab("") +
    scale_y_continuous(labels=percent) +
    scale_color_brewer(palette="Set1") +
    geom_point(aes(color=sample_id), size=5, alpha=0.8, position=position_dodge(width = 0.1)) +
    facet_grid(. ~ c_call)
plot(g4)

## ----eval=TRUE, warning=FALSE, message=FALSE----------------------------------
# Load required packages
library(alakazam)

# Load example data
data(ExampleDb)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
# Partitions the data based on the sample column
clones <- countClones(ExampleDb, group="sample_id")
head(clones, 5)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
# Partitions the data based on both the sample_id and c_call columns
# Weights the clone sizes by the duplicate_count column
clones <- countClones(ExampleDb, group=c("sample_id", "c_call"), copy="duplicate_count", clone="clone_id")
head(clones, 5)

## ----eval=TRUE, results='hide', warning=FALSE, fig.width=6, fig.height=4------
# Partitions the data on the sample column
# Calculates a 95% confidence interval via 100 bootstrap realizations
set.seed(123) # For reproducibility of example bootstrap results
curve <- estimateAbundance(ExampleDb, group="sample_id", ci=0.95, nboot=100, clone="clone_id")

## ----eval=TRUE, warning=FALSE, fig.width=6, fig.height=4----------------------
# Plots a rank abundance curve of the relative clonal abundances
sample_colors <- c("-1h"="seagreen", "+7d"="steelblue")
plot(curve, colors = sample_colors, legend_title="Sample")

## ----eval=TRUE, results='hide'------------------------------------------------
# Compare diversity curve across values in the "sample" column
# q ranges from 0 (min_q=0) to 4 (max_q=4) in 0.05 increments (step_q=0.05)
# A 95% confidence interval will be calculated (ci=0.95)
# 100 resampling realizations are performed (nboot=100)
set.seed(123) # For reproducibility of example alphaDiversity results
sample_curve <- alphaDiversity(ExampleDb, group="sample_id", clone="clone_id",
                               min_q=0, max_q=4, step_q=0.1,
                               ci=0.95, nboot=100)

# Compare diversity curve across values in the c_call column
# Analyse is restricted to c_call values with at least 30 sequences by min_n=30
# Excluded groups are indicated by a warning message
isotype_curve <- alphaDiversity(ExampleDb, group="c_call", clone="clone_id",
                                min_q=0, max_q=4, step_q=0.1,
                                ci=0.95, nboot=100)

## ----eval=TRUE, fig.width=6, fig.height=4-------------------------------------
# Plot a log-log (log_q=TRUE, log_d=TRUE) plot of sample diversity
# Indicate number of sequences resampled from each group in the title
sample_main <- paste0("Sample diversity")
sample_colors <- c("-1h"="seagreen", "+7d"="steelblue")
plot(sample_curve, colors=sample_colors, main_title=sample_main, 
     legend_title="Sample")

# Plot isotype diversity using default set of Ig isotype colors
isotype_main <- paste0("Isotype diversity")
plot(isotype_curve, colors=IG_COLORS, main_title=isotype_main, 
     legend_title="Isotype")

## ----eval=TRUE, fig.width=6, fig.height=3-------------------------------------
# Test diversity at q=0, q=1 and q=2 (equivalent to species richness, Shannon entropy, 
# Simpson's index) across values in the sample_id column
# 100 bootstrap realizations are performed (nboot=100)
set.seed(123) # For reproducibility of example alphaDiversity results
isotype_test <- alphaDiversity(ExampleDb, group="c_call", min_q=0, max_q=2, step_q=1, nboot=100, clone="clone_id")

# Print P-value table
print(isotype_test@tests)

# Plot results at q=0 and q=2
# Plot the mean and standard deviations at q=0 and q=2
plot(isotype_test, 0, colors=IG_COLORS, main_title=isotype_main, 
     legend_title="Isotype")
plot(isotype_test, 2, colors=IG_COLORS, main_title=isotype_main, 
     legend_title="Isotype")

## ----eval=TRUE, warning=FALSE, message=FALSE----------------------------------
library(alakazam)
library(dplyr)
library(airr)

db <- read_rearrangement(system.file("extdata", "example_quality.tsv", package="alakazam"))
fastq_file <- system.file("extdata", "example_quality.fastq", package="alakazam")

## -----------------------------------------------------------------------------
original_cols <- colnames(db)
db <- readFastqDb(db, fastq_file, style="both", quality_sequence=TRUE)
new_cols <- setdiff(colnames(db), original_cols)
db[,new_cols] %>% head()

## -----------------------------------------------------------------------------
quality <- getPositionQuality(db, sequence_id="sequence_id", 
                              sequence="sequence_alignment",
                              quality_num="quality_alignment_num")
head(quality)

## ----fig.cap="Sequence quality per IMGT position for one sequence.", fig.asp=0.25----
min_pos <- min(quality$position)
max_pos <- max(quality$position)

ggplot(quality, aes(x=position,
                    y=quality_alignment_num,
                    color=nt)) +
  geom_point() +
  coord_cartesian(xlim=c(110,120)) +
  xlab("IMGT position") +
  ylab("Sequencing quality") +
  scale_fill_gradient(low = "light blue",  high = "dark red") +
  scale_x_continuous(breaks=c(min_pos:max_pos)) +
  alakazam::baseTheme()

## -----------------------------------------------------------------------------
db <- maskPositionsByQuality(db, min_quality=70,
                             sequence="sequence_alignment",
                             quality="quality_alignment_num")

## ----eval=TRUE, warning=FALSE, message=FALSE----------------------------------
# Load required packages
library(alakazam)
library(dplyr)

# Subset example data
data(ExampleDb)
db <- ExampleDb[ExampleDb$sample_id == "+7d", ]

## ----eval=TRUE, warning=FALSE, fig.width=7.5, fig.height=6--------------------
db_props <- aminoAcidProperties(db, seq="junction", trim=TRUE, 
                                label="cdr3")

# The full set of properties are calculated by default
dplyr::select(db_props[1:3, ], starts_with("cdr3"))

# Define a ggplot theme for all plots
tmp_theme <- theme_bw() + theme(legend.position="bottom")

# Generate plots for all four of the properties
g1 <- ggplot(db_props, aes(x=c_call, y=cdr3_aa_length)) + tmp_theme +
    ggtitle("CDR3 length") + 
    xlab("Isotype") + ylab("Amino acids") +
    scale_fill_manual(name="Isotype", values=IG_COLORS) +
    geom_boxplot(aes(fill=c_call))
g2 <- ggplot(db_props, aes(x=c_call, y=cdr3_aa_gravy)) + tmp_theme + 
    ggtitle("CDR3 hydrophobicity") + 
    xlab("Isotype") + ylab("GRAVY") +
    scale_fill_manual(name="Isotype", values=IG_COLORS) +
    geom_boxplot(aes(fill=c_call))
g3 <- ggplot(db_props, aes(x=c_call, y=cdr3_aa_basic)) + tmp_theme +
    ggtitle("CDR3 basic residues") + 
    xlab("Isotype") + ylab("Basic residues") +
    scale_y_continuous(labels=scales::percent) +
    scale_fill_manual(name="Isotype", values=IG_COLORS) +
    geom_boxplot(aes(fill=c_call))
g4 <- ggplot(db_props, aes(x=c_call, y=cdr3_aa_acidic)) + tmp_theme +
    ggtitle("CDR3 acidic residues") + 
    xlab("Isotype") + ylab("Acidic residues") +
    scale_y_continuous(labels=scales::percent) +
    scale_fill_manual(name="Isotype", values=IG_COLORS) +
    geom_boxplot(aes(fill=c_call))

# Plot in a 2x2 grid
gridPlot(g1, g2, g3, g4, ncol=2)

## ----eval=TRUE, warning=FALSE-------------------------------------------------
db_props <- aminoAcidProperties(db, seq="junction", property=c("gravy", "charge"),
                                trim=TRUE, label="cdr3")
dplyr::select(db_props[1:3, ], starts_with("cdr3"))

## ----eval=TRUE, warning=FALSE, message=FALSE----------------------------------
# Load the relevant data objects from the seqinr package
library(seqinr)
data(aaindex)
data(pK)
h <- aaindex[["KIDA850101"]]$I
p <- setNames(pK[["Murray"]], rownames(pK))
# Rename the hydrophobicity vector to use single-letter codes
names(h) <- translateStrings(names(h), ABBREV_AA)
db_props <- aminoAcidProperties(db, seq="junction", property=c("gravy", "charge"), 
                                trim=TRUE, label="cdr3", 
                                hydropathy=h, pK=p)
dplyr::select(db_props[1:3, ], starts_with("cdr3"))

## ----eval=TRUE, warning=FALSE, message=FALSE----------------------------------
# Translate junction DNA sequences to amino acids and trim first and last codons
cdr3 <- translateDNA(db$junction[1:3], trim=TRUE)

# Grand average of hydrophobicity
gravy(cdr3)

# Average bulkiness
bulk(cdr3)

# Average polarity
polar(cdr3)

# Normalized aliphatic index
aliphatic(cdr3)
# Unnormalized aliphatic index
aliphatic(cdr3, normalize=FALSE)

# Normalized net charge
charge(cdr3)
# Unnormalized net charge
charge(cdr3, normalize=FALSE)

# Count of acidic amino acids
# Takes a named list of regular expressions
countPatterns(cdr3, nt=FALSE, c(ACIDIC="[DE]"), label="cdr3")

## ----eval=TRUE, warning=FALSE, message=FALSE----------------------------------
# Set the file paths from inside the package directory
# These files are smaller versions of the example databases previously mentioned
changeo_file <- system.file("extdata", "example_changeo.tab.gz", package="alakazam")
airr_file <- system.file("extdata", "example_airr.tsv.gz", package="alakazam")

# Read in the data
db_changeo <- alakazam::readChangeoDb(changeo_file)
db_airr <- airr::read_rearrangement(airr_file)

## ----eval=FALSE, warning=FALSE, message=FALSE---------------------------------
# # Write the data to a tab-delimited file
# alakazam::writeChangeoDb(db_changeo, "changeo.tsv")
# airr::write_rearrangement(db_airr, "airr.tsv")

#' @keywords internal
#' @aliases alakazam-package
"_PACKAGE"

# Alakazam package documentation and import directives

#' The Alakazam package
#' 
#' \code{alakazam} in a member of the Immcantation framework of tools and serves five main 
#' purposes:
#' \itemize{
#'   \item  Providing core functionality for other R packages in Immcantation. This
#'          includes common tasks such as file I/O, basic DNA sequence manipulation, and
#'          interacting with V(D)J segment and gene annotations.
#'   \item  Providing an R interface for interacting with the output of the pRESTO and 
#'          Change-O tool suites.
#'   \item  Performing clonal abundance and diversity analysis on lymphocyte repertoires.
#'   \item  Performing lineage reconstruction on clonal populations of immunoglobulin 
#'          (Ig) sequences.
#'   \item  Performing physicochemical property analyses of lymphocyte receptor sequences.
#' }
#' For additional details regarding the use of the \code{alakazam} package see the 
#' vignettes:\cr
#' \code{browseVignettes("alakazam")}
#' 
#' @section  File I/O:
#' \itemize{
#'   \item  \link{readChangeoDb}:        Input Change-O style files.
#'   \item  \link{writeChangeoDb}:       Output Change-O style files.
#' }
#' 
#' @section  Sequence cleaning:
#' \itemize{
#'   \item  \link{maskSeqEnds}:          Mask ragged ends.
#'   \item  \link{maskSeqGaps}:          Mask gap characters.
#'   \item  \link{collapseDuplicates}:   Remove duplicate sequences.
#' }
#' 
#' @section  Lineage reconstruction:
#' \itemize{
#'   \item  \link{makeChangeoClone}:     Clean sequences for lineage reconstruction.
#'   \item  \link{buildPhylipLineage}:   Perform lineage reconstruction of Ig sequences.
#' }
#' 
#' @section  Lineage topology analysis:
#' \itemize{
#'   \item  \link{tableEdges}:           Tabulate annotation relationships over edges.
#'   \item  \link{testEdges}:            Significance testing of annotation edges.
#'   \item  \link{testMRCA}:             Significance testing of MRCA annotations.
#'   \item  \link{summarizeSubtrees}:    Various summary statistics for subtrees.
#'   \item  \link{plotSubtrees}:         Plot distributions of summary statistics 
#'                                       for a population of trees.
#' }
#' 
#' @section  Diversity analysis:
#' \itemize{
#'   \item  \link{countClones}:          Calculate clonal abundance.
#'   \item  \link{estimateAbundance}:  	 Bootstrap clonal abundance curves.
#'   \item  \link{alphaDiversity}:  	 Generate clonal alpha diversity curves.
#'   \item  \link{plotAbundanceCurve}:   Plot clone size distribution as a rank-abundance 
#'   \item  \link{plotDiversityCurve}:   Plot clonal diversity curves.
#'   \item  \link{plotDiversityTest}:    Plot testing at given diversity hill indices. 
#' }
#' 
#' @section  Ig and TCR sequence annotation:
#' \itemize{
#'   \item  \link{countGenes}:           Calculate Ig and TCR allele, gene and family usage.
#'   \item  \link{extractVRegion}:       Extract CDRs and FWRs sub-sequences.
#'   \item  \link{getAllele}:            Get V(D)J allele names.
#'   \item  \link{getGene}:              Get V(D)J gene names.
#'   \item  \link{getFamily}:            Get V(D)J family names.
#'   \item  \link{junctionAlignment}: Junction alignment properties 
#' }
#' 
#' @section  Sequence distance calculation:
#' \itemize{
#'   \item  \link{seqDist}:        Calculate Hamming distance between two sequences.
#'   \item  \link{seqEqual}:       Test two sequences for equivalence.
#'   \item  \link{pairwiseDist}:   Calculate a matrix of pairwise Hamming distances for a 
#'                                 set of sequences.
#'   \item  \link{pairwiseEqual}:  Calculate a logical matrix of pairwise equivalence for a 
#'                                 set of sequences.
#' }
#' 
#' @section  Amino acid properties:
#' \itemize{
#'   \item  \link{translateDNA}:         Translate DNA sequences to amino acid sequences.
#'   \item  \link{aminoAcidProperties}:  Calculate various physicochemical properties of amino acid 
#'                                       sequences.
#'   \item  \link{countPatterns}:        Count patterns in sequences.
#'                                              
#' }
#' 
#' @name     alakazam
#' @references
#' \enumerate{
#'   \item  Vander Heiden JA, Yaari G, et al. pRESTO: a toolkit for processing 
#'            high-throughput sequencing raw reads of lymphocyte receptor repertoires. 
#'            Bioinformatics. 2014 30(13):1930-2.
#'   \item  Stern JNH, Yaari G, Vander Heiden JA, et al. B cells populating the multiple 
#'            sclerosis brain mature in the draining cervical lymph nodes. 
#'            Sci Transl Med. 2014 6(248):248ra107.
#'   \item  Wu Y-CB, et al. Influence of seasonal exposure to grass pollen on local and 
#'            peripheral blood IgE repertoires in patients with allergic rhinitis. 
#'            J Allergy Clin Immunol. 2014 134(3):604-12.
#'   \item  Gupta NT, Vander Heiden JA, et al. Change-O: a toolkit for analyzing 
#'            large-scale B cell immunoglobulin repertoire sequencing data.
#'            Bioinformatics. 2015 Oct 15;31(20):3356-8.
#' }
#'
#' @import      ggplot2
#' @import      graphics
#' @import      methods
#' @import      utils
#' @importFrom  airr        read_rearrangement write_rearrangement
#' @importFrom	ape 		read.fastq read.tree di2multi reorder.phylo root ladderize
#' @importFrom  dplyr       do n desc %>%
#'                          bind_cols bind_rows combine arrange left_join
#'                          group_by ungroup
#'                          filter slice select 
#'                          mutate mutate_at 
#' 							one_of
#'							right_join rowwise
#'                          summarize summarize_at all_of
#'                          transmute rename
#' @importFrom  igraph      V E graph_from_data_frame as_data_frame as_edgelist 
#'                          make_graph make_directed_graph make_undirected_graph
#'                          vertex_attr set_vertex_attr
#'                          degree shortest_paths all_shortest_paths distances
#'                          graph_from_adjacency_matrix components groups
#' @importFrom  Matrix      sparseMatrix rowSums
#' @importFrom  progress    progress_bar
#' @importFrom  readr       read_delim read_tsv write_delim write_tsv cols
#' @importFrom  rlang       := sym syms enquo
#' @importFrom  scales      log2_trans log10_trans trans_breaks trans_format
#'                          math_format percent scientific pretty_breaks
#' @importFrom  seqinr      translate s2c
#' @importFrom  stats       na.omit setNames ecdf sd cor cov median mad
#'                          dbinom pbinom qbinom rbinom
#'                          dnorm pnorm qnorm rnorm
#'                          dmultinom rmultinom
#' @importFrom  stringi     stri_dup stri_flatten stri_join stri_length 
#'                          stri_count_boundaries stri_count_fixed 
#'                          stri_count_regex stri_extract_all_regex 
#'                          stri_extract_first_regex stri_replace_all_regex 
#'                          stri_replace_first_regex stri_split_fixed
#'                          stri_pad_left stri_pad_right
#'                          stri_detect_fixed stri_paste
#' @importFrom  tibble      tibble
#' @importFrom  tidyr       complete gather
#' @importFrom  Rcpp        evalCpp
#' @importFrom  Biostrings         BString extractAt
#' @importFrom  GenomicAlignments  explodeCigarOps explodeCigarOpLengths                       
#' @importFrom  IRanges            IRanges
#' @useDynLib   alakazam, .registration=TRUE
NULL

# Package loading actions
.onAttach <- function(libname, pkgname) {
    msg <- citation(pkgname)
    msg <-paste(c(format(msg,"citation")),collapse="\n\n")
    packageStartupMessage(msg)
}

# Amino acid sequence properties

#### Chemical property functions ####

#' Calculates the average bulkiness of amino acid sequences
#'
#' \code{bulk} calculates the average bulkiness score of amino acid sequences. 
#' Non-informative positions are excluded, where non-informative is defined as any 
#' character in \code{c("X", "-", ".", "*")}.
#' 
#' @param    seq         vector of strings containing amino acid sequences.
#' @param    bulkiness   named numerical vector defining bulkiness scores for 
#'                       each amino acid, where names are single-letter amino acid 
#'                       character codes. If \code{NULL}, then the Zimmerman et al, 1968
#'                       scale is used.
#' 
#' @return   A vector of bulkiness scores for the sequence(s).
#' 
#' @references
#' \enumerate{
#'   \item  Zimmerman JM, Eliezer N, Simha R. The characterization of amino acid sequences 
#'            in proteins by statistical methods. J Theor Biol 21, 170-201 (1968).
#' }
#' @seealso 
#' For additional size related indices see \link[seqinr]{aaindex}.
#'
#' @examples
#' # Default bulkiness scale
#' seq <- c("CARDRSTPWRRGIASTTVRTSW", "XXTQMYVRT")
#' bulk(seq)
#'
#' # Use the Grantham, 1974 side chain volumn scores from the seqinr package
#' library(seqinr)
#' data(aaindex)
#' x <- aaindex[["GRAR740103"]]$I
#' # Rename the score vector to use single-letter codes
#' names(x) <- translateStrings(names(x), ABBREV_AA)
#' # Calculate average volume
#' bulk(seq, bulkiness=x)
#' 
#' @export
bulk <- function(seq, bulkiness=NULL) {
    # Get bulkiness scores
    if (is.null(bulkiness)) {
        bulkiness <- BULKINESS_ZIMJ68
    }
    # Remove non-informative positions
    seq <- gsub("[X\\.\\*-]", "", as.character(seq))
    # Create character vector from string
    aa <- strsplit(seq, "")
    # Calculate average bulkiness
    aa_bulk <- sapply(aa, function(x) sum(bulkiness[x]) / length(x))
    
    return(aa_bulk)
}


#' Calculates the average polarity of amino acid sequences
#'
#' \code{polar} calculates the average polarity score of amino acid sequences. 
#' Non-informative positions are excluded, where non-informative is defined as any 
#' character in \code{c("X", "-", ".", "*")}.
#' 
#' @param    seq         vector of strings containing amino acid sequences.
#' @param    polarity    named numerical vector defining polarity scores for 
#'                       each amino acid, where names are single-letter amino acid 
#'                       character codes. If \code{NULL}, then the Grantham, 1974
#'                       scale is used.
#' 
#' @return   A vector of bulkiness scores for the sequence(s).
#' 
#' @references
#' \enumerate{
#'   \item  Grantham R. Amino acid difference formula to help explain protein evolution. 
#'            Science 185, 862-864 (1974).
#' }
#' @seealso 
#' For additional size related indices see \code{\link[seqinr]{aaindex}}.
#'
#' @examples
#' # Default scale
#' seq <- c("CARDRSTPWRRGIASTTVRTSW", "XXTQMYVRT")
#' polar(seq)
#'
#' # Use the Zimmerman et al, 1968 polarity scale from the seqinr package
#' library(seqinr)
#' data(aaindex)
#' x <- aaindex[["ZIMJ680103"]]$I
#' # Rename the score vector to use single-letter codes
#' names(x) <- translateStrings(names(x), ABBREV_AA)
#' # Calculate polarity
#' polar(seq, polarity=x)
#' 
#' @export
polar <- function(seq, polarity=NULL) {
    # Get bulkiness scores
    if (is.null(polarity)) {
        polarity <- POLARITY_GRAR74
    }
    # Remove non-informative positions
    seq <- gsub("[X\\.\\*-]", "", as.character(seq))
    # Create character vector from string
    aa <- strsplit(seq, "")
    # Calculate average polarity
    aa_polar <- sapply(aa, function(x) sum(polarity[x]) / length(x))
    
    return(aa_polar)
}


#' Calculates the hydrophobicity of amino acid sequences
#'
#' \code{gravy} calculates the Grand Average of Hydrophobicity (gravy) index 
#' of amino acid sequences using the method of Kyte & Doolittle. Non-informative
#' positions are excluded, where non-informative is defined as any character in 
#' \code{c("X", "-", ".", "*")}.
#' 
#' @param    seq         vector of strings containing amino acid sequences.
#' @param    hydropathy  named numerical vector defining hydropathy index values for 
#'                       each amino acid, where names are single-letter amino acid 
#'                       character codes. If \code{NULL}, then the Kyte & Doolittle
#'                       scale is used.
#' 
#' @return   A vector of gravy scores for the sequence(s).
#' 
#' @references
#' \enumerate{
#'   \item  Kyte J, Doolittle RF. A simple method for displaying the hydropathic character 
#'            of a protein. J Mol Biol. 157, 105-32 (1982).
#' }
#' @seealso 
#' For additional hydrophobicity indices see \code{\link[seqinr]{aaindex}}.
#'
#' @examples
#' # Default scale
#' seq <- c("CARDRSTPWRRGIASTTVRTSW", "XXTQMYVRT")
#' gravy(seq)
#'
#' # Use the Kidera et al, 1985 scores from the seqinr package
#' library(seqinr)
#' data(aaindex)
#' x <- aaindex[["KIDA850101"]]$I
#' # Rename the score vector to use single-letter codes
#' names(x) <- translateStrings(names(x), ABBREV_AA)
#' # Calculate hydrophobicity
#' gravy(seq, hydropathy=x)
#' 
#' @export
gravy <- function(seq, hydropathy=NULL) {
    # Get hydrophobicity scores
    if (is.null(hydropathy)) {
        hydropathy <- HYDROPATHY_KYTJ82
    }

    # Remove non-informative positions
    seq <- gsub("[X\\.\\*-]", "", as.character(seq))
    # Create character vector from string
    aa <- strsplit(seq, "")
    # Calculate gravy
    aa_gravy <- sapply(aa, function(x) sum(hydropathy[x]) / length(x))
    
    return(aa_gravy)
}


#' Calculates the aliphatic index of amino acid sequences
#' 
#' \code{aliphatic} calculates the aliphatic index of amino acid sequences using 
#' the method of Ikai. Non-informative positions are excluded, where non-informative 
#' is defined as any character in \code{c("X", "-", ".", "*")}.
#'
#' @param    seq        vector of strings containing amino acid sequences.
#' @param    normalize  if \code{TRUE} then divide the aliphatic index of each amino acid 
#'                      sequence by the number of informative positions. Non-informative 
#'                      position are defined by the presence any character in 
#'                      \code{c("X", "-", ".", "*")}. If \code{FALSE} then return the raw
#'                      aliphatic index.
#'                      
#' @return   A vector of the aliphatic indices for the sequence(s).
#'
#' @references 
#' \enumerate{
#'   \item  Ikai AJ. Thermostability and aliphatic index of globular proteins. 
#'            J Biochem. 88, 1895-1898 (1980).
#' }
#' 
#' @examples 
#' seq <- c("CARDRSTPWRRGIASTTVRTSW", NA, "XXTQMYVRT")
#' aliphatic(seq)
#'
#' @export
aliphatic <- function(seq, normalize=TRUE) {
    # Calculate aliphatic index for valid amino acids
    ala <- countOccurrences(seq, "[A]")
    val <- countOccurrences(seq, "[V]")
    leu_ile <- countOccurrences(seq, "[LI]")
    
    aa_aliphatic = ala + 2.9 * val + 3.9 * leu_ile
    
    if (normalize) {
        aa_aliphatic <- aa_aliphatic / stri_length(gsub("[X\\.\\*-]", "", seq))
    }
    
    return(aa_aliphatic)
}


#' Calculates the net charge of amino acid sequences.
#'
#' \code{charge} calculates the net charge of amino acid sequences using 
#' the method of Moore, 1985, with exclusion of the C-terminus and N-terminus charges.
#' 
#' @param    seq        vector strings defining of amino acid sequences.
#' @param    pH         environmental pH.
#' @param    pK         named vector defining pK values for each charged amino acid,
#'                      where names are the single-letter amino acid character codes
#'                      \code{c("R", "H", "K", "D", "E", "C", "Y")}). If \code{NULL}, 
#'                      then the EMBOSS scale is used.
#' @param    normalize  if \code{TRUE} then divide the net charge of each amino acid 
#'                      sequence by the number of informative positions. Non-informative 
#'                      position are defined by the presence any character in 
#'                      \code{c("X", "-", ".", "*")}. If \code{FALSE} then return the raw
#'                      net charge.
#' 
#' @return   A vector of net charges for the sequence(s).
#'
#' @references
#' \enumerate{
#'   \item  Moore DS. Amino acid and peptide net charges: A simple calculational procedure. 
#'            Biochem Educ. 13, 10-11 (1985).
#'   \item  \url{https://emboss.sourceforge.net/apps/cvs/emboss/apps/iep.html}
#'   }
#'   
#' @seealso 
#' For additional pK scales see \code{\link[seqinr]{pK}}.
#' 
#' @examples 
#' seq <- c("CARDRSTPWRRGIASTTVRTSW", "XXTQMYVRT") 
#' # Unnormalized charge
#' charge(seq)
#' # Normalized charge
#' charge(seq, normalize=TRUE)
#' 
#' # Use the Murray et al, 2006 scores from the seqinr package
#' library(seqinr)
#' data(pK)
#' x <- setNames(pK[["Murray"]], rownames(pK))
#' # Calculate charge
#' charge(seq, pK=x)
#'
#' @export
charge <- function(seq, pH=7.4, pK=NULL, normalize=FALSE) {
    
    # Get charge data
    if(is.null(pK)) {
        pK <- PK_EMBOSS
    }
    
    # Calculate charge
    arg <- countOccurrences(seq, "R") * (1/(1 + 10^(1 * (pH - pK["R"]))))
    his <- countOccurrences(seq, "H") * (1/(1 + 10^(1 * (pH - pK["H"]))))
    lys <- countOccurrences(seq, "K") * (1/(1 + 10^(1 * (pH - pK["K"]))))
    asp <- countOccurrences(seq, "D") * (-1/(1 + 10^(-1 * (pH - pK["D"]))))
    glu <- countOccurrences(seq, "E") * (-1/(1 + 10^(-1 * (pH - pK["E"]))))
    cys <- countOccurrences(seq, "C") * (-1/(1 + 10^(-1 * (pH - pK["C"]))))
    tyr <- countOccurrences(seq, "Y") * (-1/(1 + 10^(-1 * (pH - pK["Y"]))))
    aa_charge <- arg + lys + his + asp + glu + tyr + cys
    
    if (normalize) {
        aa_charge <- aa_charge / stri_length(gsub("[X\\.\\*-]", "", seq))
    }
    
    return(aa_charge)
}

#' Validate amino acid sequences
#'
#' \code{isValidAASeq} checks that a set of sequences are valid non-ambiguous 
#' amino acid sequences. A sequence is considered valid if it contains only 
#' characters in the the non-ambiguous IUPAC character set or any characters in 
#' \code{c("X", ".", "-", "*")}.
#'  
#' @param    seq  character vector of sequences to check.
#' 
#' @return   A logical vector with \code{TRUE} for each valid amino acid sequences 
#'           and \code{FALSE} for each invalid sequence.
#' @seealso 
#' See \link{ABBREV_AA} for the set of non-ambiguous amino acid characters.
#' See \link{IUPAC_AA} for the full set of ambiguous amino acid characters.
#' 
#' @examples 
#' seq <- c("CARDRSTPWRRGIASTTVRTSW", "XXTQMYVR--XX", "CARJ", "10") 
#' isValidAASeq(seq)
#' 
#' @export
isValidAASeq <- function(seq) {
    
    # Get valid amino acids from seqinr
    # for consistency with `gravy` and other
    # amino acid properties that don't consider
    # amino acid ambiguities and special encoded amino acids
    # http://pir.georgetown.edu/resid/faq.shtml#q01
    # Also include here characters for non informative positions
    valid_AA <- c(names(ABBREV_AA), "X", ".", "*", "-")
    
    .isValid <- function(aa) {
        all(aa %in% valid_AA)
    }
    
    return(sapply(strsplit(seq, ""), .isValid))
    
#    valid_AA <- paste(c(names(ABBREV_AA),"X.*-"),collapse="")
#    valid <- !grepl(paste0("[^",valid_AA,"]"), seq) & !is.na(seq)
#    valid
    
}

# Count patterns
# 
# Counts the number of times a "pattern" occurs in "x", a string
#
# @param   x        a string (usually amino acids)
# @param   pattern  regular expression to be matched in string
#  
# @return  number of times the regular expression was found
countOccurrences <- function(x, pattern) {
    return(sapply(gregexpr(pattern, x), function(y) { sum(y > 0) }))
}


#' Count sequence patterns
#' 
#' \code{countPatterns} counts the fraction of times a set of character patterns occur 
#' in a set of sequences.
#'
#' @param   seq         character vector of either DNA or amino acid sequences.
#' @param   patterns    list of sequence patterns to count in each sequence. If the 
#'                      list is named, then names will be assigned as the column names of 
#'                      output data.frame.
#' @param   nt		    if \code{TRUE} then \code{seq} are DNA sequences and and will be 
#'                      translated before performing the pattern search.
#' @param   trim        if \code{TRUE} remove the first and last codon or amino acid from 
#'                      each sequence before the pattern search. If \code{FALSE} do
#'                      not modify the input sequences.
#' @param   label       string defining a label to add as a prefix to the output 
#'                      column names.
#'  
#' @return  A data.frame containing the fraction of times each sequence pattern was 
#'          found.
#' 
#' @examples 
#' seq <- c("TGTCAACAGGCTAACAGTTTCCGGACGTTC",
#'          "TGTCAGCAATATTATATTGCTCCCTTCACTTTC",
#'          "TGTCAAAAGTATAACAGTGCCCCCTGGACGTTC")
#' patterns <- c("A", "V", "[LI]")
#' names(patterns) <- c("arg", "val", "iso_leu")
#' countPatterns(seq, patterns, nt=TRUE, trim=TRUE, label="cdr3")
#'             
#' @export
countPatterns <- function(seq, patterns, nt=TRUE, trim=FALSE, label="region") {
    # Translate sequences if nucleotide
    region_aa <- if (nt) { translateDNA(seq, trim=trim) } else { seq }
    
    # TODO: What is the proper length normalization? With or without non-informative position?
    
    # Calculate region lengths
    aa_length <- stri_length(region_aa)
    # Count occurrence of each amino acid pattern for each sequence
    out_df <- data.frame(matrix(0, nrow=length(region_aa), ncol=length(patterns)))
    # If patterns are unnamed, make the names X1...Xn
    if(is.null(names(patterns))) { names(patterns) <- names(out_df) }
    # If region name, append to names of patterns
    if(label != '') { 
        names(out_df) <- paste(label, names(patterns), sep="_") 
    } else {
        names(out_df) <- names(patterns)
    }
    # Iterate over patterns
    for(i in 1:length(patterns)) {
        out_df[, i] <- countOccurrences(region_aa, patterns[i]) / aa_length
    }
    return(out_df)
}


#' Calculates amino acid chemical properties for sequence data
#'
#' \code{aminoAcidProperties} calculates amino acid sequence physicochemical properties, including
#' length, hydrophobicity, bulkiness, polarity, aliphatic index, net charge, acidic residue
#' content, basic residue content, and aromatic residue content.
#'
#' @param   data          \code{data.frame} containing sequence data.
#' @param   property      vector strings specifying the properties to be calculated. Defaults
#'                        to calculating all defined properties.
#' @param   seq           \code{character} name of the column containing input 
#'                        sequences.
#' @param   nt      	  boolean, TRUE if the sequences (or sequence) are DNA and will be translated.
#' @param   trim          if \code{TRUE} remove the first and last codon/amino acids from each
#'                        sequence before calculating properties. If \code{FALSE} do
#'                        not modify input sequences.
#' @param   label         name of sequence region to add as prefix to output column names.
#' @param   ...           additional named arguments to pass to the functions 
#'                        \link{gravy}, \link{bulk}, \link{aliphatic}, \link{polar} or \link{charge}.
#' 
#' @return  A modified \code{data} data.frame with the following columns:
#'          \itemize{
#'            \item  \code{*_aa_length}:     number of amino acids.
#'            \item  \code{*_aa_gravy}:      grand average of hydrophobicity (gravy) index.
#'            \item  \code{*_aa_bulk}:       average bulkiness of amino acids.
#'            \item  \code{*_aa_aliphatic}:  aliphatic index.
#'            \item  \code{*_aa_polarity}:   average polarity of amino acids.
#'            \item  \code{*_aa_charge}:     net charge.
#'            \item  \code{*_aa_basic}:      fraction of informative positions that are 
#'                                           Arg, His or Lys.
#'            \item  \code{*_aa_acidic}:     fraction of informative positions that are 
#'                                           Asp or Glu.
#'            \item  \code{*_aa_aromatic}:   fraction of informative positions that are 
#'                                           His, Phe, Trp or Tyr.
#'            
#'          }
#'          
#'          Where \code{*} is the value from \code{label} or the name specified for 
#'          \code{seq} if \code{label=NULL}.
#'          
#' @details 
#' For all properties except for length, non-informative positions are excluded, 
#' where non-informative is defined as any character in \code{c("X", "-", ".", "*")}.
#' 
#' The scores for gravy, bulkiness and polarity are calculated as simple averages of the 
#' scores for each informative positions. The basic, acid and aromatic indices are 
#' calculated as the fraction of informative positions falling into the given category.
#' 
#' The aliphatic index is calculated using the Ikai, 1980 method.
#' 
#' The net charge is calculated using the method of Moore, 1985, excluding the N-terminus and
#' C-terminus charges, and normalizing by the number of informative positions.  The default 
#' pH for the calculation is 7.4.
#' 
#' The following data sources were used for the default property scores:
#' \itemize{
#'   \item  hydropathy:  Kyte & Doolittle, 1982.  
#'   \item  bulkiness:   Zimmerman et al, 1968. 
#'   \item  polarity:    Grantham, 1974.
#'   \item  pK:          EMBOSS.
#' }
#' 
#' @references
#' \enumerate{
#'   \item  Zimmerman JM, Eliezer N, Simha R. The characterization of amino acid sequences 
#'            in proteins by statistical methods. J Theor Biol 21, 170-201 (1968).
#'   \item  Grantham R. Amino acid difference formula to help explain protein evolution. 
#'            Science 185, 862-864 (1974).
#'   \item  Ikai AJ. Thermostability and aliphatic index of globular proteins. 
#'            J Biochem 88, 1895-1898 (1980).
#'   \item  Kyte J, Doolittle RF. A simple method for displaying the hydropathic character 
#'            of a protein. J Mol Biol 157, 105-32 (1982).
#'   \item  Moore DS. Amino acid and peptide net charges: A simple calculational procedure. 
#'            Biochem Educ 13, 10-11 (1985).
#'   \item  Wu YC, et al. High-throughput immunoglobulin repertoire analysis distinguishes 
#'            between human IgM memory and switched memory B-cell populations. 
#'            Blood 116, 1070-8 (2010).
#'   \item  Wu YC, et al. The relationship between CD27 negative and positive B cell 
#'            populations in human peripheral blood. 
#'            Front Immunol 2, 1-12 (2011).
#'   \item  \url{https://emboss.sourceforge.net/apps/cvs/emboss/apps/iep.html}
#' }
#' 
#' @seealso 
#' See \link{countPatterns} for counting the occurrence of specific amino acid subsequences.
#' See \link{gravy}, \link{bulk}, \link{aliphatic}, \link{polar} and \link{charge} for functions 
#' that calculate the included properties individually.
#' 
#' @examples
#' # Subset example data
#' db <- ExampleDb[c(1,10,100), c("sequence_id", "junction")]
#' 
#' # Calculate default amino acid properties from DNA sequences
#' aminoAcidProperties(db, seq="junction")
#
#' # Calculate default amino acid properties from amino acid sequences
#' # Use a custom output column prefix
#' db$junction_aa <- translateDNA(db$junction)
#' aminoAcidProperties(db, seq="junction_aa", label="junction", nt=FALSE)
#' 
#' # Use the Grantham, 1974 side chain volume scores from the seqinr package
#' # Set pH=7.0 for the charge calculation
#' # Calculate only average volume and charge
#' # Remove the head and tail amino acids from the junction, thus making it the CDR3
#' library(seqinr)
#' data(aaindex)
#' x <- aaindex[["GRAR740103"]]$I
#' # Rename the score vector to use single-letter codes
#' names(x) <- translateStrings(names(x), ABBREV_AA)
#' # Calculate properties
#' aminoAcidProperties(db, property=c("bulk", "charge"), seq="junction", 
#'                     trim=TRUE, label="cdr3", bulkiness=x, pH=7.0)
#'
#' @export
aminoAcidProperties <- function(data, property=c("length", "gravy", "bulk",
                                                 "aliphatic","polarity","charge",
                                                 "basic","acidic", "aromatic"),
                                seq="junction", nt=TRUE, trim=FALSE, label=NULL, ...) {
    # Check arguments
    property <- match.arg(property, several.ok=TRUE)
    
    # Define the data.frame that will be returned with amino acid properties
    prop_colnames <- list(
        "length"    = "aa_length",
        "gravy"     = "aa_gravy",
        "bulk"      = "aa_bulk",
        "aliphatic" = "aa_aliphatic",
        "polarity"  = "aa_polarity",
        "charge"    = "aa_charge",
        "basic"     = "aa_basic",
        "acidic"    = "aa_acidic",
        "aromatic"  = "aa_aromatic"
    )
    # If no label, use sequence column name
    if (is.null(label)) { label <- seq }
    prop_colnames <- lapply(prop_colnames, function(x) { paste(label,x,sep="_") })
    
    out_df <- data.frame(matrix(NA, nrow=nrow(data), ncol=length(property)))
    colnames(out_df) <- prop_colnames[property]
    
    # Check if out_df column names already existed in data
    # if yes, throw warning
    check <- checkColumns(data, colnames(out_df))
    if (any(check == TRUE)) { warning("Duplicated columns found. Overwriting previous values.")}
    # Check input
    if (length(seq) > 1) {
        stop("You may specify only one sequence column; seq must be a vector of length 1.")
    }
    check <- checkColumns(data, seq)
    if (check != TRUE) { stop(check) }
    
    # Assign ellipsis arguments to correct function
    dots <- list(...)
    args_gravy <- dots[names(dots) %in% names(formals(gravy))]
    args_bulk <- dots[names(dots) %in% names(formals(bulk))]
    args_aliphatic <- dots[names(dots) %in% names(formals(aliphatic))]
    args_polar <- dots[names(dots) %in% names(formals(polar))]
    args_charge <- dots[names(dots) %in% names(formals(charge))]
    
    # Get sequence vector and translate if required
    region <- as.character(data[[seq]])
    region_aa <- if (nt) { 
        translateDNA(region, trim=trim) 
    } else {
        if (trim) {
            region <- substr(region, 2, stri_length(region) - 1)
        }
        region
    }
    
    ## Will retrieve properties for valid sequences only
    ## keep index to fill results data.frame
    valid_seq <- isValidAASeq(region_aa)
    if (any(valid_seq == F) ){
        not_valid_num <- sum(!valid_seq)
        warning(paste0("Found ", not_valid_num , " sequences with non valid amino acid symbols"))
    }
    valid_seq_idx <- which(valid_seq)
    region_aa <- region_aa[valid_seq_idx]
    
    # Calculate region lengths
    if ("length" %in% property) {
        aa_length <- stri_length(region_aa)
        out_df[valid_seq_idx , prop_colnames$length] <- aa_length
    }
    # Average hydrophobicity
    if ("gravy" %in% property) {
        #aa_gravy <- gravy(region_aa, hydropathy)
        aa_gravy <- do.call('gravy', c(list(seq=region_aa), args_gravy))
        out_df[valid_seq_idx , prop_colnames$gravy] <- aa_gravy
    }
    # Average bulkiness
    if ("bulk" %in% property) {
        #aa_bulk <- bulk(region_aa)
        aa_bulk <- do.call('bulk', c(list(seq=region_aa), args_bulk))
        out_df[valid_seq_idx , prop_colnames$bulk] <- aa_bulk
    }
    if ("aliphatic" %in% property) {
        # Normalizes aliphatic index
        aa_aliphatic <- do.call('aliphatic', c(list(seq=region_aa), args_aliphatic))
        out_df[valid_seq_idx , prop_colnames$aliphatic] <- aa_aliphatic
    }
    # Average polarity
    if ("polarity" %in% property) {
        #aa_polarity <- polar(region_aa)
        aa_polarity <- do.call('polar', c(list(seq=region_aa), args_polar))
        out_df[valid_seq_idx , prop_colnames$polarity] <- aa_polarity
    }
    # Normalized net charge
    if ("charge" %in% property) {
        #aa_charge <- charge(region_aa)
        aa_charge <- do.call('charge', c(list(seq=region_aa), args_charge))
        out_df[valid_seq_idx , prop_colnames$charge] <- aa_charge
    }
    
    # Count of informative positions
    aa_info <- stri_length(gsub("[X\\.\\*-]", "", region_aa))
    # Fraction of amino acid that are basic
    if ("basic" %in% property) {
        aa_basic <- countOccurrences(region_aa, "[RHK]") / aa_info
        out_df[valid_seq_idx , prop_colnames$basic] <- aa_basic
    }
    # Fraction of amino acid that are acidic
    if ("acidic" %in% property) {
        aa_acidic <- countOccurrences(region_aa, "[DE]") / aa_info
        out_df[valid_seq_idx , prop_colnames$acidic] <- aa_acidic
    }
    # Count fraction of aa that are aromatic
    if ("aromatic" %in% property) {
        aa_aromatic <- countOccurrences(region_aa, "[FWHY]") / aa_info
        out_df[valid_seq_idx , prop_colnames$aromatic] <- aa_aromatic
    }
    
    data_cols <- colnames(data) %in% colnames(out_df) == FALSE
    return(cbind(data[, data_cols], out_df))
}

ExampleDb <- file.path("..", "data-tests", "ExampleDb.gz")
db <- readChangeoDb(ExampleDb)

test_that("translateDNA", {
    expect_equal(translateDNA("ACTGACTCGA", trim=FALSE), "TDS")
    expect_equal(translateDNA("ACTGACTCGA", trim=TRUE), "D")
})

test_that("countOccurrences0",{
    expect_equal(countOccurrences("STSTSTS", "STS"), 2)
    #expect_equal(countOccurrences("STSTSTS","STS"), 3)
})

test_that("gravy", {
    aa_seq <- "CARDRSTPWRRGIASTTVRTSW"
    expect_equal(gravy(aa_seq), -0.918, tolerance=0.001)
})

test_that("bulk", {
    seq <- c("CARDRSTPWRRGIASTTVRTSW", "XXTQMYVRT")
    expect_equal(bulk(seq), c(14.46227, 16.58857), tolerance=0.001)
    
    data(aaindex, package="seqinr")
    x <- aaindex[["GRAR740103"]]$I
    # Rename the score vector to use single-letter codes
    names(x) <- translateStrings(names(x), ABBREV_AA)
    # Calculate average volume
    obs <- bulk(seq, bulkiness=x)
    expect_equal(obs, c(77.34091, 93.71429), tolerance=0.001)
})

test_that("polar", {
    # Default scale
    seq <- c("CARDRSTPWRRGIASTTVRTSW", "XXTQMYVRT")
    obs <- polar(seq)
    expect_equal(obs, c(8.55, 8.00), tolerance=0.001)
    # Use the Zimmerman et al, 1968 polarity scale from the seqinr package
    data(aaindex, package = "seqinr")
    x <- aaindex[["ZIMJ680103"]]$I
    # Rename the score vector to use single-letter codes
    names(x) <- translateStrings(names(x), ABBREV_AA)
    # Calculate polarity
    obs <- polar(seq, polarity=x)
    expect_equal(obs, c(14.94864,  8.86000),tolerance=0.001)
})


test_that("aliphatic", {
    seq <- c("CARDRSTPWRRGIASTTVRTSW", NA, "XXTQMYVRT")
    obs <- aliphatic(seq, normalize=TRUE)
    expect_equal(obs, c(0.4000000, NA, 0.4142857),tolerance=0.001)
    
})

test_that("charge", {
    seq <- c("CARDRSTPWRRGIASTTVRTSW", "XXTQMYVRT")
    # Unnormalized charge
    obs <- charge(seq, normalize=FALSE)
    expect_equal(obs, c(3.9266889, 0.9980008), tolerance=0.001)

    # Normalized charge
    obs_norm <- charge(seq, normalize=TRUE)
    expect_equal(obs_norm, c(0.1784859, 0.1425715),tolerance=0.001)
    
    # Use the Murray et al, 2006 scores from the seqinr package
    data(pK, package="seqinr")
    x <- setNames(pK[["Murray"]], rownames(pK))
    # Calculate charge
    obs <- charge(seq, pK=x, normalize=FALSE)
    expect_equal(obs, c(3.8946562, 0.9977872), tolerance=0.001)
})

test_that("aminoAcidProperties", {
    seq_aa <- translateDNA(db$JUNCTION[1:5])
    
    junction_gravy <- gravy(seq_aa)
    junction_bulk <- bulk(seq_aa)
    junction_polar <- polar(seq_aa)
    junction_aliphatic <- aliphatic(seq_aa)
    junction_charge <- charge(seq_aa)
    
    junction_properties <- aminoAcidProperties(db[1:5, ], seq="JUNCTION", 
                                               trim=FALSE, label="junction")
    expect_equal(junction_gravy, junction_properties$junction_aa_gravy, tolerance=0.001)
    expect_equal(junction_bulk, junction_properties$junction_aa_bulk, tolerance=0.001)
    expect_equal(junction_polar, junction_properties$junction_aa_polar, tolerance=0.001)
    expect_equal(junction_aliphatic, junction_properties$junction_aa_aliphatic, tolerance=0.001)
    expect_equal(junction_charge, junction_properties$junction_aa_charge, tolerance=0.001)
    
    
    data(aaindex, package="seqinr")
    h <- aaindex[["KIDA850101"]]$I
    # Rename the score vector to use single-letter codes
    names(h) <- translateStrings(names(h), ABBREV_AA)
    
    junction_gravy_h <- gravy(seq_aa, hydropathy =  h)
    junction_properties_h <- aminoAcidProperties(db[1:5,], seq="JUNCTION", 
                                            trim=FALSE, label="junction",
                                            hydropathy = h)
    expect_equal(junction_gravy_h, junction_properties_h$junction_aa_gravy, tolerance = .001)
    
    junction_gravy_na <- gravy(c(NA, "NA", "NULL"), hydropathy=h)
    expect_equal(junction_gravy_na, c(NA, 0.27, NA), tolerance=0.001)
    
    db[1,"JUNCTION"] <- NA
    db[2,"JUNCTION"] <- "NA"
    db[3,"JUNCTION"] <- "NULL"
    db[4,"JUNCTION"] <- "NLL"
    expect_warning(junction_properties_na <- aminoAcidProperties(db[1:4,], seq="JUNCTION", nt=FALSE,
                                                 trim=FALSE, label="junction",
                                                 hydropathy=h),
                   "Found 2 sequences with non valid amino acid symbols")
    expect_equal(junction_properties_na$junction_aa_length, c(NA, 2, NA, 3))
    expect_equal(junction_properties_na$junction_aa_gravy, tolerance=0.001,
                 c(NA,0.27,NA,-0.463))
    expect_equal(junction_properties_na$junction_aa_bulk,c(NA, 12.16, NA, 18.54), tolerance=0.001)
    expect_equal(junction_properties_na$junction_aa_aliphatic, c(NA, 0.5, NA, 2.6), tolerance=0.001)
    expect_equal(junction_properties_na$junction_aa_polarity, c(NA, 9.85, NA, 7.13), tolerance=0.001)
    expect_equal(junction_properties_na$junction_aa_charge, c(NA, 0, NA, 0), tolerance=0.001)
    expect_equal(junction_properties_na$junction_aa_basic, c(NA, 0, NA, 0), tolerance=0.001)
    expect_equal(junction_properties_na$junction_aa_acidic, c(NA, 0, NA, 0), tolerance=0.001)
    expect_equal(isValidAASeq(db$JUNCTION[1:4]), c(FALSE, TRUE, FALSE, TRUE))
    expect_warning(aminoAcidProperties(db[1:4, ], seq="JUNCTION", nt=FALSE,
                                       trim=FALSE, label="junction", property="length"),
                   "2 sequences")
    
    db <- data.frame(
        "junction"="ACTGACTCGA",
        "junction_aa"="TDS"
    )
    aa_prop_trim_F <- aminoAcidProperties(db, seq="junction_aa", nt=F, trim=F)
    expect_equal(aa_prop_trim_F$junction_aa_aa_length, 3)
    aa_prop_trim_T <- aminoAcidProperties(db, seq="junction_aa", nt=F, trim=T)
    expect_equal(aa_prop_trim_T$junction_aa_aa_length, 1)
})

test_that("validate amino acid sequences" ,{
    seq <- c("CARDRSTPWRRGIASTTVRTSW", "XXTQMYVR--XX", "CARJ", "10")
    expect_equal(isValidAASeq(seq), c(TRUE, TRUE, FALSE, FALSE))

})

test_that("countPatterns", {
    seq <- c("TGTCAACAGGCTAACAGTTTCCGGACGTTC",
             "TGTCAGCAATATTATATTGCTCCCTTCACTTTC",
             "TGTCAAAAGTATAACAGTGCCCCCTGGACGTTC")
    patterns <- c("A", "V", "[LI]")
    names(patterns) <- c("arg", "val", "iso_leu")
    obs <- countPatterns(seq, patterns, trim=TRUE, label="cdr3")
    expect_equal(obs$cdr3_arg, c(0.1250000, 0.1111111, 0.1111111), tolerance=0.001)
    expect_equal(obs$cdr3_val, c(0, 0, 0))
    expect_equal(obs$cdr3_iso_leu, c(0, 0.1111111, 0), tolerance=0.001)
})

# Ig lineage reconstruction via maximum parsimony

#' @include Classes.R
NULL


#### Preprocessing functions ####

#' Generate a ChangeoClone object for lineage construction
#' 
#' \code{makeChangeoClone} takes a data.frame with AIRR or Change-O style columns as input and 
#' masks gap positions, masks ragged ends, removes duplicate sequences, and merges 
#' annotations associated with duplicate sequences. It returns a \code{ChangeoClone} 
#' object which serves as input for lineage reconstruction. \strong{Note}: To use the 
#' most recent methods for building, visualizing and analyzing 
#' trees, use the R package [Dowser](https://dowser.readthedocs.io).
#' 
#' @param    data         data.frame containing the AIRR or Change-O data for a clone. See Details
#'                        for the list of required columns and their default values.
#' @param    id           name of the column containing sequence identifiers.
#' @param    seq          name of the column containing observed DNA sequences. All 
#'                        sequences in this column must be multiple aligned.
#' @param    germ         name of the column containing germline DNA sequences. All entries 
#'                        in this column should be identical for any given clone, and they
#'                        must be multiple aligned with the data in the \code{seq} column.
#' @param    v_call       name of the column containing V-segment allele assignments. All 
#'                        entries in this column should be identical to the gene level.
#' @param    j_call        name of the column containing J-segment allele assignments. All 
#'                        entries in this column should be identical to the gene level.
#' @param    junc_len     name of the column containing the length of the junction as a 
#'                        numeric value. All entries in this column should be identical 
#'                        for any given clone.
#' @param    locus  name of the column containing locus specification. Must be present
#'                        and only contain the value "IGH", representing heavy chains.
#' @param    clone        name of the column containing the identifier for the clone. All 
#'                        entries in this column should be identical.
#' @param    mask_char    character to use for masking and padding.
#' @param    max_mask     maximum number of characters to mask at the leading and trailing
#'                        sequence ends. If \code{NULL} then the upper masking bound will 
#'                        be automatically determined from the maximum number of observed 
#'                        leading or trailing Ns amongst all sequences. If set to \code{0} 
#'                        (default) then masking will not be performed.
#' @param    pad_end      if \code{TRUE} pad the end of each sequence with \code{mask_char}
#'                        to make every sequence the same length.
#' @param    text_fields  text annotation columns to retain and merge during duplicate removal.
#' @param    num_fields   numeric annotation columns to retain and sum during duplicate removal.
#' @param    seq_fields   sequence annotation columns to retain and collapse during duplicate 
#'                        removal. Note, this is distinct from the \code{seq} and \code{germ} 
#'                        arguments, which contain the primary sequence data for the clone
#'                        and should not be repeated in this argument.
#' @param    add_count    if \code{TRUE} add an additional annotation column called 
#'                        \code{collapse_count} during duplicate removal that indicates the 
#'                        number of sequences that were collapsed.
#' @param    verbose      passed on to \code{collapseDuplicates}. If \code{TRUE}, report the 
#'                        numbers of input, discarded and output sequences; otherwise, process
#'                        sequences silently.                        
#'
#' @return   A \link{ChangeoClone} object containing the modified clone.
#'
#' @details
#' The input data.frame (\code{data}) must columns for each of the required column name 
#' arguments: \code{id}, \code{seq}, \code{germ}, \code{v_call}, \code{j_call}, 
#' \code{junc_len}, and \code{clone}.  The default values are as follows:
#' \itemize{
#'   \item  \code{id       = "sequence_id"}:         unique sequence identifier.
#'   \item  \code{seq      = "sequence_alignment"}:  IMGT-gapped sample sequence.
#'   \item  \code{germ     = "germline_alignment"}:  IMGT-gapped germline sequence.
#'   \item  \code{v_call   = "v_call"}:              V segment allele call.
#'   \item  \code{j_call   = "j_call"}:              J segment allele call.
#'   \item  \code{junc_len = "junction_length"}:     junction sequence length.
#'   \item  \code{clone    = "clone_id"}:            clone identifier.
#' }
#' Additional annotation columns specified in the \code{text_fields}, \code{num_fields} 
#' or \code{seq_fields} arguments will be retained in the \code{data} slot of the return 
#' object, but are not required. If the input data.frame \code{data} already contains a 
#' column named \code{sequence}, which is not used as the \code{seq} argument, then that 
#' column will not be retained.
#' 
#' The default columns are IMGT-gapped sequence columns, but this is not a requirement. 
#' However, all sequences (both observed and germline) must be multiple aligned using
#' some scheme for both proper duplicate removal and lineage reconstruction. 
#'
#' The value for the germline sequence, V-segment gene call, J-segment gene call, 
#' junction length, and clone identifier are determined from the first entry in the 
#' \code{germ}, \code{v_call}, \code{j_call}, \code{junc_len} and \code{clone} columns, 
#' respectively. For any given clone, each value in these columns should be identical.
#'  
#' @seealso  Executes in order \link{maskSeqGaps}, \link{maskSeqEnds}, 
#'           \link{padSeqEnds}, and \link{collapseDuplicates}. 
#'           Returns a \link{ChangeoClone} object which serves as input to
#'           \link{buildPhylipLineage}.
#' 
#' @examples
#' # Example data
#' db <- data.frame(sequence_id=LETTERS[1:4],
#'                  sequence_alignment=c("CCCCTGGG", "CCCCTGGN", "NAACTGGN", "NNNCTGNN"),
#'                  germline_alignment="CCCCAGGG",
#'                  v_call="Homsap IGKV1-39*01 F",
#'                  j_call="Homsap IGKJ5*01 F",
#'                  junction_length=2,
#'                  clone_id=1,
#'                  locus=rep("IGH", length=4),
#'                  c_call=c("IGHM", "IGHG", "IGHG", "IGHA"),
#'                  duplicate_count=1:4,
#'                  stringsAsFactors=FALSE)
#' 
#' 
#'  # Without end masking
#'  makeChangeoClone(db, text_fields="c_call", num_fields="duplicate_count")
#' 
#'  # With end masking
#'  makeChangeoClone(db, max_mask=3, text_fields="c_call", num_fields="duplicate_count")
#'
#' @export
makeChangeoClone <- function(data, id="sequence_id", seq="sequence_alignment", 
                             germ="germline_alignment", v_call="v_call", j_call="j_call",
                             junc_len="junction_length", clone="clone_id", mask_char="N",
                             locus="locus", max_mask=0, pad_end=FALSE, text_fields=NULL, 
                             num_fields=NULL, seq_fields=NULL, add_count=TRUE, verbose=FALSE) {
    # Check for valid fields
    check <- checkColumns(data, c(id, seq, germ, v_call, j_call, junc_len, clone, 
                                  text_fields, num_fields, seq_fields, locus))
    if (check != TRUE) { stop(check) }
    
    num_clones <- length(unique(data[[clone]]))
    if (num_clones > 1) {
        stop(paste0("`data` contains ",num_clones, " clone identifiers in the `",clone,"` field. Expecting one."))
    }
    
    if(sum(is.na(data[[locus]])) > 0){
        stop(paste("Missing values found in",locus,"column"))
    }
    if(sum(data[[locus]] != "IGH") > 0){
        stop(paste("Only heavy chain (IGH) allowed in",locus,"column.",
            "Heavy+light chain trees only supported in Dowser: https://dowser.readthedocs.io"))
    }

    # Replace gaps with Ns and masked ragged ends
    tmp_df <- data[, c(id, seq, text_fields, num_fields, seq_fields)]
    tmp_df[[seq]] <- maskSeqGaps(tmp_df[[seq]], mask_char=mask_char, outer_only=FALSE)
    tmp_df[[seq]] <- maskSeqEnds(tmp_df[[seq]], mask_char=mask_char, max_mask=max_mask, trim=FALSE)
    germline <- maskSeqGaps(data[[germ]][1], mask_char=mask_char, outer_only=FALSE)

    # Pad ends
    if (pad_end) {
        tmp_df[[seq]] <- padSeqEnds(tmp_df[[seq]], pad_char=mask_char)
        germline      <- padSeqEnds(germline, pad_char=mask_char)
    }
    
    seq_len <- stringi::stri_length(tmp_df[[seq]])
    if (any(seq_len != seq_len[1])) {
        len_message <- paste0("All sequences are not the same length for data with first ", 
                     id, " = ", tmp_df[[id]][1], ".")
        if (!pad_end) {
            len_message <- paste(len_message, 
                    "Consider specifying pad_end=TRUE and verify the multiple alignment.")
        } else {
            len_message <- paste(len_message,
                    "Verify that all sequences are properly multiple-aligned.")
        }
        stop(len_message)
    }
    
    # Remove duplicates
    tmp_df <- collapseDuplicates(tmp_df, id=id, seq=seq, text_fields=text_fields, 
                                 num_fields=num_fields, seq_fields=seq_fields,
                                 add_count=add_count, verbose=verbose)
    
    # Define return object
    tmp_names <- names(tmp_df)
    if ("sequence" %in% tmp_names & seq != "sequence") {
        tmp_df <- tmp_df[, tmp_names != "sequence"]
        tmp_names <- names(tmp_df)
    }
    names(tmp_df)[tmp_names == seq] <- "sequence"
    names(tmp_df)[tmp_names == id] <- "sequence_id"
 
    if (length(unique(data[[junc_len]]))>1) {
      message("Junctions of multiple lengths found. `ChangeoClone` will use the length of the first one in slot `junc_len`.")
    }

    clone <- new("ChangeoClone", 
                 data=as.data.frame(tmp_df),
                 clone=as.character(data[[clone]][1]),
                 germline=germline, 
                 v_gene=getGene(data[[v_call]][1]), 
                 j_gene=getGene(data[[j_call]][1]), 
                 junc_len=data[[junc_len]][1])
    
    return(clone)
}


#### PHYLIP functions ####

# Create PHYLIP input files in a temporary folder
#
# @param   clone  a ChangeoClone object
# @param   path   a directory to store the write the output files to
# @return  a named vector translating sequence_id (names) to PHYLIP taxa (values)
writePhylipInput <- function(clone, path) {
    # Define PHYLIP columns
    nseq <- nrow(clone@data)
    v1 <- c(sprintf('%-9s', nseq + 1),
            sprintf("%-9s", "Germline"), 
            sprintf("SAM%-6s", 1:nseq))
    v2 <- c(stringi::stri_length(clone@germline),
            clone@germline, 
            clone@data[["sequence"]])
    phy_df <- data.frame(v1, v2, stringsAsFactors=F)
    
    # Define names vector mapping taxa names to original sequence identifiers
    id_map <- setNames(gsub("^\\s+|\\s+$", "", v1[-(1:2)]), clone@data[["sequence_id"]])
    
    # Create PHYLIP input file
    infile <- file.path(path, "infile")
    if (.Platform$OS.type == "windows") { 
        infile <- gsub("/","\\\\",infile)
    }

    write.table(phy_df, file=infile, 
                quote=F, sep=" ", col.names=F, row.names=F)    
    
    return(id_map)
}


# Run PHYLIP dnapars or dnaml application
#
# @param   path          temporary directory containing infile.
# @param   phylip_exec   path to dnapars or dnaml executable.
# @param   verbose       if TRUE suppress phylip console output.
# @param   onetree       if TRUE save only one tree
# @return  TRUE if phylip ran successfully and FALSE otherwise
runPhylip <- function(path, phylip_exec, verbose=FALSE, onetree=FALSE) {

    # Expand shell variables
    phylip_exec <- path.expand(phylip_exec)
    
    # Remove old files
    outfile <- file.path(path, "outfile")
    outtree <- file.path(path, "outtree")
    if (.Platform$OS.type == "windows") { 
        outfile <- gsub("/","\\\\",outfile)
        outtree <- gsub("/","\\\\",outtree)
    }
    if (file.exists(outfile)) { file.remove(outfile) }
    if (file.exists(outtree)) { file.remove(outtree) }    
    
    # Set platform specific options
    if (.Platform$OS.type == "windows") { 
        quiet_params <- list(stdout=FALSE, stderr=FALSE)
        invoke <- system2
    } else { 
        quiet_params <- list(stdout=FALSE, stderr=FALSE)
        invoke <- system2
    } 
    
    # Set dnapars or dnaml options
    if ( grepl("dnaml$",phylip_exec) | grepl("dnaml\\.exe$",phylip_exec)){
        phy_options <- c("I", "5")
    }else if (grepl("dnapars$",phylip_exec) | grepl("dnapars\\.exe$",phylip_exec)) {
        phy_options <- c("S", "Y", "I", "4", "5", ".")
    }else{
        stop("Executable not recognized! Must end with dnapars or dnaml")
    }
    if (onetree) {
        phy_options <- c(phy_options, "V", "1")
    }

    params <- list(phylip_exec, input=c(phy_options, "Y"), wait=TRUE)

    if (!verbose) {
        params <- append(params, quiet_params)
    }
    
    # Call phylip
    wd <- getwd()
    setwd(path)
    status <- tryCatch(do.call(invoke, params), error=function(e) e)
    setwd(wd)
    
    # Return TRUE if phylip ran successfully
    invisible(status == 0)
}


# Reads in the PHYLIP outfile
#
# @param   path  the temporary folder containing the dnapars outfile
# @return  a character vector with each item as a line in the outfile
readPhylipOutput <- function(path) {
    outfile <- file.path(path, "outfile")
    if (.Platform$OS.type == "windows") { 
        outfile <- gsub("/","\\\\",outfile)
    }
    phylip_out <- scan(outfile, what="character", sep="\n", 
            blank.lines.skip=FALSE, strip.white=FALSE, quiet=TRUE)
    return(phylip_out)
}


# Test for successful PHYLIP dnapars run by checking the outfile
#
# @param   phylip_out  a character vector returned by readPhylipOut
# @return  TRUE if trees built 
#          FALSE if no trees built
checkPhylipOutput <- function(phylip_out) {
    # Check for failed tree build
    result <- !(any(grepl('-1 trees in all found', phylip_out)))
    
    return(result)
}


# Extracts inferred sequences from PHYLIP dnapars or dnaml outfile
#
# @param   phylip_out   a character vector returned by readPhylipOutput
# @return  a list containing an id vector, a sequence vector and an annotation data.frame
getPhylipInferred <- function(phylip_out) {
  # Process dnapars and dnaml output
    pars_starts = grep("From\\s+To\\s+Any Steps\\?\\s+State at upper node", 
                          phylip_out, perl=T, fixed=F)
    ml_start = grep("node\\s+Reconstructed sequence", 
                          phylip_out, perl=T, fixed=F)
    if(length(pars_starts) > 0){
        seq_start <- min(pars_starts)
        seq_empty <- grep("^\\s*$", phylip_out[seq_start:length(phylip_out)], perl=T, fixed=F)
        seq_len <- seq_empty[min(which(seq_empty[-1] == (seq_empty[-length(seq_empty)] + 1)))]
        seq_block <- paste(phylip_out[(seq_start + 2):(seq_start + seq_len - 2)], collapse="\n")
        tc <- textConnection(seq_block)
        seq_df <- read.table(tc, as.is=T, fill=T, blank.lines.skip=F)
        close(tc)
        
        # Correct first line of block and remove blank rows
        fix.row <- c(1, which(is.na(seq_df[,1])) + 1)
        end_col <-  ncol(seq_df) - 2
        seq_df[fix.row, ] <- data.frame(cbind(0, seq_df[fix.row, 1], "no", seq_df[fix.row, 2:end_col]), stringsAsFactors=F)
        if (length(fix.row)>1) {
            seq_df <- seq_df[-(fix.row[-1] - 1), ]
        }
         # Create data.frame of inferred sequences
        inferred_num <- unique(grep("^[0-9]+$", seq_df[, 2], value=T))
        inferred_seq <- sapply(inferred_num, function(n) { 
            paste(t(as.matrix(seq_df[seq_df[, 2] == n, -c(1:3)])), collapse="") })
  
    }else if(length(ml_start) > 0){
        seq_start <- min(ml_start)
        seq_empty <- grep("^\\s*$", phylip_out[seq_start:length(phylip_out)], perl=T, fixed=F)
        seq_len <- max(seq_empty)
        seq_block <- paste(phylip_out[(seq_start + 2):(seq_start + seq_len - 2)], collapse="\n")
        tc <- textConnection(seq_block)
        seq_df <- read.table(tc, as.is=T, fill=T, blank.lines.skip=F)
        close(tc)

        # Correct first line of block and remove blank rows
        fix.row <- c(which(seq_df[,1]==""))
        if (length(fix.row)>1) {
            seq_df <- seq_df[-(fix.row), ]
        }
        # Create data.frame of inferred sequences
        inferred_num <- unique(grep("^[0-9]+$", seq_df[, 1], value=T))
        inferred_seq <- sapply(inferred_num, function(n) { 
            paste(t(as.matrix(seq_df[seq_df[, 1] == n, -1])), collapse="") })
        inferred_seq = toupper(inferred_seq)
    
    }else{
        stop("Input file format not recognized")
    }
      
    if (length(inferred_num)>0) {
        return(data.frame(sequence_id=paste0("Inferred", inferred_num), 
            sequence=inferred_seq, stringsAsFactors = FALSE))
    }
    data.frame(sequence_id=c(), sequence=c(), stringsAsFactors = FALSE)
}


# Extracts graph edge list from a PHYLIP dnapars or dnaml outfile
#
# @param   phylip_out  character vector returned by readPhylipOutput
# @param   id_map      named vector of PHYLIP taxa names (values) to sequence 
#                      identifiers (names) that will be translated. If NULL
#                      no taxa name translation is performed
# @return  a data.frame of edges with columns (from, to, weight)
getPhylipEdges <- function(phylip_out, id_map=NULL) {

    pars_start = grep('between\\s+and\\s+length', phylip_out, 
                           perl=TRUE, fixed=FALSE)
    ml_start = grep('Between\\s+And\\s+Length', phylip_out, 
                           perl=TRUE, fixed=FALSE)

    if(length(pars_start) > 0){
        # Process dnapars output
        edge_start <- min(pars_start)
        edge_len <- min(grep('^\\s*$', phylip_out[edge_start:length(phylip_out)], 
                             perl=TRUE, fixed=FALSE))
        edge_block <- paste(phylip_out[(edge_start + 2):(edge_start + edge_len - 2)], collapse='\n')
        tc <- textConnection(edge_block)
        edge_df <- read.table(tc, col.names=c('from', 'to', 'weight'), 
                              as.is=TRUE)
        close(tc)
    }else if(length(ml_start) > 0){
        edge_start <- min(ml_start)+3
        edge_len <- min(grep('^\\s*$', phylip_out[edge_start:length(phylip_out)], 
                             perl=TRUE, fixed=FALSE))
        block = phylip_out[(edge_start + 0):(edge_start + edge_len - 2)]
        block = unlist(lapply(strsplit(block,split="\\s+"),function(x){
            paste(x[1:(min(which(x == "("))-1)],collapse=" ")
        }))
        edge_block <- paste(block, collapse='\n')
        tc <- textConnection(edge_block)
        edge_df <- read.table(tc, col.names=c('from', 'to', 'weight'), 
                              as.is=TRUE)
        close(tc)
    }
    # Modify inferred taxa names to include "Inferred"
    inf_map <- unique(grep("^[0-9]+$", c(edge_df$from, edge_df$to), value=T))
    names(inf_map) <- paste0("Inferred", inf_map)
    edge_df$from <- translateStrings(edge_df$from, inf_map)
    edge_df$to <- translateStrings(edge_df$to, inf_map)
    
    if (!is.null(id_map)) {
        # Reassign PHYLIP taxa names to sequence IDs
        edge_df$from <- translateStrings(edge_df$from, id_map)
        edge_df$to <- translateStrings(edge_df$to, id_map)
    }
    
    return(edge_df)
}


# Modify edges of phylip output
#
# @param   edges     data.frame of edges returned by getPhylipEdges
# @param   clone     a ChangeoClone object containing sequence data
# @param   dist_mat  DNA character distance matrix
# @return  a list of modified edges data.frame and clone object
modifyPhylipEdges <- function(edges, clone, dist_mat=getDNAMatrix(gap=0)) {
    # Move germline to root position
    germ_idx <- which(edges$to == "Germline")
    edges[germ_idx, c('from', 'to')] <- edges[germ_idx, c('to', 'from')]

    # Calculate edge mutations
    for (i in 1:nrow(edges)) {
        if (edges$from[i] == "Germline") {
            seq1 <- clone@germline
        } else {
            seq1 <- clone@data[["sequence"]][clone@data[["sequence_id"]] == edges$from[i]]
        }
        seq2 <- clone@data[["sequence"]][clone@data[["sequence_id"]] == edges$to[i]]
        edges$weight[i] <- seqDist(seq1, seq2, dist_mat)        
    }
    
    # Find rows zero weight edges with inferred parent nodes
    remove_row <- which(edges$weight == 0 & 
                        edges$from != "Germline" & 
                        grepl('^Inferred\\d+$', edges$from))
    
    # Replace inferred parent nodes with child nodes when edge weight is zero
    while (length(remove_row) > 0) {
        # Remove first node with zero distance to parent
        r <- remove_row[1]
        r_idx <- which(edges[c('from', 'to')] == edges$from[r], arr.ind=T)
        edges[r_idx] <- edges$to[r]
        
        # Recalculate edge weights for modified rows
        r_mod <- r_idx[, 1][r_idx[, 1] != r]
        for (i in r_mod) {
            if (edges$from[i] == "Germline") {
                seq1 <- clone@germline
            } else {
                seq1 <- clone@data[["sequence"]][clone@data[["sequence_id"]] == edges$from[i]]
            }
            seq2 <- clone@data[["sequence"]][clone@data[["sequence_id"]] == edges$to[i]]
            edges$weight[i] <- seqDist(seq1, seq2, dist_mat)      
        }
        
        # Remove row
        edges <- edges[-r, ]
        
        # Re-determine rows to remove
        remove_row <- which(edges$weight == 0 & 
                            edges$from != "Germline" & 
                            grepl('^Inferred\\d+$', edges$from))      
    }
    
    # Remove rows from clone
    keep_clone <- clone@data[["sequence_id"]] %in% unique(c(edges$from, edges$to))
    clone@data <- as.data.frame(clone@data[keep_clone, ])
    
    return(list(edges=edges, clone=clone))
}

# Convert edge data.frame and clone object to igraph graph object
#
# @param   edges  data.frame of edges returned by getPhylipEdges
# @param   clone  a ChangeoClone object containing sequence data
# @return  an igraph graph object
phylipToGraph <- function(edges, clone) {
    # Create igraph object
    g <- igraph::graph_from_data_frame(edges, directed=T)
    
    # Add germline sequence
    germ_idx <- which(igraph::V(g)$name == "Germline")
    g <- igraph::set_vertex_attr(g, "sequence", index=germ_idx, clone@germline)
    
    # Add sample sequences and names
    clone_idx <- match(clone@data[["sequence_id"]], igraph::V(g)$name) 
    g <- igraph::set_vertex_attr(g, "sequence", index=clone_idx, clone@data[["sequence"]])
    
    # Add annotations
    ann_fields <- names(clone@data)[!(names(clone@data) %in% c("sequence_id", "sequence"))]
    for (n in ann_fields) {
        g <- igraph::set_vertex_attr(g, n, index=germ_idx, NA)
        g <- igraph::set_vertex_attr(g, n, index=clone_idx, clone@data[[n]])
    }
    
    # Add edge and vertex labels
    igraph::V(g)$label <- igraph::V(g)$name
    igraph::E(g)$label <- igraph::E(g)$weight
    
    # Add graph attributes
    g$clone <- clone@clone
    g$v_gene <- clone@v_gene
    g$j_gene <- clone@j_gene
    g$junc_len <- clone@junc_len
    
    return(g)
}


#' Infer an Ig lineage using PHYLIP
#' 
#' \code{buildPhylipLineage} reconstructs an Ig lineage via maximum parsimony using the 
#' dnapars application, or maximum likelihood using the dnaml application of the PHYLIP package.
#' \strong{Note}: To use the most recent methods for building, visualizing and analyzing 
#' trees, use the R package [Dowser](https://dowser.readthedocs.io).
#' 
#' @param    clone         \link{ChangeoClone} object containing clone data.
#' @param    phylip_exec   absolute path to the PHYLIP dnapars executable.
#' @param    dist_mat      character distance matrix to use for reassigning edge weights. 
#'                         Defaults to a Hamming distance matrix returned by \link{getDNAMatrix} 
#'                         with \code{gap=0}. If gap characters, \code{c("-", ".")}, are assigned 
#'                         a value of -1 in \code{dist_mat} then contiguous gaps of any run length,
#'                         which are not present in both sequences, will be counted as a 
#'                         distance of 1. Meaning, indels of any length will increase
#'                         the sequence distance by 1. Gap values other than -1 will 
#'                         return a distance that does not consider indels as a special case.
#' @param    rm_temp       if \code{TRUE} delete the temporary directory after running dnapars;
#'                         if \code{FALSE} keep the temporary directory.
#' @param    verbose       if \code{FALSE} suppress the output of dnapars; 
#'                         if \code{TRUE} STDOUT and STDERR of dnapars will be passed to 
#'                         the console.
#' @param    branch_length specifies how to define branch lengths; one of \code{"mutations"} 
#'                         or \code{"distance"}. If set to \code{"mutations"} (default), then branch 
#'                         lengths represent the number of mutations between nodes. 
#'                         If set to \code{"distance"}, then branch lengths represent
#'                         the expected number of mutations per site, unaltered from PHYLIP output.
#' @param    temp_path     specific path to temp directory if desired.
#' @param    onetree       if \code{TRUE} save only one tree.
#'                                                
#' @return   An igraph \code{graph} object defining the Ig lineage tree. Each unique input 
#'           sequence in \code{clone} is a vertex of the tree, with additional vertices being
#'           either the germline (root) sequences or inferred intermediates. The \code{graph} 
#'           object has the following attributes.
#'           
#'           Vertex attributes:
#'           \itemize{
#'             \item  \code{name}:      value in the \code{sequence_id} column of the \code{data} 
#'                                      slot of the input \code{clone} for observed sequences. 
#'                                      The germline (root) vertex is assigned the name 
#'                                      "Germline" and inferred intermediates are assigned
#'                                      names with the format \{"Inferred1", "Inferred2", ...\}.
#'             \item  \code{sequence}:  value in the \code{sequence} column of the \code{data} 
#'                                      slot of the input \code{clone} for observed sequences.
#'                                      The germline (root) vertex is assigned the sequence
#'                                      in the \code{germline} slot of the input \code{clone}.
#'                                      The sequence of inferred intermediates are extracted
#'                                      from the dnapars output.
#'             \item  \code{label}:     same as the \code{name} attribute.
#'           }
#'           Additionally, each other column in the \code{data} slot of the input 
#'           \code{clone} is added as a vertex attribute with the attribute name set to 
#'           the source column name. For the germline and inferred intermediate vertices,
#'           these additional vertex attributes are all assigned a value of \code{NA}.
#'           
#'           Edge attributes:
#'           \itemize{
#'             \item  \code{weight}:    Hamming distance between the \code{sequence} attributes
#'                                      of the two vertices.
#'             \item  \code{label}:     same as the \code{weight} attribute.
#'           }
#'           Graph attributes:
#'           \itemize{
#'             \item  \code{clone}:     clone identifier from the \code{clone} slot of the
#'                                      input \code{ChangeoClone}.
#'             \item  \code{v_gene}:    V-segment gene call from the \code{v_gene} slot of 
#'                                      the input \code{ChangeoClone}.
#'             \item  \code{j_gene}:    J-segment gene call from the \code{j_gene} slot of 
#'                                      the input \code{ChangeoClone}.
#'             \item  \code{junc_len}:  junction length (nucleotide count) from the 
#'                                      \code{junc_len} slot of the input \code{ChangeoClone}.
#' 
#'              Alternatively, this function will return an \code{phylo} object, which is compatible
#'              with the ape package. This object will contain reconstructed ancestral sequences in
#'              \code{nodes} attribute.
#'           }
#'           
#' @details
#' \code{buildPhylipLineage} builds the lineage tree of a set of unique Ig sequences via
#' maximum parsimony through an external call to the dnapars application of the PHYLIP
#' package. dnapars is called with default algorithm options, except for the search option, 
#' which is set to "Rearrange on one best tree". The germline sequence of the clone is used 
#' for the outgroup. 
#' 
#' Following tree construction using dnapars, the dnapars output is modified to allow
#' input sequences to appear as internal nodes of the tree. Intermediate sequences 
#' inferred by dnapars are replaced by children within the tree having a Hamming distance 
#' of zero from their parent node. With the default \code{dist_mat}, the distance calculation 
#' allows IUPAC ambiguous character matches, where an ambiguous character has distance zero 
#' to any character in the set of characters it represents. Distance calculation and movement of 
#' child nodes up the tree is repeated until all parent-child pairs have a distance greater than zero 
#' between them. The germline sequence (outgroup) is moved to the root of the tree and
#' excluded from the node replacement processes, which permits the trunk of the tree to be
#' the only edge with a distance of zero. Edge weights of the resultant tree are assigned 
#' as the distance between each sequence.
#' 
#' @references
#' \enumerate{
#'   \item  Felsenstein J. PHYLIP - Phylogeny Inference Package (Version 3.2). 
#'            Cladistics. 1989 5:164-166.
#'   \item  Stern JNH, Yaari G, Vander Heiden JA, et al. B cells populating the multiple 
#'            sclerosis brain mature in the draining cervical lymph nodes. 
#'            Sci Transl Med. 2014 6(248):248ra107.
#' }
#'   
#' @seealso  Takes as input a \link{ChangeoClone}. 
#'           Temporary directories are created with \link{makeTempDir}.
#'           Distance is calculated using \link{seqDist}. 
#'           See [igraph](http://www.rdocumentation.org/packages/igraph/topics/aaa-igraph-package) 
#'           and [igraph.plotting](http://www.rdocumentation.org/packages/igraph/topics/plot.common) 
#'           for working  with igraph \code{graph} objects.
#'
#' @examples
#' \dontrun{
#' # Preprocess clone
#' db <- subset(ExampleDb, clone_id == 3138)
#' clone <- makeChangeoClone(db, text_fields=c("sample_id", "c_call"), 
#'                           num_fields="duplicate_count")
#' 
#' # Run PHYLIP and process output
#' phylip_exec <- "~/apps/phylip-3.695/bin/dnapars"
#' graph <- buildPhylipLineage(clone, phylip_exec, rm_temp=TRUE)
#' 
#' # Plot graph with a tree layout
#' library(igraph)
#' plot(graph, layout=layout_as_tree, vertex.label=V(graph)$c_call, 
#'      vertex.size=50, edge.arrow.mode=0, vertex.color="grey80")
#' 
#' # To consider each indel event as a mutation, change the masking character 
#' # and distance matrix
#' clone <- makeChangeoClone(db, text_fields=c("sample_id", "c_call"), 
#'                           num_fields="duplicate_count", mask_char="-")
#' graph <- buildPhylipLineage(clone, phylip_exec, dist_mat=getDNAMatrix(gap=-1), 
#'                             rm_temp=TRUE)
#' }
#' 
#' @export
buildPhylipLineage <- function(clone, phylip_exec, dist_mat=getDNAMatrix(gap=0), 
                               rm_temp=FALSE, verbose=FALSE, temp_path=NULL, onetree=FALSE,
                               branch_length=c("mutations", "distance")) {
    # Check clone size
    if (nrow(clone@data) < 2) {
        warning("Clone ", clone@clone, " was skipped as it does not contain at least 
                2 unique sequences")
        return(NULL)
    }
    
    # determine branch length type
	  branch_length <- match.arg(branch_length)

    # Check fields
    seq_len <- unique(stringi::stri_length(clone@data[["sequence"]]))
    germ_len <- ifelse(length(clone@germline) == 0, 0, stringi::stri_length(clone@germline))
    if(germ_len == 0) {
        stop("Clone ", clone@clone, "does not contain a germline sequence.")
    }
    if(length(seq_len) != 1) {
        stop("Clone ", clone@clone, "does not contain sequences of equal length.")
    }
    if(seq_len != germ_len) {
        stop("The germline and input sequences are not the same length for clone ", clone@clone)
    }
    
    # Check dnapars access
    if (file.access(phylip_exec, mode=1) == -1) {
        stop("The file ", phylip_exec, " cannot be executed.")
    }
    
     # Create temporary directory
    if(is.null(temp_path)){
        temp_path <- makeTempDir(paste0(clone@clone, "-phylip"))
    }else{
        if(!dir.exists(temp_path)){
            dir.create(temp_path)
        }
    }
    if (verbose) {
        cat("TEMP_DIR> ", temp_path, "\n", sep="")
    }
    
    # Run PHYLIP
    id_map <- writePhylipInput(clone, temp_path)

    runPhylip(temp_path, phylip_exec, verbose=verbose, onetree=onetree)

    phylip_out <- readPhylipOutput(temp_path)
    
    # Remove temporary directory
    if (rm_temp) {
        unlink(temp_path, recursive=TRUE)
    }
    
    # Check output for trees
    if (!checkPhylipOutput(phylip_out)) {
        warning('PHYLIP failed to generate trees for clone ', clone)
        return(NULL)
    }
    
    # Extract inferred sequences from PHYLIP output
    inf_df <- getPhylipInferred(phylip_out)

    # Extract edge table from PHYLIP output 
    edges <- getPhylipEdges(phylip_out, id_map=id_map)

    clone@data <- as.data.frame(dplyr::bind_rows(clone@data, inf_df))

    # Convert edges and clone data to igraph graph object
    if( branch_length == "mutations" ){
        mod_list <- modifyPhylipEdges(edges, clone, dist_mat=dist_mat)
        graph <- phylipToGraph(mod_list$edges, mod_list$clone)
    }else{
        # or just keep everything the same	
    	germ_idx <- which(edges$to == "Germline")
    	edges[germ_idx, c('from', 'to')] <- edges[germ_idx, c('to', 'from')]
        graph <- phylipToGraph(edges, clone)
    }
   
    return(graph)
}


#' Convert a tree in ape \code{phylo} format to igraph \code{graph} format.
#' 
#' \code{phyloToGraph} converts a tree in \code{phylo} format to and 
#' \code{graph} format.
#' 
#' @param   phylo  An ape \code{phylo} object.
#' @param   germline  If specified, places specified tip sequence as the direct 
#'                    ancestor of the tree
#'                                                 
#' @return   A \code{graph} object representing the input tree. 
#'           
#' @details
#' Convert from phylo to graph object. Uses the node.label vector to label internal nodes. Nodes 
#' may rotate but overall topology will remain constant.
#' 
#' @references
#' \enumerate{
#'   \item  Hoehn KB, Lunter G, Pybus OG - A Phylogenetic Codon Substitution Model for Antibody 
#'              Lineages. Genetics 2017 206(1):417-427
#'              https://doi.org/10.1534/genetics.116.196303 
#'  \item  Hoehn KB, Vander Heiden JA, Zhou JQ, Lunter G, Pybus OG, Kleinstein SHK - 
#'              Repertoire-wide phylogenetic models of B cell molecular evolution reveal 
#'              evolutionary signatures of aging and vaccination. bioRxiv 2019  
#'              https://doi.org/10.1101/558825 
#' }
#'
#' @examples
#' \dontrun{
#'    library(igraph)
#'    library(ape)
#' 
#'    #convert to phylo
#'    phylo = graphToPhylo(graph)
#'    
#'    #plot tree using ape
#'    plot(phylo,show.node.label=TRUE)
#'    
#'    #store as newick tree
#'    write.tree(phylo,file="tree.newick")
#'    
#'    #read in tree from newick file
#'    phylo_r = read.tree("tree.newick")
#'    
#'    #convert to igraph
#'    graph_r = phyloToGraph(phylo_r,germline="Germline")
#'    
#'    #plot graph - same as before, possibly rotated
#'    plot(graph_r,layout=layout_as_tree)
#' }
#' 
#' @export
phyloToGraph <- function(phylo, germline="Germline") {
    names <- 1:length(unique(c(phylo$edge[, 1],phylo$edge[, 2])))
    for(i in 1:length(phylo$tip.label)){
        names[i] <- phylo$tip.label[i]
    }
    if(!is.null(phylo$node.label)){
        for(j in 1:length(phylo$node.label)){
            i <- i + 1
            names[i] <- phylo$node.label[j]
        }
    }
    d <- data.frame(cbind(phylo$edge,phylo$edge.length))
    names(d)=c("from", "to", "weight")
    
    if(!is.null(phylo$nodes)){
    	seqs <- unlist(lapply(phylo$nodes,function(x)x$sequence))
    	names(seqs) <- lapply(phylo$nodes,function(x)x$id)
    }
    if(!is.null(germline)){
        germnode <- which(phylo$tip.label == germline)
        phylo$uca <- phylo$edge[phylo$edge[,2] == germnode,1]
        if(sum(d$from == phylo$uca) == 2){
            d[d$from == phylo$uca, ]$from <- germnode
            d <- d[!(d$from == germnode & d$to == germnode),] 
        }else{
            row <- which(d$from == phylo$uca & d$to == germnode)
            d[row,]$to <- phylo$uca
            d[row,]$from <- germnode
        }
    }

    d$to <- as.character(d$to)
    d$from <- as.character(d$from)
    g <- igraph::graph_from_data_frame(d)
    igraph::V(g)$name <- names[as.numeric(igraph::V(g)$name)]
    igraph::E(g)$label <- igraph::E(g)$weight
    if(!is.null(phylo$nodes)){
    	igraph::V(g)$sequence <- seqs[igraph::V(g)$name]
    }
    return(g)
}


#' Convert a tree in igraph \code{graph} format to ape \code{phylo} format.
#' 
#' \code{graphToPhylo} a tree in igraph \code{graph} format to ape \code{phylo} 
#' format.
#' 
#' @param   graph  An igraph \code{graph} object.
#'
#' @return   A \code{phylo} object representing the input tree. Tip and internal node names are 
#'           stored in the \code{tip.label} and \code{node.label} vectors, respectively.
#'           
#' @details
#' Convert from igraph \code{graph} object to ape \code{phylo} object. If \code{graph} object
#' was previously rooted with the germline as the direct ancestor, this will re-attach the 
#' germline as a descendant node with a zero branch length to a new universal common ancestor (UCA) 
#' node and store the germline node ID in the \code{germid} attribute and UCA node number in 
#' the \code{uca} attribute. Otherwise these attributes will not be specified in the \code{phylo} object. 
#' Using \code{phyloToGraph(phylo, germline=phylo$germid)} creates a \code{graph} object with the germline 
#' back as the direct ancestor. Tip and internal node names are 
#' stored in the \code{tip.label} and \code{node.label} vectors, respectively.
#' 
#' @references
#' \enumerate{
#'   \item  Hoehn KB, Lunter G, Pybus OG - A Phylogenetic Codon Substitution Model for Antibody 
#'              Lineages. Genetics 2017 206(1):417-427
#'              https://doi.org/10.1534/genetics.116.196303 
#'  \item  Hoehn KB, Vander Heiden JA, Zhou JQ, Lunter G, Pybus OG, Kleinstein SHK - 
#'              Repertoire-wide phylogenetic models of B cell molecular evolution reveal 
#'              evolutionary signatures of aging and vaccination. bioRxiv 2019  
#'              https://doi.org/10.1101/558825 
#' }
#'
#' @examples
#' \dontrun{
#'    library(igraph)
#'    library(ape)
#' 
#'    #convert to phylo
#'    phylo = graphToPhylo(graph)
#'    
#'    #plot tree using ape
#'    plot(phylo,show.node.label=TRUE)
#'    
#'    #store as newick tree
#'    write.tree(phylo,file="tree.newick")
#'    
#'    #read in tree from newick file
#'    phylo_r = read.tree("tree.newick")
#'    
#'    #convert to igraph
#'    graph_r = phyloToGraph(phylo_r,germline="Germline")
#'    
#'    #plot graph - same as before, possibly rotated
#'    plot(graph_r,layout=layout_as_tree)
#' }
#' 
#' @export
graphToPhylo <- function(graph) {
    df  <- igraph::as_data_frame(graph)
    node_counts <- table(c(df$to,df$from))
    tips <- names(node_counts)[node_counts == 1]
    nodes <- names(node_counts)[node_counts > 1]
    attr <- igraph::vertex_attr(graph)
    seqs <- attr$sequence
    names(seqs) <- attr$name

    germline <- tips[tips %in% df$from]
    if(length(germline) > 0){
        ucanode <- paste0(germline,"_UCA")#max(as.numeric(nodes))+1
        nodes <- c(ucanode,nodes)
        df[df$from == germline,]$from <- ucanode
        row <- c(ucanode,germline,0.0, 0.0)
        names(row) <- c("from","to","weight", "label")
        df <- rbind(df, row)
        seqs <- c(seqs,seqs["Germline"])
        names(seqs)[length(seqs)] = paste0(germline,"_UCA")
    }
    tipn <- 1:length(tips)
    names(tipn) <- tips
    noden <- (length(tips)+1):(length(tips)+length(nodes))
    names(noden) <- nodes
    renumber <- c(tipn,noden)

    df$from <- as.numeric(renumber[df$from])
    df$to <- as.numeric(renumber[df$to])    
    
    phylo <- list()
    phylo$edge <- matrix(cbind(df$from,df$to),ncol=2)
    phylo$edge.length <- as.numeric(df$weight)
    phylo$tip.label <- tips
    phylo$node.label <- nodes
    phylo$Nnode <- length(nodes)
    phylo$node.label <- nodes
    class(phylo) <- "phylo"

    nnodes <- length(renumber)
    phylo$nodes <- lapply(1:nnodes,function(x){
        n <- list()
        n$id <- names(renumber[renumber == x])
        n$sequence <- seqs[n$id]
        n
    })

    phylo = ape::ladderize(phylo, right=FALSE)
    
    return(phylo)
}

# Reroot phylogenetic tree to have its germline sequence at a zero-length branch 
# to a node which is the direct ancestor of the tree's UCA. Assigns \code{uca}
# to be the ancestral node to the tree's germline sequence, as \code{germid} as
# the tree's germline sequence ID. 
#
# @param   tree     An ape \code{phylo} object
# @param   germid   ID of the tree's predicted germline sequence
# @param   resolve  If \code{TRUE} reroots tree to specified germline sequnece.
#                   usually not necessary with IgPhyML trees analyzed with HLP model.
rerootGermline <- function(tree, germid, resolve=FALSE){
    if(resolve) {
        tree <- ape::root(phy=tree, outgroup=germid, resolve.root=T, edge.label=TRUE)
    }
    tree <- ape::reorder.phylo(tree, "postorder")  
    edges <- tree$edge
    rootnode <- which(tree$tip.label==germid)
    rootedge <- which(edges[, 2] == rootnode)
    rootanc <- edges[edges[, 2] == rootnode, 1]
    mrcaedge <- which(edges[, 1] == rootanc & edges[, 2] != rootnode)
    if(length(mrcaedge) > 1){
            print("POLYTOMY AT ROOT!")
            quit(save="no", status=1, runLast=FALSE)
    }
    tree$edge.length[mrcaedge] <- tree$edge.length[mrcaedge] + tree$edge.length[rootedge]
    tree$edge.length[rootedge] <- 0
    tree$uca <- rootanc
    tree$germid <- germid
    
    return(tree)
}

#' Read in output from IgPhyML
#' 
#' \code{readIgphyml} reads output from the IgPhyML phylogenetics inference package for 
#' B cell repertoires
#' 
#' @param    file          IgPhyML output file (.tab).
#' @param    id            ID to assign to output object.
#' @param    format        if \code{"graph"} return trees as igraph \code{graph} objects. 
#'                         if \code{"phylo"} return trees as ape \code{phylo} objects.
#' @param    branches        if \code{"distance"} branch lengths are in expected mutations per
#'                         site. If \code{"mutations"} branches are in expected mutations.
#' @param    collapse      if \code{TRUE} transform branch lengths to units of substitutions, 
#'                         rather than substitutions per site, and collapse internal nodes
#'                         separated by branches < 0.1 substitutions. Will also remove all
#'                         internal node labels, as it makes them inconsistent.
#'                                                
#' @return   A list containing IgPhyML model parameters and estimated lineage trees. 
#'           
#'           Object attributes:
#'           \itemize{
#'             \item  \code{param}:     Data.frame of parameter estimates for each clonal 
#'                                      lineage. Columns include: \code{CLONE}, which is the 
#'                                      clone id; \code{NSEQ}, the total number of sequences in 
#'                                      the lineage; \code{NSITE}, the number of codon sites;
#'                                      \code{TREE_LENGTH}, the sum of all branch lengths in 
#'                                      the estimated lineage tree; and \code{LHOOD}, the log 
#'                                      likelihood of the clone's sequences given the tree and
#'                                      parameters. Subsequent columns are parameter estimates 
#'                                      from IgPhyML, which will depend on the model used. 
#'                                      Parameter columns ending with \code{_MLE} are maximum 
#'                                      likelihood estimates; those ending with \code{_LCI} are 
#'                                      the lower 95%% confidence interval estimate; those ending 
#'                                      with \code{_UCI} are the upper 95%% confidence interval 
#'                                      estimate. The first line of \code{param} is for clone 
#'                                      \code{REPERTOIRE}, 
#'                                      which is a summary of all lineages within the repertoire.
#'                                      For this row, \code{NSEQ} is the total number of sequences, 
#'                                      \code{NSITE} is the average number of sites, and
#'                                      \code{TREE_LENGTH} is the mean tree length. For most 
#'                                      applications, parameter values will be the same for all 
#'                                      lineages within the repertoire, so access them simply by:
#'                                      \code{<object>$param$OMEGA_CDR_MLE[1]} to, for instance,
#'                                      get the estimate of dN/dS on the CDRs at the repertoire level.
#'             \item  \code{trees}:     List of tree objects estimated by IgPhyML. If 
#'                                      \code{format="graph"} these are igraph \code{graph} objects. 
#'                                      If \code{format="phylo"}, these are ape \code{phylo} objects.
#'             \item  \code{command}:   Command used to run IgPhyML.
#'           }
#'           
#' @details
#' \code{readIgphyml} reads output from the IgPhyML repertoire phylogenetics inference package. 
#' The resulting object is divided between parameter estimates (usually under the HLP19 model),
#' which provide information about mutation and selection pressure operating on the sequences.
#' 
#' Trees returned from this function are either igraph objects or phylo objects, and each may be 
#' visualized accordingly. Further, branch lengths in tree may represent either the expected number of
#' substitutions per site (codon, if estimated under HLP or GY94 models), or the total number of 
#' expected substitutions per site. If the latter, internal nodes - but not tips - separated by branch
#' lengths less than 0.1 are collapsed to simplify viewing.
#' 
#' @references
#' \enumerate{
#'   \item  Hoehn KB, Lunter G, Pybus OG - A Phylogenetic Codon Substitution Model for Antibody 
#'              Lineages. Genetics 2017 206(1):417-427
#'              https://doi.org/10.1534/genetics.116.196303 
#'  \item  Hoehn KB, Vander Heiden JA, Zhou JQ, Lunter G, Pybus OG, Kleinstein SHK - 
#'              Repertoire-wide phylogenetic models of B cell molecular evolution reveal 
#'              evolutionary signatures of aging and vaccination. bioRxiv 2019  
#'              https://doi.org/10.1101/558825 
#' }
#'
#' @examples
#' \dontrun{
#'    # Read in and plot a tree from an igphyml run
#'    library(igraph)
#'    s1 <- readIgphyml("IB+7d_lineages_gy.tsv_igphyml_stats_hlp.tab", id="+7d")
#'    print(s1$param$OMEGA_CDR_MLE[1])
#'    plot(s1$trees[[1]], layout=layout_as_tree, edge.label=E(s1$trees[[1]])$weight)
#' }
#' 
#' @export
readIgphyml <- function(file, id=NULL, format=c("graph", "phylo"), collapse=FALSE,
    branches=c("mutations","distance")) {
    # Check arguments
    format <- match.arg(format)
    branches <- match.arg(branches)
    
    out <- list()
    trees <- list()
    df <- read.table(file, sep="\t", header=TRUE, stringsAsFactors=FALSE)
    params <- df[, !names(df) %in% c("TREE")]
    names(params) = tolower(names(params))
    out[["param"]] <- params
    out[["command"]] <- df[1, ]$TREE
    for (i in 2:nrow(df)) {
        tree <- ape::read.tree(text=df[i, ][["TREE"]])
        germ_base <- paste0(df[["CLONE"]][i], "_GERM")
        germ_id <- tree$tip.label[grepl(germ_base,tree$tip.label)]
        if(length(germ_id) > 1){
            stop("Can only be one tip of the form '<cloneid>_GERM'")
        }

        if(!ape::is.rooted(tree)){
            warning(paste("Tree",germ_base,"is not rooted and should be!"))
            tree <- rerootGermline(tree,germ_id,resolve=TRUE)
            tree$node.label <- NULL
        }
        rtree <- ape::ladderize(tree)
        rtree$germid <- germ_id
        if (branches == "mutations") {
            rtree$edge.length <- round(rtree$edge.length*df[i, ]$NSITE, digits=1)
        }
        if (collapse) {
            rtree$node.label <- NULL
            if(branches == "mutations"){
                rtree <- ape::di2multi(rtree, tol=0.1)
            }else{
                rtree <- ape::di2multi(rtree, tol=0.0001)
            }
        }
        if (format == "graph") {
            ig <- phyloToGraph(rtree, germline=rtree$germid)
            trees[[df[["CLONE"]][i]]] <- ig
        } else if (format == "phylo") {
            trees[[df[["CLONE"]][i]]] <- rtree
        } else {
            stop("Format must be either 'graph' or 'phylo'.")
        }
    }
    
    out[["trees"]] <- trees

    if (!is.null(id)) {
        out$param$id <- id
    }
    
    return(out)
}


#' Combine IgPhyML object parameters into a dataframe
#' 
#' \code{combineIgphyml} combines IgPhyML object parameters into a data.frame.
#' 
#' @param   iglist         list of igphyml objects returned by \link{readIgphyml}. 
#'                         Each must have an \code{id} column in its \code{param} attribute, 
#'                         which can be added automatically using the \code{id} option of 
#'                         \code{readIgphyml}.
#' @param   format         string specifying whether each column of the resulting data.frame
#'                         should represent a parameter (\code{wide}) or if 
#'                         there should only be three columns; i.e. id, variable, and value
#'                         (\code{long}).
#'                                                
#' @return   A data.frame containing HLP model parameter estimates for all igphyml objects.
#'           Only parameters shared among all objects will be returned.
#'           
#' @details
#' \code{combineIgphyml} combines repertoire-wide parameter estimates from multiple igphyml
#' objects produced by readIgphyml into a dataframe that can be easily used for plotting and 
#' other hypothesis testing analyses.
#' 
#' All igphyml objects used must have an "id" column in their \code{param} attribute, which
#' can be added automatically from the \code{id} flag of \code{readIgphyml}. 
#' 
#' @references
#' \enumerate{
#'   \item  Hoehn KB, Lunter G, Pybus OG - A Phylogenetic Codon Substitution Model for Antibody 
#'              Lineages. Genetics 2017 206(1):417-427
#'              https://doi.org/10.1534/genetics.116.196303 
#'  \item  Hoehn KB, Vander Heiden JA, Zhou JQ, Lunter G, Pybus OG, Kleinstein SHK - 
#'              Repertoire-wide phylogenetic models of B cell molecular evolution reveal 
#'              evolutionary signatures of aging and vaccination. bioRxiv 2019  
#'              https://doi.org/10.1101/558825 
#' }
#'
#' @seealso  \link{readIgphyml} 
#'           
#' @examples
#' \dontrun{
#'    # Read in and combine two igphyml runs
#'    s1 <- readIgphyml("IB+7d_lineages_gy.tsv_igphyml_stats_hlp.tab", id="+7d")
#'    s2 <- readIgphyml("IB+7d_lineages_gy.tsv_igphyml_stats_hlp.tab", id="s2")
#'    combineIgphyml(list(s1, s2))
#' }
#' 
#' @export
combineIgphyml <- function(iglist, format=c("wide", "long")) {
    # Check arguments
    format <- match.arg(format)
    
    ordered_params <- c(
        "id", "nseq", "nsite", "lhood", "tree_length", 
        "omega_fwr_mle", "omega_fwr_lci", "omega_fwr_uci", 
        "omega_cdr_mle", "omega_cdr_lci", "omega_cdr_uci", 
        "kappa_mle", "kappa_lci", "kappa_uci", 
        "wrc_2_mle", "wrc_2_lci", "wrc_2_uci", 
        "gyw_0_mle", "gyw_0_lci", "gyw_0_uci", 
        "wa_1_mle", "wa_1_lci", "wa_1_uci", 
        "tw_0_mle", "tw_0_lci", "tw_0_uci", 
        "syc_2_mle", "syc_2_lci", "syc_2_uci", 
        "grs_0_mle", "grs_0_lci", "grs_0_uci")
    paramCount <- table(unlist(lapply(iglist, function(x) names(x$param))))
    params <- names(paramCount[paramCount == max(paramCount)])
    params <- ordered_params[ordered_params %in% params]
    if (sum(params == "id") == 0) {
        message <- "id not specified in objects. Use 'id' flag in readIgphyml."
        stop(message)
    }
    
    repertoires <- lapply(iglist, function(x) x$param[1, params])
    combined <- dplyr::bind_rows(repertoires)
    if (format == "long") {
        combined <- tidyr::gather(combined, "variable", "value", -!!rlang::sym("id"))
        combined$variable <- factor(combined$variable, levels=params)
    }
    
    return(combined)
}

# ExampleDb <- file.path("tests", "data-tests", "ExampleDb_airr.gz")
ExampleDb <- file.path("..", "data-tests", "ExampleDb_airr.gz")
expect_warning(
    db <- readChangeoDb(ExampleDb),
    regexp="airr::read_rearrangement"
)

test_that("makeChangeoClone",{
    # Example AIRR data.frame
    db <- data.frame(sequence_id=LETTERS[1:4],
                     sequence_alignment=c("CCCCTGGG", "CCCCTGGN", "NAACTGGN", "NNNCTGNN"),
                     v_call="Homsap IGKV1-39*01 F",
                     j_call="Homsap IGKJ5*01 F",
                     junction_length=2,
                     germline_alignment="CCCCAGGG",
                     clone_id=1,
                     TYPE=c("IgM", "IgG", "IgG", "IgA"),
                     COUNT=1:4,
                     locus=rep("IGH",length=4),
                     stringsAsFactors=FALSE)
    
    # Example Change-O data.frame
   changeodb <- data.frame(SEQUENCE_ID=LETTERS[1:4],
                  SEQUENCE_IMGT=c("CCCCTGGG", "CCCCTGGN", "NAACTGGN", "NNNCTGNN"),
                  V_CALL="Homsap IGKV1-39*01 F",
                  J_CALL="Homsap IGKJ5*01 F",
                  JUNCTION_LENGTH=2,
                  GERMLINE_IMGT_D_MASK="CCCCAGGG",
                  CLONE=1,
                  TYPE=c("IgM", "IgG", "IgG", "IgA"),
                  COUNT=1:4,
                  LOCUS=rep("IGH",length=4),
                  stringsAsFactors=FALSE)
                   
    exp <- data.frame("sequence_id"=c("C", "A"),
                      "sequence"=c("NAACTGGN", "CCCCTGGG"),
                      "TYPE"=c("IgG", "IgG,IgM"),
                      "COUNT"=c(3, 3),
                      "collapse_count"=c(1, 2),
                      stringsAsFactors=FALSE)

    exp_pad <- data.frame("sequence_id"=c("C", "A"),
                      "sequence"=c("NAACTGGNN", "CCCCTGGGN"),
                      "TYPE"=c("IgG", "IgG,IgM"),
                      "COUNT"=c(3, 3),
                      "collapse_count"=c(1, 2),
                      stringsAsFactors=FALSE)
    
    # Bad data
    # Stop if data contains multiple clone identifiers
    bad_db <- db
    bad_db$clone_id <- 1:nrow(bad_db)
    expect_error(makeChangeoClone(bad_db, locus="locus"),
                 "`data` contains 4 clone identifiers in the `clone_id` field. Expecting one.")
    
    # Without end masking
    clone <- makeChangeoClone(db, text_fields="TYPE", num_fields="COUNT")
    
    changeoclone <- makeChangeoClone(changeodb, id="SEQUENCE_ID", seq="SEQUENCE_IMGT", 
           germ="GERMLINE_IMGT_D_MASK", v_call="V_CALL", j_call="J_CALL", locus="LOCUS",
           junc_len="JUNCTION_LENGTH", clone="CLONE", text_fields="TYPE", num_fields="COUNT")

    expect_true(inherits(clone, "ChangeoClone"))
    expect_equal(clone@clone, "1")
    expect_equal(clone@germline, "CCCCAGGG")
    expect_equal(clone@v_gene, "IGKV1-39")
    expect_equal(clone@j_gene, "IGKJ5")
    expect_equal(clone@junc_len, 2)
    expect_equal(clone@data, exp, tolerance=0.001)
    expect_equal(changeoclone@data, exp, tolerance=0.001)
    
    # With padding
    db_trim <- db
    db_trim$sequence_alignment <- c("CCCCTGGG", "CCCCTGG", "NAACTGG", "NNNCTG")
    clone <- makeChangeoClone(db_trim, text_fields="TYPE", num_fields="COUNT", pad_end=TRUE)

    db_trim <- changeodb
    db_trim$sequence_alignment <- c("CCCCTGGG", "CCCCTGG", "NAACTGG", "NNNCTG")
    changeoclone <- makeChangeoClone(db_trim, id="SEQUENCE_ID", seq="SEQUENCE_IMGT", 
           germ="GERMLINE_IMGT_D_MASK", v_call="V_CALL", j_call="J_CALL", locus="LOCUS",
           junc_len="JUNCTION_LENGTH", clone="CLONE",text_fields="TYPE", 
           num_fields="COUNT", pad_end=TRUE)

    expect_true(inherits(clone, "ChangeoClone"))
    expect_equal(clone@clone, "1")
    expect_equal(clone@germline, "CCCCAGGGN")
    expect_equal(clone@v_gene, "IGKV1-39")
    expect_equal(clone@j_gene, "IGKJ5")
    expect_equal(clone@junc_len, 2)
    expect_equal(clone@data, exp_pad, tolerance=0.001)
    expect_equal(changeoclone@data, exp_pad, tolerance=0.001)

    # With end masking
    clone <- makeChangeoClone(db, max_mask=3, text_fields="TYPE", num_fields="COUNT")
    changeoclone <- makeChangeoClone(changeodb, max_mask=3, id="SEQUENCE_ID",
            seq="SEQUENCE_IMGT", germ="GERMLINE_IMGT_D_MASK", v_call="V_CALL", 
            j_call="J_CALL", locus="LOCUS", junc_len="JUNCTION_LENGTH", clone="CLONE", 
            text_fields="TYPE", num_fields="COUNT")

    exp <- data.frame("sequence_id"="A",
                      "sequence"=c("NNNCTGNN"),
                      "TYPE"=c("IgA,IgG,IgM"),
                      "COUNT"=c(10),
                      "collapse_count"=c(4),
                      stringsAsFactors=F)
    
    expect_true(inherits(clone, "ChangeoClone"))
    expect_equal(clone@clone, "1")
    expect_equal(clone@germline, "CCCCAGGG")
    expect_equal(clone@v_gene, "IGKV1-39")
    expect_equal(clone@j_gene, "IGKJ5")
    expect_equal(clone@junc_len, 2)
    expect_equal(clone@data, exp, tolerance=0.001)
    expect_equal(changeoclone@data, exp, tolerance=0.001)

    # Check what happens if locus unspecified
    db$locus[1] <- "IGK"
    expect_error(
            clone <- makeChangeoClone(db)
    )
    db$locus[1] <- NA
    expect_error(
            clone <- makeChangeoClone(db)
    )
    db <- dplyr::select(db, -locus)
    expect_error(
            clone <- makeChangeoClone(db)
    )
})

test_that("buildPhylipLineage", {
    # Preprocess clone
    clone <- subset(db, clone_id == 164)
    clone <- makeChangeoClone(clone, germ="germline_alignment_d_mask", 
                              text_fields=c("sample", "isotype"), 
                              num_fields="duplicate_count")
    
    # Run PHYLIP and process output
    
    # Test for error if phylip_exec doesn't exist
    phylip_exec <- "~/dummy/phylip-3.69/dnapars"
    if (file.access(phylip_exec, mode=1) == -1) {
        expect_error(
            graph <- buildPhylipLineage(clone, phylip_exec, rm_temp=TRUE),
            "The file ~/dummy/phylip-3.69/dnapars cannot be executed"
        )
    }
    
    phylip_exec <- Sys.which('dnapars')
    # If dnapars found, run test, else, skip
    if (phylip_exec!="") {
        graph <- buildPhylipLineage(clone, phylip_exec, rm_temp=TRUE)
        
        expect_true(inherits(graph, "igraph"))
        expect_equal(igraph::vcount(graph), 5)
        expect_equal(igraph::ecount(graph), 4)
        expect_true(igraph::is_directed(graph))
        
        expect_equal(igraph::graph_attr_names(graph),
                     c("clone", "v_gene", "j_gene", "junc_len"))
        
        expect_equal(igraph::graph_attr(graph),
                     list("clone"="164", "v_gene"="IGHV3-48","j_gene"="IGHJ2",
                       "junc_len"=66))
        
        expect_equal(V(graph)$name,
                     c("GN5SHBT08J0DZP", "GN5SHBT07FM4BU", "Germline",
                       "GN5SHBT07IAWQ9", "GN5SHBT07IBCBZ"))
        expect_equal(V(graph)$name, V(graph)$label)
        expect_equal(V(graph)$sequence,
                     c("GAGGTGCAGCTGGTGGAGTCTGGGGGANNNGGCTTGGTACAGCCTGGAGGGTCCCTGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTCNNNNNNNNNNNNAGTAGTTATGAAATGAACTGGGTCCGCCAGGCTCCAGGGAAGGGGCTGGAGTGGGTTTCATACATTAGTAGTAGTNNNNNNGGTAGTACCATATACTACGCAGACTCTGTGAAGNNNGGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGACACGGCTGTTTATTACTGTGCGAGAGATTTGGCGCTCGATAGTAGTGGTTATTACCCTAGCTACTGGTACTTCGATCTCTGGGGCCGTGGCACCCTGGTCACTGTCTCCTCAG",
                       "GAGGTGCAGCTGGTGGAGTCTGGGGGANNNGGCTTGGTACAGCCTGGAGGGTCCCCGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTCNNNNNNNNNNNNAGTAGTTATGAAATGAACTGGGTCCGCCAGGCTCCAGGGAAGGGGCTGGAGTGGGTTTCATACATTAGTAGTAGTNNNNNNGGTAGTACCATATACTACGCAGACTCTGTGAAGNNNGGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGACACGGCTGTTTATTACTGTGCGAGAGATTTGGCGCTCGATAGTAGTGGTTATTACCCTAGCTACTGGTACTTCGATCTCTGGGGCCGTGGCACCCTGGTCACTGTCTCCTCAG",
                       "GAGGTGCAGCTGGTGGAGTCTGGGGGANNNGGCTTGGTACAGCCTGGAGGGTCCCTGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTCNNNNNNNNNNNNAGTAGTTATGAAATGAACTGGGTCCGCCAGGCTCCAGGGAAGGGGCTGGAGTGGGTTTCATACATTAGTAGTAGTNNNNNNGGTAGTACCATATACTACGCAGACTCTGTGAAGNNNGGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGACACGGCTGTTTATTACTGTGCGAGAGANNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNCTACTGGTACTTCGATCTCTGGGGCCGTGGCACCCTGGTCACTGTCTCCTCAG",
                       "GAGGTGCAGCTGGTGGAGTCTGGGGGANNNGGCTTGGTACAGCCTGGAGGGTCCCCGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTCNNNNNNNNNNNNAGTAGTTATGAAATGAACTGGGTCCGCCAGGCTCCAGGGAAGGGGGTGGAGTGGGTTTCATACATTAGTAGTAGTNNNNNNGGTAGTACCATATACTACGCAGACTCTGTGAAGNNNGGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGACACGGCTGTTTATTACTGTGCGAGAGATTTGGCGCTCGATAGTAGTGGTTATTACCCTAGCTACTGGTACTTCGATCTCTGGGGCCGTGGCACCCTGGTCACTGTCTCCTCAG",
                       "GAGGTGCAGCTGGTGGAGTCTGGGGGANNNGGCTTGGTACAGCCTGGAGGGTCCCTGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTCNNNNNNNNNNNNAGTAGTTATGAAATGAACTGGGTCCGCCAGGCTCCAGGGAAGGGGCTGGAGTGGGTTTCATACATTAGTAGTAGTNNNNNNGGTAGTACCATATACTACGCAGACTCTGTGAAGNNNGGCCGATCCACCATCTCCAGAGACAACGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGACACGGCTGTTTATTACTGTGCGAGAGATTTGGCGCTCGATAGTAGTGGTTATTACCCTAGCTACTGGTACTTCGATCTCTGGGGCCGTGGCACCCTGGTCACTGTCTCCTCAG"
                     ))
        
        expect_equal(igraph::vertex_attr_names(graph),
                     c("name", "sequence", "sample", "isotype", "duplicate_count", 
                       "collapse_count", "label"))

        expect_equal(E(graph)$weight,c(1,1,1,0))
        expect_equal(E(graph)$label,c(1,1,1,0))
               
    }
 
})

# Common DNA, amino acid, and gene annotation operations for Alakazam

#### Distance functions ####

#' Build a DNA distance matrix
#'
#' \code{getDNAMatrix} returns a Hamming distance matrix for IUPAC ambiguous
#' DNA characters with modifications for gap, \code{c("-", ".")}, and missing, 
#' \code{c("?")}, character values.
#' 
#' @param    gap  value to assign to characters in the set \code{c("-", ".")}.
#' 
#' @return   A \code{matrix} of DNA character distances with row and column names 
#'           indicating the character pair. By default, distances will be either 0 
#'           (equivalent), 1 (non-equivalent or missing), or -1 (gap). 
#' 
#' @seealso  Creates DNA distance matrix for \link{seqDist}.
#'           See \link{getAAMatrix} for amino acid distances.
#' 
#' @examples
#' # Set gap characters to Inf distance
#' # Distinguishes gaps from Ns
#' getDNAMatrix()
#' 
#' # Set gap characters to 0 distance
#' # Makes gap characters equivalent to Ns
#' getDNAMatrix(gap=0)
#' 
#' @export
getDNAMatrix <- function(gap=-1) {
    # Define Hamming distance matrix
    sub_mat <- diag(18)
    colnames(sub_mat) <- rownames(sub_mat) <- c(names(IUPAC_DNA), c("-", ".", "?"))
    for (i in 1:length(IUPAC_DNA)) {
        for (j in i:length(IUPAC_DNA)) {
            sub_mat[i, j] <- sub_mat[j, i] <- any(IUPAC_DNA[[i]] %in% IUPAC_DNA[[j]])
        }
    }
    
    # Add gap characters
    sub_mat[c("-", "."), c("-", ".")] <- 1 
    sub_mat[c("-", "."), 1:15] <- 1 - gap 
    sub_mat[1:15, c("-", ".")] <- 1 - gap
    
    return(1 - sub_mat)
}


#' Build an AA distance matrix
#'
#' \code{getAAMatrix} returns a Hamming distance matrix for IUPAC ambiguous
#' amino acid characters.
#' 
#' @param    gap  value to assign to characters in the set \code{c("-", ".")}.
#' 
#' @return   A \code{matrix} of amino acid character distances with row and column names 
#'           indicating the character pair.
#' 
#' @seealso  Creates an amino acid distance matrix for \link{seqDist}.
#'           See \link{getDNAMatrix} for nucleotide distances.
#' 
#' @examples
#' getAAMatrix()
#' 
#' @export
getAAMatrix <- function(gap=0) {
  # Define Hamming distance matrix
  sub_mat <- diag(27)
  colnames(sub_mat) <- rownames(sub_mat) <- c(names(IUPAC_AA), c("-", "."))
  for (i in 1:length(IUPAC_AA)) {
    for (j in i:length(IUPAC_AA)) {
      sub_mat[i, j] <- sub_mat[j, i] <- any(IUPAC_AA[[i]] %in% IUPAC_AA[[j]])
    }
  }
  
  # Add gap characters
  sub_mat[c("-", "."), c("-", ".")] <- 1 
  sub_mat[c("-", "."), c(1:27)] <- 1 - gap 
  sub_mat[c(1:27), c("-", ".")] <- 1 - gap
  
  return(1 - sub_mat)
}


#' Remove duplicate DNA sequences and combine annotations
#' 
#' \code{collapseDuplicates} identifies duplicate DNA sequences, allowing for ambiguous 
#' characters, removes the duplicate entries, and combines any associated annotations.
#'
#' @param    data         data.frame containing Change-O columns. The data.frame 
#'                        must contain, at a minimum, a unique identifier column 
#'                        and a column containing a character vector of DNA sequences.
#' @param    id           name of the column containing sequence identifiers.
#' @param    seq          name of the column containing DNA sequences.
#' @param    text_fields  character vector of textual columns to collapse. The textual 
#'                        annotations of duplicate sequences will be merged into a single 
#'                        string with each unique value alphabetized and delimited by 
#'                        \code{sep}.
#' @param    num_fields   vector of numeric columns to collapse. The numeric annotations
#'                        of duplicate sequences will be summed. 
#' @param    seq_fields   vector of nucleotide sequence columns to collapse. The sequence 
#'                        with the fewest number of non-informative characters will be 
#'                        retained. Where a non-informative character is one of 
#'                        \code{c("N", "-", ".", "?")}. Note, this is distinct from the 
#'                        \code{seq} parameter which is used to determine duplicates.
#' @param    add_count    if \code{TRUE} add the column \code{collpase_count} that 
#'                        indicates the number of sequences that were collapsed to build 
#'                        each unique entry.
#' @param    ignore       vector of characters to ignore when testing for equality.
#' @param    sep          character to use for delimiting collapsed annotations in the 
#'                        \code{text_fields} columns. Defines both the input and output 
#'                        delimiter.
#' @param    dry          if \code{TRUE} perform dry run. Only labels the sequences without 
#'                        collapsing them.
#' @param    verbose      if \code{TRUE} report the number input, discarded and output 
#'                        sequences; if \code{FALSE} process sequences silently.
#'                        
#' @return   A modified \code{data} data.frame with duplicate sequences removed and 
#'           annotation fields collapsed if \code{dry=FALSE}. If \code{dry=TRUE}, 
#'           sequences will be labeled with the collapse action, but the input will be
#'           otherwise unmodified (see Details).
#'           
#' @details
#' \code{collapseDuplicates} identifies duplicate sequences in the \code{seq} column by
#' testing for character identity, with consideration of IUPAC ambiguous nucleotide codes. 
#' A cluster of sequences are considered duplicates if they are all equivalent, and no 
#' member of the cluster is equivalent to a sequence in a different cluster. 
#' 
#' Textual annotations, specified by \code{text_fields}, are collapsed by taking the unique
#' set of values within in each duplicate cluster and delimiting those values by \code{sep}.
#' Numeric annotations, specified by \code{num_fields}, are collapsed by summing all values 
#' in the duplicate cluster. Sequence annotations, specified by \code{seq_fields}, are 
#' collapsed by retaining the first sequence with the fewest number of N characters.
#' 
#' Columns that are not specified in either \code{text_fields}, \code{num_fields}, or 
#' \code{seq_fields} will be retained, but the value will be chosen from a random entry 
#' amongst all sequences in a cluster of duplicates.
#' 
#' An ambiguous sequence is one that can be assigned to two different clusters, wherein
#' the ambiguous sequence is equivalent to two sequences which are themselves 
#' non-equivalent. Ambiguous sequences arise due to ambiguous characters at positions that
#' vary across sequences, and are discarded along with their annotations when \code{dry=FALSE}. 
#' Thus, ambiguous sequences are removed as duplicates of some sequence, but do not create a potential
#' false-positive annotation merger. Ambiguous sequences are not included in the 
#' \code{collapse_count} annotation that is added when \code{add_count=TRUE}.
#' 
#' If \code{dry=TRUE} sequences will not be removed from the input. Instead, the following columns
#' will be appended to the input defining the collapse action that would have been performed in the
#' \code{dry=FALSE} case.
#' 
#' \itemize{
#'   \item  \code{collapse_id}:     an identifier for the group of identical sequences.
#'   \item  \code{collapse_class}:  string defining how the sequence matches to the other in the set.
#'                                  one of \code{"duplicated"} (has duplicates),
#'                                  \code{"unique"} (no duplicates), \code{"ambiguous_duplicate"} 
#'                                  (no duplicates after ambiguous sequences are removed), 
#'                                  or \code{"ambiguous"} (matches multiple non-duplicate sequences).
#'   \item  \code{collapse_pass}:   \code{TRUE} for the sequences that would be retained.
#' }
#' 
#' @seealso  Equality is tested with \link{seqEqual} and \link{pairwiseEqual}. 
#'           For IUPAC ambiguous character codes see \link{IUPAC_DNA}.
#'
#' @examples
#' # Example data.frame
#' db <- data.frame(sequence_id=LETTERS[1:4],
#'                  sequence_alignment=c("CCCCTGGG", "CCCCTGGN", "NAACTGGN", "NNNCTGNN"),
#'                  c_call=c("IGHM", "IGHG", "IGHG", "IGHA"),
#'                  sample_id=c("S1", "S1", "S2", "S2"),
#'                  duplicate_count=1:4,
#'                  stringsAsFactors=FALSE)
#' 
#' # Annotations are not parsed if neither text_fields nor num_fields is specified
#' # The retained sequence annotations will be random
#' collapseDuplicates(db, verbose=TRUE)
#' 
#' # Unique text_fields annotations are combined into a single string with ","
#' # num_fields annotations are summed
#' # Ambiguous duplicates are discarded
#' collapseDuplicates(db, text_fields=c("c_call", "sample_id"), num_fields="duplicate_count", 
#'                    verbose=TRUE)
#'
#' # Use alternate delimiter for collapsing textual annotations
#' collapseDuplicates(db, text_fields=c("c_call", "sample_id"), num_fields="duplicate_count", 
#'                    sep="/", verbose=TRUE)
#' 
#' # Add count of duplicates
#' collapseDuplicates(db, text_fields=c("c_call", "sample_id"), num_fields="duplicate_count", 
#'                    add_count=TRUE, verbose=TRUE)
#' 
#' # Masking ragged ends may impact duplicate removal
#' db$sequence_alignment <- maskSeqEnds(db$sequence_alignment)
#' collapseDuplicates(db, text_fields=c("c_call", "sample_id"), num_fields="duplicate_count", 
#'                    add_count=TRUE, verbose=TRUE)
#'
#' @export
collapseDuplicates <- function(data, id="sequence_id", seq="sequence_alignment",
                               text_fields=NULL, num_fields=NULL, seq_fields=NULL,
                               add_count=FALSE, ignore=c("N", "-", ".", "?"), 
                               sep=",", dry=FALSE, verbose=FALSE) {
    # Stop if ids are not unique
    if (any(duplicated(data[[id]]))) {
        stop("All values in the id column are not unique")
    }
    # Verify column classes and exit if they are incorrect
    if (!is.null(text_fields)) {
        if (!all(sapply(subset(data, select=text_fields), is.character))) {
            stop("All text_fields columns must be of type 'character'")
        }
    }
    if (!is.null(num_fields)) {
        if (!all(sapply(subset(data, select=num_fields), is.numeric))) {
            stop("All num_fields columns must be of type 'numeric'")
        }
    }
    if (!is.null(seq_fields)) {
        if (!all(sapply(subset(data, select=seq_fields), is.character))) {
            stop("All seq_fields columns must be of type 'character'")
        }
    }
    seq_len <- stri_length(data[[seq]])
    if (any(seq_len != seq_len[1])) {
        warning("All sequences are not the same length for data with first ", 
                id, " = ", data[[id]][1])
    }
    
    # Define verbose reporting function
    .printVerbose <- function(n_total, n_unique, n_discard) {
        cat(" FUNCTION> collapseDuplicates\n", sep="")
        cat(" FIRST_ID> ", data[[id]][1], "\n", sep="")
        cat("    TOTAL> ", n_total, "\n", sep="")
        cat("   UNIQUE> ", n_unique, "\n", sep="")
        cat("COLLAPSED> ", n_total - n_unique - n_discard, "\n", sep="")
        cat("DISCARDED> ", n_discard, "\n", sep="")
        cat("\n")
    }
    
    # Define function to count informative positions in sequences
    .informativeLength <- function(x) {
        stri_length(gsub("[N\\-\\.\\?]", "", x, perl=TRUE))
    }
    
    # Initialize collapse_count with 1 for each sequence
    if(add_count) {
        data[["collapse_count"]] <- rep(1, nrow(data))
        num_fields <- c(num_fields, "collapse_count")
    }
    
    # Initialize dry run columns 
    if (dry) {
        data$collapse_id <- NA
        data$collapse_class <- NA
        data$collapse_pass <- TRUE
    }
    
    # Return input if there are no sequences to collapse
    nseq <- nrow(data)
    if (nseq <= 1) { 
        if (verbose) { .printVerbose(nseq, 1, 0) }
        if (dry) {
            data[['collapse_id']] <- 1
            data[['collapse_class']] <- "unique"
            data[['collapse_pass']] <- TRUE            
        }
        return(data)
    }
    
    # Build distance matrix
    exact_duplicates <- any(duplicated(data[[seq]]))
    d_mat <- pairwiseEqual(unique(data[[seq]]))
    colnames(d_mat) <- rownames(d_mat) <- unique(data[[seq]])
    n_uniqueseq <- nrow(d_mat)
    
    # Return input if no sequences are equal
    if (!any(d_mat[lower.tri(d_mat, diag=F)]) & !exact_duplicates) {
        if (verbose) { .printVerbose(nseq, nseq, 0) }
        if (dry) {
            data[['collapse_id']] <- 1:nrow(data)
            data[['collapse_class']] <- "unique"
            data[['collapse_pass']] <- TRUE            
        }
        return(data)
    }        
    
    # Find sequences that will cluster ambiguously
    ambig_rows <- numeric()
    for (i in 1:n_uniqueseq) {
        idx <- which(d_mat[i, ])
        tmp_mat <- d_mat[idx, idx]
        if (!all(tmp_mat)) { 
            ambig_rows <- append(ambig_rows, i) 
        }
    }
    discard_count <- length(ambig_rows)
    
    # from ambiguous rows in d_mat to 
    # ambiguous rows in data
    data_ambig_rows <- data[[seq]] %in% rownames(d_mat)[ambig_rows]
    data_discard_count <- sum(data_ambig_rows)
    
    if (dry & length(ambig_rows)>0) {
        data[["collapse_class"]][data_ambig_rows] <- "ambiguous"
        data[["collapse_pass"]][data_ambig_rows] <- FALSE
    }
    
    # Return single sequence if all or all but one sequence belong to ambiguous clusters 
	if (nrow(data) - data_discard_count <= 1) {
        inform_len <- data.frame(list("inform_len"=.informativeLength(data[[seq]])))
        # For each ambiguous cluster, return the best sequence
        g <- igraph::simplify(igraph::graph_from_adjacency_matrix(d_mat))
        inform_len$clusters <- igraph::components(g)$membership[data[[seq]]]
        inform_len$select_id <- 1:nrow(inform_len)
        selected <- inform_len %>%
            dplyr::group_by(!!rlang::sym("clusters")) %>%
            dplyr::slice(which.max(!!rlang::sym("inform_len"))) %>%
            dplyr::ungroup() %>%
            dplyr::select(!!rlang::sym("select_id")) %>% unlist() 
        
        if (verbose) { .printVerbose(nseq, 0, discard_count - 1) }
        if (dry) {
            data[["collapse_id"]] <- inform_len$clusters
            data[["collapse_pass"]][selected] <- TRUE
        } else {
            return(data[selected, ])
        }
    }
    
    # Exclude ambiguous sequences from clustering
    if (!dry & discard_count > 0) {
            d_mat <- d_mat[-ambig_rows, -ambig_rows, drop = FALSE]  # 'drop = FALSE' to keep dataframe structure if only one sequence is left 
            data <- data[!data_ambig_rows,]
    }
    
    # Cluster remaining sequences into unique and duplicate sets
    dup_taxa <-  list()
    uniq_taxa <- character()
    done_taxa <- character()
    taxa_names <- rownames(d_mat)
    collapse_id <- 1
    
    for (taxa_i in 1:length(taxa_names)) {
        
        taxa <- taxa_names[taxa_i]
        data_taxa_i <- which(data[[seq]] %in% taxa)
        
        # Skip taxa if previously assigned to a cluster
        # or if ambiguous
        # (ambiguous taxa don't get their own collapse_id)
        if (taxa %in% done_taxa) { next }
        if (dry & taxa_i %in% ambig_rows) { next }
        
        # Find all zero distance taxa
        idx <- which(d_mat[taxa, ])
        # Translate from d_mat idx to data idx
        data_idx <- which(data[[seq]] %in% colnames(d_mat)[idx])
        
        # Update vector of clustered taxa
        done_taxa <- c(done_taxa, taxa_names[idx])
        
        # Update collapse group
        if (dry) {
            data[["collapse_id"]][data_idx] <- paste(data[["collapse_id"]][data_idx], collapse_id, sep=",")
        }
        
        if (dry) {
            #idx_copy <- idx
            data_idx_copy <- data_idx
            idx <- idx[idx %in% ambig_rows == FALSE]
            data_idx <- which(data[[seq]] %in% colnames(d_mat)[idx])
        }
        
        if (length(data_idx) == 1) {
            # Assign unique sequences to unique vector
            uniq_taxa <- append(uniq_taxa, taxa_names[idx])
            if (dry) {
                if (length(data_idx_copy)==1) {
                    ## 'truly' unique
                    data[["collapse_class"]][data_taxa_i] <- "unique"    
                } else {
                    ## unique after ambiguous removal
                    data[["collapse_class"]][data_taxa_i] <- "ambiguous_duplicate"
                }
                data[["collapse_pass"]][data_taxa_i] <- TRUE
            }
        } else if (length(data_idx) > 1) {
            # Assign clusters of duplicates to duplicate list            
            dup_taxa <- c(dup_taxa, list(taxa_names[idx]))    
            if (dry) {
                # Keep collpase_pass==TRUE for the sequence with the
                # larger number of informative positions 
                # (the first one if ties)
                max_info_idx <- which.max(.informativeLength(data[[seq]][data_idx]))[1]
                data[["collapse_class"]][data_idx] <- "duplicated"
                data[["collapse_pass"]][data_idx[-max_info_idx]] <- FALSE
            }
        } else {
            # Report error (should never occur)
                stop("Error in distance matrix of collapseDuplicates")
        }
        
        collapse_id <- collapse_id + 1
    }
   
    if (dry) {
        data[["collapse_id"]] <-  sub("^NA,","",data[["collapse_id"]])
        return(data)
    }
    
    # Collapse duplicate sets and append entries to unique data.frame
    unique_list <- list(data[data[[seq]] %in% uniq_taxa, ])
    for (taxa in dup_taxa) {
        # Define row indices of identical sequences
        idx <- which(data[[seq]] %in% taxa)
        tmp_df <- data[idx[1], ]

        if (length(idx) > 1) {

            # Initialize with data from most informative sequence
            seq_set <- data[idx, c(id, seq)]
            inform_len <- .informativeLength(seq_set[[seq]])
            max_inform <- which.max(inform_len)[1] # if ties, pick first
            tmp_df <- data[idx[max_inform], ]

            # Define set of text fields for row
            for (f in text_fields) {
                f_set <- na.omit(data[[f]][idx])
                if (length(f_set) > 0) {
                    f_set <- unlist(strsplit(f_set, sep))
                    f_set <- sort(unique(f_set))
                    f_val <- paste(f_set, collapse=sep)
                } else {
                    f_val <- NA
                }
                tmp_df[, f] <- f_val
            }

            # Sum numeric fields
            for (f in num_fields) {
                f_set <- na.omit(data[[f]][idx])
                if (length(f_set) > 0) {
                    f_val <- sum(f_set)
                } else {
                    f_val <- NA
                }
                tmp_df[, f] <- f_val
            }

            # Select sequence fields with fewest Ns
            for (f in seq_fields) {
                f_set <- na.omit(data[[f]][idx])
                if (length(f_set) > 0) {
                    f_len <- .informativeLength(f_set)
                    f_val <- f_set[which.max(f_len)]
                } else {
                    f_val <- NA
                }
                tmp_df[, f] <- f_val
            }

        }

        # Add row to unique list
        unique_list <- c(unique_list, list(tmp_df))
    }

    # Combine all rows into unique data.frame
    unique_df <- as.data.frame(bind_rows(unique_list))
    
    if (verbose) { .printVerbose(nseq, nrow(unique_df), discard_count) }
    return(unique_df)
}


#### Transformation functions ####

#' Translate nucleotide sequences to amino acids
#' 
#' \code{translateDNA} translates nucleotide sequences to amino acid sequences.
#' 
#' @param   seq     vector of strings defining DNA sequence(s) to be converted to translated.
#' @param   trim    boolean flag to remove 3 nts from both ends of seq
#'          (converts IMGT junction to CDR3 region).
#' 
#' @return  A vector of translated sequence strings.
#' 
#' @seealso  \code{\link[seqinr]{translate}}.
#' 
#' @examples
#' # Translate a single sequence
#' translateDNA("ACTGACTCGA")
#'
#' # Translate a vector of sequences
#' translateDNA(ExampleDb$junction[1:3])
#' 
#' # Remove the first and last codon from the translation
#' translateDNA(ExampleDb$junction[1:3], trim=TRUE)
#' 
#' @export
translateDNA <- function (seq, trim=FALSE) {
  # Function to translate a single string
  .translate <- function(x) {
    if (stri_length(x) >= 3 & !is.na(x)) {
      stri_join(seqinr::translate(unlist(strsplit(x, "")), ambiguous=TRUE), 
                collapse="")
    } else {
      NA
    }
  }
  
  # Remove 3 nucleotides from each end
  # Eg,  "ACTGACTCGA" -> "GACT" (with "ACT" and "CGA" removed)
  if (trim) { seq <- substr(seq, 4, stri_length(seq) - 3) }
  
  # Replace gaps with N
  seq <- gsub("[-.]", "N", seq)
  
  # Apply translation
  aa <- sapply(seq, .translate, USE.NAMES=FALSE)
  
  return(aa)
}


#' Masks gap characters in DNA sequences
#' 
#' \code{maskSeqGaps} substitutes gap characters, \code{c("-", ".")}, with \code{"N"} 
#' in a vector of DNA sequences.
#'
#' @param    seq         character vector of DNA sequence strings.
#' @param    mask_char   character to use for masking.
#' @param    outer_only  if \code{TRUE} replace only contiguous leading and trailing gaps;
#'                       if \code{FALSE} replace all gap characters.
#'                       
#' @return   A modified \code{seq} vector with \code{"N"} in place of \code{c("-", ".")} 
#'           characters.
#' 
#' @seealso  See \link{maskSeqEnds} for masking ragged edges.
#'           
#' @examples
#' # Mask with Ns
#' maskSeqGaps(c("ATG-C", "CC..C"))
#' maskSeqGaps("--ATG-C-")
#' maskSeqGaps("--ATG-C-", outer_only=TRUE)
#' 
#' # Mask with dashes
#' maskSeqGaps(c("ATG-C", "CC..C"), mask_char="-")
#' 
#' @export
maskSeqGaps <- function(seq, mask_char="N", outer_only=FALSE) {
  if (outer_only) {
    for (i in 1:length(seq)) {
      head_match <- attr(regexpr("^[-\\.]+", seq[i]), "match.length")
      tail_match <- attr(regexpr("[-\\.]+$", seq[i]), "match.length")
      if (head_match > 0) { 
        seq[i] <- gsub("^[-\\.]+", 
                       paste(rep(mask_char, head_match), collapse=""), 
                       seq[i]) 
      }
      if (tail_match > 0) { 
        seq[i] <- gsub("[-\\.]+$", 
                       paste(rep(mask_char, tail_match), collapse=""), 
                       seq[i]) 
      }
    }
  } else {
    seq <- gsub("[-\\.]", mask_char, seq)
  }
  
  return(seq)
}


#' Masks ragged leading and trailing edges of aligned DNA sequences
#' 
#' \code{maskSeqEnds} takes a vector of DNA sequences, as character strings,
#' and replaces the leading and trailing characters with \code{"N"} characters to create 
#' a sequence vector with uniformly masked outer sequence segments.
#' 
#' @param    seq        character vector of DNA sequence strings.
#' @param    mask_char  character to use for masking.
#' @param    max_mask   the maximum number of characters to mask. If set to 0 then
#'                      no masking will be performed. If set to \code{NULL} then the upper 
#'                      masking bound will be automatically determined from the maximum 
#'                      number of observed leading or trailing \code{"N"} characters amongst 
#'                      all strings in \code{seq}. 
#' @param    trim       if \code{TRUE} leading and trailing characters will be cut rather 
#'                      than masked with \code{"N"} characters.
#' @return   A modified \code{seq} vector with masked (or optionally trimmed) sequences.
#' 
#' @seealso  See \link{maskSeqGaps} for masking internal gaps.
#'           See \link{padSeqEnds} for padding sequence of unequal length.
#' 
#' @examples
#' # Default behavior uniformly masks ragged ends
#' seq <- c("CCCCTGGG", "NAACTGGN", "NNNCTGNN")
#' maskSeqEnds(seq)
#'
#' # Does nothing
#' maskSeqEnds(seq, max_mask=0)
#' 
#' # Cut ragged sequence ends
#' maskSeqEnds(seq, trim=TRUE)
#'
#' # Set max_mask to limit extent of masking and trimming
#' maskSeqEnds(seq, max_mask=1)
#' maskSeqEnds(seq, max_mask=1, trim=TRUE)
#' 
#' # Mask dashes instead of Ns
#' seq <- c("CCCCTGGG", "-AACTGG-", "---CTG--")
#' maskSeqEnds(seq, mask_char="-")
#' 
#' @export
maskSeqEnds <- function(seq, mask_char="N", max_mask=NULL, trim=FALSE) {
  # Find length of leading and trailing Ns
  left_lengths <- attr(regexpr(paste0("(^", mask_char, "*)"), seq, perl=T), "capture.length")
  right_lengths <- attr(regexpr(paste0("(", mask_char, "*$)"), seq, perl=T), "capture.length")
  
  # Mask to minimal inner sequence length
  left_mask <- min(max(left_lengths[, 1]), max_mask)
  right_mask <- min(max(right_lengths[, 1]), max_mask)
  seq_lengths <- stri_length(seq)
  if (trim) {
    seq <- substr(seq, left_mask + 1, seq_lengths - right_mask)
  } else {
    substr(seq, 0, left_mask) <- paste(rep(mask_char, left_mask), collapse='')
    substr(seq, seq_lengths - right_mask + 1, seq_lengths + 1) <- 
      paste(rep(mask_char, right_mask), collapse='')
  }
  
  return(seq)
}


#' Pads ragged ends of aligned DNA sequences
#' 
#' \code{padSeqEnds} takes a vector of DNA sequences, as character strings,
#' and appends the ends of each sequence with an appropriate number of \code{"N"} 
#' characters to create a sequence vector with uniform lengths.
#' 
#' @param    seq       character vector of DNA sequence strings.
#' @param    len       length to pad to. Only applies if longer than the maximum length of
#'                     the data in \code{seq}.
#' @param    start     if \code{TRUE} pad the beginning of each sequence instead of the end. 
#' @param    pad_char  character to use for padding.
#' @param    mod3      if \code{TRUE} pad sequences to be of length multiple three.
#' 
#' @return   A modified \code{seq} vector with padded sequences.
#' 
#' @seealso  See \link{maskSeqEnds} for creating uniform masking from existing masking.
#' 
#' @examples
#' # Default behavior uniformly pads ragged ends
#' seq <- c("CCCCTGGG", "ACCCTG", "CCCC")
#' padSeqEnds(seq)
#'
#' # Pad to fixed length
#' padSeqEnds(seq, len=15)
#'
#' # Add padding to the beginning of the sequences instead of the ends
#' padSeqEnds(seq, start=TRUE)
#' padSeqEnds(seq, len=15, start=TRUE)
#' 
#' @export
padSeqEnds <- function(seq, len=NULL, start=FALSE, pad_char="N", mod3=TRUE) {
  # Set length to max input length
  width <- max(stringi::stri_length(seq),len)
  if (mod3 && width %% 3 != 0) {
    width <- width + (3 - width %% 3)
  }
  
  # Pad
  if (!start) { 
    seq <- stringi::stri_pad_right(seq, width=width, pad=pad_char)
  } else {
    seq <- stringi::stri_pad_left(seq, width=width, pad=pad_char)
  }
  
  return(seq)
}


#### Subregion functions ####

#' Extracts FWRs and CDRs from IMGT-gapped sequences
#' 
#' \code{extractVRegion} extracts the framework and complementarity determining regions of 
#' the V segment for IMGT-gapped immunoglobulin (Ig) nucleotide sequences according to the 
#' IMGT numbering scheme.
#'
#' @param     sequences  character vector of IMGT-gapped nucleotide sequences.
#' @param     region     string defining the region(s) of the V segment to extract. 
#'                       May be a single region or multiple regions (as a vector) from
#'                       \code{c("fwr1", "cdr1", "fwr2", "cdr2" ,"fwr3")}.  By default, all
#'                       regions will be returned.
#'                       
#' @return    If only one region is specified in the \code{region} argument, a character 
#'            vector of the extracted sub-sequences will be returned. If multiple regions 
#'            are specified, then a character matrix will be returned with columns 
#'            corresponding to the specified regions and a row for each entry in 
#'            \code{sequences}.
#' 
#' @seealso   IMGT-gapped region boundaries are defined in \link{IMGT_REGIONS}.
#' 
#' @references
#' \enumerate{
#'   \item  Lefranc M-P, et al. IMGT unique numbering for immunoglobulin and T cell 
#'            receptor variable domains and Ig superfamily V-like domains.
#'            Dev Comp Immunol. 2003 27(1):55-77.
#' }
#' 
#' @examples
#' # Assign example clone
#' clone <- subset(ExampleDb, clone_id == 3138)
#'
#' # Get all regions
#' extractVRegion(clone$sequence_alignment)
#' 
#' # Get single region
#' extractVRegion(clone$sequence_alignment, "fwr1")
#' 
#' # Get all CDRs
#' extractVRegion(clone$sequence_alignment, c("cdr1", "cdr2"))
#' 
#' # Get all FWRs
#' extractVRegion(clone$sequence_alignment, c("fwr1", "fwr2", "fwr3"))
#'
#' @export
extractVRegion <- function(sequences, region=c("fwr1", "cdr1", "fwr2", "cdr2", "fwr3")) {
    # Check region argument
    region <- match.arg(region, several.ok=TRUE)
    
    if (length(region) == 1) {
        sub_sequences <- substr(sequences, 
                                IMGT_REGIONS[[region]][1], 
                                IMGT_REGIONS[[region]][2])
    } else {
        sub_sequences <- sapply(region, function(x) substr(sequences, 
                                                           IMGT_REGIONS[[x]][1], 
                                                           IMGT_REGIONS[[x]][2]))
    }
    
    return(sub_sequences)
}


#' Calculate junction region alignment properties
#'
#' \code{junctionAlignment} determines the number of deleted germline nucleotides in the 
#' junction region and the number of V gene and J gene nucleotides in the CDR3.
#'
#' @param   data                \code{data.frame} containing sequence data.
#' @param   germline_db         reference germline database for the V, D and J genes.
#'                              in \code{data}
#' @param   v_call              V gene assignment column.
#' @param   d_call              D gene assignment column.
#' @param   j_call              J gene assignment column.
#' @param   v_germline_start    column containing the start position of the alignment 
#'                              in the V reference germline.
#' @param   v_germline_end      column containing the end position of the alignment in the 
#'                              V reference germline.
#' @param   d_germline_start    column containing the start position of the alignment 
#'                              in the D reference germline.
#' @param   d_germline_end      column containing the start position of the alignment 
#'                              in the D reference germline.
#' @param   j_germline_start    column containing the start position of the alignment 
#'                              in the J reference germline.
#' @param   j_germline_end      column containing the start position of the alignment 
#'                              in the J reference germline.
#' @param   np1_length          combined length of the N and P regions between the 
#'                              V and D regions (heavy chain) or V and J regions (light chain).      
#' @param   np2_length          combined length of the N and P regions between the 
#'                              D and J regions (heavy chain).            
#' @param   junction            column containing the junction sequence.
#' @param   junction_length     column containing the length of the junction region in nucleotides.
#' @param   sequence_alignment  column containing the aligned sequence.
#' 
#' @return  A modified input \code{data.frame} with the following additional columns storing 
#'          junction alignment information:
#'          \enumerate{
#'              \item  \code{e3v_length}:     number of 3' V germline nucleotides deleted.
#'              \item  \code{e5d_length}:     number of 5' D germline nucleotides deleted.
#'              \item  \code{e3d_length}:     number of 3' D germline nucleotides deleted.
#'              \item  \code{e5j_length}:     number of 5' J germline nucleotides deleted.
#'              \item  \code{v_cdr3_length}:  number of sequence_alignment V nucleotides in the CDR3.
#'              \item  \code{j_cdr3_length}:  number of sequence_alignment J nucleotides in the CDR3.
#'          }
#' 
#' @examples
#' germline_db <- list(
#' "IGHV3-11*05"="CAGGTGCAGCTGGTGGAGTCTGGGGGA...GGCTTGGTCAAGCCTGGAGGGTCCCTGAGACT
#' CTCCTGTGCAGCCTCTGGATTCACCTTC............AGTGACTACTACATGAGCTGGATCCGCCAGGCTCCAG
#' GGAAGGGGCTGGAGTGGGTTTCATACATTAGTAGTAGT......AGTAGTTACACAAACTACGCAGACTCTGTGAAG
#' ...GGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGA
#' CACGGCCGTGTATTACTGTGCGAGAGA",
#' "IGHD3-10*01"="GTATTACTATGGTTCGGGGAGTTATTATAAC",
#' "IGHJ5*02"="ACAACTGGTTCGACCCCTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAG"
#' )
#' 
#' db <- junctionAlignment(SingleDb, germline_db)
#'
#' @export
junctionAlignment <- function(data, germline_db, 
                              v_call="v_call",
                              d_call="d_call",
                              j_call="j_call",
                              v_germline_start="v_germline_start",
                              v_germline_end="v_germline_end",
                              d_germline_start="d_germline_start",
                              d_germline_end="d_germline_end",
                              j_germline_start="j_germline_start",
                              j_germline_end="j_germline_end",
                              np1_length="np1_length",
                              np2_length="np2_length",
                              junction="junction",
                              junction_length="junction_length",
                              sequence_alignment="sequence_alignment") {
  
  # Check input
  check <- checkColumns(data, 
                        c(v_call, d_call, j_call, 
                          v_germline_start, v_germline_end,
                          d_germline_start, d_germline_end,
                          j_germline_start, j_germline_end,
                          np1_length, np2_length,
                          junction, junction_length,
                          sequence_alignment))
  if (check != TRUE) { stop(check) }
  
  # Get deletions
  for (i in 1:nrow(data))  {
    v_dels <- countDeleted(data[i,], 
                           allele_call=v_call, germline_start=v_germline_start, germline_end=v_germline_end, 
                           germline_db=germline_db, junction=junction, junction_length=junction_length, 
                           sequence_alignment=sequence_alignment)
    d_dels <- countDeleted(data[i,], 
                           allele_call=d_call, germline_start=d_germline_start, germline_end=d_germline_end, 
                           germline_db=germline_db, junction=junction, junction_length=junction_length, 
                           sequence_alignment=sequence_alignment)
    j_dels <- countDeleted(data[i,], 
                           allele_call=j_call, germline_start=j_germline_start, germline_end=j_germline_end, 
                           germline_db=germline_db, junction=junction, junction_length=junction_length, 
                           sequence_alignment=sequence_alignment)
    data[['e3v_length']][i] <- v_dels[2]
    data[['e5d_length']][i] <- d_dels[1]
    data[['e3d_length']][i] <- d_dels[2]
    data[['e5j_length']][i] <- j_dels[1]
    data[['v_cdr3_length']][i] <- v_dels[3]
    data[['j_cdr3_length']][i] <- j_dels[3]
  }
  
  return(data)
}

# Junction alignment helper
#
# Report the number of deleted germline nucleotides in the alignment
#
# @param    db_row               one row from a Rearrangement database.
# @param    allele_call          column containing gene assignments.
# @param    germline_start       column containing the start position of the alignment in the reference germline.
# @param    germline_end         column containing the end position of the alignment in the reference germline.
# @param    germline_db          reference germline database for the V, D and J genes.
# @param    junction             column containing the junction sequence.
# @param    junction_length      column containing the length of the  junction region in nucleotides.
# @param    sequence_alignment   column containing the aligned sequence.
# 
# @return   Alignment deletions
countDeleted <- function(db_row, allele_call, germline_start, germline_end, 
                         germline_db, junction, junction_length,
                         sequence_alignment) {
    # db_row: one row from data
    # allele_call: one of v,d,j
    # germline_db: the reference germline database used to assign genes. 
    allele <- getAllele(db_row[[allele_call]], first=T)
    deleted <- c(NA, NA, NA)
    
    # Check for valid allele information
    if (is.na(allele)) { 
        return(deleted) 
    }
    # Check for allele in reference germlines
    tryCatch(germline <- germline_db[[allele]],
             error=function(e) { stop(allele, " not found in germline_db.") })
    
    allele_germline_start <- as.numeric(db_row[[germline_start]])
    allele_germline_end <- as.numeric(db_row[[germline_end]])
    
    germline_head <- stringi::stri_sub(germline, 1, allele_germline_start - 1)
    deleted_head <- nchar(gsub("\\.", "", germline_head))
    
    germline_tail <- stringi::stri_sub(germline, allele_germline_end+1, nchar(germline))
    deleted_tail <- nchar(gsub("\\.", "", germline_tail))
    
    deleted[1] <- deleted_head
    deleted[2] <- deleted_tail
    
    if (is.na(db_row[[junction]])) {
        warning("NA junction found.")
        return (deleted)
    }
    if (!db_row[[junction_length]]>6) {
        message("Junction length <= 6.")
        return (deleted)
    }
    
    junction_len <- db_row[[junction_length]]
    junction_start <- 310
    # junction_end <- junction_start + junction_len - 1
    
    # get aligned junction end (counting gaps)
    seq_aln <- s2c(db_row[[sequence_alignment]]) != "-"
    seq_aln[1:junction_start-1] <- 0
    junction_end <- which(cumsum(seq_aln[1:length(seq_aln)]) > junction_len)[1] - 1
    
    # For V and J alleles, calculate number of nt in the CDR3
    germ_cdr3_length <- NA
    if (grepl("[Vv]", allele)) {
        last_cdr3_pre_np <- db_row[[germline_end]] - db_row[[germline_start]] + 1 
        first_cdr3_pre_np <- junction_start + 3   # without conserved 
        # len <- last_cdr3_pre_np - first_cdr3_pre_np + 1
        #germ_seq <- stringi::stri_sub(germline, db_row[[germline_end]]+1-len, db_row[[germline_end]] )
        germ_seq <- stringi::stri_sub(db_row[[sequence_alignment]], first_cdr3_pre_np, last_cdr3_pre_np )
        germ_cdr3_length <- nchar(gsub("[\\.-]", "", germ_seq))
    } else if (grepl("[Jj]", allele))  {
        j_aln_len <- db_row[[germline_end]] - db_row[[germline_start]] + 1 
        # germ_seq <- stringi::stri_sub(germline, db_row[[germline_start]], db_row[[germline_end]]-j_tail)
        germ_seq <- stringi::stri_sub(db_row[[sequence_alignment]], 
                                      nchar(db_row[[sequence_alignment]]) - j_aln_len + 1,
                                      junction_end - 3)
        germ_cdr3_length <- nchar(gsub("-", "", germ_seq))
    } 
    
    deleted <- c(deleted_head, deleted_tail, germ_cdr3_length)
    return(deleted)
}


#### Rcpp wrappers ####

#' Calculate distance between two sequences
#' 
#' \code{seqDist} calculates the distance between two DNA sequences.
#'
#' @param    seq1      character string containing a DNA sequence.
#' @param    seq2      character string containing a DNA sequence.
#' @param    dist_mat  Character distance matrix. Defaults to a Hamming distance 
#'                     matrix returned by \link{getDNAMatrix}. If gap 
#'                     characters, \code{c("-", ".")}, are assigned a value of -1 
#'                     in \code{dist_mat} then contiguous gaps of any run length,
#'                     which are not present in both sequences, will be counted as a 
#'                     distance of 1. Meaning, indels of any length will increase
#'                     the sequence distance by 1. Gap values other than -1 will 
#'                     return a distance that does not consider indels as a special case.
#'
#' @return   Numerical distance between \code{seq1} and \code{seq2}.
#' 
#' @seealso  Nucleotide distance matrix may be built with 
#'           \link{getDNAMatrix}. Amino acid distance matrix may be built
#'           with \link{getAAMatrix}. Used by \link{pairwiseDist} for generating
#'           distance matrices. See \link{seqEqual} for testing sequence equivalence.
#'           
#' @examples
#' # Ungapped examples
#' seqDist("ATGGC", "ATGGG")
#' seqDist("ATGGC", "ATG??")
#' 
#' # Gaps will be treated as Ns with a gap=0 distance matrix
#' seqDist("ATGGC", "AT--C", dist_mat=getDNAMatrix(gap=0))
#' 
#' # Gaps will be treated as universally non-matching characters with gap=1
#' seqDist("ATGGC", "AT--C", dist_mat=getDNAMatrix(gap=1))
#' 
#' # Gaps of any length will be treated as single mismatches with a gap=-1 distance matrix
#' seqDist("ATGGC", "AT--C", dist_mat=getDNAMatrix(gap=-1))
#' 
#' # Gaps of equivalent run lengths are not counted as gaps
#' seqDist("ATG-C", "ATG-C", dist_mat=getDNAMatrix(gap=-1))
#'
#' # Overlapping runs of gap characters are counted as a single gap
#' seqDist("ATG-C", "AT--C", dist_mat=getDNAMatrix(gap=-1))
#' seqDist("A-GGC", "AT--C", dist_mat=getDNAMatrix(gap=-1))
#' seqDist("AT--C", "AT--C", dist_mat=getDNAMatrix(gap=-1))
#' 
#' # Discontiguous runs of gap characters each count as separate gaps
#' seqDist("-TGGC", "AT--C", dist_mat=getDNAMatrix(gap=-1))
#' 
#' @export
seqDist <- function(seq1, seq2, dist_mat=getDNAMatrix()) {
    seqDistRcpp(seq1, seq2, dist_mat)
}


#' Calculate pairwise distances between sequences
#' 
#' \code{pairwiseDist} calculates all pairwise distance between a set of sequences.
#'
#' @param    seq       character vector containing a DNA sequences.
#' @param    dist_mat  Character distance matrix. Defaults to a Hamming distance 
#'                     matrix returned by \link{getDNAMatrix}. If gap 
#'                     characters, \code{c("-", ".")}, are assigned a value of -1 
#'                     in \code{dist_mat} then contiguous gaps of any run length,
#'                     which are not present in both sequences, will be counted as a 
#'                     distance of 1. Meaning, indels of any length will increase
#'                     the sequence distance by 1. Gap values other than -1 will 
#'                     return a distance that does not consider indels as a special case.
#'
#' @return   A matrix of numerical distance between each entry in \code{seq}. 
#'           If \code{seq} is a named vector, row and columns names will be added 
#'           accordingly.
#' 
#'           Amino acid distance matrix may be built with \link{getAAMatrix}. 
#'           Uses \link{seqDist} for calculating distances between pairs.
#'           See \link{pairwiseEqual} for generating an equivalence matrix.
#'           
#' @examples
#' # Gaps will be treated as Ns with a gap=0 distance matrix
#' pairwiseDist(c(A="ATGGC", B="ATGGG", C="ATGGG", D="AT--C"), 
#'              dist_mat=getDNAMatrix(gap=0))
#' 
#' # Gaps will be treated as universally non-matching characters with gap=1
#' pairwiseDist(c(A="ATGGC", B="ATGGG", C="ATGGG", D="AT--C"), 
#'              dist_mat=getDNAMatrix(gap=1))
#' 
#' # Gaps of any length will be treated as single mismatches with a gap=-1 distance matrix
#' pairwiseDist(c(A="ATGGC", B="ATGGG", C="ATGGG", D="AT--C"), 
#'              dist_mat=getDNAMatrix(gap=-1))
#' 
#' @export
pairwiseDist <- function(seq, dist_mat=getDNAMatrix()) {
    pairwiseDistRcpp(seq, dist_mat)
}


#' Calculate pairwise distances between sequences
#' 
#' \code{nonsquareDist} calculates all pairwise distance between a set of sequences and a subset of it.
#'
#' @param    seq       character vector containing a DNA sequences. The sequence vector needs to
#'                     be named.
#' @param    indx      numeric vector containing the indices (a subset of indices of \code{seq}).
#' @param    dist_mat  Character distance matrix. Defaults to a Hamming distance 
#'                     matrix returned by \link{getDNAMatrix}. If gap 
#'                     characters, \code{c("-", ".")}, are assigned a value of -1 
#'                     in \code{dist_mat} then contiguous gaps of any run length,
#'                     which are not present in both sequences, will be counted as a 
#'                     distance of 1. Meaning, indels of any length will increase
#'                     the sequence distance by 1. Gap values other than -1 will 
#'                     return a distance that does not consider indels as a special case.
#'
#' @return   A matrix of numerical distance between each entry in \code{seq} and 
#'           sequences specified by \code{indx} indices. 
#'           
#'           Note that the input subsampled indices will be ordered ascendingly. Therefore, 
#'           it is necessary to assign unique names to the input sequences, \code{seq}, 
#'           to recover the input order later. Row and columns names will be added accordingly.
#' 
#'           Amino acid distance matrix may be built with \link{getAAMatrix}. 
#'           Uses \link{seqDist} for calculating distances between pairs.
#'           See \link{pairwiseEqual} for generating an equivalence matrix.
#'           
#' @examples
#' # Gaps will be treated as Ns with a gap=0 distance matrix
#' seq <- c(A="ATGGC", B="ATGGG", C="ATGGG", D="AT--C")
#' pairwiseDist(seq, 
#'              dist_mat=getDNAMatrix(gap=0))
#' 
#' nonsquareDist(seq, indx=c(1,3), 
#'              dist_mat=getDNAMatrix(gap=0))
#' 
#' @export
nonsquareDist <- function(seq, indx, dist_mat=getDNAMatrix()) {
    nonsquareDistRcpp(seq, indx, dist_mat)
}

# load(file.path("tests", "data-tests", "junctionAlignment.rda"))
load(file.path("..", "data-tests", "junctionAlignment.rda"))
load(file.path("..", "data-tests", "junctionAlignment_changeo.rda"))

germline_db <- c(
    "IGHV3-30*04"="CAGGTGCAGCTGGTGGAGTCTGGGGGA...GGCGTGGTCCAGCCTGGGAGGTCCCTGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTC............AGTAGCTATGCTATGCACTGGGTCCGCCAGGCTCCAGGCAAGGGGCTGGAGTGGGTGGCAGTTATATCATATGAT......GGAAGTAATAAATACTACGCAGACTCCGTGAAG...GGCCGATTCACCATCTCCAGAGACAATTCCAAGAACACGCTGTATCTGCAAATGAACAGCCTGAGAGCTGAGGACACGGCTGTGTATTACTGTGCGAGAGA",
    "IGHD6-19*01"="GGGTATAGCAGTGGCTGGTAC",
    "IGHJ6*02"="ATTACTACTACTACTACGGTATGGACGTCTGGGGCCAAGGGACCACGGTCACCGTCTCCTCA",
    "IGHV4-39*01"="CAGCTGCAGCTGCAGGAGTCGGGCCCA...GGACTGGTGAAGCCTTCGGAGACCCTGTCCCTCACCTGCACTGTCTCTGGTGGCTCCATCAGC......AGTAGTAGTTACTACTGGGGCTGGATCCGCCAGCCCCCAGGGAAGGGGCTGGAGTGGATTGGGAGTATCTATTATAGT.........GGGAGCACCTACTACAACCCGTCCCTCAAG...AGTCGAGTCACCATATCCGTAGACACGTCCAAGAACCAGTTCTCCCTGAAGCTGAGCTCTGTGACCGCCGCAGACACGGCTGTGTATTACTGTGCGAGACA",
    "IGHD3-10*01"="GTATTACTATGGTTCGGGGAGTTATTATAAC",
    "IGHJ4*02"="ACTACTTTGACTACTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAG",
    "IGHV3-48*01"="GAGGTGCAGCTGGTGGAGTCTGGGGGA...GGCTTGGTACAGCCTGGGGGGTCCCTGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTC............AGTAGCTATAGCATGAACTGGGTCCGCCAGGCTCCAGGGAAGGGGCTGGAGTGGGTTTCATACATTAGTAGTAGT......AGTAGTACCATATACTACGCAGACTCTGTGAAG...GGCCGATTCACCATCTCCAGAGACAATGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGACACGGCTGTGTATTACTGTGCGAGAGA",
    "IGHD3-16*02"="GTATTATGATTACGTTTGGGGGAGTTATCGTTATACC",
    "IGHJ1*01"="GCTGAATACTTCCAGCACTGGGGCCAGGGCACCCTGGTCACCGTCTCCTCAG",
    "IGLV1-47*01"="CAGTCTGTGCTGACTCAGCCACCCTCA...GCGTCTGGGACCCCCGGGCAGAGGGTCACCATCTCTTGTTCTGGAAGCAGCTCCAACATC............GGAAGTAATTATGTATACTGGTACCAGCAGCTCCCAGGAACGGCCCCCAAACTCCTCATCTATAGGAAT.....................AATCAGCGGCCCTCAGGGGTCCCT...GACCGATTCTCTGGCTCCAAG......TCTGGCACCTCAGCCTCCCTGGCCATCAGTGGGCTCCGGTCCGAGGATGAGGCTGATTATTACTGTGCAGCATGGGATGACAGCCTGAGTGGTCC",
    "IGLJ1*01"="TTATGTCTTCGGAACTGGGACCAAGGTCACCGTCCTAG",
    "IGHV3-11*03"="CAGGTGCAGCTGTTGGAGTCTGGGGGA...GGCTTGGTCAAGCCTGGAGGGTCCCTGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTC............AGTGACTACTACATGAGCTGGATCCGCCAGGCTCCAGGGAAGGGGCTGGAGTGGGTTTCATACATTAGTAGTAGT......AGTAGTTACACAAACTACGCAGACTCTGTGAAG...GGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGACACGGCCGTGTATTACTGTGCGAGA",
    "IGHV2-10*01"="CAGGTCACCTTGAAGGAGTCTGGTCCT...GCACTGGTGAAACCCACACAGACCCTCATGCTGACCTGCACCTTCTCTGGGTTCTCACTCAGC......ACTTCTGGAATGGGTGTGGGTTAGATCTGTCAGCCCTCAGCAAAGGCCCTGGAGTGGCTTGCACACATTTATTAGAAT.........GATAATAAATACTACAGCCCATCTCTGAAG...AGTAGGCTCATTATCTCCAAGGACACCTCCAAGAATGAAGTGGTTCTAACAGTGATCAACATGGACATTGTGGACACAGCCACACATTACTGTGCAAGGAGAC",
    "IGHD2-21*02"="AGCATATTGTGGTGGTGACTGCTATTCC",
    "IGHV3-23*01"="GAGGTGCAGCTGTTGGAGTCTGGGGGA...GGCTTGGTACAGCCTGGGGGGTCCCTGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTT............AGCAGCTATGCCATGAGCTGGGTCCGCCAGGCTCCAGGGAAGGGGCTGGAGTGGGTCTCAGCTATTAGTGGTAGT......GGTGGTAGCACATACTACGCAGACTCCGTGAAG...GGCCGGTTCACCATCTCCAGAGACAATTCCAAGAACACGCTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGACACGGCCGTATATTACTGTGCGAAAGA",
    "IGHD1-26*01"="GGTATAGTGGGAGCTACTAC",
    "IGHV3-64*01"="GAGGTGCAGCTGGTGGAGTCTGGGGGA...GGCTTGGTCCAGCCTGGGGGGTCCCTGAGACTCTCCTGTGCAGCCTCTGGATTCACCTTC............AGTAGCTATGCTATGCACTGGGTCCGCCAGGCTCCAGGGAAGGGACTGGAATATGTTTCAGCTATTAGTAGTAAT......GGGGGTAGCACATATTATGCAAACTCTGTGAAG...GGCAGATTCACCATCTCCAGAGACAATTCCAAGAACACGCTGTATCTTCAAATGGGCAGCCTGAGAGCTGAGGACATGGCTGTGTATTACTGTGCGAGAGA",
    "IGHD5-12*01"="GTGGATATAGTGGCTACGATTAC",
    "IGHJ4*01"="ACTACTTTGACTACTGGGGCCAAGGAACCCTGGTCACCGTCTCCTCAG",
    "IGHV3-49*03"="GAGGTGCAGCTGGTGGAGTCTGGGGGA...GGCTTGGTACAGCCAGGGCGGTCCCTGAGACTCTCCTGTACAGCTTCTGGATTCACCTTT............GGTGATTATGCTATGAGCTGGTTCCGCCAGGCTCCAGGGAAGGGGCTGGAGTGGGTAGGTTTCATTAGAAGCAAAGCTTATGGTGGGACAACAGAATACGCCGCGTCTGTGAAA...GGCAGATTCACCATCTCAAGAGATGATTCCAAAAGCATCGCCTATCTGCAAATGAACAGCCTGAAAACCGAGGACACAGCCGTGTATTACTGTACTAGAGA",
    "IGLV3-10*01"="TCCTATGAGCTGACACAGCCACCCTCG...GTGTCAGTGTCCCCAGGACAAACGGCCAGGATCACCTGCTCTGGAGATGCATTGCCA..................AAAAAATATGCTTATTGGTACCAGCAGAAGTCAGGCCAGGCCCCTGTGCTGGTCATCTATGAGGAC.....................AGCAAACGACCCTCCGGGATCCCT...GAGAGATTCTCTGGCTCCAGC......TCAGGGACAATGGCCACCTTGACTATCAGTGGGGCCCAGGTGGAGGATGAAGCTGACTACTACTGTTACTCAACAGACAGCAGTGGTAATCATAG",
    "IGLJ2*01"="TGTGGTATTCGGCGGAGGGACCAAGCTGACCGTCCTAG",
    "IGLV2-23*01"="CAGTCTGCCCTGACTCAGCCTGCCTCC...GTGTCTGGGTCTCCTGGACAGTCGATCACCATCTCCTGCACTGGAACCAGCAGTGATGTTGGG.........AGTTATAACCTTGTCTCCTGGTACCAACAGCACCCAGGCAAAGCCCCCAAACTCATGATTTATGAGGGC.....................AGTAAGCGGCCCTCAGGGGTTTCT...AATCGCTTCTCTGGCTCCAAG......TCTGGCAACACGGCCTCCCTGACAATCTCTGGGCTCCAGGCTGAGGACGAGGCTGATTATTACTGCTGCTCATATGCAGGTAGTAGCACTTTAC",
    "IGLJ3*02"="TTGGGTGTTCGGCGGAGGGACCAAGCTGACCGTCCTAG",
    "IGKV2D-30*01"="GATGTTGTGATGACTCAGTCTCCACTCTCCCTGCCCGTCACCCTTGGACAGCCGGCCTCCATCTCCTGCAGGTCTAGTCAAAGCCTCGTATACAGT...GATGGAAACACCTACTTGAATTGGTTTCAGCAGAGGCCAGGCCAATCTCCAAGGCGCCTAATTTATAAGGTT.....................TCTAACTGGGACTCTGGGGTCCCA...GACAGATTCAGCGGCAGTGGG......TCAGGCACTGATTTCACACTGAAAATCAGCAGGGTGGAGGCTGAGGATGTTGGGGTTTATTACTGCATGCAAGGTACACACTGGCCTCC",
    "IGKJ1*01"="GTGGACGTTCGGCCAAGGGACCAAGGTGGAAATCAAAC",
    "IGKV2-30*01"="GATGTTGTGATGACTCAGTCTCCACTCTCCCTGCCCGTCACCCTTGGACAGCCGGCCTCCATCTCCTGCAGGTCTAGTCAAAGCCTCGTATACAGT...GATGGAAACACCTACTTGAATTGGTTTCAGCAGAGGCCAGGCCAATCTCCAAGGCGCCTAATTTATAAGGTT.....................TCTAACCGGGACTCTGGGGTCCCA...GACAGATTCAGCGGCAGTGGG......TCAGGCACTGATTTCACACTGAAAATCAGCAGGGTGGAGGCTGAGGATGTTGGGGTTTATTACTGCATGCAAGGTACACACTGGCCTCC"
)

test_that("junctionAlignment: counts deleted and in cdr3 germline nucleotides", {
    
    expect_warning(aln_db <- junctionAlignment(db, germline_db ))
    
    expect_equal(aln_db[['e3v_length']][1],0)
    expect_equal(aln_db[['e5d_length']][1],12)
    expect_equal(aln_db[['e3d_length']][1],1)
    expect_equal(aln_db[['e5j_length']][1],6)
    expect_equal(aln_db[['v_cdr3_length']][1],8)
    expect_equal(aln_db[['j_cdr3_length']][1],23)
    
    # plotJunctionAlignment(db[1,],germline_db)$plot
    
    # Has insertion 
    # IMGT:
    # Nucleotide insertions have been detected and automatically removed for this analysis
    # from nt position in user submitted sequence: 253
    # nb of inserted nt: 9
    # causing frameshift: no
    #
    # And deletion
    # from nt position in user submitted sequence: 181
    # nb of deleted nt	: 6
    # causing frameshift: no
    expect_equal(aln_db[['e3v_length']][2],2)
    expect_equal(aln_db[['e5d_length']][2],9)
    expect_equal(aln_db[['e3d_length']][2],5)
    expect_equal(aln_db[['e5j_length']][2],4)
    expect_equal(aln_db[['v_cdr3_length']][2],6)
    expect_equal(aln_db[['j_cdr3_length']][2],10)
    
    # plotJunctionAlignment(db[2,],germline_db)$plot 
    
    
    # Partial: NA junction
    expect_equal(aln_db[['e3v_length']][3],240)
    expect_equal(aln_db[['e5d_length']][3],13)
    expect_equal(aln_db[['e3d_length']][3],9)
    expect_equal(aln_db[['e5j_length']][3],19)
    expect_true(is.na(aln_db[['v_cdr3_length']][3]))
    expect_true(is.na(aln_db[['j_cdr3_length']][3]))
    
    # plotJunctionAlignment(db[3,],germline_db)$plot 
    
    
    # Partial short junction
    # Junction is one nt ("A")
    # IMGT V-quest: no results
    # Should MakeDb.py --partial fail sequences like this?
    # Is v_germline_start correct?
    expect_equal(aln_db[['e3v_length']][4],128)
    expect_equal(aln_db[['e5d_length']][4],26)
    expect_equal(aln_db[['e3d_length']][4],0)
    expect_equal(aln_db[['e5j_length']][4],24)
    expect_true(is.na(aln_db[['v_cdr3_length']][4]))
    expect_true(is.na(aln_db[['j_cdr3_length']][4]))
    
    # plotJunctionAlignment(db[4,],germline_db)$plot 
    
    
    # Partial
    # Nucleotide deletions have been detected in the CDR3
    # Check makeRegionDefinition works with this sequence
    expect_equal(aln_db[['e3v_length']][5],3)
    expect_true(is.na(aln_db[['e5d_length']][5]))
    expect_true(is.na(aln_db[['e3d_length']][5]))
    expect_equal(aln_db[['e5j_length']][5],3)
    expect_equal(aln_db[['v_cdr3_length']][5],24)
    expect_equal(aln_db[['j_cdr3_length']][5],4)
    

    # plotJunctionAlignment(db[5,],germline_db)$plot
    
    # IMGT, for these 2 sequences, IMGT/HighV-QUEST does not find a junction,
    # but IgBLAST does
    #
    # plotJunctionAlignment(db[6,],germline_db)$plot #IgBLAST
    # plotJunctionAlignment(db[9,],germline_db)$plot #IMGT
    # Can't create germline for AGTCCTTAGCCAACAAT_imgt (9), CreateGermlines gives
    # error "Germline sequence differs in length from input sequence by 22 characters."
    expect_equal(aln_db[['e3v_length']][6],2)
    expect_equal(aln_db[['e5d_length']][6],6)
    expect_equal(aln_db[['e3d_length']][6],3)
    expect_equal(aln_db[['e5j_length']][6],12)
    expect_equal(aln_db[['v_cdr3_length']][6],6)
    expect_equal(aln_db[['j_cdr3_length']][6],18)
    
    expect_equal(aln_db[['e3v_length']][9],2)
    expect_true(is.na(aln_db[['e5d_length']][9]))
    expect_true(is.na(aln_db[['e3d_length']][9]))
    expect_equal(aln_db[['e5j_length']][9],0)
    expect_true(is.na(aln_db[['v_cdr3_length']][9]))
    expect_true(is.na(aln_db[['j_cdr3_length']][9]))
    
    # For this nonproductive seq, the junction is all np1 nucleotides (22nt)
    # plotJunctionAlignment(db[7,],germline_db)$plot #IgBLAST
    # plotJunctionAlignment(db[10,],germline_db)$plot #IMGT
    expect_equal(aln_db[['e3v_length']][7],32)
    expect_true(is.na(aln_db[['e5d_length']][7]))
    expect_true(is.na(aln_db[['e3d_length']][7]))
    expect_equal(aln_db[['e5j_length']][7],7)
    expect_equal(aln_db[['v_cdr3_length']][7],0)
    expect_equal(aln_db[['j_cdr3_length']][7],0)
    
    
    # db$sequence_alignment[c(8,11)]
    # > db[c(8,11), "junction"]
    # A tibble: 2 x 1
    # junction                         
    # <chr>                            
    # 1 TGCTCCTCATACAGTAGCACTTATTGGGTGTTC
    # 2 TGCTCCTCATACAGTAGCACTTATTGGGTGTTC
    #  db$sequence[8]
    # [1] "GGGGGTCACAAGAGGCAGCGCTCTCGGGACGTCTCCACCATGGCCTGGGCTCTGCTGCTCCTCACTCTCCTCACTCAGGACACAGGGTCCTGGGCCCAGTCTGCCCTGACTCAGCCTGCCTCCGTGTCTGGGTCTCCTGGACAGTCGATCACCATCTCCTGCACTGGAACCAGCAGTGATTTTGGGAATTATAATTTTGTCTCCTGGTATCAACAACACCCAGACAAAGCCCCCAAACTCATTATTTATGAGGGCAGTAAGCGGCCCTCAGGGGTTTCTAATCACTTCTCTGTCTCCAAGTCTGGCAACACGGCCTCCCTGACAATCTCTGGGCTCCAGGCTGACGACGAGGCTGATTATTACTGCTCCTCATACAGTAGCACTTATTGGGTGTTCGGCGGAGGGACCAAGCTGACCGTCCTGGGTCAGCCCAAGGCTGCCCCCTCG"
    #                                                                                                                                                                                                                                                                                                                                                                                             |- position 376
    # This sequence has IMGT gaps in the junction as shown in sequence_alignment, but only when aligned with IMGT.
    # The alignment gaps are not reported in junction
    # No alignment gaps in sequence_alignment when aligned with IgBLAST
    # IMGT/V-QUEST reports a 6nt indel in the input sequence that is located
    # in the junction. This is from position 376 in the input sequence.
    # The 6nt in the reference v germline are gcaggt. Can't create a germline, CreateGermlines.py 
    # gives the error "Germline sequence differs in length from input sequence by 6 characters."
    # "CAGTCTGCCCTGACTCAGCCTGCCTCC...GTGTCTGGGTCTCCTGGACAGTCGATCACCATCTCCTGCACTGGAACCAGCAGTGATTTTGGG.........AATTATAATTTTGTCTCCTGGTATCAACAACACCCAGACAAAGCCCCCAAACTCATTATTTATGAGGGC.....................AGTAAGCGGCCCTCAGGGGTTTCT...AATCACTTCTCTGTCTCCAAG......TCTGGCAACACGGCCTCCCTGACAATCTCTGGGCTCCAGGCTGACGACGAGGCTGATTATTACTGCTCCTCATACAGTAGCACTTATTGGGTGTTCGGCGGAGGGACCAAGCTGACCGTCCT"        
    # "CAGTCTGCCCTGACTCAGCCTGCCTCC...GTGTCTGGGTCTCCTGGACAGTCGATCACCATCTCCTGCACTGGAACCAGCAGTGATTTTGGG.........AATTATAATTTTGTCTCCTGGTATCAACAACACCCAGACAAAGCCCCCAAACTCATTATTTATGAGGGC.....................AGTAAGCGGCCCTCAGGGGTTTCT...AATCACTTCTCTGTCTCCAAG......TCTGGCAACACGGCCTCCCTGACAATCTCTGGGCTCCAGGCTGACGACGAGGCTGATTATTACTGCTCCTCATAC......AGTAGCACTTATTGGGTGTTCGGCGGAGGGACCAAGCTGACCGTCCTGG"
    # plotJunctionAlignment(db[8,],germline_db)$plot
    # plotJunctionAlignment(db[11,],germline_db)$plot
    
    # This sequence has two . ('..') in the "junction" (#13, IMGT alignment)
    # and junction is too short
    # db$junction_length[12:13]
    # [1] 2 0
    # plotJunctionAlignment(db[12,],germline_db)$plot
    # plotJunctionAlignment(db[13,],germline_db)$plot
    
    expect_equal(aln_db[['e3v_length']][12],48)
    expect_true(is.na(aln_db[['e5d_length']][12]))
    expect_true(is.na(aln_db[['e3d_length']][12]))
    expect_equal(aln_db[['e5j_length']][12],3)
    # Junction is too short, no cdr3
    expect_true(is.na(aln_db[['v_cdr3_length']][12]))
    expect_true(is.na(aln_db[['j_cdr3_length']][12]))
    
    # No junction, expect all NA
    expect_equal(aln_db[['e3v_length']][13],0)
    expect_true(is.na(aln_db[['e5d_length']][13]))
    expect_true(is.na(aln_db[['e3d_length']][13]))
    expect_equal(aln_db[['e5j_length']][13],0)
    expect_true(is.na(aln_db[['v_cdr3_length']][13]))
    expect_true(is.na(aln_db[['j_cdr3_length']][13])) ## NOTE: alignment is off, due to changeo issue 178

    db_changeo <- junctionAlignment(db_changeo, 
                          germline_db,sequence_alignment="SEQUENCE_IMGT", 
                          v_call="V_CALL", d_call="D_CALL", j_call="J_CALL",
                          v_germline_start="V_GERM_START_IMGT", v_germline_end="V_GERM_END",
                          d_germline_start="D_GERM_START", d_germline_end="D_GERM_END",
                          j_germline_start="J_GERM_START", j_germline_end="J_GERM_END",
                          np1_length="NP1_LENGTH",np2_length="NP2_LENGTH",
                          junction="JUNCTION", junction_length="JUNCTION_LENGTH")
    
    expect_equal(db_changeo[['e3v_length']][2],3)
    expect_equal(db_changeo[['e5d_length']][2],6)
    expect_equal(db_changeo[['e3d_length']][2],7)
    expect_equal(db_changeo[['e5j_length']][2],9)
    expect_equal(db_changeo[['v_cdr3_length']][2],5)
    expect_equal(db_changeo[['j_cdr3_length']][2],19)
    
})
                                      

ExampleDb <- file.path("..", "data-tests", "ExampleDb.gz")
db <- readChangeoDb(ExampleDb)

#ExampleDb_airr <- file.path("tests", "data-tests", "ExampleDb_airr.gz")
#db_airr <- airr::read_rearrangement(ExampleDb_airr)

# Load test database
e1 <- new.env()
#load(file.path("tests", "data-tests", "TestDb.rda"), envir=e1)
load(file.path("..", "data-tests", "TestDb.rda"), envir=e1)
db2 <- get("TestDb", envir=e1)
rm(e1)

# Load test database airr format
#e1 <- new.env()
#load(file.path("tests", "data-tests", "TestDb.rda"), envir=e1)
#load(file.path("..", "data-tests", "TestDb_airr.rda"), envir=e1)
#db2_airr <- get("db2_airr", envir=e1)
#rm(e1)

#### seqDist ####

test_that("seqDist: short toy sequences", {
    
    expect_equal(seqDist("AC-A", "AC-G", getDNAMatrix(gap=0)), 1)
    expect_equal(seqDist("AC--A", "ACATG", getDNAMatrix(gap=0)), 1)
    expect_equal(seqDist("AC--A", "ACATG", getDNAMatrix(gap=1)), 3)
    expect_equal(seqDist("AC--A", "ACATG", getDNAMatrix(gap=-1)), 2)
    
    # AC--AAC--A
    # ACATGACATG
    # **--.**--.
    expect_equal(seqDist("AC--AAC--A", "ACATGACATG", getDNAMatrix(gap=0)), 2)
    expect_equal(seqDist("AC--AAC--A", "ACATGACATG", getDNAMatrix(gap=1)), 6)
    expect_equal(seqDist("AC--AAC--A", "ACATGACATG", getDNAMatrix(gap=-1)), 4)
    
    
    # Ungapped examples
    expect_equal(seqDist("ATGGC", "ATGGG"), 1)
    expect_equal(seqDist("ATGGC", "ATG??"), 2)
    
    # Gaps will be treated as Ns with a gap=0 distance matrix
    expect_equal(
        seqDist("ATGGC", "AT--C", dist_mat=getDNAMatrix(gap=0)),
        0)
    
    # Gaps will be treated as universally non-matching characters with gap=1
    expect_equal(
        seqDist("ATGGC", "AT--C", dist_mat=getDNAMatrix(gap=1)),
        2)
    
    # Gaps of any length will be treated as single mismatches with a gap=-1 distance matrix
    expect_equal(
        seqDist("ATGGC", "AT--C", dist_mat=getDNAMatrix(gap=-1)),
        1)
    
    # Gaps of equivalent run lengths are not counted as gaps
    expect_equal(
        seqDist("ATG-C", "ATG-C", dist_mat=getDNAMatrix(gap=-1)),
        0)
    
    # Overlapping runs of gap characters are counted as a single gap
    expect_equal(
        seqDist("ATG-C", "AT--C", dist_mat=getDNAMatrix(gap=-1)),
        1) 
    
    expect_equal(
        seqDist("A-GGC", "AT--C", dist_mat=getDNAMatrix(gap=-1)),
        1)
    
    expect_equal(
        seqDist("AT--C", "AT--C", dist_mat=getDNAMatrix(gap=-1)),
        0)
    
    # Discontiguous runs of gap characters each count as separate gaps
    expect_equal(
        seqDist("-TGGC", "AT--C", dist_mat=getDNAMatrix(gap=-1)),
        2)
})

test_that("seqDist: long IMGT-gapped sequences", {
    # Define test sequence set
    seq <- substr(db2$SEQUENCE_IMGT[1:4], 1, 312)
    germ <- substr(db2$GERMLINE_IMGT_D_MASK[1:4], 1, 312)
    # Replace dots with Ns
    seq_n <- gsub(".", "N", seq, fixed=TRUE)
    germ_n <- gsub(".", "N", germ, fixed=TRUE)
    
    # Region ranges
    regions <- c(list("SEQ"=c(1, 312)),
                     IMGT_REGIONS)
    
    
    # Expected mutations
    expected <- list("SEQ"=c(13, 19, 19, 55),
                     "fwr1"=c(2, 6, 6, 4),
                     "cdr1"=c(1, 5, 5, 3),
                     "fwr2"=c(2, 1, 1, 3),
                     "cdr2"=c(3, 3, 3, 6),
                     "fwr3"=c(5, 4, 4, 39))
    
    # Test full V region sequence
    #cat("Full V-region (", paste(regions[["SEQ"]], collapse=":"), ") distance:\n", sep="")
    
    #cat("  With dots:\n")
    d <- mapply(seqDist, seq, germ, MoreArgs=list(dist_mat=getDNAMatrix(gap=0)), 
                USE.NAMES=FALSE)
    #d <- mapply(function(x, y) { sum(seqinr::s2c(x) != seqinr::s2c(y)) }, seq, germ, USE.NAMES=FALSE)
    expect_equal(d, expected[["SEQ"]], info="Full V-region with dots")
    
    #cat("  With Ns:\n")
    d <- mapply(seqDist, seq_n, germ_n, MoreArgs=list(dist_mat=getDNAMatrix(gap=0)), 
                USE.NAMES=FALSE)
    #d <- mapply(function(x, y) { sum(seqinr::s2c(x) != seqinr::s2c(y)) }, seq, germ, USE.NAMES=FALSE)
    expect_equal(d, expected[["SEQ"]], info="Full V-region with Ns")
    
    # Test by region
    for (n in c("fwr1", "cdr1", "fwr2", "cdr2", "fwr3")) {
        #cat(n, " (", paste(regions[[n]], collapse=":"), ") distance:\n", sep="")
        
        # Define substrings
        seq_sub <- extractVRegion(seq, n)
        germ_sub <- extractVRegion(germ, n)
        seq_n_sub <- extractVRegion(seq_n, n)
        germ_n_sub <- extractVRegion(germ_n, n)
        
        #cat("  With dots:\n")
        d <- mapply(seqDist, seq_sub, germ_sub, MoreArgs=list(dist_mat=getDNAMatrix(gap=0)), 
                    USE.NAMES=FALSE)
        #d <- mapply(function(x, y) { sum(seqinr::s2c(x) != seqinr::s2c(y)) }, seq_sub, germ_sub, USE.NAMES=FALSE)
        expect_equal(d, expected[[n]], info=paste(n, "with dots"))
        
        #cat("  With Ns:\n")
        d <- mapply(seqDist, seq_n_sub, germ_n_sub, MoreArgs=list(dist_mat=getDNAMatrix(gap=0)), 
                    USE.NAMES=FALSE)
        #d <- mapply(function(x, y) { sum(seqinr::s2c(x) != seqinr::s2c(y)) }, seq_sub, germ_sub, USE.NAMES=FALSE)
        expect_equal(d, expected[[n]], info=paste(n, "with Ns"))
    }
})

#### pairwiseDist ####

test_that("pairwiseDist Nucleotide", {
    seq <- c(A="ATGGC", B="ATGGG", C="ATGGG", D="AT--C")
    # Gaps will be treated as Ns with a gap=0 distance matrix
    obs <- pairwiseDist(seq, 
                 dist_mat=getDNAMatrix(gap=0))
    expect_equal(obs,
                 matrix(c(0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0),ncol=4),
                 check.attributes=F)
    # Gaps will be treated as universally non-matching characters with gap=1
    obs <- pairwiseDist(seq, 
                 dist_mat=getDNAMatrix(gap=1))
    expect_equal(obs,
                 matrix(c(0, 1, 1, 2, 1, 0, 0, 3, 1, 0, 0, 3, 2, 3, 3, 0),ncol=4),
                 check.attributes=F)
    
    # Gaps of any length will be treated as single mismatches with a gap=-1 distance matrix
    obs <- pairwiseDist(seq, 
                 dist_mat=getDNAMatrix(gap=-1))
    expect_equal(obs,
                 matrix(c(0, 1, 1, 1, 1, 0, 0, 2, 1, 0, 0, 2, 1, 2, 2, 0),ncol=4),
                 check.attributes=F)
    
    # test subset pairwise function
    seq <- c("TGTGCGAGAGGCCTCACGGAACCTCGCACA", "TGTGCGAGACAGGTTTACTATGATAGTAGT", "TGTGCGAGAGGTTTGGTAGATACAGCTATG", "TGTGCGAGAGGCTTCTCCGGGAGGGGGCCG", 
             "TGTGCGAGAGTCCGGGGTCGCAAGCTGCAG", "TGTGCGAGACCTTATTCGCACGACGGTGGT", "TGTGCGAGAGATTACCGGCTACAGGCGTGC", "TGTGCGAAAGATCTCTCCCACCGCCTAGGA")
    seq <- setNames(seq, seq)
    obs <- pairwiseDist(seq, 
                        dist_mat=getDNAMatrix(gap=0))
    subobs <- nonsquareDist(seq, indx=c(7,2,5),
                              dist_mat=getDNAMatrix(gap=0))
    expect_equal(subobs,
                 obs[rownames(subobs),],
                 check.attributes=F)
})


test_that("pairwiseDist AminoAcid", {
    # Create a toy junction set with stop codon
    seq_uniq <- c("ACGTACGTACGTACGTACGTACGTACGTATCGT", 
                  "ACGTACGTACGTACGTACGTACGTACGTATTGA")  # <- stop codon: TGA
    seq_uniq <- setNames(alakazam::translateDNA(seq_uniq), seq_uniq)
    dist_mat <- alakazam::pairwiseDist(seq_uniq, dist_mat=alakazam::getAAMatrix())
    
    expect_equal(dist_mat[1,2], 1, tolerance=0)
    expect_equal(dist_mat[2,1], 1, tolerance=0)
    
    # Create a toy junction set with N
    seq_uniq <- c("ACGTACGTACGTACGTACGTACGTACGTATCGT", 
                  "ACGTACGTACGTACGTACGTACGTACGTATAAT")  # <- N <- AAT 
    
    seq_uniq <- setNames(alakazam::translateDNA(seq_uniq), seq_uniq)
    dist_mat <- alakazam::pairwiseDist(seq_uniq, dist_mat=alakazam::getAAMatrix())
    
    expect_equal(dist_mat[1,2], 1, tolerance=0)
    expect_equal(dist_mat[2,1], 1, tolerance=0)
    
    # Create a toy junction set with X
    seq_uniq <- c("ACGTACGTACGTACGTACGTACGTACGTATCGT", 
                  "ACGTACGTACGTACGTACGTACGTACGTATNNN")  # <- X <- NNN
    
    seq_uniq <- setNames(alakazam::translateDNA(seq_uniq), seq_uniq)
    dist_mat <- alakazam::pairwiseDist(seq_uniq, dist_mat=alakazam::getAAMatrix())
    
    expect_equal(dist_mat[1,2], 0, tolerance=0)
    expect_equal(dist_mat[2,1], 0, tolerance=0)
    
    # Create a toy junction set with X from gaps
    seq_uniq <- c("ACGTACGTACGTACGTACGTACGTACGTATCGT", 
                  "ACGTACGTACGTACGTACGTACGTACGTAT---")  # X <- ---
    
    seq_uniq <- setNames(alakazam::translateDNA(seq_uniq), seq_uniq)
    dist_mat <- alakazam::pairwiseDist(seq_uniq, dist_mat=alakazam::getAAMatrix())

    expect_equal(dist_mat[1,2], 0, tolerance=0)
    expect_equal(dist_mat[2,1], 0, tolerance=0)

    # Create a toy junction set with all possible characters
    aa_uniq <- c("-ABCDEFGHIJKLMNPQRSTVWXYZ*", 
                 "-.BCDEFGHIJKLMNPQRSTVWXYZ*")
    
    dist_mat <- alakazam::pairwiseDist(aa_uniq, dist_mat=alakazam::getAAMatrix())
    
    expect_equal(dist_mat[1,2], 0, tolerance=0)
    expect_equal(dist_mat[2,1], 0, tolerance=0)
})

#### seqEqual ####

test_that("seqEqual", {
    # Ignore gaps
    expect_true(seqEqual("ATG-C", "AT--C"))
    expect_true(seqEqual("ATGGC", "ATGGN"))
    expect_false(seqEqual("AT--T", "ATGGC"))
    
    # Ignore only Ns
    expect_false(seqEqual("ATG-C", "AT--C", ignore="N"))
    expect_true(seqEqual("ATGGC", "ATGGN", ignore="N"))
    expect_false(seqEqual("AT--T", "ATGGC", ignore="N"))
})

#### translateDNA ####

test_that("translateDNA", {
    expect_equal(
        translateDNA(db$JUNCTION[1:3]),
        c("CARDRSTPWRRGIASTTVRTSW", "CARDLLWSVLLTGYYSYGMDAW", "CARDLLWSVLLTGYYSYGMDAW"))
    
    expect_equal(
        translateDNA(db$JUNCTION[1:3], trim=TRUE),
        c("ARDRSTPWRRGIASTTVRTS", "ARDLLWSVLLTGYYSYGMDA", "ARDLLWSVLLTGYYSYGMDA"))
    expect_equal(translateDNA("ACTGACTCGA"), "TDS")

    # test NAs
    expect_equal(translateDNA(NA), NA)
})

#### maskSeqGaps ####

test_that("maskSeqGaps", {
    expect_equal(maskSeqGaps(c("ATG-C", "CC..C")),
                 c("ATGNC", "CCNNC"))
    
    expect_equal(maskSeqGaps("--ATG-C-"), "NNATGNCN")
    expect_equal(maskSeqGaps("--ATG-C-", outer_only=TRUE), "NNATG-CN")
})

#### maskSeqEnds ####

test_that("maskSeqEnds", {
    # Default behavior uniformly masks ragged ends
    seq <- c("CCCCTGGG", "NAACTGGN", "NNNCTGNN")
    expect_equal(maskSeqEnds(seq),c("NNNCTGNN", "NNNCTGNN", "NNNCTGNN"))
    
    # Does nothing
    expect_equal(maskSeqEnds(seq, max_mask=0), c("CCCCTGGG", "NAACTGGN", "NNNCTGNN"))
    
    # Cut ragged sequence ends
    expect_equal(maskSeqEnds(seq, trim=TRUE), c("CTG", "CTG", "CTG"))
    
    # Set max_mask to limit extent of masking and trimming
    maskSeqEnds(seq, max_mask=1)
    expect_equal(maskSeqEnds(seq, max_mask=1, trim=TRUE), c("CCCTGG", "AACTGG", "NNCTGN"))
})

#### padSeqEnds ####

test_that("padSeqEnds", {
    seq <- c("CCCCTGGG", "ACCCTG", "CCCC")
    
    # Default behavior pads ends to longest length
    expect_equal(padSeqEnds(seq), 
                 c("CCCCTGGGN", "ACCCTGNNN", "CCCCNNNNN"))
    
    # start argument pads beginning instead of end
    expect_equal(padSeqEnds(seq, start=TRUE), 
                 c("NCCCCTGGG", "NNNACCCTG", "NNNNNCCCC"))
    
    # len argument pads to defined length
    expect_equal(padSeqEnds(seq, len=15), 
                 c("CCCCTGGGNNNNNNN", "ACCCTGNNNNNNNNN", "CCCCNNNNNNNNNNN"))
    expect_equal(padSeqEnds(seq, len=15, start=TRUE), 
                 c("NNNNNNNCCCCTGGG", "NNNNNNNNNACCCTG", "NNNNNNNNNNNCCCC"))

    # Invalid length same as default
    expect_equal(padSeqEnds(seq, len=2), 
                 c("CCCCTGGGN", "ACCCTGNNN", "CCCCNNNNN"))
})

#### collapseDuplicates ####

test_that("collapseDuplicates", {
    # Example Change-O data.frame
    db <- data.frame(SEQUENCE_ID=LETTERS[1:5],
                     SEQUENCE_IMGT=c("CCCCTGGG", "CCCCTGGN", "NAACTGGN", "NNNCTGNN", "NAACTGNG"),
                     TYPE=c("IgM", "IgG", "IgG", "IgA", "IgG"),
                     SAMPLE=c("S1", "S1", "S2", "S2","S2"),
                     COUNT=1:5,
                     stringsAsFactors=FALSE)
    
    # Annotations are not parsed if neither text_fields nor num_fields is specified
    # The retained sequence annotations will be random
    obs <- collapseDuplicates(db, verbose=F, id="SEQUENCE_ID", seq="SEQUENCE_IMGT")
    exp <- data.frame(
        "SEQUENCE_ID" = c("C", "A"),
        "SEQUENCE_IMGT" = c("NAACTGGN", "CCCCTGGG"),
        "TYPE" = c("IgG","IgM"),
        "SAMPLE" = c("S2", "S1"),
        "COUNT" = c(3,1),
        stringsAsFactors = F
    )
    expect_equivalent(obs, exp[2:1,])
    
    obs_dry <- collapseDuplicates(db[-5,], verbose=F, dry=T, id="SEQUENCE_ID", seq="SEQUENCE_IMGT")
    expect_equal(obs_dry$collapse_class, c("duplicated", "duplicated", "ambiguous_duplicate", "ambiguous"))
    
    expect_equal(sort(obs_dry[obs_dry$collapse_pass,"SEQUENCE_ID"]), 
                 sort(obs$SEQUENCE_ID))
    
    ## Try messing up order
    ## C comes first
    obs_dry <- collapseDuplicates(db, verbose=F, dry=T, id="SEQUENCE_ID", seq="SEQUENCE_IMGT")
    expect_equal(sort(obs_dry[obs_dry$collapse_pass,"SEQUENCE_ID"]), 
                 sort(obs$SEQUENCE_ID))
    ## E comes first
    obs_dry <- collapseDuplicates(db[nrow(db):1,], verbose=F, dry=T, id="SEQUENCE_ID", seq="SEQUENCE_IMGT")
    expect_equal(sort(obs_dry[obs_dry$collapse_pass,"SEQUENCE_ID"]), 
                 c("A","E"))
    
    # Unique text_fields annotations are combined into a single string with ","
    # num_fields annotations are summed
    # Ambiguous duplicates are discarded
    obs <- collapseDuplicates(db[-5,], text_fields=c("TYPE", "SAMPLE"), num_fields="COUNT", 
                       verbose=F, id="SEQUENCE_ID", seq="SEQUENCE_IMGT")
    exp$TYPE <- c("IgG","IgG,IgM")
    exp$COUNT <- c(3,3)
    expect_equal(obs, exp)
    
    
    # Use alternate delimiter for collapsing textual annotations
    obs <- collapseDuplicates(db[-5,], text_fields=c("TYPE", "SAMPLE"), num_fields="COUNT", 
                       sep="/", verbose=F, id="SEQUENCE_ID", seq="SEQUENCE_IMGT")
    exp$TYPE <- c("IgG","IgG/IgM")
    expect_equal(obs, exp)
    
    # Add count of duplicates
    obs <- collapseDuplicates(db[-5,], text_fields=c("TYPE", "SAMPLE"), num_fields="COUNT", 
                       add_count=TRUE, verbose=F, id="SEQUENCE_ID", seq="SEQUENCE_IMGT")
    exp$TYPE <- c("IgG","IgG,IgM")
    exp$collapse_count <- c(1,2)
    expect_equal(obs, exp)
    
    # Masking ragged ends may impact duplicate removal
    db$SEQUENCE_IMGT <- maskSeqEnds(db$SEQUENCE_IMGT)
    obs <- collapseDuplicates(db[-5,], text_fields=c("TYPE", "SAMPLE"), num_fields="COUNT", 
                       add_count=TRUE, verbose=F, id="SEQUENCE_ID", seq="SEQUENCE_IMGT")
    exp <- data.frame(
        "SEQUENCE_ID" = "A",
        "SEQUENCE_IMGT" = "NNNCTGNN",
        "TYPE" = "IgA,IgG,IgM",
        "SAMPLE" = "S1,S2",
        "COUNT" = 10,
        "collapse_count" = 4,
        stringsAsFactors = F
    ) 
    expect_equal(obs, exp)
    
    # Test dry
    obs <- collapseDuplicates(db[-5,], text_fields=c("TYPE", "SAMPLE"), num_fields="COUNT", 
                              add_count=TRUE, verbose=F, id="SEQUENCE_ID", seq="SEQUENCE_IMGT",
                              dry=T)
    exp <- obs
    exp$collapse_class <- "duplicated"
    exp$collapse_pass <- c(T,F,F,F)
    expect_equal(obs, exp)
    
    
    
    ## All sequences are ambiguous,
    ## but belong to two independent ambiguous clusters
    ## Should return two sequences
    test <- data.frame(
        list("SEQUENCE_IMGT" = c("AANCTANNT",
                                 "AAACNNNNT",
                                 "AATCTNNNT",
                                 "AANCTCTTT")),
        stringsAsFactors = F
    )
    
    test2 <- data.frame(
        list("SEQUENCE_IMGT" = c("ATATCNNNT",
                                 "ATNNCANNT",
                                 "ATNTCTNNT",
                                 "ATCTCNTTT")),
        stringsAsFactors = F
    )
    test <- bind_rows(test, test2)
    test$SEQUENCE_ID <- test$SEQUENCE_IMGT
    # d_mat <- pairwiseEqual(test$SEQUENCE_IMGT)
    # colnames(d_mat) <- rownames(d_mat) <- test$SEQUENCE_ID
    # g <- graph_from_adjacency_matrix(d_mat)
    # plot(g)
    expect <- c("AANCTCTTT", "ATCTCNTTT")
    col <- collapseDuplicates(test, id="SEQUENCE_ID", seq="SEQUENCE_IMGT")
    expect_equal(col$SEQUENCE_ID, expect)
    col_dry <- collapseDuplicates(test, dry=T, id="SEQUENCE_ID", seq="SEQUENCE_IMGT")
    expect_equal(col_dry$SEQUENCE_ID[col_dry$collapse_pass], expect)
    
})

#### extractVRegion ####

test_that("extractVRegion", {
    clone <- subset(db, CLONE == 164)
    
    # Get all regions
    obs <- extractVRegion(clone$SEQUENCE_IMGT)
    expect_equal(dim(unique(obs)), c(7, 5))
    expect_equal(colnames(obs),
                 c("fwr1", "cdr1", "fwr2", "cdr2", "fwr3"))
    
    fwr1 <- c(
        "GAGGTGCAGCTGGTGGAGTCTGG.GGA...GGCTTGGTACAGCCTGGAGGGTCCCTGAGACTCTCCTGTGCAGCCTCT",
        "GAGGTGCAGCTGGTGGAGTCTGG.GGA...GGCTTGGTACAGCCTGGAGGGTCCCTGAGACTCTCCTGTGCAGCCTCT",
        "GAGGTGCAGCTGGTGGAGTCTGG.GGA...GGCTTGGTACAGCCTGGAGGGTCCCTGAGACTCTCCTGTGCAGCCTCT"                     
    )
    expect_equal(obs[1:3,"fwr1"], fwr1)
    
    cdr1 <- c(
        "GGATTCACCTTC............AGTAGTTATGAA",
        "GGATTCACCTTC............AGTAGTTATGAA",
        "GGATTCACCTTC............AGTAGTTATGAA",
        "GGATTCACCTTC............AGTAGTTATGAA"
    )
    expect_equal(obs[4:7,"cdr1"], cdr1)
    
    fwr2 <- c(
        "ATGAACTGGGTCCGCCAGGCTCCAGGGAAGGGGCTGGAGTGGGTTTCATAC",
        "ATGAACTGGGTCCGCCAGGCTCCAGGGAAGGGGCTGGAGTGGGTTTCATAC",
        "ATGAACTGGGTCCGCCAGGCTCCAGGGAAGGGGCTGGAGTGGGTTTCATAC"
    )
    expect_equal(obs[8:10,"fwr2"], fwr2)
    
    cdr2 <- c(
        "ATTAGTAGTAGT......GGTAGTACCATA",
        "ATTAGTAGTAGT......GGTAGTACCATA",
        "ATTAGTAGTAGT......GGTAGTACCATA",
        "ATTAGTAGTAGT......GGTAGTACCATA"
    )
    expect_equal(obs[11:14,"cdr2"], cdr2)
    
    fwr3 <- c(
        "TACTACGCAGACTCTGTGAAG...GGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGACACGGCTGTTTATTACTGT",
        "TACTACGCAGACTCTGTGAAG...GGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGACACGGCTGTTTATTACTGT",
        "TACTACGCAGACTCTGTGAAG...GGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGACACGGCTGTTTATTACTGT"
    )
    expect_equal(obs[15:17,"fwr3"], fwr3)
    
    # Get single region
    obs <- extractVRegion(clone$SEQUENCE_IMGT[1:3], "fwr1")
    expect_equal(obs[1:3], fwr1)
    
    # Get all CDRs
    obs <- extractVRegion(clone$SEQUENCE_IMGT, c("cdr1", "cdr2"))
    expect_equal(obs[1:4,"cdr1"],cdr1)
    expect_equal(obs[11:14,"cdr2"],cdr2)
    
    # Get all FWRs
    obs <- extractVRegion(clone$SEQUENCE_IMGT, c("fwr1", "fwr2", "fwr3"))
    expect_equal(obs[1:3,"fwr1"],fwr1)
    expect_equal(obs[8:10,"fwr2"],fwr2)
    expect_equal(obs[15:17,"fwr3"],fwr3)
})

# Deprecated and defunct functions

#' @include Classes.R
#' @include Diversity.R
NULL

#### Deprecated ####

#' Generate a clonal diversity index curve
#'
#' \code{rarefyDiversity} divides a set of clones by a group annotation,
#' resamples the sequences from each group, and calculates diversity
#' scores (\eqn{D}) over an interval of diversity orders (\eqn{q}).
#' 
#' @param    data      data.frame with Change-O style columns containing clonal assignments.
#' @param    group     name of the \code{data} column containing group identifiers.
#' @param    clone     name of the \code{data} column containing clone identifiers.
#' @param    copy      name of the \code{data} column containing copy numbers for each 
#'                     sequence. If \code{copy=NULL} (the default), then clone abundance
#'                     is determined by the number of sequences. If a \code{copy} column
#'                     is specified, then clone abundances is determined by the sum of 
#'                     copy numbers within each clonal group.
#' @param    min_q     minimum value of \eqn{q}.
#' @param    max_q     maximum value of \eqn{q}.
#' @param    step_q    value by which to increment \eqn{q}.
#' @param    min_n     minimum number of observations to sample.
#'                     A group with less observations than the minimum is excluded.
#' @param    max_n     maximum number of observations to sample. If \code{NULL} then no 
#'                     maximum is set.
#' @param    ci        confidence interval to calculate; the value must be between 0 and 1.
#' @param    nboot     number of bootstrap realizations to generate.
#' @param    uniform   if \code{TRUE} then uniformly resample each group to the same 
#'                     number of observations. If \code{FALSE} then allow each group to
#'                     be resampled to its original size or, if specified, \code{max_size}.
#' @param    cell_id   name of the \code{data} column containing cell identifiers. 
#' @param    progress  if \code{TRUE} show a progress bar.
#' 
#' @return   A \link{DiversityCurve} object summarizing the diversity scores.
#' 
#' @details
#' Clonal diversity is calculated using the generalized diversity index (Hill numbers) 
#' proposed by Hill (Hill, 1973). See \link{calcDiversity} for further details.
#'
#' Diversity is calculated on the estimated complete clonal abundance distribution.
#' This distribution is inferred by using the Chao1 estimator to estimate the number
#' of seen clones, and applying the relative abundance correction and unseen clone
#' frequency described in Chao et al, 2015.
#'
#' To generate a smooth curve, \eqn{D} is calculated for each value of \eqn{q} from
#' \code{min_q} to \code{max_q} incremented by \code{step_q}.  When \code{uniform=TRUE}
#' variability in total sequence counts across unique values in the \code{group} column 
#' is corrected by repeated resampling from the estimated complete clonal distribution to a 
#' common number of sequences.
#' 
#' The diversity index (\eqn{D}) for each group is the mean value of over all resampling 
#' realizations. Confidence intervals are derived using the standard deviation of the 
#' resampling realizations, as described in Chao et al, 2015.
#' 
#' @references
#' \enumerate{
#'   \item  Hill M. Diversity and evenness: a unifying notation and its consequences. 
#'            Ecology. 1973 54(2):427-32.
#'   \item  Chao A. Nonparametric Estimation of the Number of Classes in a Population. 
#'            Scand J Stat. 1984 11, 265270.
#'   \item  Chao A, et al. Rarefaction and extrapolation with Hill numbers: 
#'            A framework for sampling and estimation in species diversity studies. 
#'            Ecol Monogr. 2014 84:45-67.
#'   \item  Chao A, et al. Unveiling the species-rank abundance distribution by 
#'            generalizing the Good-Turing sample coverage theory. 
#'            Ecology. 2015 96, 11891201.
#' }
#'  
#' @seealso  \link{alphaDiversity}
#' 
#' @examples
#' \dontrun{
#' # Group by sample identifier
#' div <- rarefyDiversity(ExampleDb, "sample_id", step_q=1, max_q=10, nboot=100)
#' plotDiversityCurve(div, legend_title="Sample")
#'                    
#' # Grouping by isotype rather than sample identifier
#' div <- rarefyDiversity(ExampleDb, "c_call", min_n=40, step_q=1, max_q=10, 
#'                        nboot=100)
#' plotDiversityCurve(div, legend_title="Isotype")
#' }
#' @export
rarefyDiversity <- function(data, group, clone="CLONE", copy=NULL,
                            min_q=0, max_q=4, step_q=0.05, min_n=30, max_n=NULL,
                            ci=0.95, nboot=2000, uniform=TRUE, 
                            cell_id="cell_id",
                            progress=FALSE) {
    .Deprecated("alphaDiversity")
    bootstrap_obj <- estimateAbundance(data, group=group, clone=clone, copy=copy, nboot=nboot, min_n=min_n, max_n=max_n, uniform=uniform, ci=ci, cell_id=cell_id)
    diversity_obj <- alphaDiversity(bootstrap_obj, ci=ci, min_q=min_q, max_q=max_q, step_q)

    return(diversity_obj)
}


#' Pairwise test of the diversity index
#' 
#' \code{testDiversity} performs pairwise significance tests of the diversity index 
#' (\eqn{D}) at a given diversity order (\eqn{q}) for a set of annotation groups via
#' rarefaction and bootstrapping.
#'
#' @param    data      data.frame with Change-O style columns containing clonal assignments.
#' @param    q         diversity order to test.
#' @param    group     name of the \code{data} column containing group identifiers.
#' @param    clone     name of the \code{data} column containing clone identifiers.
#' @param    copy      name of the \code{data} column containing copy numbers for each 
#'                     sequence. If \code{copy=NULL} (the default), then clone abundance
#'                     is determined by the number of sequences. If a \code{copy} column
#'                     is specified, then clone abundances is determined by the sum of 
#'                     copy numbers within each clonal group.
#' @param    min_n     minimum number of observations to sample.
#'                     A group with less observations than the minimum is excluded.
#' @param    max_n     maximum number of observations to sample. If \code{NULL} the maximum
#'                     if automatically determined from the size of the largest group.
#' @param    nboot     number of bootstrap realizations to perform.
#' @param    ci        confidence interval to calculate; the value must be between 0 and 1.
#' @param    progress  if \code{TRUE} show a progress bar.
#' @param    cell_id   the name of the \code{data} column containing cell identifiers.
#' 
#' @return   A \link{DiversityCurve} object containing slot test with p-values and summary 
#'             statistics.
#' 
#' @details
#' Clonal diversity is calculated using the generalized diversity index proposed by 
#' Hill (Hill, 1973). See \link{calcDiversity} for further details.
#' 
#' Diversity is calculated on the estimated complete clonal abundance distribution.
#' This distribution is inferred by using the Chao1 estimator to estimate the number
#' of seen clones, and applying the relative abundance correction and unseen clone
#' frequency described in Chao et al, 2014.
#'
#' Variability in total sequence counts across unique values in the \code{group} column is 
#' corrected by repeated resampling from the estimated complete clonal distribution to 
#' a common number of sequences. The diversity index estimate (\eqn{D}) for each group is 
#' the mean value of over all bootstrap realizations. 
#' 
#' Significance of the difference in diversity index (\eqn{D}) between groups is tested by 
#' constructing a bootstrap delta distribution for each pair of unique values in the 
#' \code{group} column. The bootstrap delta distribution is built by subtracting the diversity 
#' index \eqn{Da} in \eqn{group-a} from the corresponding value \eqn{Db} in \eqn{group-b}, 
#' for all bootstrap realizations, yielding a distribution of \code{nboot} total deltas; where 
#' \eqn{group-a} is the group with the greater mean \eqn{D}. The p-value for hypothesis 
#' \eqn{Da  !=  Db} is the value of \eqn{P(0)} from the empirical cumulative distribution 
#' function of the bootstrap delta distribution, multiplied by 2 for the two-tailed correction.
#' 
#' @note
#' This method may inflate statistical significance when clone sizes are uniformly small,
#' such as when most clones sizes are 1, sample size is small, and \code{max_n} is near
#' the total count of the smallest data group. Use caution when interpreting the results 
#' in such cases. We are currently investigating this potential problem.
#' 
#' @references
#' \enumerate{
#'   \item  Hill M. Diversity and evenness: a unifying notation and its consequences. 
#'            Ecology. 1973 54(2):427-32.
#'   \item  Chao A. Nonparametric Estimation of the Number of Classes in a Population. 
#'            Scand J Stat. 1984 11, 265270.
#'   \item  Wu Y-CB, et al. Influence of seasonal exposure to grass pollen on local and 
#'            peripheral blood IgE repertoires in patients with allergic rhinitis. 
#'            J Allergy Clin Immunol. 2014 134(3):604-12.
#'   \item  Chao A, et al. Rarefaction and extrapolation with Hill numbers: 
#'            A framework for sampling and estimation in species diversity studies. 
#'            Ecol Monogr. 2014 84:45-67.
#'   \item  Chao A, et al. Unveiling the species-rank abundance distribution by 
#'            generalizing the Good-Turing sample coverage theory. 
#'            Ecology. 2015 96, 11891201.
#' }
#' 
#' @seealso  \link{alphaDiversity}
#' 
#' @examples  
#' \dontrun{
#' # Groups under the size threshold are excluded and a warning message is issued.
#' testDiversity(ExampleDb, "sample_id", q=0, min_n=30, nboot=100)
#' }
#' 
#' @export
testDiversity <- function(data, q, group, clone="CLONE", copy=NULL,
                          min_n=30, max_n=NULL, nboot=2000, progress=FALSE, ci=0.95,
                          cell_id="cell_id") {
    .Deprecated("alphaDiversity")

    abundance_obj <- estimateAbundance(data, group=group, clone=clone, copy=copy, nboot=nboot, min_n=min_n, max_n=max_n, ci=ci, cell_id=cell_id)
    diversity_obj <- alphaDiversity(abundance_obj, min_q=q, max_q=q, step_q=1, ci=ci)

    return(diversity_obj)
}

#### Defunct ####

# Gene usage analysis

#### Calculation functions ####

#' Tabulates V(D)J allele, gene or family usage within each locus. 
#'
#' Determines the count and relative abundance of V(D)J alleles, genes or families within
#' groups. If sequences from multiple loci are present, the frequency is calculated within 
#' each locus.
#'
#' @param    data    data.frame with AIRR-format or Change-O style columns.
#' @param    gene    column containing allele assignments. Only the first allele in the
#'                   column will be considered when \code{mode} is "gene", "family" or
#'                   "allele". The value will be used as it is with \code{mode="asis"}.
#' @param    groups  columns containing grouping variables. If \code{NULL} do not group.
#' @param    copy    name of the \code{data} column containing copy numbers for each
#'                   sequence. If this value is specified, then total copy abundance
#'                   is determined by the sum of copy numbers within each gene.
#'                   This argument is ignored if \code{clone} is specified.
#' @param    clone   name of the \code{data} column containing clone identifiers for each
#'                   sequence. If this value is specified, then one gene will be considered
#'                   for each clone. Note, this is accomplished by using the most
#'                   common gene within each \code{clone} identifier. As such,
#'                   ambiguous alleles within a clone will not be accurately represented.
#' @param    mode    one of \code{c("gene", "family", "allele", "asis")} defining
#'                   the degree of specificity regarding allele calls. Determines whether
#'                   to return counts for genes (calling \code{getGene}),
#'                   families (calling \code{getFamily}), alleles (calling
#'                   \code{getAllele}) or using the value as it is in the column
#'                   \code{gene}, without any processing.
#' @param    fill    logical of \code{c(TRUE, FALSE)} specifying when if groups (when specified)
#'                   lacking a particular gene should be counted as 0 if TRUE or not (omitted).
#' @param    first   if TRUE return only the first allele/gene/family call for computing the frequency; if FALSE return all calls delimited by commas.
#' @param    collapse  if TRUE check for duplicates and return only unique allele/gene/family assignments per sequence; if
#'                   FALSE return all assignments (faster). Has no effect if first=TRUE.
#' @param   remove_na    removes rows with \code{NA} values in the gene column if \code{TRUE} and issues a warning.
#'                        Otherwise, keeps those rows and considers \code{NA} as a gene in the final counts
#'                        and relative abundances.
#' @param   cell_id  name of the \code{data} column containing the cell identifiers for each sequence.
#'
#' @return   A data.frame summarizing family, gene or allele counts and frequencies
#'           with columns:
#'           \itemize{
#'             \item \code{locus}:        locus of the gene (IGH, IGK, IGL, TRA, TRB, TRD, TRG). Note that frequencies are calculated within each locus.
#'             \item \code{gene}:         name of the family, gene or allele.
#'             \item \code{seq_count}:    total number of sequences for the gene in the locus.
#'             \item \code{locus_count}:  total number of sequences in the locus.
#'             \item \code{seq_freq}:     frequency of the gene as a fraction of the total
#'                                        number of sequences within each grouping.
#'             \item \code{copy_count}:   sum of the copy counts in the \code{copy} column.
#'                                        for each gene. Only present if the \code{copy}
#'                                        argument is specified.
#'             \item \code{locus_copy_count}: sum of the copy counts in the \code{copy} column.
#'                                        for all gene in the locus. Only present if the 
#'                                        \code{copy} argument is specified.
#'             \item \code{copy_freq}:    frequency of the gene as a fraction of the total
#'                                        copy number within each group. Only present if
#'                                        the \code{copy} argument is specified.
#'             \item \code{clone_count}:  total number of clones for the gene. Only present if
#'                                        the \code{clone} argument is specified.
#'             \item \code{clone_freq}:   frequency of the gene as a fraction of the total
#'                                        number of clones within each grouping. Only present if
#'                                        the \code{clone} argument is specified.
#'           }
#'           Additional columns defined by the \code{groups} argument will also be present.
#'
#' @examples
#' # Without copy numbers
#' genes <- countGenes(ExampleDb, gene = "v_call", groups = "sample_id", mode = "family")
#' genes <- countGenes(ExampleDb, gene = "v_call", groups = "sample_id", mode = "gene")
#' genes <- countGenes(ExampleDb, gene = "v_call", groups = "sample_id", mode = "allele")
#'
#' # With copy numbers and multiple groups
#' genes <- countGenes(ExampleDb,
#'     gene = "v_call", groups = c("sample_id", "c_call"),
#'     copy = "duplicate_count", mode = "family"
#' )
#'
#' # Count by clone
#' genes <- countGenes(ExampleDb,
#'     gene = "v_call", groups = c("sample_id", "c_call"),
#'     clone = "clone_id", mode = "family"
#' )
#'
#' # Count absent genes
#' genes <- countGenes(ExampleDb,
#'     gene = "v_call", groups = "sample_id",
#'     mode = "allele", fill = TRUE
#' )
#'
#' @export
countGenes <- function(data, gene, groups = NULL, copy = NULL, clone = NULL, fill = FALSE, first = TRUE, collapse = TRUE,
                       mode = c("gene", "allele", "family", "asis"), cell_id = "cell_id", remove_na = TRUE) {

    ## DEBUG
    # data=ExampleDb; gene="c_call"; groups=NULL; mode="gene"; clone="clone_id"
    # data=subset(db, clond_id == 3138)
    # Hack for visibility of dplyr variables
    # . <- NULL

    # Check input
    mode <- match.arg(mode)
    check <- checkColumns(data, c(gene, groups, copy))
    if (check != TRUE) {
        warning(check) # instead of throwing an error and potentially disrupting a workflow
    }

    # Don't allow a copy column for single-cell data as it doesn't make sense to count copies
    if (cell_id %in% names(data) & !is.null(copy)) {
        stop("Copy column specification not allowed for single-cell or mixed bulk and single-cell data.
        A cell_id column is present in the dataframe single-cell or mixed bulk and single-cell data is assumed.")
    }

    # Handle NAs
    if (remove_na) {
        bool_na <- is.na(data[, gene])
        if (any(bool_na)) {
            if (!all(bool_na)) {
                msg <- paste0(
                    "NA(s) found in ", sum(bool_na), " row(s) of the ", gene,
                    " column and excluded from tabulation"
                )
                warning(msg)
            }
            data <- data[!bool_na, ]
        }
    }

    # Extract gene, allele or family assignments
    if (mode != "asis") {
        gene_func <- switch(mode,
            allele = getAllele,
            gene = getGene,
            family = getFamily
        )
        data[[gene]] <- gene_func(data[[gene]], first = first, collapse = collapse)
    }

    # Tabulate abundance
    if (cell_id %in% names(data)) {
        # Handle single-cell and mixed bulk and single-cell data
        data_sc <- data %>% dplyr::filter(!is.na(!!rlang::sym(cell_id)))
        data_sc[[cell_id]] <- as.character(data_sc[[cell_id]])
        data_blk <- data %>% dplyr::filter(is.na(!!rlang::sym(cell_id)))
        if (nrow(data_blk) > 0) {
            warning(paste0(
                "Mixed bulk and single-cell data detected.\n",
                "Sequences with '", cell_id, "' NA will be counted individually towards the total number of sequences.\n",
                "Consider filtering these sequences first for heavy or light chains. \n"
            ))
            data_blk[[cell_id]] <- paste0("bulk_", 1:nrow(data_blk)) # dummy cell_id for bulk data
        }
        data <- bind_rows(data_sc, data_blk)

        if (is.null(clone)) {
            # Tabulate sequence abundance
            cell_num <- data %>%
                dplyr::select(!!!rlang::syms(c(groups, cell_id))) %>%
                dplyr::distinct() %>%
                dplyr::group_by(!!!rlang::syms(c(groups))) %>%
                dplyr::summarize(cell_count = n()) %>%
                dplyr::ungroup() %>%
                dplyr::select(!!!rlang::syms(c(groups, "cell_count")))
            gene_tab_count <- data %>%
                dplyr::select(!!!rlang::syms(c(groups, gene, cell_id))) %>%
                dplyr::distinct() %>%
                dplyr::group_by(!!!rlang::syms(c(groups, gene))) %>%
                dplyr::summarize(seq_count = n())
            if (nrow(cell_num) > 1) {
                gene_tab <- gene_tab_count %>%
                    dplyr::left_join(cell_num, by = groups) %>%
                    dplyr::mutate(seq_freq = !!rlang::sym("seq_count") / !!rlang::sym("cell_count")) %>%
                    dplyr::arrange(desc(!!rlang::sym("seq_count")))
            } else {
                gene_tab <- gene_tab_count %>%
                    dplyr::mutate(seq_freq = !!rlang::sym("seq_count") / cell_num[["cell_count"]]) %>%
                    dplyr::arrange(desc(!!rlang::sym("seq_count")))
            }
        } else {
            # Get unique count per cell and
            # Find count of genes within each clone and keep first with maximum count
            gene_tab <- data %>%
                dplyr::group_by(!!!rlang::syms(c(groups, gene, cell_id))) %>%
                dplyr::distinct() %>%
                dplyr::group_by(!!!rlang::syms(c(groups, clone, gene))) %>%
                dplyr::mutate(clone_gene_count = n()) %>%
                dplyr::ungroup() %>%
                dplyr::group_by(!!!rlang::syms(c(groups, clone))) %>%
                dplyr::slice(which.max(!!rlang::sym("clone_gene_count"))) %>%
                dplyr::ungroup() %>%
                dplyr::group_by(!!!rlang::syms(c(groups, gene))) %>%
                dplyr::summarize(clone_count = n()) %>%
                dplyr::mutate(clone_freq = !!rlang::sym("clone_count") / sum(!!rlang::sym("clone_count"), na.rm = TRUE)) %>%
                dplyr::arrange(!!rlang::sym("clone_count"))
        }
    } else {
        if (is.null(copy) & is.null(clone)) {
            # Tabulate sequence abundance by locus
            if (!"locus" %in% colnames(data)){
              data$locus <- substr(data[[gene]], 1, 3)
            }
            locus_tab <- data %>%
                dplyr::group_by(!!!rlang::syms(unique(c(groups, "locus")))) %>%
                dplyr::summarize(locus_count = n(), .groups = "drop")
            gene_tab_count <- data %>%
                dplyr::group_by(!!!rlang::syms(unique(c(groups, "locus", gene)))) %>%
                dplyr::summarize(seq_count = n(), .groups = "drop")
            if (nrow(locus_tab) > 1) {
                gene_tab <- gene_tab_count %>%
                    dplyr::left_join(locus_tab, by = unique(c(groups, "locus"))) %>%
                    dplyr::mutate(seq_freq = !!rlang::sym("seq_count") / !!rlang::sym("locus_count")) %>%
                    dplyr::arrange(desc(!!rlang::sym("seq_count")))
            } else {
                gene_tab <- gene_tab_count %>%
                    dplyr::mutate(seq_freq = !!rlang::sym("seq_count") / locus_tab$locus_count) %>%
                    dplyr::arrange(desc(!!rlang::sym("seq_count")))
            }
        } else if (!is.null(clone) & is.null(copy)) {
            # test if no clone IDs are given
            if (all(is.na(data[[clone]]))) {
                stop("No clone IDs are present in the data.")
            } else if (any(is.na(data[[clone]]))) {
                num_na_clone <- sum(is.na(data[[clone]]))
                warning(paste0(
                    "Found ", num_na_clone, " sequences without clonal assignments. ",
                    "These sequences will be removed before counting clone frequencies."
                ))
                data <- data %>%
                    dplyr::filter(!is.na(!!rlang::sym(clone)))
            }
            # get locus
            if (!"locus" %in% colnames(data)){
               data$locus <- substr(data[[gene]], 1, 3)
            }
            # Find count of genes within each clone and keep first with maximum count
            clone_gene_tab <- data %>%
                dplyr::group_by(!!!rlang::syms(unique(c(groups, "locus", clone, gene)))) %>%
                dplyr::mutate(clone_gene_count = n()) %>%
                dplyr::ungroup() %>%
                dplyr::group_by(!!!rlang::syms(unique(c(groups, "locus", clone)))) %>%
                dplyr::slice(which.max(!!rlang::sym("clone_gene_count"))) %>%
                dplyr::ungroup()
            locus_tab <- clone_gene_tab %>%
                dplyr::group_by(!!!rlang::syms(unique(c(groups, "locus")))) %>%
                dplyr::summarize(locus_clone_count = n(), .groups = "drop")
            gene_tab_count <- clone_gene_tab %>%
                dplyr::group_by(!!!rlang::syms(unique(c(groups, "locus", gene)))) %>%
                dplyr::summarize(clone_count = n(), .groups = "drop")
            if (nrow(locus_tab) > 1) {
                gene_tab <- gene_tab_count %>%
                    dplyr::left_join(locus_tab, by = unique(c(groups, "locus"))) %>%
                    dplyr::mutate(clone_freq = !!rlang::sym("clone_count") / !!rlang::sym("locus_clone_count")) %>%
                    dplyr::arrange(desc(!!rlang::sym("clone_count")))
            } else {
                gene_tab <- gene_tab_count %>%
                    dplyr::mutate(clone_freq = !!rlang::sym("clone_count") / locus_tab$locus_clone_count) %>%
                    dplyr::arrange(desc(!!rlang::sym("clone_count")))
            }
        } else {
            if (!is.null(clone) & !is.null(copy)) {
                warning(
                    "Specifying both 'copy' and 'clone' columns is not meaningful. ",
                    "The 'clone' argument will be ignored."
                )
            }
            # get locus
            if (!"locus" %in% colnames(data)){
              data$locus <- substr(data[[gene]], 1, 3)
            }
            # Tabulate copy abundance by locus and gene
            locus_tab <- data %>%
                dplyr::group_by(!!!rlang::syms(unique(c(groups, "locus")))) %>%
                dplyr::summarize(
                    locus_count = length(!!rlang::sym(gene)),
                    locus_copy_count = sum(!!rlang::sym(copy), na.rm = TRUE), .groups = "drop"
                )
            gene_tab_count <- data %>%
                dplyr::group_by(!!!rlang::syms(unique(c(groups, "locus", gene)))) %>%
                dplyr::summarize(
                    seq_count = length(!!rlang::sym(gene)),
                    copy_count = sum(!!rlang::sym(copy), na.rm = TRUE), .groups = "drop"
                )
            if (nrow(locus_tab) > 1) {
                gene_tab <- gene_tab_count %>%
                    dplyr::left_join(locus_tab, by = unique(c(groups, "locus"))) %>%
                    dplyr::mutate(
                        seq_freq = !!rlang::sym("seq_count") / !!rlang::sym("locus_count"),
                        copy_freq = !!rlang::sym("copy_count") / !!rlang::sym("locus_copy_count")
                    ) %>%
                    dplyr::arrange(desc(!!rlang::sym("copy_count")))
            } else {
                gene_tab <- gene_tab_count %>%
                    dplyr::mutate(
                        seq_freq = !!rlang::sym("seq_count") / locus_tab$locus_count,
                        copy_freq = !!rlang::sym("copy_count") / locus_tab$locus_copy_count
                    ) %>%
                    dplyr::arrange(desc(!!rlang::sym("copy_count")))
            }
        }
    }

    # If a gene is present in one GROUP but not another, will fill the COUNT and FREQ with 0s
    if (fill) {
        gene_tab <- gene_tab %>%
            ungroup() %>%
            tidyr::complete(!!!rlang::syms(as.list(c(groups, gene))),
                fill = list(seq_count = 0, seq_freq = 0, copy_count = 0, copy_freq = 0, clone_count = 0, clone_freq = 0)
            )
        # fill in the loci
        gene_tab$locus <- substr(gene_tab[[gene]], 1, 3)
        # TODO: correct locus count, locus copy count, locus clone count if needed?
    }

    # Rename gene column
    gene_tab <- gene_tab %>% rename(dplyr::all_of(c("gene" = gene)))

    return(gene_tab)
}


#### Annotation functions ####

#' Get Ig segment allele, gene and family names
#'
#' \code{getSegment} performs generic matching of delimited segment calls with a custom
#' regular expression. \link{getAllele}, \link{getGene} and \link{getFamily} extract
#' the allele, gene and family names, respectively, from a character vector of
#' immunoglobulin (Ig) or TCR segment allele calls in IMGT format.
#'
#' @param     segment_call    character vector containing segment calls delimited by commas.
#' @param     segment_regex   string defining the segment match regular expression.
#' @param     first           if \code{TRUE} return only the first call in
#'                            \code{segment_call}; if \code{FALSE} return all calls
#'                            delimited by commas.
#' @param     collapse        if \code{TRUE} check for duplicates and return only unique
#'                            segment assignments; if \code{FALSE} return all assignments
#'                            (faster). Has no effect if \code{first=TRUE}.
#' @param     strip_d         if \code{TRUE} remove the "D" from the end of gene annotations
#'                            (denoting a duplicate gene in the locus);
#'                            if \code{FALSE} do not alter gene names.
#' @param     omit_nl         if \code{TRUE} remove non-localized (NL) genes from the result.
#'                            Only applies at the gene or allele level.
#' @param     sep             character defining both the input and output segment call
#'                            delimiter.
#'
#' @return    A character vector containing allele, gene or family names.
#'
#' @references
#'   \url{https://www.imgt.org/}
#'
#' @seealso  \link{countGenes}
#'
#' @examples
#' # Light chain examples
#' kappa_call <- c(
#'     "Homsap IGKV1D-39*01 F,Homsap IGKV1-39*02 F,Homsap IGKV1-39*01",
#'     "Homsap IGKJ5*01 F"
#' )
#'
#' getAllele(kappa_call)
#' getAllele(kappa_call, first = FALSE)
#' getAllele(kappa_call, first = FALSE, strip_d = FALSE)
#'
#' getGene(kappa_call)
#' getGene(kappa_call, first = FALSE)
#' getGene(kappa_call, first = FALSE, strip_d = FALSE)
#'
#' getFamily(kappa_call)
#' getFamily(kappa_call, first = FALSE)
#' getFamily(kappa_call, first = FALSE, collapse = FALSE)
#' getFamily(kappa_call, first = FALSE, strip_d = FALSE)
#'
#' getLocus(kappa_call)
#' getChain(kappa_call)
#'
#' # Heavy chain examples
#' heavy_call <- c(
#'     "Homsap IGHV1-69*01 F,Homsap IGHV1-69D*01 F",
#'     "Homsap IGHD1-1*01 F",
#'     "Homsap IGHJ1*01 F"
#' )
#'
#' getAllele(heavy_call, first = FALSE)
#' getAllele(heavy_call, first = FALSE, strip_d = FALSE)
#'
#' getGene(heavy_call, first = FALSE)
#' getGene(heavy_call, first = FALSE, strip_d = FALSE)
#'
#' getFamily(heavy_call)
#' getLocus(heavy_call)
#' getChain(heavy_call)
#'
#' # Filtering non-localized genes
#' nl_call <- c(
#'     "IGHV3-NL1*01,IGHV3-30-3*01,IGHV3-30*01",
#'     "Homosap IGHV3-30*01 F,Homsap IGHV3-NL1*01 F",
#'     "IGHV1-NL1*01"
#' )
#'
#' getAllele(nl_call, first = FALSE, omit_nl = TRUE)
#' getGene(nl_call, first = FALSE, omit_nl = TRUE)
#' getFamily(nl_call, first = FALSE, omit_nl = TRUE)
#'
#' # Temporary designation examples
#' tmp_call <- c("IGHV9S3*01", "IGKV10S12*01")
#'
#' getAllele(tmp_call)
#' getGene(tmp_call)
#' getFamily(tmp_call)
#'
#' @export
getSegment <- function(segment_call, segment_regex, first = TRUE, collapse = TRUE,
                       strip_d = TRUE, omit_nl = FALSE, sep = ",") {
    # Define boundaries of individual segment calls
    edge_regex <- paste0("[^", sep, "]*")

    # Remove NL genes
    if (omit_nl) {
        # Clean segment_call to keep only the name (remove species)
        allele_regex <- "((IG[HKL]|TR[ABDG])[VDJADEGMC][A-R0-9\\(\\)]*[-/\\w]*[-\\*]*[\\.\\w]+)"
        segment_call <- gsub(paste0(edge_regex, "(", allele_regex, ")", edge_regex), "\\1",
            segment_call,
            perl = T
        )
        # non-localized regex
        nl_regex <- paste0(
            "(IG[HKL]|TR[ABDG])[VDJADEGMC][0-9]+-NL[0-9]([-/\\w]*[-\\*][\\.\\w]+)*(",
            sep, "|$)"
        )
        # delete non-localized calls
        segment_call <- gsub(nl_regex, "", segment_call, perl = TRUE)
    }

    # Extract calls
    r <- gsub(paste0(edge_regex, "(", segment_regex, ")", edge_regex), "\\1",
        segment_call,
        perl = T
    )

    # Strip D from gene names if required
    if (strip_d) {
        strip_regex <- paste0("(?<=[A-Z0-9][0-9])D(?=\\*|-|", sep, "|$)")
        r <- gsub(strip_regex, "", r, perl = TRUE)
    }

    # Collapse to unique set if required
    if (first) {
        r <- gsub(paste0(sep, ".*$"), "", r)
    } else if (collapse) {
        r <- sapply(strsplit(r, sep), function(x) paste(unique(x), collapse = sep))
    }

    return(r)
}


#' @rdname getSegment
#' @export
getAllele <- function(segment_call, first = TRUE, collapse = TRUE,
                      strip_d = TRUE, omit_nl = FALSE, sep = ",") {
    allele_regex <- "((IG[HKL]|TR[ABDG])[VDJADEGMC][A-R0-9\\(\\)]*[-/\\w]*[-\\*]*[\\.\\w]+)"
    r <- getSegment(segment_call, allele_regex,
        first = first, collapse = collapse,
        strip_d = strip_d, omit_nl = omit_nl, sep = sep
    )

    return(r)
}


#' @rdname getSegment
#' @export
getGene <- function(segment_call, first = TRUE, collapse = TRUE,
                    strip_d = TRUE, omit_nl = FALSE, sep = ",") {
    gene_regex <- "((IG[HKL]|TR[ABDG])[VDJADEGMC][A-R0-9\\(\\)]*[-/\\w]*)"
    r <- getSegment(segment_call, gene_regex,
        first = first, collapse = collapse,
        strip_d = strip_d, omit_nl = omit_nl, sep = sep
    )

    return(r)
}


#' @rdname getSegment
#' @export
getFamily <- function(segment_call, first = TRUE, collapse = TRUE,
                      strip_d = TRUE, omit_nl = FALSE, sep = ",") {
    family_regex <- "((IG[HKL]|TR[ABDG])[VDJADEGMC][A-R0-9\\(\\)]*)"
    r <- getSegment(segment_call, family_regex,
        first = first, collapse = collapse,
        strip_d = strip_d, omit_nl = omit_nl, sep = sep
    )

    return(r)
}


#' @rdname getSegment
#' @export
getLocus <- function(segment_call, first = TRUE, collapse = TRUE,
                     strip_d = TRUE, omit_nl = FALSE, sep = ",") {
    locus_regex <- "((IG[HLK]|TR[ABDG]))"
    r <- getSegment(segment_call, locus_regex,
        first = first, collapse = collapse,
        strip_d = strip_d, omit_nl = omit_nl, sep = sep
    )

    return(r)
}


#' @rdname getSegment
#' @export
getChain <- function(segment_call, first = TRUE, collapse = TRUE,
                     strip_d = TRUE, omit_nl = FALSE, sep = ",") {
    r <- getLocus(segment_call,
        first = first, collapse = collapse,
        strip_d = strip_d, omit_nl = omit_nl, sep = sep
    )
    r <- gsub("(IGH)|(TR[BD])", "VH", r)
    r <- gsub("(IG[KL])|(TR[AG])", "VL", r)

    return(r)
}

#### Utility functions ####

# Get all VJ(L) combinations from one or more chains of the same type
#
# Input:
# V annotation, J annotation, and optionally junction length of
# one of more chains of the same type
# - v: V annotation
# - j: J annotation
# - l: junction length (optional)
# - sep_chain: character separting multiple chains
# - sep_anno: character separating multiple/ambiguous annotations within each chain
# - first: to be passed to getGene()
#
# Output:
# A vector containing all unique VJ(L) combinations represented
#
# Assumption:
# 1) number of chains match across v, j, l
# 2) if length value is supplied, each chain has a single length value
#
# Example input
# v <- "Homsap IGLV2-20*09 F,Homsap IGLV3-30*09 F;Homsap IGKV1-27*01 F;Homsap IGKV1-25*01 F,Homsap IGKV1-25*38 F,Homsap IGKV1-32*02 F"
# j <- "Homsap IGLJ3*02 F;Homsap IGKJ5*02 F;Homsap IGKJ3*03 F,Homsap IGKJ9*03 F"
# l <- "36;39;60"
# getAllVJL(v, j, l, ";", ",", FALSE)
# - 3 light chains
# - number of V annotations per chain: 2/1/3
# - number of J annotations per chain: 1/1/2
# - Expected supremum of the number of VJL combinations: 2*1 + 1*1 + 3*2 = 9
# - Note that this is the supremum because the ACTUAL number (7) CAN be lower due to presence
#   of different alleles from the SAME gene (since computation is performed at the gene level)
getAllVJL <- function(v, j, l, sep_chain, sep_anno, first) {
    # is l NULL?
    l_NULL <- is.null(l)

    # are there multiple chains?
    # assumes that number of chains match across v, j, l
    # (pre-checked in groupGenes)
    multi_chain <- stringi::stri_detect_fixed(str = v, pattern = sep_chain)

    # are there multiple annotations per chain?
    multi_anno_v <- stringi::stri_detect_fixed(str = v, pattern = sep_anno)
    multi_anno_j <- stringi::stri_detect_fixed(str = j, pattern = sep_anno)

    # separate chains
    # gets a vector of strings
    # each vector entry corresponds to a chain
    if (multi_chain) {
        v <- stringi::stri_split_fixed(str = v, pattern = sep_chain)[[1]]
        j <- stringi::stri_split_fixed(str = j, pattern = sep_chain)[[1]]

        if (!l_NULL) {
            l <- stringi::stri_split_fixed(str = l, pattern = sep_chain)[[1]]
        }
    }

    # separate annotations
    # gets a list
    # each list entry has a vector of one or more strings
    if (multi_anno_v) {
        v <- stringi::stri_split_fixed(str = v, pattern = sep_anno)
        # take care of 'first' here because getGene will not split by ","
        if (first) {
            v <- lapply(v, function(x) {
                x[1]
            })
        }
    }
    if (multi_anno_j) {
        j <- stringi::stri_split_fixed(str = j, pattern = sep_anno)
        if (first) {
            j <- lapply(j, function(x) {
                x[1]
            })
        }
    }

    # get gene
    # gets a list
    # each list entry corresponds to a chain, and is a vector of one or more strings
    v <- sapply(v, getGene, collapse = TRUE, simplify = FALSE, USE.NAMES = FALSE)
    j <- sapply(j, getGene, collapse = TRUE, simplify = FALSE, USE.NAMES = FALSE)

    # if there is multiple chains and/or multiple annotations
    if (multi_chain | (multi_anno_v & first) | (multi_anno_j & first)) {
        # gets a list
        # each list entry is a vector of one or more strings
        # do if/else outside sapply so it only gets evaluated once
        if (!l_NULL) {
            exp <- sapply(1:length(v), function(i) {
                # eg_df <- expand.grid(v[[i]], j[[i]], l[i])
                # eg_vec <- apply(eg_df, 1, stringi::stri_paste, collapse="@")
                n_v <- length(v[[i]])
                n_j <- length(j[[i]])
                eg_vec <- stringi::stri_paste(rep.int(v[[i]], times = n_j),
                    rep(j[[i]], each = n_v),
                    rep.int(l[i], times = n_v * n_j),
                    sep = "@"
                )
                return(eg_vec)
            }, simplify = FALSE, USE.NAMES = FALSE)
        } else {
            exp <- sapply(1:length(v), function(i) {
                # eg_df = expand.grid(v[[i]], j[[i]])
                # eg_vec = apply(eg_df, 1, stringi::stri_paste, collapse="@")
                n_v <- length(v[[i]])
                n_j <- length(j[[i]])
                eg_vec <- stringi::stri_paste(rep.int(v[[i]], times = n_j),
                    rep(j[[i]], each = n_v),
                    sep = "@"
                )
                return(eg_vec)
            }, simplify = FALSE, USE.NAMES = FALSE)
        }

        # concat and convert to vector; keep distinct values
        exp <- unique(unlist(exp, use.names = FALSE))
    } else {
        n_v <- length(v[[1]])
        n_j <- length(j[[1]])

        if (!l_NULL) {
            exp <- stringi::stri_paste(rep.int(v[[1]], times = n_j),
                rep(j[[1]], each = n_v),
                rep.int(l[1], times = n_v * n_j),
                sep = "@"
            )
        } else {
            exp <- stringi::stri_paste(rep.int(v[[1]], times = n_j),
                rep(j[[1]], each = n_v),
                sep = "@"
            )
        }
    }

    return(exp)
}

# Check if the input data has heavy chains
#
# Input:
# a data.frame containing the AIRR or Change-O data for a clone. See Details
# for the list of required columns and their default values.
#
# locus: The column specifying the sequence locus.
#
# Output:
# A logical vector indicating if the row entry is a heavy chain or not.
isHeavyChain <- function(data, locus = "locus") {
    check <- data[[locus]] %in% c("IGH", "TRB", "TRD")
    return(check)
}

# Check if the input data has light chains
#
# Input:
# a data.frame containing the AIRR or Change-O data for a clone. See Details
# for the list of required columns and their default values.
#
# locus: The column specifying the sequence locus.
#
# Output:
# A logical vector indicating if the row entry is a light chain or not.
isLightChain <- function(data, locus = "locus") {
    check <- data[[locus]] %in% c("IGK", "IGL", "TRA", "TRG")
    return(check)
}

# Check for common single cell problems
#
# Input:
# a data.frame containing the AIRR or Change-O data for a clone. See Details
# for the list of required columns and their default values.
#
# locus: The column specifying the sequence locus.
#
# cell_id: The column specifying the sequence cell id.
# Output:
# A data.frame containing the AIRR or Change-O data.
singleCellValidation <- function(data, locus = "locus", cell_id = "cell_id") {
    heavy <- data[isHeavyChain(data, locus = locus), ]
    light <- data[isLightChain(data, locus = locus), ]

    # check for multiple heavy chains
    heavy_count <- table(heavy[[cell_id]])
    multi_heavy_cells <- names(heavy_count[heavy_count > 1])
    if (length(multi_heavy_cells) != 0) {
        stop(paste(
            "Only one heavy chain is allowed per cell. Remove cells with",
            "multiple heavy chains."
        ))
    }
    heavy <- heavy[!heavy[[cell_id]] %in% multi_heavy_cells, ]

    # check for unpaired light chains
    pairedIndx <- light[[cell_id]] %in% heavy[[cell_id]]
    if (FALSE %in% pairedIndx) {
        stop("Unpaired light chains were found in the data. Please remove them.")
        light <- light[pairedIndx, ]
    }
    data <- rbind(heavy, light)
    return(data)
}

#' Group sequences by gene assignment
#'
#' \code{groupGenes} will group rows by shared V and J gene assignments,
#' and optionally also by junction lengths. IGH:IGK/IGL, TRB:TRA, and TRD:TRG
#' paired single-cell BCR/TCR sequencing and unpaired bulk sequencing
#' (IGH, TRB, TRD chain only) are supported. In the case of ambiguous (multiple)
#' gene assignments, the grouping may be specified to be a union across all
#' ambiguous V and J gene pairs, analogous to single-linkage clustering
#' (i.e., allowing for chaining).
#'
#' @param    data          data.frame containing sequence data.
#' @param    v_call        name of the column containing the heavy/long chain
#'                         V-segment allele calls.
#' @param    j_call        name of the column containing the heavy/long chain
#'                         J-segment allele calls.
#' @param    junc_len      name of column containing the junction length.
#'                         If \code{NULL} then 1-stage partitioning is perform
#'                         considering only the V and J genes is performed.
#'                         See Details for further clarification.
#' @param    sequence_alignment name of the column containing the sequence alignment.
#' @param    cell_id       name of the column containing cell identifiers or barcodes.
#'                         If specified, grouping will be performed in single-cell mode
#'                         with the behavior governed by the \code{locus} and
#'                         \code{only_heavy} arguments. If set to \code{NULL} then the
#'                         bulk sequencing data is assumed.
#' @param    locus         name of the column containing locus information.
#'                         Only applicable to single-cell data.
#'                         Ignored if \code{cell_id=NULL}.
#' @param    only_heavy    This is deprecated. Only heavy chains will be used in clustering.
#'                         Use only the IGH (BCR) or TRB/TRD (TCR) sequences
#'                         for grouping. Only applicable to single-cell data.
#'                         Ignored if \code{cell_id=NULL}.
#' @param    split_light   A deprecated parameter. This would split clones by the light chain.
#'                         For similar function use dowser::resolveLightChains
#' @param    first         if \code{TRUE} only the first call of the gene assignments
#'                         is used. if \code{FALSE} the union of ambiguous gene
#'                         assignments is used to group all sequences with any
#'                         overlapping gene calls.
#'
#' @return   Returns a modified data.frame with disjoint union indices
#'           in a new \code{vj_group} column.
#'
#'           If \code{junc_len} is supplied, the grouping this \code{vj_group}
#'           will have been based on V, J, and junction length simultaneously. However,
#'           the output column name will remain \code{vj_group}.
#'
#'           The output \code{v_call}, \code{j_call}, \code{cell_id}, and \code{locus}
#'           columns will be converted to type \code{character} if they were of type
#'           \code{factor} in the input \code{data}.
#'
#' @details
#' To invoke single-cell mode the \code{cell_id} argument must be specified and the \code{locus}
#' column must be correct. Otherwise, \code{groupGenes} will be run with bulk sequencing assumptions,
#' using all input sequences regardless of the values in the \code{locus} column.
#'
#' Values in the \code{locus} column must be one of \code{c("IGH", "IGI", "IGK", "IGL")} for BCR
#' or \code{c("TRA", "TRB", "TRD", "TRG")} for TCR sequences. Otherwise, the function returns an
#' error message and stops.
#'
#' Under single-cell mode with paired chained sequences, there was a choice of whether
#' grouping should be done by (a) using IGH (BCR) or TRB/TRD (TCR) sequences only or
#' (b) using IGH plus IGK/IGL (BCR) or TRB/TRD plus TRA/TRG (TCR).
#' This was governed by the \code{only_heavy} argument, now deprecated.
#'
#' Specifying \code{junc_len} will force \code{groupGenes} to perform a 1-stage partitioning of the
#' sequences/cells based on V gene, J gene, and junction length simultaneously.
#' If \code{junc_len=NULL} (no column specified), then \code{groupGenes} performs only the first
#' stage of a 2-stage partitioning in which sequences/cells are partitioned in the first stage
#' based on V gene and J gene, and then in the second stage further splits the groups based on
#' junction length (the second stage must be performed independently, as this only returns the
#' first stage results).
#'
#' In the input \code{data}, the \code{v_call}, \code{j_call}, \code{cell_id}, and \code{locus}
#' columns, if present, must be of type \code{character} (as opposed to \code{factor}).
#'
#' It is assumed that ambiguous gene assignments are separated by commas.
#'
#' All rows containing \code{NA} values in any of the \code{v_call}, \code{j_call}, and \code{junc_len}
#' (if \code{junc_len != NULL}) columns will be removed. A warning will be issued when a row
#' containing an \code{NA} is removed.
#'
#' @section Expectations for single-cell data:
#'
#' Single-cell paired chain data assumptions:
#'   \itemize{
#'      \item every row represents a sequence (chain).
#'      \item heavy/long and light/short chains of the same cell are linked by \code{cell_id}.
#'      \item the value in \code{locus} column indicates whether the chain is the heavy/long or light/short chain.
#'      \item each cell possibly contains multiple heavy/long and/or light/short chains.
#'      \item every chain has its own V(D)J annotation, in which ambiguous V(D)J
#'            annotations, if any, are separated by a comma.
#'   }
#'
#' Single-cell example:
#'   \itemize{
#'      \item A cell has 1 heavy chain and 2 light chains.
#'      \item There should be 3 rows corresponding to this cell.
#'      \item One of the light chains may have an ambiguous V annotation which looks like \code{"Homsap IGKV1-39*01 F,Homsap IGKV1D-39*01 F"}.
#'   }
#'
#' @examples
#' # Group by genes
#' db <- groupGenes(ExampleDb)
#' head(db$vj_group)
#'
#' @export
groupGenes <- function(data, v_call = "v_call", j_call = "j_call", junc_len = NULL,
                    sequence_alignment = NULL, cell_id = NULL, split_light = FALSE,
                    locus = "locus", only_heavy = TRUE, first = FALSE) {
    # CGJ 6/24/24 -- onlyHeavy warning
    # Deprecation/Removal of only_heavy = FALSE
    if (!only_heavy) {
        warning("only_heavy = FALSE is deprecated. Running as if only_heavy = TRUE")
        only_heavy <- TRUE
    }

    if (split_light) {
        warning(paste(
            "split_light = TRUE is deprecated. Please use split_light = FALSE.",
            "After clonal identification, light chain groups can be found with dowser::resolveLightChains"
        ))
        split_light <- FALSE
    }

    # CGJ 6/24/24 mixed data check -- with the lc options removed we want it to go
    # through the SC pathway
    # Let the user know that data seems to have single cell sequences, but they didn't set up
    # the function to run in single cell mode
    if (is.null(cell_id) & "cell_id" %in% colnames(data) & !is.null(locus)) {
        nmissing <- sum(is.na(data$cell_id))
        if (nmissing < nrow(data) & nmissing > 0) {
            # cell_id values are partially missing
            stop(paste(
                "A cell_id column was found in the data, but was not specified.",
                "Additionally, the data appears to have paired and unpaired cell data.",
                "This data type requires the single cell workflow, please specify the cell_id and rerun."
            ))
        } else if (nmissing == 0) {
            # all cell_id values are present
            stop(paste(
                "A cell_id column was found in the data, but was not specified.",
                "This data type requires the single cell workflow, please specify the cell_id and rerun."
            ))
        } else {
            # all cell_id values are missing
            warning(paste(
                "A cell_id column was found in the data, but was not specified. All values are NA."
            ))
        }
    }

    # Check base input
    # CGJ 4/12/24 added locus to this check to not repeat later
    check <- checkColumns(data, c(v_call, j_call, junc_len, sequence_alignment, locus))
    if (check != TRUE) {
        stop(paste(
            "A column or some combination of columns v_call, j_call,",
            "junc_len, sequence_alignment, and locus were not found in the data"
        ))
    }
    # CGJ 4/12/24
    # check for factors here rather than after the indexing/one-to-one annotation check
    # NULL will disappear when doing c()
    # c(NULL,NULL) gives NULL still
    cols_for_grouping <- c(v_call, j_call, junc_len)
    # cols cannot be factor
    if (any(sapply(cols_for_grouping, function(x) {
        class(data[[x]]) == "factor"
    }))) {
        stop(
            "one or more of { ", v_call, ", ", j_call,
            ifelse(is.null(junc_len), " ", ", "), junc_len,
            "} is factor. Must be character.\nIf using read.table(), make sure to set stringsAsFactors=FALSE.\n"
        )
    }

    # if necessary, cast select columns to character (factor not allowed later on)
    if (!is(data[[v_call]], "character")) {
        data[[v_call]] <- as.character(data[[v_call]])
    }
    if (!is(data[[j_call]], "character")) {
        data[[j_call]] <- as.character(data[[j_call]])
    }

    # e.g.: "Homsap IGHV3-7*01 F,Homsap IGHV3-6*01 F;Homsap IGHV1-4*01 F"
    separator_within_seq <- ","
    separator_between_seq <- ";"

    # single-cell mode? CGJ 6/24/24 -- SC mode is not needed?
    # Initialize mixed
    mixed <- FALSE
    single_cell <- FALSE
    if (!is.null(cell_id) & !is.null(locus)) {
        # single cell fields exist
        if (sum(is.na(data[[cell_id]])) == 0) {
            # all rows have cell_id data
            single_cell <- TRUE

            if (!is(data[[cell_id]], "character")) {
                data[[cell_id]] <- as.character(data[[cell_id]])
            }
            if (!is(data[[locus]], "character")) {
                data[[locus]] <- as.character(data[[locus]])
            }

            # check locus column
            valid_loci <- c("IGH", "IGI", "IGK", "IGL", "TRA", "TRB", "TRD", "TRG")
            check <- !all(unique(data[[locus]]) %in% valid_loci)
            if (check) {
                stop("The locus column contains invalid loci annotations.")
            }
        } else if (any(!is.na(data[[cell_id]]))) {
            # some rows have cell_id data, and some NA
            single_cell <- TRUE
            mixed <- TRUE

            if (!is(data[[cell_id]], "character")) {
                data[[cell_id]] <- as.character(data[[cell_id]])
            }
            if (!is(data[[locus]], "character")) {
                data[[locus]] <- as.character(data[[locus]])
            }

            # check locus column
            valid_loci <- c("IGH", "IGI", "IGK", "IGL", "TRA", "TRB", "TRD", "TRG")
            check <- !all(unique(data[[locus]]) %in% valid_loci)
            if (check) {
                stop("The locus column contains invalid loci annotations.")
            }
        }
    }

    # only set if `single_cell` & `only_heavy`
    v_call_light <- NULL
    j_call_light <- NULL
    junc_len_light <- NULL

    # single-cell mode
    if (single_cell) {
        # make a copy
        data_orig <- data
        # regardless of using heavy only, or using both heavy and light
        # for each cell
        # - index wrt data of heavy chain
        # - index wrt data of light chain(s)
        cell_id_uniq <- unique(data[[cell_id]])
        if (mixed) {
            cell_seq_idx <- sapply(cell_id_uniq, function(x) {
                if (is.na(x)) {
                    # heavy chain
                    idx_h <- which(is.na(data[[cell_id]]) & data[[locus]] %in% c("IGH", "TRB", "TRD"))
                } else {
                    # heavy chain
                    idx_h <- which(data[[cell_id]] == x & data[[locus]] %in% c("IGH", "TRB", "TRD"))
                }
                # light chain
                idx_l <- which(data[[cell_id]] == x & data[[locus]] %in% c("IGK", "IGL", "TRA", "TRG"))
                return(list(heavy = idx_h, light = idx_l))
            }, USE.NAMES = FALSE, simplify = FALSE)
        } else {
            cell_seq_idx <- sapply(cell_id_uniq, function(x) {
                # heavy chain
                idx_h <- which(data[[cell_id]] == x & data[[locus]] %in% c("IGH", "TRB", "TRD"))
                # light chain
                idx_l <- which(data[[cell_id]] == x & data[[locus]] %in% c("IGK", "IGL", "TRA", "TRG"))

                return(list(heavy = idx_h, light = idx_l))
            }, USE.NAMES = FALSE, simplify = FALSE)
        }


        # use heavy chains only
        # Straightforward subsetting like below won't work in cases
        #     where multiple HCs are present for a cell
        # CGJ 6/16/23
        # subset to heavy only -- for both B and T cells
        data <- data_orig[isHeavyChain(data_orig, locus = locus), ]

        # Check for cells with two heavy chains
        # singleCellValdiation 4/12/24
        if (!mixed) {
            data <- singleCellValidation(data, locus = locus, cell_id = cell_id)
        }

        # flatten data
        cols <- c(cell_id, v_call, j_call, junc_len)
        data <- data.frame(matrix(NA, nrow = length(cell_seq_idx), ncol = length(cols)))
        colnames(data) <- cols

        for (i_cell in 1:length(cell_seq_idx)) {
            i_cell_h <- cell_seq_idx[[i_cell]][["heavy"]]

            data[[cell_id]][i_cell] <- cell_id_uniq[i_cell]

            # heavy chain V, J, junc_len
            # If the cell doesn't have heavy chain data, the gene
            # calls assigned with paste0 will be "" (empty strings)
            data[[v_call]][i_cell] <- paste0(data_orig[[v_call]][i_cell_h],
                collapse = separator_between_seq
            )
            data[[j_call]][i_cell] <- paste0(data_orig[[j_call]][i_cell_h],
                collapse = separator_between_seq
            )
            if (!is.null(junc_len)) {
                data[[junc_len]][i_cell] <- paste0(data_orig[[junc_len]][i_cell_h],
                    collapse = separator_between_seq
                )
            }
        }
    }

    # if(mixed){
    # remove the entry(s) from data and cell_seq_idx that are cells with only light chains
    # indx <- which(data[[v_call]] == "")
    # data <- data[-indx,]
    # cell_seq_idx <- cell_seq_idx[-indx]
    # }

    # one-to-one annotation-to-chain correspondence for both V and J (heavy)
    # for each cell/row, number of between_seq separators in heavy V annotation and in heavy J annotation must match
    # (in theory, there should be 1 heavy chain per cell; but 10x can return cell with >1 heavy chains and
    #  you never know if the user will supply this cell as input)
    n_separator_btw_seq_v_heavy <- stringi::stri_count_fixed(str = data[[v_call]], pattern = separator_between_seq)
    n_separator_btw_seq_j_heavy <- stringi::stri_count_fixed(str = data[[j_call]], pattern = separator_between_seq)
    if (any(n_separator_btw_seq_v_heavy != n_separator_btw_seq_j_heavy)) {
        stop("Requirement not met: one-to-one annotation-to-chain correspondence for both V and J (heavy)")
    }


    # NULL will disappear when doing c()
    # c(NULL,NULL) gives NULL still
    cols_for_grouping_heavy <- c(v_call, j_call, junc_len)
    cols_for_grouping_light <- c(v_call_light, j_call_light, junc_len_light)

    # Check NA(s) in columns
    bool_na <- rowSums(is.na(data[, c(cols_for_grouping_heavy, cols_for_grouping_light)])) > 0
    # Check for empty strings as well. They happen
    # if a single cell cell_id only has light chain data.
    # If not removed, the group identifiers can have numbering not correlative e.g G2, G3, and no G1.
    bool_empty <- rowSums(data[, c(cols_for_grouping_heavy, cols_for_grouping_light)] == "") > 0
    # consider empty as NA
    bool_na <- bool_na | bool_empty
    if (any(bool_na)) {
        entityName <- ifelse(single_cell, " cell(s)", " sequence(s)")
        msg <- paste0(
            "NA(s) found in one or more of { ",
            v_call, ", ", j_call,
            ifelse(is.null(junc_len), "", ", "), junc_len,
            ifelse(is.null(v_call_light), "", ", "), v_call_light,
            ifelse(is.null(j_call_light), "", ", "), j_call_light,
            ifelse(is.null(junc_len_light), "", ", "), junc_len_light,
            " } columns. ", sum(bool_na), entityName, " removed.\n"
        )
        warning(msg)
        data <- data[!bool_na, ]
        if (single_cell) {
            # maintain one-to-one relationship between
            # rows of data, cell_id_uniq, and cell_seq_idx
            cell_id_uniq <- cell_id_uniq[!bool_na]
            cell_seq_idx <- cell_seq_idx[!bool_na]
        }
    }

    ### expand

    # speed-up strategy
    # compute expanded VJL combos for unique rows
    # then distribute back to all rows

    # unique combinations of VJL
    # heavy chain seqs only
    if ((!single_cell) | (single_cell & only_heavy)) {
        combo_unique <- unique(data[, cols_for_grouping_heavy])

        # unique components
        v_unique <- unique(combo_unique[[v_call]])
        j_unique <- unique(combo_unique[[j_call]])

        # map each row in full data to unique combo
        m_v <- match(data[[v_call]], v_unique)
        m_j <- match(data[[j_call]], j_unique)

        # CGJ 4/15/24 put together instead of back to back steps for same condition
        if (is.null(junc_len)) {
            combo_unique_full_idx <- sapply(1:nrow(combo_unique), function(i) {
                idx_v <- which(v_unique == combo_unique[[v_call]][i])
                idx_j <- which(j_unique == combo_unique[[j_call]][i])
                idx <- which(m_v == idx_v & m_j == idx_j)
                return(idx)
            }, simplify = FALSE, USE.NAMES = FALSE)
            exp_lst <- sapply(1:nrow(combo_unique), function(i) {
                getAllVJL(
                    v = combo_unique[[v_call]][i], j = combo_unique[[j_call]][i],
                    l = NULL, first = first,
                    sep_anno = separator_within_seq, sep_chain = separator_between_seq
                )
            }, simplify = F, USE.NAMES = FALSE)
        } else {
            l_unique <- unique(combo_unique[[junc_len]])
            m_l <- match(data[[junc_len]], l_unique)
            combo_unique_full_idx <- sapply(1:nrow(combo_unique), function(i) {
                idx_v <- which(v_unique == combo_unique[[v_call]][i])
                idx_j <- which(j_unique == combo_unique[[j_call]][i])
                idx_l <- which(l_unique == combo_unique[[junc_len]][i])
                idx <- which(m_v == idx_v & m_j == idx_j & m_l == idx_l)
                return(idx)
            }, simplify = FALSE, USE.NAMES = FALSE)
            exp_lst <- sapply(1:nrow(combo_unique), function(i) {
                getAllVJL(
                    v = combo_unique[[v_call]][i], j = combo_unique[[j_call]][i],
                    l = combo_unique[[junc_len]][i], first = first,
                    sep_anno = separator_within_seq, sep_chain = separator_between_seq
                )
            }, simplify = F, USE.NAMES = FALSE)
        }
    }

    # one-to-one correspondence btw exp_lst and combo_unique_full_idx
    # exp_lst: VJL combinations
    # combo_unique_full_idx: rows in data carrying each exp_lst
    # exp_lst may not be all unique because gene-level info is kept instead of allele-level
    # make exp_lst unique

    exp_lst_uniq <- unique(exp_lst)
    exp_lst_uniq_full_idx <- sapply(exp_lst_uniq, function(x) {
        # wrt exp_lst, therefore also wrt combo_unique_full_idx
        idx_lst <- which(unlist(lapply(exp_lst, function(y) {
            length(y) == length(x) && all(y == x)
        })))
        # merge

        unlist(combo_unique_full_idx[idx_lst], use.names = FALSE)
    }, simplify = FALSE, USE.NAMES = FALSE)

    stopifnot(length(unique(unlist(exp_lst_uniq_full_idx, use.names = FALSE))) == nrow(data))
    # tip: unlist with use.names=F makes it much faster (>100x)
    # https://www.r-bloggers.com/speed-trick-unlist-use-namesfalse-is-heaps-faster/
    exp_uniq <- sort(unique(unlist(exp_lst_uniq, use.names = FALSE)))
    n_cells_or_seqs <- nrow(data)

    # notes on implementation

    # regular/dense matrix is more straightforward to implement but very costly memory-wise
    # sparse matrix is less straightforward to implement but way more memory efficient

    # sparse matrix is very slow to modify to on-the-fly (using a loop like for dense matrix)
    # way faster to construct in one go

    # (DO NOT DELETE)
    # for illustrating the concept
    # this is the way to go if using regular matrix (memory-intensive)
    # same concept implemented using sparse matrix

    # mtx_cell_VJL <- matrix(0, nrow=nrow(data), ncol=length(exp_uniq))
    # colnames(mtx_cell_VJL) <- exp_uniq
    #
    # mtx_adj <- matrix(0, nrow=length(exp_uniq), ncol=length(exp_uniq))
    # rownames(mtx_adj) <- exp_uniq
    # colnames(mtx_adj) <- exp_uniq
    #
    # outdated:
    # for (i_cell in 1:length(exp_lst)) {
    #     #if (i_cell %% 1000 == 0) { cat(i_cell, "\n") }
    #     cur_uniq <- unique(exp_lst[[i_cell]])
    #     mtx_cell_VJL[i_cell, cur_uniq] <- 1
    #     mtx_adj[cur_uniq, cur_uniq] <- 1
    # }

    # actual implementation using sparse matrix from Matrix package

    ### matrix indicating relationship between cell and VJ(L) combinations
    # row: cell
    # col: unique heavy VJ(L) (and light VJ(L))

    # row indices
    m1_i <- lapply(1:length(exp_lst_uniq), function(i) {
        rep(exp_lst_uniq_full_idx[[i]], each = length(exp_lst_uniq[[i]]))
    })
    m1_i_v <- unlist(m1_i, use.names = FALSE)

    # column indices
    m1_j <- lapply(1:length(exp_lst_uniq), function(i) {
        # wrt exp_uniq
        idx <- match(exp_lst_uniq[[i]], exp_uniq)
        # stopifnot( all.equal( exp_uniq[idx], exp_lst_uniq[[i]] ) )

        rep.int(idx, length(exp_lst_uniq_full_idx[[i]]))
    })
    m1_j_v <- unlist(m1_j, use.names = FALSE)

    stopifnot(length(m1_i_v) == length(m1_j_v))

    # no particular need for this to be not of class "nsparseMatrix"
    # so no need to specify x=rep(1, length(m1_i))
    # not specifying makes it even more space-efficient
    mtx_cell_VJL <- Matrix::sparseMatrix(
        i = m1_i_v, j = m1_j_v,
        dims = c(n_cells_or_seqs, length(exp_uniq)),
        symmetric = F, triangular = F, index1 = T,
        dimnames = list(NULL, exp_uniq)
    )

    ### adjacency matrix
    # row and col: unique heavy VJ(L) (and light VJ(L))

    # row indices
    m2_i <- lapply(1:length(exp_lst_uniq), function(i) {
        # wrt exp_uniq
        idx <- match(exp_lst_uniq[[i]], exp_uniq)
        # stopifnot( all.equal( exp_uniq[idx], exp_lst_uniq[[i]] ) )

        rep(idx, each = length(exp_lst_uniq[[i]]))
    })
    m2_i_v <- unlist(m2_i, use.names = FALSE)

    # col indices
    m2_j <- lapply(1:length(exp_lst_uniq), function(i) {
        # wrt exp_uniq
        idx <- match(exp_lst_uniq[[i]], exp_uniq)
        # stopifnot( all.equal( exp_uniq[idx], exp_lst_uniq[[i]] ) )

        rep.int(idx, length(exp_lst_uniq[[i]]))
    })
    m2_j_v <- unlist(m2_j, use.names = FALSE)

    stopifnot(length(m2_i_v) == length(m2_j_v))

    # important: x must be specified for mtx_adj in order to make it not of class "nsparseMatrix"
    # this is because igraph accepts sparse matrix from Matrix but not the "pattern" matrices variant
    mtx_adj <- Matrix::sparseMatrix(
        i = m2_i_v, j = m2_j_v, x = rep(1, length(m2_i_v)),
        dims = c(length(exp_uniq), length(exp_uniq)),
        symmetric = F, triangular = F, index1 = T,
        dimnames = list(exp_uniq, exp_uniq)
    )

    rm(m1_i, m1_j, m2_i, m2_j, m1_i_v, m1_j_v, m2_i_v, m2_j_v)

    ### identify connected components based on adjcencey matrix
    # this is the grouping
    # source: https://stackoverflow.com/questions/35772846/obtaining-connected-components-in-r

    # Create igraph object from adjacency matrix
    if (sum(rowSums(mtx_adj) > 0) == nrow(mtx_adj)) {
        # This "if" is to ignore a warning in the special case
        # that all sequences are isolated (only diagonal elements).
        # When mtx_adj is a diagonal matrix and diag=FALSE (no self-links), 
        # the diagonal will be zeroed out and max, used somewhere in the igraph function, 
        # will return an error like "no non-missing arguments to max"
        withCallingHandlers(
            g <- igraph::graph_from_adjacency_matrix(adjmatrix = mtx_adj, mode = "undirected", diag = FALSE),
            warning = function(w) {
                if (grepl("no non-missing arguments to max", w$message)) {
                    invokeRestart("muffleWarning")
                }
            }
        )
    } else {
        g <- igraph::graph_from_adjacency_matrix(adjmatrix = mtx_adj, mode = "undirected", diag = FALSE)
    }
    # plot(g, vertex.size=10, vertex.label.cex=1, vertex.color="skyblue", vertex.label.color="black", vertex.frame.color="transparent", edge.arrow.mode=0)


    connected <- igraph::components(g)
    VJL_groups <- igraph::groups(connected)
    names(VJL_groups) <- paste0("G", 1:length(VJL_groups))

    ### identify cells associated with each connected component (grouping)

    # each entry corresponds to a group/partition
    # each element within an entry is a cell

    cellIdx_byGroup_lst <- lapply(VJL_groups, function(x) {
        if (length(x) > 1) {
            # matrix
            # important to specify rowSums from Matrix package
            # base::rowSums will NOT work
            cell_idx <- which(Matrix::rowSums(mtx_cell_VJL[, x, drop = F]) > 0)
        } else {
            # vector
            cell_idx <- which(mtx_cell_VJL[, x] > 0)
        }
        return(cell_idx)
    })

    # sanity check: there should be perfect/disjoint partitioning
    # (each cell has exactly one group assignment)
    stopifnot(n_cells_or_seqs == length(unique(unlist(cellIdx_byGroup_lst, use.names = FALSE))))

    # assign
    data$vj_group <- NA
    for (i in 1:length(cellIdx_byGroup_lst)) {
        data[["vj_group"]][cellIdx_byGroup_lst[[i]]] <- names(VJL_groups)[i]
    }
    stopifnot(!any(is.na(data[["vj_group"]])))

    if (!single_cell) {
        return(data)
    } else {
        data_orig$vj_group <- NA

        # map back to data_orig
        if (!mixed) {
            for (i_cell in 1:nrow(data)) {
                # wrt data_orig
                i_orig_h <- cell_seq_idx[[i_cell]][["heavy"]]
                i_orig_l <- cell_seq_idx[[i_cell]][["light"]]
                # sanity check
                stopifnot(all(data_orig[[cell_id]][c(i_orig_h, i_orig_l)] == cell_id_uniq[i_cell]))
                # grouping
                data_orig$vj_group[c(i_orig_h, i_orig_l)] <- data$vj_group[i_cell]
            }
        } else {
            for (i_cell in 1:nrow(data)) {
                i_orig_h <- cell_seq_idx[[i_cell]][["heavy"]]
                i_orig_l <- cell_seq_idx[[i_cell]][["light"]]

                # if it's just the light chain give in an NA and skip
                if (length(i_orig_h) == 0 & length(i_orig_l) == 1) {
                    data_orig$vj_group[i_orig_l] <- NA
                    next
                }

                # sanity check
                # both chains are present in the cell
                # ssnn: TODO BUG This sanity check doesn't work for mixed data
                # Error in if (!is.na(data_orig[[cell_id]][i_orig_h]) & length(i_orig_l) !=  :
                #              the condition has length > 1
                # Debug shows:
                # !is.na(data_orig[[cell_id]][i_orig_h])
                # [1] FALSE FALSE
                # Browse[1]> data
                #   cell_id                           v_call            j_call vj_group
                # 1       1                       IGHV1-1*01          IGHJ2*01       G2
                # 2       2                       IGHV1-1*01          IGHJ1*01       G1
                # 3       3                       IGHV1-2*01          IGHJ1*01       G1
                # 4    <NA> IGHV1-1*01,IGHV1-2*01;IGHV1-2*01 IGHJ1*01;IGHJ1*01       G1
                # Browse[1]> data_orig
                #   subject_id                v_call   j_call     junction locus cell_id junction_length vj_group
                # 1         S1            IGHV1-1*01 IGHJ2*01 TGTAAAAAATGG   IGH       1              12       G2
                # 2         S1            IGHV1-1*01 IGHJ1*01 TGTAAAAAATGG   IGH       2              12       G1
                # 3         S1            IGHV1-2*01 IGHJ1*01 TGTAAAACCTGG   IGH       3              12       G1
                # 4         S1 IGHV1-1*01,IGHV1-2*01 IGHJ1*01 TGTAAACCCTGG   IGH    <NA>              12     <NA>
                # 5         S1            IGHV1-2*01 IGHJ1*01 TGTAAACCCTGG   IGH    <NA>              12     <NA>
                # 6         S1            IGKV1-1*01 IGKJ1*01 TGTCCCCCCTGG   IGK       1              12       G2
                # 7         S1            IGKV1-1*01 IGKJ1*01 TGTCCCCCCTGG   IGK    <NA>              12     <NA>
                # account for the NA cell_ids length being longer than 1
                for (i in 1:length(i_orig_h)) {
                    i_orig_h_temp <- i_orig_h[i]
                    if (!is.na(data_orig[[cell_id]][i_orig_h_temp]) & length(i_orig_l) != 0) {
                        # sanity check
                        stopifnot(all(data_orig[[cell_id]][c(i_orig_h_temp, i_orig_l)] == cell_id_uniq[i_cell]))
                    } else {
                        # just the heavy chain
                        if (is.na(data_orig[[cell_id]][i_orig_h_temp])) {
                            stopifnot(is.na(cell_id_uniq[i_cell]))
                        } else {
                            stopifnot(data_orig[[cell_id]][i_orig_h_temp] == cell_id_uniq[i_cell])
                        }
                    }
                }
                # grouping
                data_orig$vj_group[c(i_orig_h, i_orig_l)] <- data$vj_group[i_cell]
            }
        }

        return(data_orig)
    }
}

#' Sort V(D)J genes
#'
#' \code{sortGenes} sorts a vector of V(D)J gene names by either lexicographic ordering
#' or locus position.
#'
#' @param    genes    vector of strings representing V(D)J gene names.
#' @param    method   string defining the method to use for sorting genes. One of:
#'                    \itemize{
#'                      \item \code{"name"}:      sort in lexicographic order. Order is by
#'                                                family first, then gene, and then allele.
#'                      \item \code{"position"}:  sort by position in the locus, as
#'                                                determined by the final two numbers
#'                                                in the gene name. Non-localized genes
#'                                                are assigned to the highest positions.
#'                    }
#'
#' @return   A sorted character vector of gene names.
#'
#' @seealso  See \code{getAllele}, \code{getGene} and \code{getFamily} for parsing
#'           gene names.
#'
#' @examples
#' # Create a list of allele names
#' genes <- c(
#'     "IGHV1-69D*01", "IGHV1-69*01", "IGHV4-38-2*01", "IGHV1-69-2*01",
#'     "IGHV2-5*01", "IGHV1-NL1*01", "IGHV1-2*01,IGHV1-2*05",
#'     "IGHV1-2", "IGHV1-2*02", "IGHV1-69*02"
#' )
#'
#' # Sort genes by name
#' sortGenes(genes)
#'
#' # Sort genes by position in the locus
#' sortGenes(genes, method = "pos")
#'
#' @export
sortGenes <- function(genes, method = c("name", "position")) {
    ## DEBUG
    # method="name"

    # Check arguments
    method <- match.arg(method)

    # Build sorting table
    sort_tab <- tibble(CALL = sort(getAllele(genes, first = FALSE, strip_d = FALSE))) %>%
        # Determine the gene and family
        mutate(
            FAMILY = getFamily(!!rlang::sym("CALL"), first = TRUE, strip_d = FALSE),
            GENE = getGene(!!rlang::sym("CALL"), first = TRUE, strip_d = FALSE),
            ALLELE = getAllele(!!rlang::sym("CALL"), first = TRUE, strip_d = FALSE)
        ) %>%
        # Identify first gene number, second gene number and allele number
        mutate(
            G1 = gsub("[^-]+-([^-\\*D]+).*", "\\1", !!rlang::sym("GENE")),
            G1 = as.numeric(gsub("[^0-9]+", "99", !!rlang::sym("G1"))),
            G2 = gsub("[^-]+-[^-]+-?", "", !!rlang::sym("GENE")),
            G2 = as.numeric(gsub("[^0-9]+", "99", !!rlang::sym("G2"))),
            A1 = as.numeric(sub("[^\\*]+\\*|[^\\*]+$", "", !!rlang::sym("ALLELE")))
        )

    # Convert missing values to 0
    sort_tab[is.na(sort_tab)] <- 0

    # Sort
    if (method == "name") {
        sorted_genes <- arrange(sort_tab, !!!rlang::syms(c("FAMILY", "G1", "G2", "A1")))[["CALL"]]
    } else if (method == "position") {
        sorted_genes <- arrange(
            sort_tab,
            desc(!!rlang::sym("G1")),
            desc(!!rlang::sym("G2")),
            !!rlang::sym("FAMILY"),
            !!rlang::sym("A1")
        )[["CALL"]]
    }

    return(sorted_genes)
}

ExampleDb <- file.path("..", "data-tests", "ExampleDb.gz")
db <- readChangeoDb(ExampleDb)
expect_warning(
    db_gg <- readChangeoDb(file.path("..", "data-tests", "db_test.tsv")),
    "db_test.tsv is not in the Change-O format"
)

### countGenes ####

test_that("countGenes", {
    # Without copy numbers
    genes <- countGenes(db, gene = "V_CALL", groups = "SAMPLE", mode = "family")
    expect_equal(genes$seq_freq,
        c(0.76, 0.87, 0.24, 0.05, 0.05, 0.02, 0.01),
        tolerance = 0.001
    )

    genes <- countGenes(db, gene = "V_CALL", groups = "SAMPLE", mode = "gene")
    expect_equal(genes$seq_freq,
        c(
            0.41, 0.35, 0.24, 0.33, 0.26, 0.11, 0.10, 0.05, 0.05,
            0.04, 0.02, 0.02, 0.01, 0.01
        ),
        tolerance = 0.001
    )

    genes <- countGenes(db, gene = "V_CALL", groups = "SAMPLE", mode = "allele")
    expect_equal(genes$seq_freq,
        c(
            0.350, 0.325, 0.215, 0.330, 0.085, 0.150, 0.100, 0.100, 0.070,
            0.050, 0.050, 0.025, 0.040, 0.030, 0.020, 0.020,
            0.010, 0.010, 0.010, 0.010
        ),
        tolerance = 0.01
    )

    # With copy numbers and multiple groups
    genes <- countGenes(db,
        gene = "V_CALL", groups = c("SAMPLE", "ISOTYPE"),
        copy = "DUPCOUNT", mode = "family"
    )
    expect_equal(genes$seq_freq,
        c(
            0.78, 0.95, 0.95, 0.22, 0.63, 0.37, 0.60, 0.20, 0.40, 0.67,
            0.60, 0.80, 0.40, 0.33, 0.03, 0.05, 0.01, 0.01
        ),
        tolerance = 0.01
    )

    genes <- countGenes(db,
        gene = "V_CALL", groups = c("SAMPLE", "ISOTYPE"),
        copy = "DUPCOUNT", mode = "gene"
    )
    expect_equal(round(genes$seq_freq, 2)[1:12],
        c(
            0.61, 0.75, 0.22, 0.54, 0.40, 0.37, 0.17,
            0.20, 0.15, 0.20, 0.60, 0.20
        ),
        tolerance = 0.01
    )

    genes <- countGenes(db,
        gene = "V_CALL", groups = c("SAMPLE", "ISOTYPE"),
        copy = "DUPCOUNT", mode = "allele"
    )
    expect_equal(genes$seq_freq[1:12],
        c(0.61, 0.72, 0.16, 0.41, 0.40, 0.37, 0.06, 0.20, 0.08, 0.20, 0.50, 0.08),
        tolerance = 0.01
    )

    # Testing of fill

    # Without copy numbers
    genes <- countGenes(db, gene = "V_CALL", groups = "SAMPLE", mode = "family", fill = T)
    expect_equal(genes$seq_freq,
        c(0.02, 0.87, 0.05, 0.05, 0.01, 0, 0.76, 0, 0.24, 0),
        tolerance = 0.001
    )

    genes <- countGenes(db, gene = "V_CALL", groups = "SAMPLE", mode = "gene", fill = T)
    expect_equal(genes$seq_freq,
        c(
            0.02, 0.04, 0.01, 0.1, 0.11, 0.26, 0.33, 0.02, 0, 0.05, 0.05, 0.01, 0, 0,
            0, 0, 0.41, 0, 0, 0, 0.35, 0, 0.24, 0
        ),
        tolerance = 0.001
    )

    genes <- countGenes(db, gene = "V_CALL", groups = "SAMPLE", mode = "allele", fill = T)
    expect_equal(genes$seq_freq,
        c(
            0.02, 0.01, 0.03, 0.01, 0.1, 0.07, 0.04, 0.01, 0.1, 0.15, 0.33, 0.02, 0,
            0.05, 0.05, 0, 0.01, 0, 0, 0, 0, 0, 0.085, 0.325, 0, 0, 0, 0, 0, 0.35, 0, 0.215, 0.025, 0
        ),
        tolerance = 0.01
    )

    # With copy numbers and multiple groups
    genes <- countGenes(db,
        gene = "V_CALL", groups = c("SAMPLE", "ISOTYPE"),
        copy = "DUPCOUNT", mode = "family", fill = T
    )
    expect_equal(round(genes$seq_freq, 2)[1:12],
        c(0, 0.8, 0, 0.2, 0, 0, 0.6, 0.4, 0, 0, 0, 0.6),
        tolerance = 0.01
    )

    genes <- countGenes(db,
        gene = "V_CALL", groups = c("SAMPLE", "ISOTYPE"),
        copy = "DUPCOUNT", mode = "gene", fill = T
    )
    expect_equal(round(genes$seq_freq, 2)[1:12],
        c(0, 0, 0, 0.2, 0, 0, 0.6, 0, 0, 0, 0.2, 0),
        tolerance = 0.01
    )

    genes <- countGenes(db,
        gene = "V_CALL", groups = c("SAMPLE", "ISOTYPE"),
        copy = "DUPCOUNT", mode = "allele", fill = T
    )
    expect_equal(genes$seq_freq[1:12],
        c(0, 0, 0, 0, 0.2, 0, 0, 0, 0, 0, 0.6, 0),
        tolerance = 0.01
    )

    # Test how NAs are handled
    db_some_na <- data.frame(
        sequence_id = c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10),
        sample = c("S1", "S2", "S4", "S3", "S2", "S3", "S4", "S4", "S3", "S4"),
        v_call = c("IGHV1-1", NA, "IGHV1-1,IGHV1-2", "IGHV1-2,IGHV1-3", "IGHV1-2", "IGHV1-3", "IGHV1-3", "IGHV1-1,IGHV1-2", "IGHV1-2", NA),
        dupcount = c(1, 1, 1, 1, 1, 1, 1, 1, 1, 1),
        clone = c(1, 2, 2, 3, 3, 3, 4, 4, 4, 4),
        stringsAsFactors = F
    )

    expect_warning(genes <- countGenes(db_some_na, gene = "v_call", mode = "gene"), NULL)
    expect_equal(genes$seq_freq,
        c(0.375, 0.375, 0.25),
        tolerance = 0.001
    )

    genes <- countGenes(db_some_na, gene = "v_call", mode = "gene", remove_na = FALSE)
    # A locus cannot be determined if the gene is NA. When determining the denominator (locus_count),
    # these sequences will be considered separately as their own "NA" locus.
    # Example:
    #  locus gene    seq_count locus_count seq_freq
    # <chr> <chr>       <int>       <int>    <dbl>
    # 1 IGH   IGHV1-1         3           8    0.375
    # 2 IGH   IGHV1-2         3           8    0.375
    # 3 IGH   IGHV1-3         2           8    0.25
    # 4 NA    NA              2           2    1
    expect_equal(genes$seq_freq,
        c(0.375, 0.375, 0.25, 1),
        tolerance = 0.001
    )

    genes <- countGenes(db_some_na, gene = "v_call", mode = "gene", copy = "clone", remove_na = FALSE)
    expect_equal(genes$copy_count,
        c(10, 7, 7, 6),
        tolerance = 0.001
    )

    genes <- countGenes(db_some_na, gene = "v_call", mode = "gene", groups = "sample", fill = TRUE, remove_na = FALSE)
    expect_equal(genes$seq_count,
        c(1, 0, 0, 0, 0, 1, 0, 1, 0, 2, 1, 0, 2, 0, 1, 1),
        tolerance = 0.001
    )

    db_all_na <- data.frame(
        sequence_id = c(1, 2, 3),
        sample = c("S1", "S2", "S2"),
        v_call = c(NA, NA, NA),
        dupcount = c(1, 1, 1),
        clone = c(1, 1, 1),
        stringsAsFactors = F
    )

    expect_warning(genes <- countGenes(db_all_na, gene = "v_call", mode = "gene"), "The column v_call contains no data")
    expect_equal(nrow(genes),
        0,
        tolerance = 0.001
    )

    expect_warning(genes <- countGenes(db_all_na, gene = "v_call", mode = "gene", remove_na = FALSE), "The column v_call contains no data")
    expect_equal(genes$seq_count,
        3,
        tolerance = 0.001
    )

    expect_warning(
        genes <- countGenes(db_all_na, gene = "v_call", mode = "gene", groups = "sample", copy = "dupcount", fill = TRUE, remove_na = FALSE),
        "The column v_call contains no data"
    )
    expect_equal(genes$copy_count,
        c(1, 2),
        tolerance = 0.001
    )

    # testing addition of collapse and first flags

    genes <- countGenes(db, gene = "J_CALL", mode = "allele", first = FALSE)
    expect_equal(genes$seq_freq,
        c(0.283, 0.213, 0.183, 0.133, 0.123, 0.0233, 0.020, 0.010, 0.010),
        tolerance = 0.001
    )

    genes <- countGenes(db, gene = "V_CALL", mode = "gene", first = FALSE, collapse = FALSE)
    expect_equal(genes$seq_freq,
        c(
            0.31, 0.233, 0.17, 0.11, 0.0867, 0.0333, 0.0167, 0.0133, 0.00667,
            0.00667, 0.00667, 0.0033, 0.0033
        ),
        tolerance = 0.001
    )

    # test bulk percentages now being by locus
    # need data with mixed loci
    db_hl <- data.frame(
        sequence_id = c(1:30),
        sample = c(rep("S1", 15), rep("S2", 15)),
        v_call = c(
            "IGHV1-2,IGHV1-3", "IGHV1-2", "IGHV1-2", "IGHV1-3", "IGHV1-3", "IGHV1-3", "IGHV1-3", "IGHV2-1,IGHV2-2", "IGHV2-2", "IGHV2-2",
            "IGKV1-1", "IGKV1-2", "IGKV1-1,IGKV1-2", "IGLV1-1", "IGLV2-1",
            "IGHV1-2", "IGHV1-2", "IGHV3-1", "IGHV3-1", "IGHV2-1,IGHV2-2", "IGHV2-1", "IGHV2-2", "IGHV2-2", "IGHV2-2", "IGHV3-1",
            "IGKV1-1", "IGKV1-2", "IGKV2-1", "IGLV1-1", "IGLV1-1"
        ),
        j_call = c(
            "IGHJ1", "IGHJ1", "IGHJ1", "IGHJ1", "IGHJ2", "IGHJ2", "IGHJ2", "IGHJ2", "IGHJ2", "IGHJ2",
            "IGKJ1", "IGKJ1", "IGKJ1", "IGLJ1", "IGLJ1",
            "IGHJ1", "IGHJ1", "IGHJ1", "IGHJ1", "IGHJ2", "IGHJ2", "IGHJ2", "IGHJ2", "IGHJ2", "IGHJ2",
            "IGKJ1", "IGKJ2", "IGKJ2", "IGLJ1", "IGLJ1"
        ),
        dupcount = c(
            1, 1, 5, 1, 1, 1, 5, 1, 5, 1, 1, 4, 1, 4, 1,
            1, 1, 1, 1, 1, 2, 2, 4, 4, 1, 1, 4, 1, 4, 1
        ),
        clone = c(
            1, 1, 1, 1, 2, 2, 2, 3, 3, 3, NA, NA, NA, NA, NA,
            4, 4, 5, 5, 6, 6, 6, 6, 6, 7, NA, NA, NA, NA, NA
        ),
        stringsAsFactors = F
    )

    genes <- countGenes(db_hl, gene = "v_call", groups = "sample", mode = "family")
    expect_equal(genes$seq_freq,
        c(
            0.7, 0.5, 0.3, 1, 0.3, 0.2, 0.667, 1, 0.5, 0.5, 0.333
        ),
        tolerance = 0.001
    )

    genes <- countGenes(db_hl, gene = "v_call", groups = "sample", mode = "gene")
    expect_equal(genes$seq_freq,
        c(
            0.4, 0.3, 0.3, 0.3, 0.2, 0.667, 0.2, 0.2, 1, 0.1, 0.333, 0.5, 0.5, 0.333, 0.333, 0.333
        ),
        tolerance = 0.001
    )

    genes <- countGenes(db_hl, gene = "j_call", groups = "sample", mode = "gene")
    expect_equal(genes$seq_freq,
        c(
            0.6, 0.6, 0.4, 0.4, 1, 1, 0.667, 1, 0.333
        ),
        tolerance = 0.001
    )

    genes <- countGenes(db_hl, gene = "v_call", groups = "sample", mode = "gene", copy = "dupcount")
    expect_equal(genes$copy_freq,
        c(
            0.556, 0.364, 0.318, 0.273, 1, 0.667, 0.8, 0.667, 0.167, 0.167, 0.333, 0.111, 0.045, 0.2, 0.167, 0.167
        ),
        tolerance = 0.001
    )

    genes <- countGenes(db_hl, gene = "j_call", groups = "sample", mode = "gene", copy = "dupcount")
    expect_equal(genes$copy_freq,
        c(
            0.636, 0.778, 0.364, 1, 1, 0.833, 1, 0.222, 0.167
        ),
        tolerance = 0.001
    )

    # case with all clone IDs provided
    genes <- countGenes(db_hl %>% dplyr::filter(!is.na(clone)), gene = "v_call", groups = "sample", mode = "gene", clone = "clone")
    expect_equal(genes$clone_freq,
        c(
            0.5, 0.333, 0.333, 0.333, 0.25, 0.25
        ),
        tolerance = 0.001
    )

    # case with some clone IDs provided
    expect_warning(
        genes <- countGenes(db_hl, gene = "v_call", groups = "sample", mode = "gene", clone = "clone"),
        "Found 10 sequences without clonal assignments. These sequences will be removed before counting clone frequencies."
    )
    expect_equal(genes$clone_freq,
        c(
            0.5, 0.333, 0.333, 0.333, 0.25, 0.25
        ),
        tolerance = 0.001
    )

    expect_warning(
        genes <- countGenes(db_hl, gene = "j_call", groups = "sample", mode = "gene", clone = "clone"),
        "Found 10 sequences without clonal assignments. These sequences will be removed before counting clone frequencies."
    )
    expect_equal(genes$clone_freq,
        c(
            0.667, 0.5, 0.5, 0.333
        ),
        tolerance = 0.001
    )

    # test if all NA
    db_hl$clone_na <- rep(NA, nrow(db_hl))
    expect_error(
        genes <- countGenes(db_hl, gene = "v_call", groups = "sample", mode = "gene", clone = "clone_na"),
        "No clone IDs are present in the data."
    )
})

test_that("countGenes works for single-cell", {
    # Without copy numbers
    expect_warning(genes <- countGenes(db_gg, gene = "v_call", groups = "subject_id", mode = "family"), "Mixed bulk and single-cell data")
    expect_equal(round(genes$seq_freq, 2),
        c(0.83, 0.50, 0.11, 0.11, 0.06, 0.06),
        tolerance = 0.001
    )

    expect_warning(genes <- countGenes(db_gg, gene = "v_call", groups = "subject_id", mode = "gene"), "Mixed bulk and single-cell data")
    expect_equal(round(genes$seq_freq, 2),
        c(0.83, 0.50, 0.11, 0.11, 0.06, 0.06),
        tolerance = 0.001
    )

    added_cell <- data.frame(
        subject_id = c("S2", "S2", "S2"),
        v_call = c("IGHV1-1", "IGKV1-1", "IGKV1-1"),
        j_call = c("IGHJ1-1", "IGKJ1-1", "IGKJ1-1"),
        cell_id = c(30, 30, 30)
    )

    # Specific use case for identical light V gene within 1 cell that should be counted only once
    db_gg_multilight <- db_gg %>%
        dplyr::select(c("subject_id", "v_call", "j_call", "cell_id")) %>%
        dplyr::bind_rows(db_gg, added_cell)
    expect_warning(genes <- countGenes(db_gg_multilight, gene = "v_call", groups = "subject_id", mode = "gene"), "Mixed bulk and single-cell data")
    expect_equal(genes$seq_freq,
        c(0.8, 0.5, 0.1, 0.1, 0.05, 0.05, 1, 1),
        tolerance = 0.001
    )
})

### getSegment ####

test_that("getSegment", {
    # Light chain tests
    kappa_call <- c(
        "Homsap IGKV1D-39*01 F,Homsap IGKV1-39*02 F,Homsap IGKV1-39*01",
        "Homsap IGKJ5*01 F"
    )

    expect_equal(
        getAllele(kappa_call),
        c("IGKV1-39*01", "IGKJ5*01")
    )
    expect_equal(
        getAllele(kappa_call, first = FALSE),
        c("IGKV1-39*01,IGKV1-39*02", "IGKJ5*01")
    )
    expect_equal(
        getAllele(kappa_call, first = FALSE, strip_d = FALSE),
        c("IGKV1D-39*01,IGKV1-39*02,IGKV1-39*01", "IGKJ5*01")
    )

    expect_equal(
        getGene(kappa_call),
        c("IGKV1-39", "IGKJ5")
    )
    expect_equal(
        getGene(kappa_call, first = FALSE),
        c("IGKV1-39", "IGKJ5")
    )
    expect_equal(
        getGene(kappa_call, first = FALSE, strip_d = FALSE),
        c("IGKV1D-39,IGKV1-39", "IGKJ5")
    )
    expect_equal(
        getGene(kappa_call, first = FALSE, strip_d = TRUE),
        c("IGKV1-39", "IGKJ5")
    )

    expect_equal(
        getFamily(kappa_call),
        c("IGKV1", "IGKJ5")
    )
    expect_equal(
        getFamily(kappa_call, first = FALSE),
        c("IGKV1", "IGKJ5")
    )
    expect_equal(
        getFamily(kappa_call, first = FALSE, collapse = FALSE),
        c("IGKV1,IGKV1,IGKV1", "IGKJ5")
    )
    expect_equal(
        getFamily(kappa_call, first = FALSE, strip_d = FALSE),
        c("IGKV1D,IGKV1", "IGKJ5")
    )

    expect_equal(
        getLocus(kappa_call, first = FALSE, strip_d = FALSE, collapse = TRUE),
        c("IGK", "IGK")
    )
    expect_equal(
        getLocus(kappa_call, first = FALSE, strip_d = FALSE, collapse = FALSE),
        c("IGK,IGK,IGK", "IGK")
    )

    expect_equal(
        getChain(kappa_call, first = FALSE, strip_d = FALSE, collapse = TRUE),
        c("VL", "VL")
    )
    expect_equal(
        getChain(kappa_call, first = FALSE, strip_d = FALSE, collapse = FALSE),
        c("VL,VL,VL", "VL")
    )

    # Heavy chain tests
    heavy_call <- c(
        "Homsap IGHV1-69*01 F,Homsap IGHV1-69D*01 F",
        "Homsap IGHD1-1*01 F",
        "Homsap IGHJ1*01 F"
    )

    expect_equal(
        getAllele(heavy_call, first = FALSE),
        c("IGHV1-69*01", "IGHD1-1*01", "IGHJ1*01")
    )
    expect_equal(
        getAllele(heavy_call, first = FALSE, strip_d = FALSE),
        c("IGHV1-69*01,IGHV1-69D*01", "IGHD1-1*01", "IGHJ1*01")
    )

    expect_equal(
        getGene(heavy_call, first = FALSE),
        c("IGHV1-69", "IGHD1-1", "IGHJ1")
    )
    expect_equal(
        getGene(heavy_call, first = FALSE, strip_d = FALSE),
        c("IGHV1-69,IGHV1-69D", "IGHD1-1", "IGHJ1")
    )
    expect_equal(
        getGene(heavy_call, first = FALSE, strip_d = TRUE),
        c("IGHV1-69", "IGHD1-1", "IGHJ1")
    )

    expect_equal(
        getLocus(heavy_call, first = FALSE, strip_d = FALSE, collapse = TRUE),
        c("IGH", "IGH", "IGH")
    )
    expect_equal(
        getLocus(heavy_call, first = FALSE, strip_d = FALSE, collapse = FALSE),
        c("IGH,IGH", "IGH", "IGH")
    )

    expect_equal(
        getChain(heavy_call, first = FALSE, strip_d = FALSE, collapse = TRUE),
        c("VH", "VH", "VH")
    )
    expect_equal(
        getChain(heavy_call, first = FALSE, strip_d = FALSE, collapse = FALSE),
        c("VH,VH", "VH", "VH")
    )

    # Filtering non-localized genes
    nl_call <- c(
        "IGHV3-NL1*01,IGHV3-30-3*01,IGHV3-30*01",
        "Homosap IGHV2-30*01 F,Homsap IGHV3-NL1*01 F",
        "IGHV1-NL1*01"
    )

    expect_equal(
        getAllele(nl_call, first = FALSE, omit_nl = TRUE),
        c("IGHV3-30-3*01,IGHV3-30*01", "IGHV2-30*01", "")
    )

    expect_equal(
        getGene(nl_call, first = FALSE, omit_nl = TRUE),
        c("IGHV3-30-3,IGHV3-30", "IGHV2-30", "")
    )

    expect_equal(
        getFamily(nl_call, first = FALSE, omit_nl = TRUE),
        c("IGHV3", "IGHV2", "")
    )

    expect_equal(
        getLocus(nl_call, first = FALSE, omit_nl = TRUE),
        c("IGH", "IGH", "")
    )

    expect_equal(
        getChain(nl_call, first = FALSE, omit_nl = TRUE),
        c("VH", "VH", "")
    )

    # Test for issue found in TIgGER
    # If there's no allele info,
    # getAllele should return the same input value
    # Before the change in allele_regex used by getAllele,
    # getAllele("TRAV38-2/DV8") returned TRAV38-2
    # which in TIgGER introduced NAs and an error, as TRAV38-2 was not a valid name
    # for the list of alleles, the valid name was TRAV38-2/DV8.
    # getAllele("IGHV1-01") returned IGHV1-01
    getAllele(getGene("IGHV1-01*02")) == getGene("IGHV1-01*02")
    getAllele(getGene("TRAV38-2/DV8*01")) == getGene("TRAV38-2/DV8*01")

    # Test that getSegment works well with c_call and strip_d=TRUE
    # and it doesn't convert IGHD into IGH
    expect_equal(getGene("IGHD"), "IGHD")
    expect_equal(getAllele("IGHD"), "IGHD")
    gene <- countGenes(data.frame("c_call" = "IGHD", stringsAsFactors = F), gene = "c_call", mode = "gene")
    asis <- countGenes(data.frame("c_call" = "IGHD", stringsAsFactors = F), gene = "c_call", mode = "asis")
    expect_equal(gene[["gene"]], "IGHD")
    expect_equal(asis[["gene"]], "IGHD")

    # Test for IMGT temporary designation nomenclature
    tmp_call <- c("IGHV9S3*01", "IGKV10S12*01")
    expect_equal(
        getAllele(tmp_call),
        c("IGHV9S3*01", "IGKV10S12*01")
    )
    expect_equal(
        getGene(tmp_call),
        c("IGHV9S3", "IGKV10S12")
    )
    expect_equal(
        getFamily(tmp_call),
        c("IGHV9", "IGKV10")
    )


    # Test Mus musculus
    # ADEGMC
    mm_alleles <- c(
        "IGHA*01",
        "IGHD*01", "IGHD5-1-1*01",
        "IGHE*01",
        "IGHG1*01", "IGHG2A*01",
        "IGHM*01",
        "IGKC*01"
    )
    mm_genes <- c(
        "IGHA",
        "IGHD", "IGHD5-1-1",
        "IGHE",
        "IGHG1", "IGHG2A",
        "IGHM",
        "IGKC"
    )
    mm_families <- c(
        "IGHA",
        "IGHD", "IGHD5",
        "IGHE",
        "IGHG1", "IGHG2A",
        "IGHM",
        "IGKC"
    )
    expect_equal(getGene(mm_alleles), mm_genes)
    expect_equal(getFamily(mm_alleles), mm_families)


    # Test for TCR
    expect_equal(
        getGene(c("TRBV29-1*01", "TRAV12-1*01")),
        c("TRBV29-1", "TRAV12-1")
    )
    expect_equal(
        getFamily(c("TRBV29-1*01", "TRAV12-1*01")),
        c("TRBV29", "TRAV12")
    )
})

### sortGenes ####

test_that("sortGenes", {
    genes <- c(
        "IGHV1-69D*01", "IGHV1-69*01", "IGHV4-38-2*01", "IGHV1-69-2*01",
        "IGHV2-5*01", "IGHV1-NL1*01", "IGHV1-2*01,IGHV1-2*05",
        "IGHV1-2", "IGHV1-2*02", "IGHV1-69*02"
    )
    # Sort genes by name
    expected <- c(
        "IGHV1-2", "IGHV1-2*01,IGHV1-2*05", "IGHV1-2*02",
        "IGHV1-69*01", "IGHV1-69D*01", "IGHV1-69*02", "IGHV1-69-2*01",
        "IGHV1-NL1*01", "IGHV2-5*01", "IGHV4-38-2*01"
    )
    sorted <- sortGenes(genes)
    expect_equal(sorted, expected)

    # Sort genes by position in the locus
    expected_locus <- c(
        "IGHV1-NL1*01", "IGHV1-69-2*01", "IGHV1-69*01",
        "IGHV1-69D*01", "IGHV1-69*02", "IGHV4-38-2*01",
        "IGHV2-5*01", "IGHV1-2", "IGHV1-2*01,IGHV1-2*05",
        "IGHV1-2*02"
    )
    sorted_locus <- sortGenes(genes, method = "pos")
    expect_equal(sorted_locus, expected_locus)
})

### groupGenes ####

test_that("groupGenes heavy only", {
    # make a data frame
    db <- data.frame(
        SEQUENCE_ID = c(1, 2, 3, 4, 5, 6, 7, 8),
        V_CALL = c("IGHV1-1", "IGHV1-1,IGHV1-2", "IGHV1-2,IGHV1-3", "IGHV1-2", "IGHV1-3", "IGHV1-3", "IGHV1-1,IGHV1-2", "IGHV1-2"),
        J_CALL = c("IGHJ1", "IGHJ2", "IGHJ1", "IGHJ1", "IGHJ1", "IGHJ2", "IGHJ1", "IGHJ2"),
        stringsAsFactors = F
    )
    db$LOCUS <- getLocus(db$V_CALL)
    # group VJ genes
    db <- groupGenes(db,
        v_call = "V_CALL",
        j_call = "J_CALL",
        locus = "LOCUS",
        first = F
    )
    # test
    # expected <- c(2, 3, 2, 2, 2, 1, 2, 3)
    # same underlying spirit as before
    expected <- c("G1", "G2", "G1", "G1", "G1", "G3", "G1", "G2")
    expect_equal(db$vj_group, expected)
})


test_that("groupGenes, when 1 row ", {
    # Tests fix for rowSums when 1 row
    db <- structure(list(
        sample_id = "UW0343__S", sequence_id = "assemble12_0",
        subject_id = "S1", v_call = "IGHV3-30*02,IGHV3-30-5*02",
        j_call = "IGHJ6*03", junction = "TGTGCGAAAGTCCCCGTGGGGACTGCCTCTTACATGGACGCCTGG",
        cell_id = NA_character_, single_cell = FALSE, locus = "IGH",
        junction_length = 45L
    ), row.names = c(NA, -1L), class = c("tbl_df", "tbl", "data.frame"))
    ft <- groupGenes(db,
        junc_len = "junction_length",
        cell_id = "cell_id", locus = "locus", only_heavy = T,
        first = TRUE
    )
    ff <- groupGenes(db,
        junc_len = "junction_length",
        cell_id = "cell_id", locus = "locus", only_heavy = T,
        first = FALSE
    )
    expect_equal(ft, ff)
})


test_that("groupGenes all cell_id values are NA", {
    db <- data.frame(
        subject_id = c("S1", "S1", "S1", "S1", "S1", "S1", "S1", "S1"),
        v_call = c("IGHV1-1*01", "IGHV1-1*01", "IGHV1-2*01", "IGHV1-1*01,IGHV1-2*01", "IGHV1-2*01", "IGKV1-1*01", "IGKV1-1*01","IGKV1-1*01"),
        j_call = c("IGHJ2*01", "IGHJ1*01", "IGHJ1*01", "IGHJ1*01", "IGHJ1*01", "IGKJ1*01", "IGKJ1*01", "IGKJ1*01"),
        junction = c("TGTAAAAAATGG", "TGTAAAAAATGG", "TGTAAAACCTGG", "TGTAAACCCTGG", "TGTAAACCCTGG", "TGTCCCCCCTGG", "TGTCCCCCCTGG", "TGTCCCCCCTGG"),
        locus = c("IGH", "IGH", "IGH", "IGH", "IGH", "IGK", "IGK", "IGK"),
        cell_id = c(NA, NA, NA, NA, NA, NA, NA, NA),
        junction_length = 12
    )

    expect_warning(
        groupGenes(db,
                        v_call = "v_call", j_call = "j_call", junc_len = NULL,
                        locus = "locus", only_heavy = TRUE,
                        first = FALSE),
        "A cell_id column was found in the data, but was not specified. All values are NA."
    )    
})



test_that("groupGenes, single-cell mode, heavy only", {
    # (in theory, should be 1 heavy per cell; but code-wise there is no restriction for groupGenes)
    load(file.path("..", "data-tests", "db_sc.rda")) # data_t1

    # manual deduction
    # 1
    # IGHV3-11 IGHJ4 93
    # 2
    # IGHV3-7 IGHJ1 24
    # IGHV3-6 IGHJ1 24  (first=F)
    # IGHV1-4 IGHJ3 27
    # IGHV1-4 IGHJ4 27  (first=F)
    #
    # 3 IGHV3-7 IGHJ5 90
    # 4 IGHV3-7 IGHJ5 90
    # 5 IGHV3-7 IGHJ5 90
    #
    # 6 IGHV4-59 IGHJ1 84
    # 7 IGHV4-59 IGHJ1 84
    # 8 IGHV4-59 IGHJ1 84
    # 9 IGHV4-59 IGHJ1 84
    #
    # 10 IGHV4-59 IGHJ1 84
    # 11 IGHV4-59 IGHJ1 84

    # Cell with CELL_ID==B has 2 heavy and 2 light chains
    # expect error
    expect_error(
        singleCellValidation(data_t1, cell_id = "CELL_ID", locus = "LOCUS"),
        regexp = "Only one heavy chain is allowed per cell"
    )
    expect_error(
        gg1 <- groupGenes(data_t1,
            v_call = "V_CALL", j_call = "J_CALL",
            junc_len = "LEN", cell_id = "CELL_ID", locus = "LOCUS",
            only_heavy = TRUE, first = FALSE
        ),
        regexp = "Only one heavy chain is allowed per cell"
    )

    # 1 by itself, 2 by itself, 3-5 together, 6-11 together
    gg1_expect <- c(
        rep("G1", 2),
        # rep("G1", 4), # Cell B removed
        rep("G2", 2 + 2 + 2),
        rep("G3", 2 + 3 + 2 + 2 + 2 + 2)
    )

    gg1 <- groupGenes(data_t1 %>% dplyr::filter(CELL_ID != "B"),
        v_call = "V_CALL", j_call = "J_CALL",
        junc_len = "LEN", cell_id = "CELL_ID", locus = "LOCUS",
        only_heavy = TRUE, first = FALSE
    )

    expect_equal(gg1[["vj_group"]], gg1_expect)
})

test_that("groupGenes, mixed bulk and single cell", {
    # TODO: add other sc cell with different L to test only_heavy T/F and fix tests
    db <- data.frame(
        subject_id = c("S1", "S1", "S1", "S1", "S1", "S1", "S1", "S1"),
        v_call = c("IGHV1-1*01", "IGHV1-1*01", "IGHV1-2*01", "IGHV1-1*01,IGHV1-2*01", "IGHV1-2*01", "IGKV1-1*01", "IGKV1-1*01","IGKV1-1*01"),
        j_call = c("IGHJ2*01", "IGHJ1*01", "IGHJ1*01", "IGHJ1*01", "IGHJ1*01", "IGKJ1*01", "IGKJ1*01", "IGKJ1*01"),
        junction = c("TGTAAAAAATGG", "TGTAAAAAATGG", "TGTAAAACCTGG", "TGTAAACCCTGG", "TGTAAACCCTGG", "TGTCCCCCCTGG", "TGTCCCCCCTGG", "TGTCCCCCCTGG"),
        locus = c("IGH", "IGH", "IGH", "IGH", "IGH", "IGK", "IGK", "IGK"),
        cell_id = c(1, 2, 3, NA, NA, 1, NA, 4),
        junction_length = 12
    )

    # subset to single-cell sequences
    # expect warning, one cell has only light chain and will be assigned group NA
    expect_warning(
        d <- groupGenes(db %>% dplyr::filter(!is.na(cell_id)),
                        v_call = "v_call", j_call = "j_call", junc_len = NULL,
                        cell_id = "cell_id", locus = "locus", only_heavy = TRUE,
                        first = FALSE),
                        pattern="NA(s) found in one or more of "
        )
    expect_equal(d[["vj_group"]], c("G2", "G1", "G3", "G2", NA))

    expect_warning(
        d_oh <- groupGenes(db %>% dplyr::filter(!is.na(cell_id)),
            v_call = "v_call", j_call = "j_call", junc_len = NULL,
            cell_id = "cell_id", locus = "locus", only_heavy = FALSE,
            first = FALSE
        ),
        "only_heavy = FALSE is deprecated"
    )

    expect_warning(
        d_sl <- groupGenes(db %>% dplyr::filter(!is.na(cell_id)),
            v_call = "v_call", j_call = "j_call", junc_len = NULL,
            cell_id = "cell_id", locus = "locus", only_heavy = TRUE,
            first = FALSE, split_light = TRUE
        ),
        "split_light = TRUE is deprecated"
    )
    expect_equal(d[["vj_group"]], d_oh[["vj_group"]], d_sl[["vj_group"]])

    # bulk data
    d_bulk <- groupGenes(db %>% dplyr::filter(is.na(cell_id)),
        v_call = "v_call", j_call = "j_call", junc_len = NULL,
        cell_id = "cell_id", locus = "locus", only_heavy = TRUE,
        first = FALSE
    )

    # mixed data
    expect_warning(
        d_mixed <- groupGenes(db,
                              v_call = "v_call", j_call = "j_call", junc_len = NULL,
                              cell_id = "cell_id", locus = "locus", only_heavy = TRUE,
                              first = FALSE
        ),
        pattern="NA(s) found in one or more of "
    )
    expect_equal(d_mixed[["vj_group"]], c("G2", "G1", "G1", "G1", "G1", "G2", NA, NA))
    # TODO SSNN 7/16/2024
    # Run tests with the same data used in scoper

    # Bulk mode cell_id=NULL
    ## first=T
    gg <- groupGenes(db_gg %>% select(-cell_id), cell_id = NULL, first = T)
    expect_equal(
        gg[["vj_group"]],
        gg[["expected_group_cell_id-null_first-T"]]
    )
    ## first=F
    gg <- groupGenes(db_gg %>% select(-cell_id), cell_id = NULL, first = F)
    expect_equal(
        gg[["vj_group"]],
        gg[["expected_group_cell_id-null_first-F"]]
    )
})


#### AIRR-format migration tests ####

test_that("countGenes, AIRR-format migration", {
    db_c <- readChangeoDb(file.path("..", "data-tests", "ExampleDb.gz"))
    expect_warning(
        db_a <- readChangeoDb(
            file.path("..", "data-tests", "ExampleDb_airr.gz")
        ),
        regexp = "airr::read_rearrangement"
    )

    # 1
    genes_c <- countGenes(db_c, gene = "V_CALL", groups = "SAMPLE", mode = "allele")
    genes_a <- countGenes(db_a, gene = "v_call", groups = "sample", mode = "allele")

    expect_true(all(genes_c == genes_a))

    # 2
    genes_c <- countGenes(db_c,
        gene = "V_CALL", groups = c("SAMPLE", "ISOTYPE"),
        copy = "DUPCOUNT", mode = "family"
    )
    genes_a <- countGenes(db_a,
        gene = "v_call", groups = c("sample", "isotype"),
        copy = "duplicate_count", mode = "family"
    )

    expect_true(all(genes_c == genes_a))

    # 3
    genes_c <- countGenes(db_c,
        gene = "V_CALL", groups = "SAMPLE",
        mode = "allele", fill = TRUE
    )
    genes_a <- countGenes(db_a,
        gene = "v_call", groups = "sample",
        mode = "allele", fill = TRUE
    )

    # genes_c uses "SAMPLE", genes_a uses "sample"
    # make colnames same case for comparison
    colnames(genes_c) <- tolower(colnames(genes_c))
    expect_true(all.equal(genes_c, genes_a[rownames(genes_c), names(genes_c)]))
})

test_that("groupGenes, AIRR-format migration", {
    # 1
    db_c <- readChangeoDb(file.path("..", "data-tests", "ExampleDb.gz"))
    expect_warning(
        db_a <- readChangeoDb(file.path("..", "data-tests", "ExampleDb_airr.gz")),
        regexp = "airr::read_rearrangement"
    )

    db_c$LOCUS <- getLocus(db_c$V_CALL)

    newDb_c <- groupGenes(db_c, v_call = "V_CALL", j_call = "J_CALL", locus = "LOCUS")
    newDb_a <- groupGenes(db_a, v_call = "v_call", j_call = "j_call")

    # newDb_c and newDb_a not directly comparable b/c diff ncol
    expect_true(all(newDb_c[["vj_group"]] == newDb_a[["vj_group"]]))

    rm(db_c, db_a, newDb_c, newDb_a)

    # 2
    # Cell with CELL_ID==B has 2 heavy and 2 light chains
    load(file.path("..", "data-tests", "db_sc.rda")) # data_t1, data_t2
    load(file.path("..", "data-tests", "db_sc_airr.rda")) # data_t1_airr, data_t2_airr

    newDb_c <- groupGenes(data_t1 %>% dplyr::filter(CELL_ID != "B"),
        v_call = "V_CALL", j_call = "J_CALL",
        junc_len = "LEN", cell_id = "CELL_ID", locus = "LOCUS",
        only_heavy = TRUE, first = FALSE
    )
    newDb_a <- groupGenes(data_t1_airr %>% dplyr::filter(cell_id != "B"),
        v_call = "v_call", j_call = "j_call",
        junc_len = "len", cell_id = "cell_id", locus = "locus",
        only_heavy = TRUE, first = FALSE
    )

    expect_true(all(newDb_c[["vj_group"]] == newDb_a[["vj_group"]]))
})

# Ig lineage topology analysis

#' @include Classes.R
NULL


#### Graph analysis functions ####

#' Generate subtree summary statistics for a tree
#'
#' \code{summarizeSubtrees} calculates summary statistics for each node of a tree. Includes
#' both node properties and subtree properties.
#'
#' @param    graph   igraph object containing an annotated lineage tree.
#' @param    fields  annotation fields to add to the output.
#' @param    root    name of the root (germline) node.
#' 
#' @return   A data.frame with columns: 
#'           \itemize{
#'             \item  \code{name}:             node name.
#'             \item  \code{parent}:           name of the parent node.
#'             \item  \code{outdegree}:        number of edges leading from the node.
#'             \item  \code{size}:             total number of nodes within the subtree rooted 
#'                                             at the node.
#'             \item  \code{depth}:            the depth of the subtree that is rooted at 
#'                                             the node.
#'             \item  \code{pathlength}:       the maximum pathlength beneath the node.
#'             \item  \code{outdegree_norm}:   \code{outdegree} normalized by the total 
#'                                             number of edges.
#'             \item  \code{size_norm}:        \code{size} normalized by the largest
#'                                             subtree size (the germline).
#'             \item  \code{depth_norm}:       \code{depth} normalized by the largest
#'                                             subtree depth (the germline).
#'             \item  \code{pathlength_norm}:  \code{pathlength} normalized by the largest
#'                                             subtree pathlength (the germline).
#'           }
#'           An additional column corresponding to the value of \code{field} is added when
#'           specified.
#' 
#' @seealso  See \link{buildPhylipLineage} for generating input trees. 
#'           See \link{getPathLengths} for calculating path length to nodes.
#' 
#' @examples
#' # Summarize a tree
#' graph <- ExampleTrees[[23]]
#' summarizeSubtrees(graph, fields="c_call", root="Germline")
#' 
#' @export
summarizeSubtrees <- function(graph, fields=NULL, root="Germline") {
    ## DEBUG
    # root="Germline"; fields=NULL
    # TODO:  should probably include a means to exclude inferred from substree size
    
    # Define node attribute data.frame    
    node_df <- data.frame(name=V(graph)$name, stringsAsFactors=F)
    for (f in fields) { 
        node_df[[f]] <- vertex_attr(graph, name=f) 
    }

    # Get edges
    edges <- igraph::as_edgelist(graph)
    # Get unweighted paths
    paths_step <- suppressWarnings(igraph::distances(graph, mode="out", algorithm="unweighted"))
    paths_step[!is.finite(paths_step)] <- NA
    # Get weighted paths
    paths_length <- igraph::distances(graph, mode="out", algorithm="dijkstra")
    paths_length[!is.finite(paths_length)] <- NA
    
    # Define each node's parent
    node_df$parent <- edges[, 1][match(node_df$name, edges[, 2])]
    # Define each node's outdegree
    node_df$outdegree <- igraph::degree(graph, mode="out")
    # Define the number of nodes in each subtree (child count + 1)
    node_df$size <- apply(paths_step, 1, function(x) length(na.omit(x)))
    # Define number of levels below each node
    node_df$depth <- apply(paths_step, 1, max, na.rm=TRUE) + 1
    # Define the maximum shortest path length (genetic distance) to a leaf from each node
    node_df$pathlength <- apply(paths_length, 1, max, na.rm=TRUE)
    
    # Normalize
    node_df <- node_df %>%
        dplyr::mutate(
            outdegree_norm=!!rlang::sym("outdegree")/sum(!!rlang::sym("outdegree"), na.rm=TRUE),
            size_norm=!!rlang::sym("size")/max(!!rlang::sym("size"), na.rm=TRUE),
            depth_norm=!!rlang::sym("depth")/max(!!rlang::sym("depth"), na.rm=TRUE),
            pathlength_norm=!!rlang::sym("pathlength")/max(!!rlang::sym("pathlength"), na.rm=TRUE))

    return(node_df)
}


#' Calculate path lengths from the tree root
#'
#' \code{getPathLengths} calculates the unweighted (number of steps) and weighted (distance) 
#' path lengths from the root of a lineage tree.
#'
#' @param    graph     igraph object containing an annotated lineage tree.
#' @param    root      name of the root (germline) node.
#' @param    field     annotation field to use for exclusion of nodes from step count.
#' @param    exclude   annotation values specifying which nodes to exclude from step count. 
#'                     If \code{NULL} consider all nodes. This does not affect the weighted
#'                     (distance) path length calculation.
#'                     
#' @return   A data.frame with columns:
#'           \itemize{
#'             \item  \code{name}:      node name
#'             \item  \code{steps}:     path length as the number of nodes traversed
#'             \item  \code{distance}:  path length as the sum of edge weights
#'           }
#' 
#' @seealso  See \link{buildPhylipLineage} for generating input trees. 
#' 
#' @examples
#' # Define example graph
#' graph <- ExampleTrees[[24]]
#' 
#' # Consider all nodes
#' getPathLengths(graph, root="Germline")
#' 
#' # Exclude nodes without an isotype annotation from step count
#' getPathLengths(graph, root="Germline", field="c_call", exclude=NA)
#' 
#' @export
getPathLengths <- function(graph, root="Germline", field=NULL, exclude=NULL) {
    # Define path length data.frame
    path_df <- data.frame(name=V(graph)$name, stringsAsFactors=FALSE)

    # Get indices of excluded vertices
    skip_idx <- which(path_df$name == root)
    if (!is.null(field)) {
        g <- vertex_attr(graph, name=field)
        skip_idx <- union(skip_idx, which(g %in% exclude))
    }
    
    # Get paths
    step_list <- shortest_paths(graph, root, mode="out", weights=NA, output="vpath")
    step_list <- step_list$vpath
    
    # Get path lengths
    for (i in 1:length(step_list)) {
        v <- step_list[[i]]
        path_df[i, "steps"] <- sum(!(v %in% skip_idx)) 
        path_df[i, "distance"] <- sum(E(graph, path=v)$weight)
    }
    
    return(path_df)
}


#' Retrieve the first non-root node of a lineage tree
#' 
#' \code{getMRCA} returns the set of lineage tree nodes with the minimum weighted or 
#' unweighted path length from the root (germline) of the lineage tree, allowing for 
#' exclusion of specific groups of nodes.
#'
#' @param    graph    igraph object containing an annotated lineage tree.
#' @param    path     string defining whether to use unweighted (steps) or weighted (distance) 
#'                    measures for determining the founder node set.. 
#' @param    root     name of the root (germline) node.
#' @param    field    annotation field to use for both unweighted path length exclusion and
#'                    consideration as an MRCA node. If \code{NULL} do not exclude any nodes.
#' @param    exclude  vector of annotation values in \code{field} to exclude from the potential 
#'                    MRCA set. If \code{NULL} do not exclude any nodes. Has no effect if 
#'                    \code{field=NULL}.
#'                    
#' @return   A data.frame of the MRCA node(s) containing the columns:
#'           \itemize{
#'             \item  \code{name}:      node name
#'             \item  \code{steps}:     path length as the number of nodes traversed
#'             \item  \code{distance}:  path length as the sum of edge weights
#'           }
#'           Along with additional columns corresponding to the 
#'           annotations of the input graph.
#'           
#' @seealso  Path lengths are determined with \link{getPathLengths}.
#' 
#' @examples
#' # Define example graph
#' graph <- ExampleTrees[[23]]
#' 
#' # Use unweighted path length and do not exclude any nodes
#' getMRCA(graph, path="steps", root="Germline")
#'
#' # Exclude nodes without an isotype annotation and use weighted path length
#' getMRCA(graph, path="distance", root="Germline", field="c_call", exclude=NA)
#' 
#' @export
getMRCA <- function(graph, path=c("distance", "steps"), root="Germline", 
                    field=NULL, exclude=NULL) {
    # Check arguments
    path <- match.arg(path)
    
    # Get distance from root
    path_df <- getPathLengths(graph, root=root, field=field, exclude=exclude)
    
    # Get indices of excluded vertices
    skip_idx <- which(path_df$name == root)
    if (!is.null(field)) {
        g <- vertex_attr(graph, name=field)
        skip_idx <- union(skip_idx, which(g %in% exclude))
    }
    
    # Get founder nodes
    if (path == "distance") { 
        path_len <- setNames(path_df$distance, 1:nrow(path_df))
    } else if (path == "steps") {
        path_len <- setNames(path_df$steps, 1:nrow(path_df))
    } else {
        stop("Invalid value for 'path' parameter. Must be one of c('distance', 'steps').\n")
    }
    
    path_len <- path_len[-skip_idx]
    root_idx <- as.numeric(names(path_len)[which(path_len == min(path_len))])
    root_df <- igraph::as_data_frame(graph, what="vertices")[root_idx, ]
    root_df$steps <- path_df$steps[root_idx]
    root_df$distance <- path_df$distance[root_idx]
    
    # Switch name column to uppercase
    names(root_df)[names(root_df) == "name"] <- "name"
    
    return(root_df)
}


#' Tabulate the number of edges between annotations within a lineage tree
#'
#' \code{tableEdges} creates a table of the total number of connections (edges) for each 
#' unique pair of annotations within a tree over all nodes.
#' 
#' @param    graph     igraph object containing an annotated lineage tree.
#' @param    field     string defining the annotation field to count.
#' @param    indirect  if \code{FALSE} count direct connections (edges) only. If 
#'                     \code{TRUE} walk through any nodes with annotations specified in 
#'                     the \code{argument} to count indirect connections. Specifying
#'                     \code{indirect=TRUE} with \code{exclude=NULL} will have no effect.
#' @param    exclude   vector of strings defining \code{field} values to exclude from counts.
#'                     Edges that either start or end with the specified annotations will not
#'                     be counted. If \code{NULL} count all edges.
#'                     
#' @return   A data.frame defining total annotation connections in the tree with columns:
#'           \itemize{
#'             \item  \code{parent}:  parent annotation
#'             \item  \code{child}:   child annotation
#'             \item  \code{count}:   count of edges for the parent-child relationship
#'           }
#'           
#' @seealso  See \link{testEdges} for performed a permutation test on edge relationships.
#'           
#' @examples
#' # Define example graph
#' graph <- ExampleTrees[[23]]
#' 
#' # Count direct edges between isotypes including inferred nodes
#' tableEdges(graph, "c_call")
#' 
#' # Count direct edges excluding edges to and from germline and inferred nodes
#' tableEdges(graph, "c_call", exclude=c("Germline", NA))
#' 
#' # Count indirect edges walking through germline and inferred nodes
#' tableEdges(graph, "c_call", indirect=TRUE, exclude=c("Germline", NA))
#' 
#' @export
tableEdges <- function(graph, field, indirect=FALSE, exclude=NULL) {
    # Function to retrieve the name if x is exactly one vertex index and NULL otherwise
    .getSingleVertex <- function(x) {
        if (length(x) == 1) { 
            vertex_attr(graph, name=field, index=x[1]) 
        } else { 
            NULL 
        }
    }

    if (indirect) {
        # Get indices of excluded and retained vertices
        if (!is.null(exclude)) {
            f <- vertex_attr(graph, name=field)
            skip_idx <- which(f %in% exclude)
            keep_idx <- as.numeric(V(graph))[-skip_idx]
        } else {
            skip_idx <- NULL
            keep_idx <- as.numeric(V(graph))
        }
        
        # Iterate over nodes and count indirect parent-child connections
        edge_list <- list()
        for (i in keep_idx) { 
            # Get parent annotation
            parent <- vertex_attr(graph, name=field, index=i)
            
            # Get indirect child node annotations
            step_list <- suppressWarnings(shortest_paths(graph, V(graph)[i], mode="out", weights=NA, output="vpath"))
            step_list <- unique(lapply(step_list$vpath, function(x) x[!(x %in% c(i, skip_idx))]))
            children <- unlist(lapply(step_list, .getSingleVertex))
                
            # Define data.frame of connections
            if (length(children) > 0) {
                edge_list[[i]] <- data.frame("parent"=parent, "child"=children, 
                                             stringsAsFactors=FALSE)
            }
        }
        
        # Merge edge list into data.frame
        edge_df <- bind_rows(edge_list)        
    } else {
        # Get adjacency list
        edge_mat <- as_edgelist(graph, names=FALSE)
        edge_mat <- vertex_attr(graph, name=field, index=edge_mat)
        edge_mat <- matrix(edge_mat, ncol=2, dimnames=list(NULL, c("parent", "child")))

        # Build and subset edge data.frame
        edge_df <- as.data.frame(edge_mat, stringsAsFactors=FALSE)
        edge_df <- edge_df[!(edge_df$parent %in% exclude) & !(edge_df$child %in% exclude), ]
    }
    
    # Count edges
    edge_tab <- edge_df %>%
        group_by(!!!rlang::syms(c("parent", "child"))) %>%
        dplyr::summarize(count=n())

    return(edge_tab)
}


#' Permute the node labels of a tree
#' 
#' \code{permuteLabels} permutes the node annotations of a lineage tree.
#'
#' @param    graph    igraph object containing an annotated lineage tree.
#' @param    field    string defining the annotation field to permute.
#' @param    exclude  vector of strings defining \code{field} values to exclude 
#'                    from permutation.
#' 
#' @return   A modified igraph object with vertex annotations permuted.
#' 
#' @seealso  \link{testEdges}.
#' 
#' @examples
#' # Define and plot example graph
#' library(igraph)
#' graph <- ExampleTrees[[23]]
#' plot(graph, layout=layout_as_tree, vertex.label=V(graph)$c_call, 
#'      vertex.size=50, edge.arrow.mode=0, vertex.color="grey80")
#' 
#' # Permute annotations and plot new tree
#' set.seed(123)
#' g <- permuteLabels(graph, "c_call")
#' plot(g, layout=layout_as_tree, vertex.label=V(g)$c_call,
#'      vertex.size=50, edge.arrow.mode=0, vertex.color="grey80")
#' 
#' @export
permuteLabels <- function(graph, field, exclude=c("Germline", NA)) {
    # Determine which nodes to permute
    labels <- vertex_attr(graph, name=field)
    i <- which(!(labels %in% exclude))
    
    # Return input on insufficient number of nodes
    if (length(i) < 2) { 
        warning("Only 1 node to permute\n")
        return(graph) 
    }
    
    # Sample and reassign field values
    s <- sample(i)
    perm <- set_vertex_attr(graph, name=field, index=i, value=labels[s])
    
    return(perm)
}



#### Test functions ####


#' Tests for MRCA annotation enrichment in lineage trees
#' 
#' \code{testMRCA} performs a permutation test on a set of lineage trees to determine
#' the significance of an annotation's association with the MRCA position of the lineage
#' trees.
#' 
#' @param    graphs    list of igraph object containing annotated lineage trees.
#' @param    field     string defining the annotation field to test.
#' @param    root      name of the root (germline) node.
#' @param    exclude   vector of strings defining \code{field} values to exclude from the
#'                     set of potential founder annotations.
#' @param    nperm     number of permutations to perform.
#' @param    progress  if \code{TRUE} show a progress bar.
#' 
#' @return   An \link{MRCATest} object containing the test results and permutation
#'           realizations.
#'           
#' @seealso  Uses \link{getMRCA} and \link{getPathLengths}. 
#'           See \link{plotMRCATest} for plotting the permutation distributions.
#'           
#' @examples
#' \donttest{
#' # Define example tree set
#' graphs <- ExampleTrees[1:10]
#' 
#' # Perform MRCA test on isotypes
#' x <- testMRCA(graphs, "c_call", nperm=10)
#' print(x)
#' }
#' 
#' @export
testMRCA <- function(graphs, field, root="Germline", exclude=c("Germline", NA), 
                     nperm=200, progress=FALSE) {
    # Function to resolve ambiguous founders
    # @param  x      data.frame from getMRCA
    # @param  field  annotation field
    .resolveMRCA <- function(x, field) {
        x %>% filter(!duplicated(!!rlang::sym(field))) %>%
            filter(length(!!rlang::sym(field)) == 1)
    }
    
    # Function to count MRCAs
    # @param  x        list of graphs
    # @param  field    annotation field
    # @param  exclude  vector of annotation values to exclude
    .countMRCA <- function(x, field, exclude) {
        # Get MRCAs
        mrca_list <- lapply(x, getMRCA, path="distance", field=field, 
                            exclude=exclude)
        # Resolve ambiguous MRCAs
        mrca_list <- lapply(mrca_list, .resolveMRCA, field=field)
        # Summarize MRCA counts
        mrca_sum <- bind_rows(mrca_list, .id="GRAPH") %>%
            select(!!!rlang::syms(c("GRAPH", field))) %>%
            rename(dplyr::all_of(c("annotation"=field))) %>%
            group_by(!!rlang::sym("annotation")) %>%
            dplyr::summarize(count=n())
        
        return(mrca_sum)
    }
    
    # Assign numeric names if graphs is an unnamed list
    if (is.null(names(graphs))) { names(graphs) <- 1:length(graphs) }
    
    # Summarize observed MRCA counts
    obs_sum <- .countMRCA(graphs, field=field, exclude=exclude)

    # Generate edge null distribution via permutation
    if (progress) { 
        pb <- progressBar(nperm) 
    }
    perm_list <- list()
    for (i in 1:nperm) {
        # Permute labels
        tmp_list <- lapply(graphs, permuteLabels, field=field, exclude=exclude)
        # Summarize MRCA counts
        tmp_sum <- .countMRCA(tmp_list, field=field, exclude=exclude)
        # Update permutation set
        tmp_sum$iter <- i
        perm_list[[i]] <- tmp_sum
        
        if (progress) { pb$tick() }
    }
    cat("\n")
    perm_sum <- bind_rows(perm_list)
    
    # Test observed against permutation distribution
    for (i in 1:nrow(obs_sum)) {
        x <- obs_sum$annotation[i]
        # Annotation count distribution
        d <- perm_sum$count[perm_sum$annotation == x]
        # Expected mean
        obs_sum[i, "expected"] <- mean(d)
        # P-value for observed > expected
        f <- ecdf(d)
        obs_sum[i, "pvalue"] <- 1 - f(obs_sum$count[i])
    }
 
    # Generate return object
    mrca_test <- new("MRCATest", 
                     tests=as.data.frame(obs_sum), 
                     permutations=as.data.frame(perm_sum),
                     nperm=nperm)
    
    return(mrca_test)
}


#' Tests for parent-child annotation enrichment in lineage trees
#' 
#' \code{testEdges} performs a permutation test on a set of lineage trees to determine
#' the significance of an annotation's association with parent-child relationships.
#'
#' @param    graphs    list of igraph objects with vertex annotations.
#' @param    field     string defining the annotation field to permute.
#' @param    indirect  if \code{FALSE} count direct connections (edges) only. If 
#'                     \code{TRUE} walk through any nodes with annotations specified in 
#'                     the \code{argument} to count indirect connections. Specifying
#'                     \code{indirect=TRUE} with \code{exclude=NULL} will have no effect.
#' @param    exclude   vector of strings defining \code{field} values to exclude from 
#'                     permutation.
#' @param    nperm     number of permutations to perform.
#' @param    progress  if \code{TRUE} show a progress bar.
#' 
#' @return   An \link{EdgeTest} object containing the test results and permutation
#'           realizations.
#' 
#' @seealso  Uses \link{tableEdges} and \link{permuteLabels}. 
#'           See \link{plotEdgeTest} for plotting the permutation distributions.
#'           
#' @examples
#' \donttest{
#' # Define example tree set
#' graphs <- ExampleTrees[1:10]
#' 
#' # Perform edge test on isotypes
#' x <- testEdges(graphs, "c_call", nperm=10)
#' print(x)
#' }
#' 
#' @export
testEdges <- function(graphs, field, indirect=FALSE, exclude=c("Germline", NA), nperm=200, 
                      progress=FALSE) {
    ## DEBUG
    # field="c_call"; exclude=c("Germline", NA); nperm=200
    
    # Assign numeric names if graphs is an unnamed list
    if (is.null(names(graphs))) { names(graphs) <- 1:length(graphs) }
    
    # Function to count edge annotations
    # @param  x        list of graphs
    # @param  field    annotation field
    # @param  exclude  vector of annotation values to exclude
    .countEdges <- function(x, field, exclude) {
        edge_list <- lapply(x, tableEdges, field=field, indirect=indirect, exclude=exclude)
        edge_sum <- bind_rows(edge_list) %>%
            group_by(!!!rlang::syms(c("parent", "child"))) %>%
            dplyr::summarize(count=sum(!!rlang::sym("count"), na.rm=TRUE))
        return(edge_sum)
    }
    
    # Count edges of observed data
    obs_sum <- .countEdges(graphs, field, exclude)
    if (nrow(obs_sum) == 0) {
        stop("No valid edges found in graphs")
    }
    
    # Generate edge null distribution via permutation
    if (progress) { 
        pb <- progressBar(nperm) 
    }
    perm_list <- list()
    for (i in 1:nperm) {
        # Permute annotations
        tmp_list <- lapply(graphs, permuteLabels, field=field, exclude=exclude)
        # Count edges
        tmp_sum <- .countEdges(tmp_list, field, exclude)
        # Update permutation set
        tmp_sum$iter <- i
        perm_list[[i]] <- tmp_sum
        
        if (progress) { pb$tick() }
    }
    perm_sum <- bind_rows(perm_list)
    
    # Test observed against permutation distribution
    for (i in 1:nrow(obs_sum)) {
        x <- obs_sum$parent[i]
        y <- obs_sum$child[i]
        # Edge count distribution
        d <- perm_sum$count[perm_sum$parent == x & perm_sum$child == y]
        # Expected mean
        obs_sum[i, "expected"] <- mean(d)
        # P-value for observed > expected
        f <- ecdf(d)
        obs_sum[i, "pvalue"] <- 1 - f(obs_sum$count[i])
    }
    
    # Generate return object
    edge_test <- new("EdgeTest", 
                 tests=as.data.frame(obs_sum), 
                 permutations=as.data.frame(perm_sum),
                 nperm=nperm)
    
    return(edge_test)
}


#### Plotting functions #####

#' Plot the results of an edge permutation test
#' 
#' \code{plotEdgeTest} plots the results of an edge permutation test performed with 
#' \code{testEdges} as either a histogram or cumulative distribution function.
#'
#' @param    data        \link{EdgeTest} object returned by \link{testEdges}.
#' @param    color       color of the histogram or lines.
#' @param    main_title  string specifying the plot title.
#' @param    style       type of plot to draw. One of:
#'                       \itemize{
#'                         \item \code{"histogram"}:  histogram of the edge count 
#'                                                    distribution with a red dotted line
#'                                                    denoting the observed value.
#'                         \item \code{"cdf"}:        cumulative distribution function 
#'                                                    of edge counts with a red dotted 
#'                                                    line denoting the observed value and
#'                                                    a blue dotted line indicating the 
#'                                                    p-value.
#'                       }
#' @param    silent      if \code{TRUE} do not draw the plot and just return the ggplot2 
#'                       object; if \code{FALSE} draw the plot.
#' @param    ...         additional arguments to pass to ggplot2::theme.
#'
#' @return   A \code{ggplot} object defining the plot.
#' 
#' @seealso  See \link{testEdges} for performing the test.
#' 
#' @examples
#' \donttest{
#' # Define example tree set
#' graphs <- ExampleTrees[6:10]
#' 
#' # Perform edge test on isotypes
#' x <- testEdges(graphs, "c_call", nperm=6)
#' 
#' # Plot
#' plotEdgeTest(x, color="steelblue", style="hist")
#' plotEdgeTest(x, style="cdf")
#' }
#' 
#' @export
plotEdgeTest <- function(data, color="black", main_title="Edge Test", 
                         style=c("histogram", "cdf"), silent=FALSE, ...) {
    # Check arguments
    style <- match.arg(style)
    
    # Extract plot data
    obs_sum <- rename(data@tests, "Parent"="parent", "Child"="child")
    perm_sum <- rename(data@permutations, "Parent"="parent", "Child"="child")

    if (style == "histogram") {
        # Plot edge null distribution
        p1 <- ggplot(perm_sum, aes(x=!!rlang::sym("count"))) +
            baseTheme() + 
            ggtitle(main_title) +
            xlab("Number of edges") +
            ylab("Number of realizations") + 
            geom_histogram(bins=50, fill=color, color=NA) +
            geom_vline(data=obs_sum, aes(xintercept=!!rlang::sym("count")), 
                       color="firebrick", linetype=3, linewidth=0.75) + 
            facet_grid("Child ~ Parent", labeller=label_both, scales="free")
    } else if (style == "cdf") {    
        # Plot ECDF of edge null distribution
        p1 <- ggplot(perm_sum, aes(x=!!rlang::sym("count"))) +
            baseTheme() + 
            ggtitle(main_title) +
            xlab("Number of edges") +
            ylab("P-value") +
            stat_ecdf(color=color, linewidth=1) +
            geom_vline(data=obs_sum, aes(xintercept=!!rlang::sym("count")), color="firebrick", 
                       linetype=3, linewidth=0.75) + 
            geom_hline(data=obs_sum, aes(yintercept=!!rlang::sym("pvalue")), color="steelblue", 
                       linetype=3, linewidth=0.75) + 
            facet_grid("Child ~ Parent", labeller=label_both, scales="free")
    }
    
    # Add additional theme elements
    p1 <- p1 + do.call(theme, list(...))
    
    # Plot
    if (!silent) { plot(p1) }
    
    invisible(p1)
}


#' Plot the results of a founder permutation test
#' 
#' \code{plotMRCATest} plots the results of a founder permutation test performed with 
#' \code{testMRCA}.
#'
#' @param    data        \link{MRCATest} object returned by \link{testMRCA}.
#' @param    color       color of the histogram or lines.
#' @param    main_title  string specifying the plot title.
#' @param    style       type of plot to draw. One of:
#'                       \itemize{
#'                         \item \code{"histogram"}:  histogram of the annotation count 
#'                                                    distribution with a red dotted line
#'                                                    denoting the observed value.
#'                         \item \code{"cdf"}:        cumulative distribution function 
#'                                                    of annotation counts with a red dotted 
#'                                                    line denoting the observed value and
#'                                                    a blue dotted line indicating the 
#'                                                    p-value.
#'                       }
#' @param    silent      if \code{TRUE} do not draw the plot and just return the ggplot2 
#'                       object; if \code{FALSE} draw the plot.
#' @param    ...         additional arguments to pass to ggplot2::theme.
#'
#' @return   A \code{ggplot} object defining the plot.
#' 
#' @seealso  See \link{testEdges} for performing the test.
#' 
#' @examples
#' \donttest{
#' # Define example tree set
#' graphs <- ExampleTrees[1:10]
#' 
#' # Perform MRCA test on isotypes
#' set.seed(123)
#' x <- testMRCA(graphs, "c_call", nperm=10)
#' 
#' # Plot
#' plotMRCATest(x, color="steelblue", style="hist")
#' plotMRCATest(x, style="cdf")
#' }
#' 
#' @export
plotMRCATest <- function(data, color="black", main_title="MRCA Test", 
                         style=c("histogram", "cdf"), silent=FALSE, ...) {
    # Check arguments
    style <- match.arg(style)
    
    # Extract plot data
    obs_sum <- rename(data@tests, "Annotation"="annotation")
    perm_sum <- rename(data@permutations, "Annotation"="annotation")
    
    if (style == "histogram") {
        # Plot MRCA null distribution
        p1 <- ggplot(perm_sum, aes(x=!!rlang::sym("count"))) +
            baseTheme() + 
            ggtitle(main_title) +
            xlab("Number of MRCAs") +
            ylab("Number of realizations") + 
            geom_histogram(fill=color, color=NA, bins=50) +
            geom_vline(data=obs_sum, aes(xintercept=!!rlang::sym("count")), 
                       color="firebrick", linetype=3, linewidth=0.75) + 
            facet_wrap("Annotation", ncol=1, scales="free_y")
    } else if (style == "cdf") {
        # Plot ECDF of MRCA null distribution
        p1 <- ggplot(perm_sum, aes(x=!!rlang::sym("count"))) +
            baseTheme() + 
            ggtitle(main_title) +
            xlab("Number of MRCAs") +
            ylab("P-value") +
            stat_ecdf(color=color, linewidth=1) +
            geom_vline(data=obs_sum, aes(xintercept=!!rlang::sym("count")), 
                       color="firebrick", linetype=3, linewidth=0.75) + 
            geom_hline(data=obs_sum, aes(yintercept=!!rlang::sym("pvalue")), 
                       color="steelblue", linetype=3, linewidth=0.75) + 
            facet_wrap("Annotation", nrow=1, scales="free_y")
    }
    
    # Add additional theme elements
    p1 <- p1 + do.call(theme, list(...))
    
    # Plot
    if (!silent) { plot(p1) }
    
    invisible(p1)
}


#' Plots subtree statistics for multiple trees
#' 
#' \code{plotSubtree} plots distributions of normalized subtree statistics for a 
#' set of lineage trees, broken down by annotation value.
#'
#' @param    graphs        list of igraph objects containing annotated lineage trees.
#' @param    field         string defining the annotation field.
#' @param    stat          string defining the subtree statistic to plot. One of:
#'                         \itemize{
#'                           \item  \code{outdegree}:   distribution of normalized node 
#'                                                      outdegrees.
#'                           \item  \code{size}:        distribution of normalized subtree sizes.
#'                           \item  \code{depth}:       distribution of subtree depths.
#'                           \item  \code{pathlength}:  distribution of maximum pathlength 
#'                                                      beneath nodes.
#'                         }
#' @param    root          name of the root (germline) node.
#' @param    exclude       vector of strings defining \code{field} values to exclude from
#'                         plotting.
#' @param    colors        named vector of colors for values in \code{field}, with 
#'                         names defining annotation names \code{field} column and values
#'                         being colors. Also controls the order in which values appear on the
#'                         plot. If \code{NULL} alphabetical ordering and a default color palette 
#'                         will be used.
#' @param    main_title    string specifying the plot title.
#' @param    legend_title  string specifying the legend title.
#' @param    style         string specifying the style of plot to draw. One of:
#'                         \itemize{
#'                           \item \code{"histogram"}:  histogram of the annotation count 
#'                                                      distribution with a red dotted line
#'                                                      denoting the observed value.
#'                           \item \code{"cdf"}:        cumulative distribution function 
#'                                                      of annotation counts with a red 
#'                                                      dotted line denoting the observed 
#'                                                      value and a blue dotted line 
#'                                                      indicating the p-value.
#'                         }
#' @param    silent        if \code{TRUE} do not draw the plot and just return the ggplot2 
#'                         object; if \code{FALSE} draw the plot.
#' @param    ...           additional arguments to pass to ggplot2::theme.
#'
#' @return   A \code{ggplot} object defining the plot.
#' 
#' @seealso  Subtree statistics are calculated with \link{summarizeSubtrees}.
#' 
#' @examples
#' # Define example tree set
#' graphs <- ExampleTrees[1:10]
#' 
#' # Violin plots of node outdegree by sample
#' plotSubtrees(graphs, "sample_id", "out", style="v")
#'
#' # Violin plots of subtree size by sample
#' plotSubtrees(graphs, "sample_id", "size", style="v")
#' 
#' # Boxplot of node depth by isotype
#' plotSubtrees(graphs, "c_call", "depth", style="b")
#' 
#' @export
plotSubtrees <- function(graphs, field, stat, root="Germline", exclude=c("Germline", NA), 
                         colors=NULL, main_title="Subtrees", legend_title="Annotation", 
                         style=c("box", "violin"), silent=FALSE, ...) {
    # Hack for visibility of special ggplot variables
    #..x.. <- NULL
    
    ## DEBUG
    # graphs=ExampleTrees; field="c_call"; colors=IG_COLORS; main_title="Outdegree"; root="Germline"; exclude=c("Germline", NA); style="box"
    # Check arguments
    style <- match.arg(style, several.ok=FALSE)
    stat <- match.arg(stat, choices=c("outdegree", "size", "depth", "pathlength"), 
                      several.ok=FALSE)

    # Set stat column and axis labels
    if (stat == "outdegree") {
        stat_col <- "outdegree_norm"
        y_lab <- "Node outdegree"
    } else if (stat == "size") {
        stat_col <- "size_norm"
        y_lab <- "Substree size"
    } else if (stat == "depth") {
        stat_col <- "depth_norm"
        y_lab <- "Depth under node"
    } else if (stat == "pathlength") {
        stat_col <- "pathlength_norm"
        y_lab <- "Path length under node"
    } else {
        stop("Invalid value for 'stat'. How did you get here?")
    }
    
    # Assign numeric names if graphs is an unnamed list
    if (is.null(names(graphs))) { names(graphs) <- 1:length(graphs) }
    
    # Get subtree summarizes and filter excluded annotations
    sum_list <- lapply(graphs, summarizeSubtrees, fields=field, root=root)
    sum_df <- bind_rows(sum_list, .id="GRAPH") %>%
        filter(!(!!rlang::sym(field) %in% exclude),
                is.finite(!!rlang::sym(stat_col)))
    
    # Set ordering based on color names
    if (!is.null(colors)) {
        # Assign missing levels to grey
        x <- unique(sum_df[[field]])
        x <- sort(x[!(x %in% names(colors))])
        if (length(x) > 0) {
            warning("The following are missing from the 'colors' argument and will be colored grey: ", 
                    paste(x, collapse=" "))
            x <- setNames(rep("grey", length(x)), x)
            colors <- c(colors, x)
        }
        # Cast to factor
        sum_df[[field]] <- factor(sum_df[[field]], levels=names(colors))
    } else {
        sum_df[[field]] <- factor(sum_df[[field]])
    }
    
    # Make plot object
    p1 <- ggplot(sum_df, aes(x=!!rlang::sym(field), y=!!rlang::sym(stat_col))) + 
        baseTheme() + 
        ggtitle(main_title) + 
        xlab("") +
        ylab(y_lab) +
        scale_y_continuous(labels=percent) +
        expand_limits(y=0)
    
    # Add distributions style
    if (style == "box") {
        p1 <- p1 + geom_boxplot(aes(fill=!!rlang::sym(field)), width=0.7, alpha=0.8)
    } else if (style == "violin") {
        p1 <- p1 + geom_violin(aes(fill=!!rlang::sym(field)), adjust=1.5, scale="width", trim=T, 
                               width=0.7, alpha=0.8) +
            geom_errorbarh(aes(xmin=after_stat(x) - 0.4, xmax=after_stat(x) + 0.4), color="black", 
                           stat="summary", fun="mean", linewidth=1.25, height=0, alpha=0.9)
    }

    # Set colors and legend
    if (!is.null(colors)) {
        p1 <- p1 + scale_fill_manual(name=legend_title, values=colors)
    } else {
        p1 <- p1 + scale_fill_discrete(name=legend_title)
    }
    
    # Add additional theme elements
    p1 <- p1 + do.call(theme, list(...))
    
    # Plot
    if (!silent) { plot(p1) }
    
    invisible(p1)
}

ExampleTrees <- file.path("..", "data-tests", "ExampleTrees.rda")
load(ExampleTrees)

graph <- ExampleTrees[[5]]
graph_2 <- graph

test_that("Test summarizeSubtrees",{
    obs_graph_sum_sub <- summarizeSubtrees(graph, fields="isotype", root="Germline")
    expected <- file.path("..","data-tests", "exp_graph_sum_sub.rda")
    load(expected)
    names(exp_graph_sum_sub)[2] = "isotype"
    expect_equal(obs_graph_sum_sub, exp_graph_sum_sub, tolerance=1)
})

test_that("Test getPathLengths",{
    
    # Consider all nodes
    pl <- getPathLengths(graph, root="Germline")
    expect_equal(pl$steps, c(1,0,2,2))
    expect_equal(pl$distance, c(28,0,32,34))
    
    # Exclude nodes without an isotype annotation from step count
    pl_exclude <- getPathLengths(graph_2, root="Germline", field="isotype", 
                                 exclude=NA)
    expect_equal(pl_exclude$steps, c(0,0,1,1))
    expect_equal(pl_exclude$distance, c(28,0,32,34))
})

test_that("Test getMRCA and testMRCA",{
    
    mrca <- getMRCA(graph, path="steps", root="Germline")
    expect_equal(mrca$name, "Inferred1")
    
    mrca_2 <- getMRCA(graph_2, path="distance", root="Germline", 
                      field="isotype", exclude=NA)
    expect_equal( mrca_2$name,  c("GN5SHBT03CABH1"))
    
    graphs <- ExampleTrees[1-10]
     
    # Perform MRCA test on isotypes
    # ensure older version of sample() used 
    # sample.kind="Rounding"
    # Will show warning: non-uniform 'Rounding' sampler used
    expect_warning(set.seed(8, sample.kind="Rounding"),"non-uniform 'Rounding' sampler used")
    x <- testMRCA(graphs, "isotype", nperm=10)
    x_tests <- slot(x, "tests")
    expect_equal(x_tests$annotation, c("IgA","IgA,IgG","IgG"))
    expect_equal(x_tests$count, c(16, 1, 31))
    expect_equal(x_tests$expected, c(13.6, 1.2, 33.8))
    expect_equal(x_tests$pvalue, c(0, 0.2, 1) )
    
})

test_that("Test tableEdges",{
    graph <- ExampleTrees[[23]]
    
    # Count direct edges between isotypes including inferred nodes
    tab <- tableEdges(graph, "isotype")
    expect_equal(tab$parent, c( "IgA", "IgA,IgG", "IgA,IgG", NA))
    expect_equal(tab$child, c( "IgA,IgG", "IgA", "IgG", "IgA"))
    expect_equal(tab$count, c( 1, 1, 3, 1))
    
    # Count direct edges excluding edges to and from germline and inferred nodes
    tab <- tableEdges(graph, "isotype", exclude=c("Germline", NA))
    expect_equal(tab$parent, c( "IgA", "IgA,IgG", "IgA,IgG"))
    expect_equal(tab$child, c( "IgA,IgG", "IgA", "IgG"))
    expect_equal(tab$count, c( 1, 1, 3))
    
    # Count indirect edges walking through germline and inferred nodes
    tab <- tableEdges(graph, "isotype", indirect=TRUE, exclude=c("Germline", NA))
    expect_equal(tab$parent, c( "IgA", "IgA,IgG", "IgA,IgG"))
    expect_equal(tab$child, c( "IgA,IgG", "IgA", "IgG"))
    expect_equal(tab$count, c( 1, 1, 3))    
         
})

test_that("Test permuteLabels",{
    graph <- ExampleTrees[[23]]

    # Permute annotations and plot new tree
    expect_warning(set.seed(43, sample.kind="Rounding"),"non-uniform 'Rounding' sampler used")
    g <- permuteLabels(graph, "isotype")
    expect_equal(V(g)$isotype,
                 c("IgG", "IgG", NA, "IgA", "IgA", "IgG", "IgA,IgG")
    )
})

test_that("Test testEdges", {
    # Define example tree set
    graphs <- ExampleTrees[1-10]

    # Perform edge test on isotypes
    expect_warning(set.seed(4, sample.kind="Rounding"),"non-uniform 'Rounding' sampler used")
    x <- slot(testEdges(graphs, "isotype", nperm=10), "tests")
    expect_equal(x$parent[1:5], c("IgA", "IgA", "IgA", "IgA,IgG", "IgA,IgG"))
    expect_equal(x$child[5:10], c("IgA,IgG", "IgG", "IgG", "IgA", "IgD,IgG", "IgG"))
    expect_equal(x$count[1:5], c(39, 3, 2, 29, 1))
    expect_equal(x$expected[5:10], c(2, 6.25, 1.33, 7, 1.00, 133.7), tolerance=0.01)
    expect_equal(x$pvalue[3:8], c(0.7, 0.0, 0.5, 0.0, 0.0, 1.0), tolerance=0.01)
})

#' Load sequencing quality scores from a FASTQ file
#' 
#' \code{readFastqDb} adds the sequencing quality scores to a data.frame
#' from a FASTQ file. Matching is done by `sequence_id`.
#'
#' @param    data            \code{data.frame} containing sequence data.
#' @param    fastq_file      path to the fastq file
#' @param    quality_offset  offset value to be used by ape::read.fastq. It is 
#'                           the value to be added to the quality scores 
#'                           (the default -33 applies to the Sanger format and 
#'                           should work for most recent FASTQ files).
#' @param    header          FASTQ file header format; one of \code{"presto"} or 
#'                           \code{"asis"}. Use \code{"presto"} to specify 
#'                           that the fastq file headers are using the pRESTO
#'                           format and can be parsed to extract 
#'                           the \code{sequence_id}. Use \code{"asis"} to skip 
#'                           any processing and use the sequence names as they are.
#' @param    sequence_id     column in \code{data} that contains sequence 
#'                           identifiers to be matched to sequence identifiers in 
#'                           \code{fastq_file}. 
#' @param    sequence        column in \code{data} that contains sequence data.
#' @param    sequence_alignment   column in \code{data} that contains IMGT aligned sequence data.      
#' @param    v_cigar         column in \code{data} that contains CIGAR 
#'                           strings for the V gene alignments.     
#' @param    d_cigar         column in \code{data} that contains CIGAR 
#'                           strings for the D gene alignments.   
#' @param    j_cigar         column in \code{data} that contains CIGAR 
#'                           strings for the J gene alignments.     
#' @param    np1_length      column in \code{data} that contains the number
#'                           of nucleotides between the V gene and first D gene 
#'                           alignments or between the V gene and J gene alignments.
#' @param    np2_length      column in \code{data} that contains the number
#'                           of nucleotides between either the first D gene and J 
#'                           gene alignments or the first D gene and second D gene
#'                           alignments.
#' @param    v_sequence_end  column in \code{data} that contains the 
#'                           end position of the V gene in \code{sequence}.
#' @param    d_sequence_end  column in \code{data} that contains the 
#'                           end position of the D gene in \code{sequence}.                      
#' @param    style           how the sequencing quality should be returned;
#'                           one of \code{"num"}, \code{"phred"}, or \code{"both"}.
#'                           Specify \code{"num"} to store the quality scores as strings of 
#'                           comma separated numeric values. Use \code{"phred"} to have
#'                           the function return the scores as Phred (ASCII) scores. 
#'                           Use \code{"both"} to retrieve both.
#' @param    quality_sequence     specify \code{TRUE} to keep the quality scores for 
#'                                \code{sequence}. If false, only the quality score
#'                                for \code{sequence_alignment} will be added to \code{data}.
#'                                
#' @return   Modified \code{data} with additional fields:
#'           \enumerate{
#'                 \item \code{quality_alignment}:     A character vector with ASCII Phred 
#'                                                     scores for \code{sequence_alignment}.
#'                 \item \code{quality_alignment_num}: A character vector, with comma separated 
#'                                                     numerical quality values for each 
#'                                                     position in \code{sequence_alignment}.
#'                 \item \code{quality}:      A character vector with ASCII Phred 
#'                                                     scores for \code{sequence}.
#'                 \item \code{quality_num}:  A character vector, with comma separated 
#'                                                     numerical quality values for each 
#'                                                     position in \code{sequence}.
#'           }
#' @seealso \link{maskPositionsByQuality} and \link{getPositionQuality}
#' 
#' @examples
#' db <- airr::read_rearrangement(system.file("extdata", "example_quality.tsv", package="alakazam"))
#' fastq_file <- system.file("extdata", "example_quality.fastq", package="alakazam")
#' db <- readFastqDb(db, fastq_file, quality_offset=-33)
#'
#' @export
readFastqDb <- function(data, fastq_file, quality_offset=-33, 
                        header=c("presto", "asis"), 
                        sequence_id="sequence_id",
                        sequence="sequence",
                        sequence_alignment="sequence_alignment",
                        v_cigar="v_cigar",
                        d_cigar="d_cigar",
                        j_cigar="j_cigar",
                        np1_length="np1_length",
                        np2_length="np2_length",
                        v_sequence_end="v_sequence_end",
                        d_sequence_end="d_sequence_end",
                        style=c("num", "ascii", "both"),
                        quality_sequence=FALSE) {
   
   check_cols <- c(sequence_id, 
                   sequence, 
                   sequence_alignment, 
                   v_cigar, d_cigar, j_cigar, 
                   np1_length, np2_length, 
                   v_sequence_end, d_sequence_end)
   
   alakazam::checkColumns(data, check_cols)
   
   style <- match.arg(style)
   
   # Process the fastq file
   header <- match.arg(header)
   fastq <- ape::read.fastq(fastq_file, offset=quality_offset) #default: -33 (pRESTO)
   fastq_db <- data.frame(
      "quality_num"=as.vector(sapply(attr(fastq, "QUAL"), paste0, collapse=",")),
      stringsAsFactors = F)
   
   fastq_db$quality <- sapply(fastq_db[["quality_num"]], 
                                     function(qual, quality_offset) {
                                        paste0(sapply(strsplit(qual, ",")[[1]], function(x,quality_offset) {
                                           y <- as.numeric(x) - quality_offset
                                           rawToChar(as.raw(y))
                                        },quality_offset), sep="",collapse="")
                                     }, quality_offset)
   
   fastq_db[[sequence_id]] <- attr(fastq, "names")
   if (header=="presto") {
      fastq_db[[sequence_id]] <- gsub("\\|.+","",attr(fastq, "names"))
   }

   # Merge
   by <- sequence_id
   names(by) <- sequence_id
   data <- data %>%
      left_join(fastq_db, by=by)
   
   data <- sequenceAlignmentQuality(data,
                                  sequence=sequence,
                                  sequence_id=sequence_id,
                                  sequence_alignment=sequence_alignment,
                                  quality="quality",
                                  quality_num="quality_num",
                                  v_cigar=v_cigar,
                                  d_cigar=d_cigar,
                                  j_cigar=j_cigar,
                                  np1_length=np1_length,
                                  np2_length=np2_length,
                                  v_sequence_end=v_sequence_end,
                                  d_sequence_end=d_sequence_end,
                                  raw=FALSE)
   
   if (!quality_sequence) {
      data[['quality']] <- NULL
      data[['quality_num']] <- NULL
   }
   
   if (style != "both") {
      if (style == "phred") {
         data[['quality_alignment_num']] <- NULL
      } else {
         data[['quality_alignment']] <- NULL
      }
   }
   
   data 
}

# Thanks!:
# https://drive5.com/usearch/manual/cigar.html &
# https://jef.works/blog/2017/03/28/CIGAR-strings-for-dummies/
# M	 	Match (alignment column containing two letters). This could contain two
#        letters (mismatch) or two identical letters. USEARCH generates CIGAR strings
#        containing Ms rather than X's and ='s (see below).
# N	   Alignment gap 	Next x positions on ref don’t match (Deletion in query?)
# D	 	Deletion (gap in the target sequence).
# I	 	Insertion (gap in the query sequence).
# S	 	Segment of the query sequence that does not appear in the alignment.
#        This is used with soft clipping, where the full-length query sequence
#        is given (field 10 in the SAM record). In this case, S operations specify
#        segments at the start and/or end of the query that do not appear in a
#        local alignment.
# H	 	Segment of the query sequence that does not appear in the alignment.
#        This is used with hard clipping, where only the aligned segment of the
#        query sequences is given (field 10 in the SAM record). In this case, H
#        operations specify segments at the start and/or end of the query that
#        do not appear in the SAM record.
# =	 	Alignment column containing two identical letters. USEARCH can read
#        CIGAR strings using this operation, but does not generate them.
# X	 	Alignment column containing a mismatch, i.e. two different letters.
#        USEARCH can read CIGAR strings using this operation, but does not generate them.

# Tested with IgBlast output, not with IMGT
calcSequenceAlignmentQuality <- function(sequence_db,
                                        sequence="sequence",
                                        sequence_id="sequence_id",
                                        sequence_alignment="sequence_alignment",
                                        quality="quality",
                                        quality_num="quality_num",
                                        v_cigar="v_cigar",
                                        d_cigar="d_cigar",
                                        j_cigar="j_cigar",
                                        np1_length="np1_length",
                                        np2_length="np2_length",
                                        v_sequence_end="v_sequence_end",
                                        d_sequence_end="d_sequence_end",
                                        raw=FALSE) {
   # query sequence
   sequence <- sequence_db[[sequence]]
   quality_phred <- strsplit(sequence_db[[quality]],"")[[1]]
   quality_num_values <- strsplit(sequence_db[[quality_num]],",")[[1]]
   
   v_cigar <- sequence_db[[v_cigar]]
   vd_pseudo_cigar <- NA
   if (!is.na(sequence_db[[np1_length]])) {
      if (sequence_db[[np1_length]]>0) {
         vd_pseudo_cigar <- paste0(sequence_db[[v_sequence_end]],"S",sequence_db[[np1_length]],"X")
      }
   }
   d_cigar <- sequence_db[[d_cigar]]
   dj_pseudo_cigar <- NA
   if (!is.na(sequence_db[[np2_length]])) {
      if (sequence_db[[np2_length]]>0){
         dj_pseudo_cigar <- paste0(sequence_db[[d_sequence_end]],"S",sequence_db[[np2_length]],"X")
      }
   }
   j_cigar <- sequence_db[[j_cigar]]
   cigars <-  c(v_cigar, vd_pseudo_cigar, d_cigar, dj_pseudo_cigar, j_cigar)
   cigars <- cigars[!is.na(cigars)]

   use_cigarillo <- utils::packageVersion("GenomicAlignments") >= package_version("1.45.5")
   if (use_cigarillo && !requireNamespace("cigarillo", quietly=TRUE)) {
      stop("The 'cigarillo' package is required when using GenomicAlignments >= 1.45.5, ",
           "which replaced explodeCigarOps()/explodeCigarOpLengths() with functions from cigarillo. ",
           "Please install it with: BiocManager::install(\"cigarillo\")")
   }
   ranges <- bind_rows(lapply(cigars, function(cigar) {
      if (use_cigarillo) {
         ops <- cigarillo::explode_cigar_ops(cigar)[[1]]
         lengths <- cigarillo::explode_cigar_oplens(cigar)[[1]]
      } else {
         ops <- GenomicAlignments::explodeCigarOps(cigar)[[1]]
         lengths <- GenomicAlignments::explodeCigarOpLengths(cigar)[[1]]
      }
      keep <- ops %in% c("N", "I") == F
      ops <- ops[keep]
      lengths <- lengths[keep]

      ranges <- data.frame(
         "start"=rep(NA, length(ops)),
         "end"=NA,
         "width"=lengths,
         "operator"=ops,
         stringsAsFactors = F)
      ranges[['start']][1] <- 1

      for (i in 1:nrow(ranges)) {
         if (ranges[['operator']][i] %in% c("S","=","X","D")) {
            ranges[['end']][i] <- ranges[['start']][i]+ranges[['width']][i]-1
         }
         if (i+1<=nrow(ranges)) {
            ranges[['start']][i+1] <- ranges[['end']][i]+1
         }
      }

      ranges <- ranges %>%
         filter(!!rlang::sym("operator")  %in% c("S","D") == FALSE)
      ranges
   }))

   iranges <- IRanges(start=ranges[['start']], end=ranges[['end']], width=ranges[['width']])
   reconstruced_sequence_alignment <- extractAt(BString(sequence), iranges)
   reconstruced_sequence_alignment <- paste0(sapply(reconstruced_sequence_alignment, toString), collapse="")
   positions <- unlist(sapply(1:nrow(ranges), function(i) { ranges$start[i]:ranges$end[i]}))

   quality_df <- data.frame(
      "reconstructed_sequence_alignment"=paste0(sapply(reconstruced_sequence_alignment, toString),collapse=""),
      "sequence_position"=positions,
      "sequence_alignment_position"=NA,
      stringsAsFactors = F
   ) %>%
   mutate (
      !!rlang::sym(quality) := quality_phred[positions],
      !!rlang::sym(quality_num) := as.numeric(quality_num_values)[positions],
      !!rlang::sym(sequence_id) := sequence_db[[sequence_id]])

   # Sanity check. Just to be sure reconstruction is working correctly,
   # and be sure I will later transfer the quality scores correctly
   sequence_alignment <- sequence_db[[sequence_alignment]]
   expected <- gsub("[\\.-]","",sequence_alignment)
   if (reconstruced_sequence_alignment != expected) {
      stop("Reconstructed sequence_alignment from cigar doesn't match db sequence_alignment.")
   }

   # map position numbering: sequence input positions <--> aligned positions
   nt_aln <- strsplit(sequence_alignment,"")[[1]]
   for ( aln_position in 1:length(nt_aln)) {
      if (nt_aln[aln_position] %in% c(".","-") == FALSE ) {
         pos <- sum(nt_aln[1:aln_position] %in% c(".","-") == F)
         quality_df[['sequence_alignment_position']][pos] <- aln_position
         quality_df[['sequence_alignment_nt']][pos] <- nt_aln[aln_position]
      } 
   }
   
   if (raw) {
      quality_df %>%
         select(-!!rlang::sym("reconstructed_sequence_alignment"))
   } else {
      qual_num <- rep(NA, length(nt_aln))
      qual_num[quality_df[['sequence_alignment_position']]] <- quality_df[[quality_num]]
      qual_num <- paste0(qual_num, sep="", collapse=",")
      
      qual_phred <- rep(" ", length(nt_aln))
      qual_phred[quality_df[['sequence_alignment_position']]] <- quality_df[[quality]]
      qual_phred <- paste0(qual_phred, sep="", collapse="")
      
      ret <- data.frame(
                 "quality_alignment_num"=qual_num,
                 "quality_alignment"=qual_phred,
                 stringsAsFactors = F
                 ) %>%
      mutate(!!rlang::sym(sequence_id) := sequence_db[[sequence_id]])
      ret
   }

}


# Retrieve sequencing quality scores from tabular data
# 
# \code{sequenceAlignmentQuality} is used internally by \code{readFastqDb} to 
# process the sequencing quality scores loaded from a \code{fastq} file. 
# 
# Once a repertoire \code{data.frame} has been processed with \link{readFastqDb} and 
# contains the fields \code{quality} and \code{quality_num},
# \code{sequenceAlignmentQuality} can be used to retrieve the quality scores 
# from the already present field \code{quality_num}, without requiring 
# again the \code{fastq} file, and report them as a \code{data.frame} with sequencing 
# qualities per position, not as a string. This is done setting \code{raw=TRUE}. 
# This \code{data.frame} with qualities per position can be used to generate figures, 
# for example.
# 
# @param    data            \code{data.frame} containing sequence data.
# @param    sequence_id     column in \code{data} that contains sequence 
#                           identifiers to be matched to sequence identifiers in 
#                           \code{fastq_file}. 
# @param    sequence        column in \code{data} that contains sequence data.
# @param    sequence_alignment     column in \code{data} that contains 
#                                  IMGT aligned sequence data.      
# @param    quality       column in \code{data} that contains 
#                                  sequencing quality as Phred scores.     
# @param    quality_num   column in \code{data} that contains 
#                                  sequencing quality as a comma separated string.                
# @param    v_cigar         column in \code{data} that contains CIGAR 
#                           strings for the V gene alignments.     
# @param    d_cigar         column in \code{data} that contains CIGAR 
#                           strings for the D gene alignments.   
# @param    j_cigar         column in \code{data} that contains CIGAR 
#                           strings for the J gene alignments.     
# @param    np1_length      column in \code{data} that contains the number
#                           of nucleotides between the V gene and first D gene 
#                           alignments or between the V gene and J gene alignments.
# @param    np2_length      column in \code{data} that contains the number
#                           of nucleotides between either the first D gene and J 
#                           gene alignments or the first D gene and second D gene
#                           alignments.
# @param    v_sequence_end  column in \code{data} that contains the 
#                           end position of the V gene in \code{sequence}.
# @param    d_sequence_end  column in \code{data} that contains the 
#                           end position of the D gene in \code{sequence}.                      
# @param    raw             specify how the sequencing quality should be returned. 
#                           If \code{TRUE}, return a \code{data.frame} with 
#                           quality information per position, where each row is
#                           a position. This \code{data.frame} has columns
#                           "sequence_position", "sequence_alignment_position", 
#                           "quality", "quality_num", "sequence_id"
#                           and "sequence_alignment_nt" If \code{FALSE}, for each sequence,
#                           concatenate the position qualities in a string, and the
#                           quality information to \code{data}
sequenceAlignmentQuality <- function(data, 
                                     sequence_id="sequence_id",
                                     sequence="sequence",
                                     sequence_alignment="sequence_alignment",
                                     quality="quality",
                                     quality_num="quality_num",
                                     v_cigar="v_cigar",
                                     d_cigar="d_cigar",
                                     j_cigar="j_cigar",
                                     np1_length="np1_length",
                                     np2_length="np2_length",
                                     v_sequence_end="v_sequence_end",
                                     d_sequence_end="d_sequence_end",
                                     raw=FALSE) {
   pb <- progressBar(nrow(data))
   qual <- bind_rows(lapply(1:nrow(data),function(i) {
      pb$tick()
      calcSequenceAlignmentQuality(data[i,],
      sequence=sequence,
      sequence_id=sequence_id,
      sequence_alignment=sequence_alignment,
      quality=quality,
      quality_num=quality_num,
      v_cigar=v_cigar,
      d_cigar=d_cigar,
      j_cigar=j_cigar,
      np1_length=np1_length,
      np2_length=np2_length,
      v_sequence_end=v_sequence_end,
      d_sequence_end=d_sequence_end,
      raw=raw)
   }))
   
   if (raw) {
      qual   
   } else {
      data %>%
         dplyr::left_join(qual, by=sequence_id)
   }
}


#' Mask sequence positions with low quality
#' 
#' \code{maskPositionsByQuality} will replace positions that 
#' have a sequencing quality score lower that \code{min_quality} with an
#' \code{"N"} character.
#' 
#' 
#' @param    data          \code{data.frame} containing sequence data.
#' @param    min_quality   minimum quality score. Positions with sequencing quality 
#'                         less than \code{min_qual} will be masked.
#' @param    sequence      column in \code{data} with sequence data to be masked.
#' @param    quality_num   column in \code{data} with quality scores (a
#'                         string of numeric values, comma separated) that can
#'                         be used to mask \code{sequence}. 
#'                         
#' @return   Modified \code{data} data.frame with an additional field containing 
#'           quality masked sequences. The  name of this field is created 
#'           concatenating the \code{sequence} name and \code{"_masked"}.
#'           
#' @seealso \link{readFastqDb} and \link{getPositionQuality}
#' 
#' @examples
#' db <- airr::read_rearrangement(system.file("extdata", "example_quality.tsv", package="alakazam"))
#' fastq_file <- system.file("extdata", "example_quality.fastq", package="alakazam")
#' db <- readFastqDb(db, fastq_file, quality_offset=-33)
#' maskPositionsByQuality(db, min_quality=90, quality_num="quality_alignment_num")
#' 
#' @export
maskPositionsByQuality <- function(data, min_quality=70,
                                   sequence="sequence_alignment",
                                   quality_num="quality_alignment_num") {
   
   required_cols <- c(sequence,quality_num)
   checkColumns(data, required_cols)
   sequence_masked <- paste0(sequence,"_masked")
   num_masked_seqs <- 0
   data <- bind_rows(lapply(1:nrow(data), function(i) {
      db_row <- data[i,]
      seq_qual <- strsplit(db_row[[quality_num]],",")[[1]]
      low_seq_qual <- which(sapply(seq_qual, function(x) {
         if (x != "NA") {
            as.numeric(x) < min_quality
         } else {
            NA
         }
      }, USE.NAMES = FALSE))
      if (length(low_seq_qual)>0) {
         num_masked_seqs <<- num_masked_seqs + 1
         seq <- strsplit(db_row[[sequence]],"")[[1]]
         seq[low_seq_qual] <- "N"
         seq <- paste0(seq, collapse="")
         db_row[[sequence_masked]] <- seq
      }
      db_row
   }))
   message("Number of masked sequences: ", num_masked_seqs)
   data
} 


#' Get a data.frame with sequencing qualities per position
#' 
#' \code{getPositionQuality} takes a data.frame with sequence quality scores 
#' in the form of a strings of comma separated numeric values, split the quality 
#' scores values by \code{","},  and returns a data.frame with the values
#' for each position.
#' 
#' 
#' @param    data          \code{data.frame} containing sequence data.
#' @param    sequence_id   column in \code{data} with sequence identifiers.
#' @param    sequence      column in \code{data} with sequence data. 
#' @param    quality_num   column in \code{data} with quality scores (as
#'                         strings of numeric values, comma separated) for \code{sequence}.
#'                         
#' @return   \code{data} with one additional field with masked sequences. The 
#'           name of this field is created concatenating \code{sequence} 
#'           and '_masked'.
#'
#' @seealso \link{readFastqDb} and \link{maskPositionsByQuality}
#'           
#' @examples
#' db <- airr::read_rearrangement(system.file("extdata", "example_quality.tsv", package="alakazam"))
#' fastq_file <- system.file("extdata", "example_quality.fastq", package="alakazam")
#' db <- readFastqDb(db, fastq_file, quality_offset=-33)
#' head(getPositionQuality(db))
#
#' @export          
getPositionQuality <- function(data, 
                               sequence_id="sequence_id",
                               sequence="sequence_alignment", 
                               quality_num="quality_alignment_num") {
   
   checkColumns(data, c(sequence, quality_num))
   
   bind_rows(lapply(1:nrow(data), function(i) {
      seq_id <- data[[sequence_id]][i]
      seq_len <- nchar(data[[sequence]][i])
      qual_parts <- strsplit(data[[quality_num]][i],",")[[1]]
      qual_values <- as.numeric(replace(qual_parts, qual_parts == "NA", NA))
      nt <- strsplit(data[[sequence]],"")[[1]]
      if (seq_len != length(qual_values)) {
         stop("Different length, for sequence: ", seq_id,". seq: ", sequence, ". qual: ", quality_num)
      }
      data.frame(
         "position"=1:seq_len, stringsAsFactors = F) %>%
         mutate(!!rlang::sym(quality_num) := qual_values,
                !!rlang::sym(sequence_id) := seq_id,
                nt = nt)
   }))
   
}

# Documentation and definitions for data and constants

#### Sysdata ####

# 1x20 vector of default amino acid hydropathy scores
# HYDROPATHY_KYTJ82

# 1x20 vector of default amino acid bulkiness scores
# BULKINESS_ZIMJ68

# 1x20 vector of default amino acid polarity scores
# POLARITY_GRAR74

# 1x7 vector of default amino acid pK values
# PK_EMBOSS

#### Data ####

#' Example AIRR database
#'
#' A small example database subset from Laserson and Vigneault et al, 2014.
#'
#' @format   A data.frame with the following AIRR style columns:
#'   \itemize{
#'     \item  \code{sequence_id}:                Sequence identifier
#'     \item  \code{sequence_alignment}:         IMGT-gapped observed sequence.
#'     \item  \code{germline_alignment}:         IMGT-gapped germline sequence.
#'     \item  \code{germline_alignment_d_mask}:  IMGT-gapped germline sequence with N, P and 
#'                                               D regions masked.
#'     \item  \code{v_call}:                     V region allele assignments.
#'     \item  \code{v_call_genotyped}:           TIgGER corrected V region allele assignment.
#'     \item  \code{d_call}:                     D region allele assignments.
#'     \item  \code{j_call}:                     J region allele assignments.
#'     \item  \code{c_call}:                     Isotype (C region) assignment.
#'     \item  \code{junction}:                   Junction region sequence.
#'     \item  \code{junction_length}:            Length of the junction region in nucleotides.
#'     \item  \code{np1_length}:                 Combined length of the N and P regions proximal
#'                                               to the V region.
#'     \item  \code{np2_length}:                 Combined length of the N and P regions proximal
#'                                               to the J region.
#'     \item  \code{duplicate_count}:            Copy count (number of duplicates) of the sequence.
#'     \item  \code{clone_id}:                   Change-O assignment clonal group identifier.
#'     \item  \code{sample_id}:                  Sample identifier. Time in relation to vaccination.
#' }
#' 
#' @seealso \link{ExampleDbChangeo} \link{ExampleTrees}
#' 
#' @references
#' \enumerate{
#'   \item  Laserson U and Vigneault F, et al. High-resolution antibody dynamics of 
#'            vaccine-induced immune responses. 
#'            Proc Natl Acad Sci USA. 2014 111:4928-33.
#' }
"ExampleDb"

#' Single sequence AIRR database
#'
#' A database with just one sequence from \code{ExampleDb} and additional AIRR Rearrangement fields 
#' containing alignment information. The sequence was reanalyzed with a recent versions of 
#' alignment software (IgBLAST 1.16.0) and reference germlines (IMGT 2020-08-12).
#' 
#' @seealso \link{ExampleDb}
"SingleDb"

#' Example Change-O database
#'
#' A small example database subset from Laserson and Vigneault et al, 2014.
#'
#' @format   A data.frame with the following Change-O style columns:
#'   \itemize{
#'     \item  \code{SEQUENCE_ID}:           Sequence identifier
#'     \item  \code{SEQUENCE_IMGT}:         IMGT-gapped observed sequence.
#'     \item  \code{GERMLINE_IMGT_D_MASK}:  IMGT-gapped germline sequence with N, P and 
#'                                          D regions masked.
#'     \item  \code{V_CALL}:                V region allele assignments.
#'     \item  \code{V_CALL_GENOTYPED}:      TIgGER corrected V region allele assignment.
#'     \item  \code{D_CALL}:                D region allele assignments.
#'     \item  \code{J_CALL}:                J region allele assignments.
#'     \item  \code{JUNCTION}:              Junction region sequence.
#'     \item  \code{JUNCTION_LENGTH}:       Length of the junction region in nucleotides.
#'     \item  \code{NP1_LENGTH}:            Combined length of the N and P regions proximal
#'                                          to the V region.
#'     \item  \code{NP2_LENGTH}:            Combined length of the N and P regions proximal
#'                                          to the J region.
#'     \item  \code{SAMPLE}:                Sample identifier. Time in relation to vaccination.
#'     \item  \code{ISOTYPE}:               Isotype assignment.
#'     \item  \code{DUPCOUNT}:              Copy count (number of duplicates) of the sequence.
#'     \item  \code{CLONE}:                 Change-O assignment clonal group identifier.
#' }
#' 
#' @seealso \link{ExampleDb} \link{ExampleTrees}
#' 
#' @references
#' \enumerate{
#'   \item  Laserson U and Vigneault F, et al. High-resolution antibody dynamics of 
#'            vaccine-induced immune responses. 
#'            Proc Natl Acad Sci USA. 2014 111:4928-33.
#' }
"ExampleDbChangeo"

#' Small example 10x Genomics Ig V(D)J sequences from CD19+ B cells isolated from PBMCs of a healthy 
#' human donor. Down-sampled from data provided by 10x Genomics under a Creative Commons Attribute license,
#' and processed with their Cell Ranger pipeline.
#'
#' @format   A data.frame with the following AIRR style columns:
#'   \itemize{
#'     \item  \code{sequence_id}:                Sequence identifier
#'     \item  \code{sequence_alignment}:         IMGT-gapped observed sequence.
#'     \item  \code{germline_alignment}:         IMGT-gapped germline sequence.
#'     \item  \code{v_call}:                     V region allele assignments.
#'     \item  \code{d_call}:                     D region allele assignments.
#'     \item  \code{j_call}:                     J region allele assignments.
#'     \item  \code{c_call}:                     Isotype (C region) assignment.
#'     \item  \code{junction}:                   Junction region sequence.
#'     \item  \code{junction_length}:            Length of the junction region in nucleotides.
#'     \item  \code{np1_length}:                 Combined length of the N and P regions proximal
#'                                               to the V region.
#'     \item  \code{np2_length}:                 Combined length of the N and P regions proximal
#'                                               to the J region.
#'     \item  \code{umi_count}:                  Number of unique molecular identifies atttributed to sequence.
#'     \item  \code{cell_id}:                    Cell identifier.
#'     \item  \code{locus}:                      Genomic locus of sequence.
#' }
#' 
#' 
#' @references
#' \enumerate{
#'   \item  Data source: https://support.10xgenomics.com/single-cell-vdj/datasets/2.2.0/vdj_v1_hs_cd19_b
#'   \item  License: https://creativecommons.org/licenses/by/4.0/
#' }
"Example10x"

#' Example Ig lineage trees
#'
#' A set of Ig lineage trees generated from the \code{ExampleDb} file, subset to
#' only those trees with at least four nodes.
#'
#' @format   A list of igraph objects output by \link{buildPhylipLineage}.
#'           Each node of each tree has the following annotations (vertex attributes):
#'   \itemize{
#'     \item  \code{sample_id}:          Sample identifier(s). Time in relation to vaccination.
#'     \item  \code{c_call}:             Isotype assignment(s). 
#'     \item  \code{duplication_count}:  Copy count (number of duplicates) of the sequence.
#'   }
#'   
#' @seealso \link{ExampleTrees}
"ExampleTrees"


#### Constants ####

#' Default colors
#' 
#' Default color palettes for DNA characters, Ig isotypes, and TCR chains.
#' 
#' @format  Named character vectors with hexcode colors as values.
#' \itemize{
#'   \item  \code{DNA_COLORS}:  DNA character colors 
#'                              \code{c("A", "C", "G", "T")}.
#'   \item  \code{IG_COLORS}:   Ig isotype colors 
#'                              \code{c("IGHA", "IGHD", "IGHE", "IGHG", "IGHM", "IGHK", "IGHL")}.
#'   \item  \code{TR_COLORS}:   TCR chain colors 
#'                              \code{c("TRA", "TRB", "TRD", "TRG")}.
#' }
#' 
#' @examples 
#' # IG_COLORS as an isotype color set for ggplot
#' isotype <- c("IGHG", "IGHM", "IGHM", "IGHA")
#' db <- data.frame(x=1:4, y=1:4, iso=isotype)
#' g1 <- ggplot(db, aes(x=x, y=y, color=iso)) + 
#'     scale_color_manual(name="Isotype", values=IG_COLORS) +
#'     geom_point(size=10)
#' plot(g1)
#' 
#' # DNA_COLORS to translate nucleotide values to a vector of colors 
#' # for use in base graphics plots
#' seq <- c("A", "T", "T", "C")
#' colors <- translateStrings(seq, setNames(names(DNA_COLORS), DNA_COLORS))
#' plot(1:4, 1:4, col=colors, pch=16, cex=6)
#' 
#' @name   DEFAULT_COLORS
NULL

#' @rdname   DEFAULT_COLORS
#' @export
DNA_COLORS <- c("A"="#64F73F", 
                "C"="#FFB340", 
                "G"="#EB413C", 
                "T"="#3C88EE")

#' @rdname DEFAULT_COLORS
#' @export
IG_COLORS <- c("IGHA"="#377EB8", 
               "IGHD"="#FF7F00", 
               "IGHE"="#E41A1C", 
               "IGHG"="#4DAF4A", 
               "IGHM"="#984EA3",
               "IGHK"="#E5C494",
               "IGHL"="#FFD92F")

#' @rdname DEFAULT_COLORS
#' @export
TR_COLORS <- c("TRA"="#CBD5E8", 
               "TRB"="#F4CAE4", 
               "TRD"="#FDCDAC", 
               "TRG"="#E6F5C9")

#' IUPAC ambiguous characters
#'
#' A translation list mapping IUPAC ambiguous characters code to corresponding nucleotide
#' amino acid characters.
#' 
#' @format  A list with single character codes as names and values containing character 
#'          vectors that define the set of standard characters that match to each each 
#'          ambiguous character.
#' \itemize{
#'   \item  \code{IUPAC_DNA}:  DNA ambiguous character translations.
#'   \item  \code{IUPAC_AA}:   Amino acid ambiguous character translations.
#'   \item  \code{DNA_IUPAC}:  Ordered DNA to ambiguous characters
#' }
#' 
#' @name    IUPAC_CODES
NULL

#' @rdname  IUPAC_CODES
#' @export
IUPAC_DNA <- list("A"="A", 
                  "C"="C", 
                  "G"="G", 
                  "T"="T",
                  "M"=c("A","C"), 
                  "R"=c("A","G"), 
                  "W"=c("A","T"), 
                  "S"=c("C","G"), 
                  "Y"=c("C","T"), 
                  "K"=c("G","T"), 
                  "V"=c("A","C","G"), 
                  "H"=c("A","C","T"), 
                  "D"=c("A","G","T"), 
                  "B"=c("C","G","T"),
                  "N"=c("A","C","G","T"))

#' @rdname    IUPAC_CODES
#' @export
IUPAC_AA <-  list("A"="A", 
                  "B"=c("N","R"),
                  "C"="C", 
                  "D"="D",
                  "E"="E",
                  "F"="F",
                  "G"="G",
                  "H"="H",
                  "I"="I",
                  "J"=c("I","L"),
                  "K"="K",
                  "L"="L",
                  "M"="M",
                  "N"="N",
                  "P"="P",
                  "Q"="Q",
                  "R"="R",
                  "S"="S",
                  "T"="T",
                  "V"="V",
                  "W"="W",
                  "X"=c("A","B","C","D","E","F","G","H",
                        "I","J","K","L","M","N","P","Q",
                        "R","S","T","V","W","X","Y","Z",
                        "*"),
                  "Y"="Y",
                  "Z"=c("E","Q"),
                  "*"="*")

#' @rdname    IUPAC_CODES
#' @export
DNA_IUPAC <- list(
                  "A"="A", 
                  "C"="C", 
                  "G"="G", 
                  "T"="T",
                  "AC"="M", 
                  "AG"="R", 
                  "AT"="W", 
                  "CG"="S", 
                  "CT"="Y", 
                  "GT"="K", 
                  "ACG"="V", 
                  "ACT"="H", 
                  "AGT"="D", 
                  "CGT"="B",
                  "ACGT"="N")

#' Amino acid abbreviation translations
#' 
#' Mappings of amino acid abbreviations.
#' 
#' @format  Named character vector defining single-letter character codes to 
#'          three-letter abbreviation mappings.
#' 
#' @name   ABBREV_AA
#' 
#' @examples 
#' aa <- c("Ala", "Ile", "Trp")
#' translateStrings(aa, ABBREV_AA)
#' 
#' @export
ABBREV_AA <- c("A"="Ala",
               "R"="Arg",
               "N"="Asn",
               "D"="Asp",
               "C"="Cys",
               "Q"="Gln",
               "E"="Glu",
               "G"="Gly",
               "H"="His",
               "I"="Ile",
               "L"="Leu",
               "K"="Lys",
               "M"="Met",
               "F"="Phe",
               "P"="Pro",
               "S"="Ser",
               "T"="Thr",
               "W"="Trp",
               "Y"="Tyr",
               "V"="Val")


#' IMGT V-segment regions
#'
#' A list defining the boundaries of V-segment framework regions (FWRs) and complementarity 
#' determining regions (CDRs) for IMGT-gapped immunoglobulin (Ig) nucleotide sequences 
#' according to the IMGT numbering scheme.
#' 
#' @format  A list with regions named one of \code{c("fwr1", "cdr1", "fwr2", "cdr2", "fwr3")} 
#'          with values containing a numeric vector of length two defining the 
#'          \code{c(start, end)} positions of the named region.
#'          
#' @references
#'   \url{https://www.imgt.org/}
#' 
#' @export
IMGT_REGIONS <- list("fwr1"=c(1, 78),
                     "cdr1"=c(79, 114),
                     "fwr2"=c(115, 165),
                     "cdr2"=c(166, 195),
                     "fwr3"=c(196, 312))

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' Test DNA sequences for equality.
#' 
#' \code{seqEqual} checks if two DNA sequences are identical.
#'
#' @param    seq1    character string containing a DNA sequence.
#' @param    seq2    character string containing a DNA sequence.
#' @param    ignore  vector of characters to ignore when testing for equality.
#'                   Default is to ignore c("N",".","-","?")
#' 
#' @return   Returns \code{TRUE} if sequences are equal and \code{FALSE} if they are not.
#'           Sequences of unequal length will always return \code{FALSE} regardless of
#'           their character values.
#' 
#' @seealso  Used by \link{pairwiseEqual} within \link{collapseDuplicates}.
#'           See \link{seqDist} for calculation Hamming distances between sequences.
#' 
#' @examples
#' # Ignore gaps
#' seqEqual("ATG-C", "AT--C")
#' seqEqual("ATGGC", "ATGGN")
#' seqEqual("AT--T", "ATGGC")
#' 
#' # Ignore only Ns
#' seqEqual("ATG-C", "AT--C", ignore="N")
#' seqEqual("ATGGC", "ATGGN", ignore="N")
#' seqEqual("AT--T", "ATGGC", ignore="N")
#' 
#' @export
seqEqual <- function(seq1, seq2, ignore = as.character( c("N","-",".","?"))) {
    .Call(`_alakazam_seqEqual`, seq1, seq2, ignore)
}

#' Calculate pairwise equivalence between sequences
#' 
#' \code{pairwiseEqual} determined pairwise equivalence between a pairs in a 
#' set of sequences, excluding ambiguous positions (Ns and gaps).
#'
#' @param    seq  character vector containing a DNA sequences.
#'
#' @return   A logical matrix of equivalence between each entry in \code{seq}. 
#'           Values are \code{TRUE} when sequences are equivalent and \code{FALSE}
#'           when they are not.
#' 
#' @seealso  Uses \link{seqEqual} for testing equivalence between pairs.
#'           See \link{pairwiseDist} for generating a sequence distance matrix.
#'           
#' @examples
#' # Gaps and Ns will match any character
#' seq <- c(A="ATGGC", B="ATGGG", C="ATGGG", D="AT--C", E="NTGGG")
#' d <- pairwiseEqual(seq)
#' rownames(d) <- colnames(d) <- seq
#' d
#' 
#' @export
pairwiseEqual <- function(seq) {
    .Call(`_alakazam_pairwiseEqual`, seq)
}

seqDistRcpp <- function(seq1, seq2, dist_mat) {
    .Call(`_alakazam_seqDistRcpp`, seq1, seq2, dist_mat)
}

pairwiseDistRcpp <- function(seq, dist_mat) {
    .Call(`_alakazam_pairwiseDistRcpp`, seq, dist_mat)
}

nonsquareDistRcpp <- function(seq, indx, dist_mat) {
    .Call(`_alakazam_nonsquareDistRcpp`, seq, indx, dist_mat)
}

# Common input/output and data structure manipulation functions for Alakazam

#### File I/O functions ####

#' Read a Change-O tab-delimited database file
#' 
#' \code{readChangeoDb} reads a tab-delimited database file created by a Change-O tool 
#' into a data.frame.
#'
#' @param    file       tab-delimited database file output by a Change-O tool.
#' @param    select     columns to select from database file.
#' @param    drop       columns to drop from database file.
#' @param    seq_upper  if \code{TRUE} convert sequence columns to upper case;
#'                      if \code{FALSE} do not alter sequence columns. See Value 
#'                      for a list of which columns are effected.

#' 
#' @return   A data.frame of the database file. Columns will be imported as is, except for 
#'           the following columns which will be explicitly converted into character 
#'           values:
#'           \itemize{
#'             \item  SEQUENCE_ID
#'             \item  CLONE
#'             \item  SAMPLE
#'           }
#'           And the following sequence columns which will converted to upper case if
#'           \code{seq_upper=TRUE} (default).
#'           \itemize{
#'             \item  SEQUENCE_INPUT
#'             \item  SEQUENCE_VDJ
#'             \item  SEQUENCE_IMGT
#'             \item  JUNCTION
#'             \item  GERMLINE_IMGT
#'             \item  GERMLINE_IMGT_D_MASK
#'           }
#'                   
#' @seealso  Wraps \link[readr]{read_delim}. 
#'           See \link{writeChangeoDb} for writing to Change-O files.
#'           See \link[airr:read_tabular]{read_rearrangement} and \link[airr:write_tabular]{write_rearrangement}
#'           to read and write AIRR-C Standard formatted repertoires.
#' 
#' @examples
#' \dontrun{
#'     # Read all columns in and convert sequence fields to upper case
#'     db <- readChangeoDb("changeo.tsv")
#' 
#'     # Subset columns and convert sequence fields to upper case
#'     db <- readChangeoDb("changeo.tsv", select=c("SEQUENCE_ID", "SEQUENCE_IMGT"))
#' 
#'     # Drop columns and do not alter sequence field case
#'     db <- readChangeoDb("changeo.tsv", drop=c("D_CALL", "DUPCOUNT"), 
#'                         seq_upper=FALSE)
#' }
#' 
#' @export
readChangeoDb <- function(file, select=NULL, drop=NULL, seq_upper=TRUE) {
    # Define column data types
    seq_columns <- c("SEQUENCE_INPUT", "SEQUENCE_VDJ", "SEQUENCE_IMGT", 
                     "JUNCTION", "JUNCTION_AA", "CDR3_IGBLAST_NT", "CDR3_IGBLAST_AA",
                     "GERMLINE_VDJ", "GERMLINE_VDJ_V_REGION", "GERMLINE_VDJ_D_MASK",
                     "GERMLINE_IMGT", "GERMLINE_IMGT_V_REGION", "GERMLINE_IMGT_D_MASK",
                     "FWR1_IMGT", "FWR2_IMGT", "FWR3_IMGT", "FWR4_IMGT",
                     "CDR1_IMGT", "CDR2_IMGT", "CDR3_IMGT")

    # Define types
    header <- names(suppressMessages(readr::read_tsv(file, n_max=1)))
    types <- do.call(readr::cols, CHANGEO[intersect(names(CHANGEO), header)])
    
    # Check if ChangeO format
    if (length(types$cols)==0) {
        airr_columns <- intersect(names(airr::RearrangementSchema), header)
        if (length(airr_columns)>0) {
            warning(paste0(
            basename(file),
            " is not in the Change-O format.\n",
            "If you are trying to read an AIRR-C Standard formatted file,\n",
            "use airr::read_rearrangement for correct type casting.
            "))
        }
    }
    
    # Read file    
    db <- suppressMessages(readr::read_tsv(file, col_types=types, na=c("", "NA", "None")))

    # Select columns
    select_columns <- colnames(db)
    if(!is.null(select)) { select_columns <- intersect(select_columns, select) }
    if(!is.null(drop)) { select_columns <- setdiff(select_columns, drop) }
    db <- select(db, dplyr::all_of(select_columns))
    
    # Convert sequence fields to upper case
    upper_cols <- intersect(seq_columns, names(db))
    if (seq_upper & length(upper_cols) > 0) {
        db <- mutate_at(db, upper_cols, toupper)
    }
    
    return(db)
}


#' Write a Change-O tab-delimited database file
#' 
#' \code{writeChangeoDb} is a simple wrapper around \link[readr]{write_delim} with defaults 
#' appropriate for writing a Change-O tab-delimited database file from a data.frame.
#'
#' @param    data  data.frame of Change-O data.
#' @param    file  output file name.
#' 
#' @return   NULL
#' 
#' @seealso  Wraps \link[readr]{write_delim}. See \link{readChangeoDb} for reading to Change-O files.
#'           See \link[airr:read_tabular]{read_rearrangement} and \link[airr:write_tabular]{write_rearrangement}
#'           to read and write AIRR-C Standard formatted repertoires.
#' 
#' @examples
#' \dontrun{
#'   # Write a database
#'   writeChangeoDb(data, "changeo.tsv")
#' }
#' 
#' @export
writeChangeoDb <- function(data, file) {
    write_tsv(data, file, na="NA")
}


#' Create a temporary folder
#'
#' \code{makeTempDir} creates a randomly named temporary folder in the 
#' system temp location.
#' 
#' @param    prefix  prefix name for the folder.
#' 
#' @return   The path to the temporary folder.
#' 
#' @seealso  This is just a wrapper for \link{tempfile} and 
#'           \link{dir.create}.
#' 
#' @examples
#' \dontrun{
#' makeTempDir("Clone50")
#' }
#' 
#' @export
makeTempDir <- function(prefix) {
    temp_path <- tempfile(paste0(prefix, "-temp-"))

    dir.create(temp_path)
    
    return(temp_path)
}


#### Data manipulation functions ####

#' Translate a vector of strings
#' 
#' \code{translateStrings} modifies a character vector by substituting one or more 
#' strings with a replacement string.
#'
#' @param    strings      vector of character strings to modify.
#' @param    translation  named character vector or a list of character vectors specifying 
#'                        the strings to replace (values) and their replacements (names).
#' 
#' @return   A modified \code{strings} vector.
#' 
#' @details
#' Does not perform partial replacements. Each translation value must match a complete 
#' \code{strings} value or it will not be replaced.  Values that do not have a replacement
#' named in the \code{translation} parameter will not be modified.
#' 
#' Replacement is accomplished using \link{gsub}.
#' 
#' @seealso  See \link{gsub} for single value replacement in the base package.
#' 
#' @examples
#' # Using a vector translation
#' strings <- LETTERS[1:5]
#' translation <- c("POSITION1"="A", "POSITION5"="E")
#' translateStrings(strings, translation)
#' 
#' # Using a list translation
#' strings <- LETTERS[1:5]
#' translation <- list("1-3"=c("A","B","C"), "4-5"=c("D","E"))
#' translateStrings(strings, translation)
#' 
#' @export
translateStrings <- function(strings, translation) {
    # TODO:  use match instead for complete matching?  Currently regex characters in values will mess up the matching.
    for (n in names(translation)) {
        rep_regex <- paste(translation[[n]], collapse='|')
        strings <- gsub(paste0("^(", rep_regex, ")$"), n, strings)
    }
    
    return(strings)
}


#' Check data.frame for valid columns and issue message if invalid
#'
#' @param   data     data.frame to check.
#' @param   columns  vector of column names to check.
#' @param   logic    one of \code{"all"} or \code{"any"} controlling whether all,
#'                   or at least one, of the columns must be valid, respectively.
#' @return  \code{TRUE} if columns are valid and a string message if not.
# 
#' @examples
#' df <- data.frame(A=1:3, B=4:6, C=rep(NA, 3))
#' checkColumns(df, c("A", "B"), logic="all")
#' checkColumns(df, c("A", "B"), logic="any")
#' checkColumns(df, c("A", "C"), logic="all")
#' checkColumns(df, c("A", "C"), logic="any")
#' checkColumns(df, c("A", "D"), logic="all")
#' checkColumns(df, c("A", "D"), logic="any")
#' 
#' @export
checkColumns <- function(data, columns, logic=c("all", "any")) {
    ## DEBUG
    # data=df; columns=c("A", "D"); logic="any"
    
    # Check arguments
    logic <- match.arg(logic)
    
    data_names <- names(data)
    if (logic == "all") {
        # Check that all columns exist
        for (f in columns) {
            if (!(f %in% data_names)) { 
                msg <- paste("The column", f, "was not found") 
                return(msg)
            }
        }        
        # Check that all values are not NA
        for (f in columns) {
            if (all(is.na(data[[f]]))) { 
                msg <- paste("The column", f, "contains no data") 
                return(msg)
            }
        }
    } else if (logic == "any") {
        # Check that columns exist
        if (!any(columns %in% data_names)) {
            msg <- paste("Input must contain at least one of the columns:", 
                         paste(columns, collapse=", "))
            return(msg)
        }
        # Check that all values are not NA
        
        columns_found <- columns[columns %in% data_names]
        invalid <- sapply(columns_found, function(f) all(is.na(data[[f]])))
        if (all(invalid)) { 
            msg <- paste("None of the columns", paste(columns_found, collapse=", "), 
                         "contain data") 
            return(msg)
        }
    }
    
    # Return TRUE if all checks pass
    return(TRUE)
}

#### Plotting and progress functions ####

#' Standard progress bar
#' 
#' \code{progressBar} defines a common progress bar format.
#' 
#' @param   n  maximum number of ticks
#' 
#' @return  A \link[progress]{progress_bar} object.
#'
#' @export
progressBar <- function(n) {
    pb <- progress::progress_bar$new(format="  PROGRESS> [:bar] :percent :elapsed",
                                     width=40, clear=FALSE, stream=stdout(), force=TRUE,
                                     total=n)
    return(pb)
}


#' Standard ggplot settings
#' 
#' \code{baseTheme} defines common ggplot theme settings for plotting.
#'
#' @param    sizing  defines the style and sizing of the theme. One of 
#'                   \code{c("figure", "window")} where \code{sizing="figure"} is appropriately
#'                   sized for pdf export at 7 to 7.5 inch width, and \code{sizing="window"}
#'                   is sized for an interactive session.
#'
#' @return   A ggplot2 object.
#' 
#' @seealso \link[ggplot2]{theme}.
#' 
#' @export
baseTheme <- function(sizing=c("figure", "window")) {
    # Check arguments
    sizing <- match.arg(sizing)
    
    base_theme <- theme_bw() +
        theme(strip.background=element_blank(),
              plot.background=element_blank(),
              panel.grid.major=element_blank(), 
              panel.grid.minor=element_blank())
              
    # Define universal plot settings appropriate for PDF figures
    if (sizing == "figure") {
        base_theme <- base_theme + 
            theme(text=element_text(size=8),
                  plot.title=element_text(size=8),
                  strip.text=element_text(size=7, face="bold"),
                  axis.title=element_text(size=8, vjust=0.25),
                  axis.text.x=element_text(size=8, vjust=0.5, hjust=0.5),
                  axis.text.y=element_text(size=8),
                  legend.text=element_text(size=7),
                  legend.title=element_text(size=7),
                  legend.key.height=grid::unit(10, "points"), 
                  legend.key.width=grid::unit(10, "points"))
    } else if (sizing == "window") {
        # Define universal plot settings appropriate for an interactive session
        base_theme <- base_theme + 
            theme(text=element_text(size=14),
                  plot.title=element_text(size=16),
                  strip.text=element_text(size=14, face="bold"),
                  axis.title=element_text(size=16, vjust=0.25),
                  axis.text.x=element_text(size=16, vjust=0.5, hjust=0.5),
                  axis.text.y=element_text(size=16),
                  legend.text=element_text(size=14),
                  legend.title=element_text(size=14),
                  legend.key.height=grid::unit(18, "points"), 
                  legend.key.width=grid::unit(18, "points"))
    }
    
    return(base_theme)
}


#' Plot multiple ggplot objects
#' 
#' Plots multiple ggplot objects in an equally sized grid.
#' 
#' @param   ...    ggplot objects to plot.
#' @param   ncol   number of columns in the plot.
#' @return  NULL
#' 
#' @seealso \link[ggplot2]{ggplot}.
#' 
#' @references
#' Modified from:
#' http://www.cookbook-r.com/Graphs/Multiple_graphs_on_one_page_(ggplot2)
#' 
#' @export
gridPlot <- function(..., ncol=1) {
    p <- list(...)
    n <- length(p)
    layout <- matrix(seq(1, ncol*ceiling(n/ncol)), ncol=ncol, nrow=ceiling(n/ncol))
    
    # Plot
    if (n == 1) {
        plot(p[[1]])
    } else {
        grid::grid.newpage()
        grid::pushViewport(grid::viewport(layout=grid::grid.layout(nrow(layout), ncol(layout))))
        for (i in 1:n) {
            idx <- as.data.frame(which(layout == i, arr.ind=T))
            plot(p[[i]], vp=grid::viewport(layout.pos.row = idx$row, layout.pos.col=idx$col))
        }
    }
}

#### Multiprocessing functions ####

#' Available CPU cores
#'
#' \code{cpuCount} determines the number of CPU cores available.
#' 
#' @return  Count of available cores. Returns 1 if undeterminable.
#'
#' @examples
#' cpuCount()
#' 
#' @export
cpuCount <-function(){
    if (.Platform$OS.type == "windows") {
        nproc <- as.numeric(Sys.getenv('NUMBER_OF_PROCESSORS'))
    } else if (.Platform$OS.type == "unix") { 
        nproc <- parallel::detectCores()
    } else {
        nproc <- 1    
    }
	
	# in case an NA is returned
	if(is.na(nproc)){nproc <- 1}
    
    return(nproc)
}

#### Generic statistical functions ####

#' Weighted meta-analysis of p-values via Stouffer's method
#'
#' \code{stoufferMeta} combines multiple weighted p-values into a meta-analysis p-value
#' using Stouffer's Z-score method.
#' 
#' @param    p   numeric vector of p-values.
#' @param    w   numeric vector of weights.
#' 
#' @return   A named numeric vector with the combined Z-score and p-value in the form
#'           \code{c(Z, pvalue)}.
#' 
#' @examples
#' # Define p-value and weight vectors
#' p <- c(0.1, 0.05, 0.3)
#' w <- c(5, 10, 1)
#'
#' # Unweighted
#' stoufferMeta(p)
#' 
#' # Weighted
#' stoufferMeta(p, w)
#' 
#' @export
stoufferMeta <- function(p, w=NULL) {
    if (is.null(w)) {
        w <- rep(1, length(p))/length(p)
    } else {
        if (length(w) != length(p)) { stop("Length of p and w must equal.") }
        w <- w/sum(w)
    }
    x <- qnorm(1 - p)
    Z  <- sum(w*x) / sqrt(sum(w^2))
    pvalue <- 1 - pnorm(Z)
    
    return(c(Z=Z, pvalue=pvalue))
}


# Stirling's approximation of the binomial coefficient
# 
# \code{lchooseStirling} calculates Stirling's approximation of the binomial coefficient 
# for large numbers.
#
# @param    n  vector of n.
# @param    k  vector of k.
# @return   The approximation of log(n choose k). For n < 100 \link{lchoose} is used.
#
# @seealso  \link{lchoose}.
#
# @examples
# alakazam:::lchooseStirling(10e9, 10e4)
#
# @export
lchooseStirling <- function(n, k) {
    if (any(n < k)) {
        stop("n must be >= k")
    }
    
    n_len <- length(n)
    k_len <- length(k)
    nCk <- rep(NA, max(n_len, k_len))
    nCk[n == k] <- 0
    
    # a = index n_small
    # i = index k_small
    # x = index nCk_small
    #
    # b = index n_large
    # j = index k_large
    # y = index nCk_large
    #
    # Check for vector inputs and assign indexing
    if (n_len >= 1 & k_len >= 1 & n_len == k_len) {
        a <- i <- x <- (n < 100 & n != k)
        b <- j <- y <- (n >= 100 & n != k)
    } else if (n_len > 1 & k_len == 1) {
        a <- x <- (n < 100 & n != k)
        b <- y <- (n >= 100 & n != k)
        i <- j <- TRUE
    } else if (n_len == 1 & k_len > 1) {
        a <- (n < 100)
        b <- !a
        i <- j <- (n != k)
        x <- if (n < 100) { i } else { NULL }
        y <- if (n >= 100) { i } else { NULL }
    } else {
        stop("Inputs are wrong. n and k must have the same length or be length one.")
    } 
    
    # Small n
    nCk[x] <-  lchoose(n[a], k[i])
    
    # Large n indices
    nCk[y] <- n[b]*log(n[b]) - k[j]*log(k[j]) - (n[b] - k[j])*log(n[b] - k[j]) + 
        0.5*(log(n[b]) - log(k[j]) - log(n[b] - k[j]) - log(2*pi))
    
    #     .nCk <- function(n, k) {
    #         n*log(n) - k*log(k) - (n - k)*log(n - k) + 
    #         0.5*(log(n) - log(k) - log(n - k) - log(2*pi))
    #     }
    
    return(nCk)
}

test_that("translateStrings", {
    # Using a vector translation
    strings <- LETTERS[1:5]
    translation <- c("POSITION1"="A", "POSITION5"="E")
    obs <- translateStrings(strings, translation)
    exp <-  c("POSITION1", "B", "C", "D", "POSITION5")
    expect_equal(obs, exp)
    
    # Using a list translation
    translation <- list("1-3"=c("A","B","C"), "4-5"=c("D","E"))
    obs <- translateStrings(strings, translation)
    exp <- c("1-3", "1-3", "1-3", "4-5", "4-5")
    expect_equal(obs, exp)
})


test_that("stoufferMeta", {
    # Define p-value and weight vectors
    p <- c(0.1, 0.05, 0.3)
    w <- c(5, 10, 1)
    
    # Unweighted
    obs <- stoufferMeta(p)
    exp <- c("Z"=1.99232360, "pvalue"=0.02316778)
    expect_equal(obs, exp, tolerance=0.001)
    
    # Weighted
    obs <- stoufferMeta(p, w)    
    exp <- c("Z"=2.08291783, "pvalue"=0.01862936)
    expect_equal(obs, exp, tolerance=0.001)
})

# Clonal diversity analysis

#' @include Classes.R
NULL

#### Coverage functions ####

#' Calculate sample coverage
#' 
#' \code{calcCoverage} calculates the sample coverage estimate, a measure of sample 
#' completeness, for varying orders using the method of Chao et al, 2015, falling back 
#' to the Chao1 method in the first order case.
#'
#' @param    x  numeric vector of abundance counts.
#' @param    r  coverage order to calculate.
#' 
#' @return   The sample coverage of the given order \code{r}.
#' 
#' @references
#' \enumerate{
#'   \item  Chao A. Nonparametric Estimation of the Number of Classes in a Population. 
#'            Scand J Stat. 1984 11, 265270.
#'   \item  Chao A, et al. Unveiling the species-rank abundance distribution by 
#'            generalizing the Good-Turing sample coverage theory. 
#'            Ecology. 2015 96, 11891201.
#' }
#' 
#' @seealso  
#' Used by \link{alphaDiversity}.
#'           
#' @examples
#' # Calculate clone sizes
#' clones <- countClones(ExampleDb, groups="sample_id")
#' 
#' # Calculate 1first order coverage for a single sample
#' calcCoverage(clones$seq_count[clones$sample_id == "+7d"])
#'
#' @export
calcCoverage <- function(x, r=1) {
    # Use traditional calculation for 1st order coverage
    if (r == 1) { return(calcChao1Coverage(x)) }
    
    # Use general form for 2nd order and higher coverage
    x <- x[x >= 1]
    n <- sum(x)
    fr <- sum(x == r)
    fs <- sum(x == r + 1)
    
    if (fr == 0) {
        stop("Cannot calculate coverage of order ", r, ". No abundance data with count=", r, ".")
    }
    if (fs == 0) {
        stop("Cannot calculate coverage of order ", r, ". No abundance data with count=", r + 1, ".")
    }
    
    a <- factorial(r)*fr / sum(x[x >= r]^r)
    b <- ((n - r)*fr / ((n - r)*fr + (r + 1)*fs))^r
    rC <- 1 - a*b
    
    return(rC)
}


# Calculate first order coverage
#
# @param    x  a numeric vector of species abundance as counts
#
# @returns  Coverage estimate.
calcChao1Coverage <- function(x) {
    x <- x[x >= 1]
    n <- sum(x)
    f1 <- sum(x == 1)
    f2 <- sum(x == 2)
    
    if (f2 > 0) {
        rC1 <- 1 - (f1 / n) * (((n - 1) * f1) / ((n - 1) * f1 + 2 * f2))
    } else {
        rC1 <- 1 - (f1 / n) * (((n - 1) * (f1 - 1)) / ((n - 1) * (f1 - 1) + 2))
    }
    
    return(rC1)
}


# Calculates diversity under rarefaction
# 
# Calculates Hill numbers under rarefaction
#
# @param    x  vector of observed abundance counts.
# @param    m  the sequence count to rarefy to.
#
# @return   The first order coverage estimate
inferRarefiedCoverage <- function(x, m) {
    x <- x[x >= 1]
    n <- sum(x)
    if (m > n) {
        stop("m must be <= the total count of observed sequences.")
    }
    
    # Unrarefied case
    if (m == n) {
        return(calcCoverage(x, r=1))
    }
    
    # Calculate rarefied coverage
    # TODO: Read up on this and fix
    #rC1 <- iNEXT:::Chat.Ind(x, m)
    y <- x[(n - x) >= m]
    rC1 <- 1 - sum(y/n * exp(lgamma(n - y + 1) - lgamma(n - m - y + 1) - lgamma(n) + lgamma(n - m)))
    
    return(rC1)
}


#### Abundance functions ####

# Calculate undetected species
# 
# Calculates the lower bound of undetected species counts using the Chao1 estimator.
#
# @param    x  vector of observed abundance counts.
# 
# @return   The count of undetected species.
inferUnseenCount <- function(x) {
    x <- x[x >= 1]
    n <- sum(x)
    f1 <- sum(x == 1)
    f2 <- sum(x == 2)
    
    if (f2 > 0) {
        f0 <- ceiling(((n - 1) * f1^2) / (n * 2 * f2))
    } else {
        f0 <- ceiling(((n - 1) * f1 * (f1 - 1)) / (n * 2))
    }
    
    return(f0)
}


# Define undetected species relative abundances
#
# @param    x  vector of detected species abundance counts.
# 
# @return   An adjusted detected species relative abundance distribution.
inferUnseenAbundance <- function(x) {
    x <- x[x >= 1]
    
    # Coverage
    rC1 <- calcCoverage(x, r=1)
    
    # Unseen count
    f0 <- inferUnseenCount(x)
    
    # Assign unseen relative abundance
    p <- rep((1 - rC1) / f0, f0)
    
    return(p)
}


# Adjustement to observed relative abundances
#
# @param    x  vector of observed abundance counts
#
# @return   An adjusted observed species relative abundance distribution.
adjustObservedAbundance <- function(x) {
    x <- x[x >= 1]
    n <- sum(x)
    
    # Coverage
    rC1 <- calcCoverage(x, r=1)
    
    # Calculate tuning parameter
    lambda <- (1 - rC1) / sum(x/n * exp(-x))
    
    # Define adjusted relative abundance
    p <- x/n * (1 -  lambda * exp(-x))
    
    return(p)
}


# Combined unseen inferrence and observed abundance adjustment
#
# @param    x  named vector of observed abundance counts by clone.
#
# @return   A vector containing the complete inferred abundance distribution.
#           Unseen species will be denote by a clone name starting with "U".
inferCompleteAbundance <- function(x) {
    # Infer complete abundance distribution
    p1 <- adjustObservedAbundance(x)
    p2 <- inferUnseenAbundance(x)
    names(p2) <- if (length(p2) > 0) { paste0("U", 1:length(p2)) } else { NULL }
    
    return(c(p1, p2))
}

#' Tabulates clones sizes
#' 
#' \code{countClones} determines the number of sequences and total copy number of 
#' clonal groups.
#'
#' @param    data    data.frame with columns containing clonal assignments.
#' @param    groups  character vector defining \code{data} columns containing grouping 
#'                   variables. If \code{groups=NULL}, then do not group data.
#' @param    copy    name of the \code{data} column containing copy numbers for each 
#'                   sequence. If this value is specified, then total copy abundance
#'                   is determined by the sum of copy numbers within each clonal group.
#' @param    clone   name of the \code{data} column containing clone identifiers.
#' @param    cell_id name of the \code{data} column containing cell identifiers. If
#'                   \code{cell_id} column is not present the function will assume bulk data.
#' @param    remove_na    removes rows with \code{NA} values in the clone column if \code{TRUE} and issues a warning. 
#'                        Otherwise, keeps those rows and considers \code{NA} as a clone in the final counts 
#'                        and relative abundances.
#' 
#' @return   A data.frame summarizing clone counts and frequencies with columns:
#'           \itemize{
#'             \item \code{clone_id}:    clone identifier. This is the default column
#'                                       name, specified with \code{clone='clone_id'}.
#'                                       If the function call uses Change-O 
#'                                       formatted data and \code{clone='CLONE'}, this
#'                                       column will have name \code{CLONE}.
#'             \item \code{seq_count}:   total number of sequences for the clone.
#'             \item \code{seq_freq}:    frequency of the clone as a fraction of the total
#'                                       number of sequences within each group.
#'             \item \code{copy_count}:  sum of the copy counts in the \code{copy} column.
#'                                       Only present if the \code{copy} argument is 
#'                                       specified.
#'             \item \code{copy_freq}:   frequency of the clone as a fraction of the total
#'                                       copy number within each group. Only present if 
#'                                       the \code{copy} argument is specified.
#'           }
#'           Also includes additional columns specified in the \code{groups} argument.
#' 
#' @examples
#' # Without copy numbers
#' clones <- countClones(ExampleDb, groups="sample_id")
#'
#' # With copy numbers and multiple groups
#' clones <- countClones(ExampleDb, groups=c("sample_id", "c_call"), copy="duplicate_count")
#' 
#' @export
countClones <- function(data, groups=NULL, copy=NULL, clone="clone_id", cell_id="cell_id", 
                        remove_na=TRUE) {
    # Check input
    check <- checkColumns(data, c(clone, copy, groups))
    if (check != TRUE) { 
        warning(check) # instead of throwing an error and potentially disrupting a workflow
    }

    # Don't allow a copy column for single-cell data as it doesn't make sense to count copies
    if (cell_id %in% names(data) & !is.null(copy)) {
        stop("Copy column specification not allowed for single-cell or mixed bulk and single-cell data. 
        A cell_id column is present in the dataframe single-cell or mixed bulk and single-cell data is assumed.")
    }

    # Handle NAs
    if (remove_na) {
        bool_na <- is.na(data[, clone])
        if (any(bool_na)) {
            if  (!all(bool_na)){
                msg <- paste0("NA(s) found in ", sum(bool_na), " row(s) of the ", clone, 
                            " column and excluded from tabulation")
                warning(msg)
            }
            data <- data[!bool_na, ]
        }
    }

    # Tabulate clonal abundance
    if (cell_id %in% names(data)) {
        # Handle single-cell and mixed bulk and single-cell case
        data_sc <- data %>% dplyr::filter(!is.na(!!rlang::sym(cell_id)))
        data_sc[[cell_id]] <- as.character(data_sc[[cell_id]])
        data_blk <- data %>% dplyr::filter(is.na(!!rlang::sym(cell_id)))
        if (nrow(data_blk) > 0){
            data_blk[[cell_id]] <- paste0("bulk", 1:nrow(data_blk)) #dummy cell_id for bulk data
        }
        data <- bind_rows(data_sc, data_blk)

        clone_tab <- data %>% 
            dplyr::select(!!!rlang::syms(c(groups, clone, cell_id))) %>%
            dplyr::distinct() %>%
            dplyr::group_by(!!!rlang::syms(c(groups, clone))) %>%
            dplyr::summarize(seq_count=n()) %>%
            dplyr::mutate(seq_freq=!!rlang::sym("seq_count")/sum(!!rlang::sym("seq_count"), na.rm=TRUE)) %>%
            dplyr::arrange(desc(!!rlang::sym("seq_count")))
    } else {
        if (is.null(copy)) {
            clone_tab <- data %>% 
                dplyr::group_by(!!!rlang::syms(c(groups, clone))) %>%
                dplyr::summarize(seq_count=n()) %>%
                dplyr::mutate(seq_freq=!!rlang::sym("seq_count")/sum(!!rlang::sym("seq_count"), na.rm=TRUE)) %>%
                dplyr::arrange(desc(!!rlang::sym("seq_count")))

        } else {
            clone_tab <- data %>% 
                dplyr::group_by(!!!rlang::syms(c(groups, clone))) %>%
                dplyr::summarize(seq_count=length(.data[[clone]]),
                                copy_count=sum(.data[[copy]], na.rm=TRUE)) %>%
                dplyr::mutate(seq_freq=!!rlang::sym("seq_count")/sum(!!rlang::sym("seq_count"), na.rm=TRUE),
                            copy_freq=!!rlang::sym("copy_count")/sum(!!rlang::sym("copy_count"), na.rm=TRUE)) %>%
                dplyr::arrange(desc(!!rlang::sym("copy_count"))) 
        }
    }

    return(clone_tab)
}


# Perform boostrap abundance calculation
# 
# @param    x       named vector of observed abundance values.
# @param    n       number of samples to draw from the estimate complete abundance distribution.
# @param    nboot   number of bootstrap realizations.
# @param    method  complete abundance inferrence method. 
#                   One of "before", "after" or "none" for complete abundance distribution
#                   inferrence before sampling, after sampling, or uncorrected, respectively.
# 
# @return   A matrix of bootstrap results.
bootstrapAbundance <- function(x, n, nboot=200, method="before") {
    ## DEBUG
    # x=abund_obs; method="before"
    # Check arguments
    method <- match.arg(method)
  
    if (method == "before") {
        # Calculate estimated complete abundance distribution
        p <- inferCompleteAbundance(x)
        # Bootstrap abundance
        boot_mat <- rmultinom(nboot, n, p) / n
    } else if (method == "after") {
        # Calculate estimated complete abundance distribution
        p <- x / sum(x, na.rm=TRUE)
        boot_sam <- rmultinom(nboot, n, p)
        boot_list <- apply(boot_sam, 2, inferCompleteAbundance)
        
        # Convert to matrix
        boot_names <- unique(unlist(sapply(boot_list, names)))
        boot_mat <- matrix(0, nrow=length(boot_names), ncol=nboot)
        rownames(boot_mat) <- boot_names
        for (i in 1:nboot) {
            boot_mat[names(boot_list[[i]]), i] <- boot_list[[i]]
        } 
    } else if (method == "none") {
        # Raw sampling of input
        p <- x / sum(x, na.rm=TRUE)
        boot_sam <- rmultinom(nboot, n, p)
    } else {
        stop("Invalid method: ", method)
    }
      
    return(boot_mat)
}

#' Estimates the complete clonal relative abundance distribution
#' 
#' \code{estimateAbundance} estimates the complete clonal relative abundance distribution 
#' and confidence intervals on clone sizes using bootstrapping.
#' 
#' @param    data      data.frame with Change-O style columns containing clonal assignments.
#' @param    clone     name of the \code{data} column containing clone identifiers.
#' @param    copy      name of the \code{data} column containing copy numbers for each 
#'                     sequence. If \code{copy=NULL} (the default), then clone abundance
#'                     is determined by the number of sequences. If a \code{copy} column
#'                     is specified, then clone abundances is determined by the sum of 
#'                     copy numbers within each clonal group.
#' @param    group     name of the \code{data} column containing group identifiers. 
#'                     If \code{NULL} then no grouping is performed and the \code{group} 
#'                     column of the output will contain the value \code{NA} for each row.
#' @param    min_n     minimum number of observations to sample.
#'                     A group with less observations than the minimum is excluded.
#' @param    max_n     maximum number of observations to sample. If \code{NULL} then no 
#'                     maximum is set.
#' @param    uniform   if \code{TRUE} then uniformly resample each group to the same 
#'                     number of observations. If \code{FALSE} then allow each group to
#'                     be resampled to its original size or, if specified, \code{max_size}.
#' @param    ci        confidence interval to calculate; the value must be between 0 and 1.
#' @param    nboot     number of bootstrap realizations to generate.
#' @param    cell_id name of the \code{data} column containing cell identifiers. If
#'                     \code{cell_id=NULL} then the function will assume bulk data.
#' @param    progress  if \code{TRUE} show a progress bar. 
#' 
#' @return   A \link{AbundanceCurve} object summarizing the abundances.
#'           
#' @references
#' \enumerate{
#'   \item  Chao A. Nonparametric Estimation of the Number of Classes in a Population. 
#'            Scand J Stat. 1984 11, 265270.
#'   \item  Chao A, et al. Rarefaction and extrapolation with Hill numbers: 
#'            A framework for sampling and estimation in species diversity studies. 
#'            Ecol Monogr. 2014 84:45-67.
#'   \item  Chao A, et al. Unveiling the species-rank abundance distribution by 
#'            generalizing the Good-Turing sample coverage theory. 
#'            Ecology. 2015 96, 11891201.
#' }
#' 
#' @seealso  
#' See \link{plotAbundanceCurve} for plotting of the abundance distribution.
#' See \link{alphaDiversity} for a similar application to clonal diversity.
#'           
#' @examples
#' abund <- estimateAbundance(ExampleDb, group="sample_id", nboot=100)
#'
#' @export
estimateAbundance <- function(data, clone="clone_id", copy=NULL, group=NULL, 
                              min_n=30, max_n=NULL, uniform=TRUE, ci=0.95, nboot=200,
                              cell_id="cell_id", progress=FALSE) {
    
    # TODO:
    # Add alakazam style cell_id=NULL,
    # similar to distToNearest
    # filter based on locusValues
    # if cell_id
    #    for rows that have unique cell_id, ok
    #    if rows have cell_id not unique, count only once
    # if not cell_id, count heavy chains (locusValues will be IGH)
    # If mixed bulk and sc do calculation but raise warning because different type of abundances
    
    ## DEBUG
    # data=ExampleDb; group="sample_id"; clone="clone_id"; copy=NULL; min_n=1; max_n=NULL; ci=0.95; uniform=F; nboot=100
    # copy="duplicate_count"
    # group=NULL

    # Hack for visibility of dplyr variables
    #. <- NULL
    
    # Check input
    if (!is.data.frame(data)) {
        stop("Input data is not a data.frame")
    }
    
    # Check columns that are reported are real columns (can be NULL)
    check <- checkColumns(data, c(clone, copy, group))
    if (check != TRUE) { stop(check) }
    
    # Set confidence interval
    ci_z <- ci + (1 - ci) / 2
    ci_x <- qnorm(ci_z)
    
    # Tabulate clonal abundance
    count_col <- if (!is.null(copy)) { "copy_count" } else { "seq_count" }
    clone_tab <- countClones(data, copy=copy, clone=clone, groups=group, 
                                cell_id=cell_id) %>%
        dplyr::mutate(clone_count=!!rlang::sym(count_col))

    # Tabulate group sizes
    if (!is.null(group)) {
        # Summarize groups
        group_tab <- clone_tab %>%
            group_by(!!rlang::sym(group)) %>%
            dplyr::summarize(count=sum(!!rlang::sym("clone_count"), na.rm=TRUE)) %>%
            rename(group=!!rlang::sym(group))
    } else {
        group_tab <- data.frame(v="All", count=sum(clone_tab$clone_count, na.rm=T))
        names(group_tab)[1] <- "group"
    }
    group_all <- as.character(group_tab$group)
    group_tab <- group_tab[group_tab$count >= min_n, ]
    group_keep <- as.character(group_tab$group)
    
    # Set number of sampled sequence
    if (uniform) {
        nsam <- min(group_tab$count, max_n)
        nsam <- setNames(rep(nsam, length(group_keep)), group_keep)
    } else {
        nsam <- if (is.null(max_n)) { group_tab$count } else { pmin(group_tab$count, max_n) }
        nsam <- setNames(nsam, group_keep)
    }
    
    # Warn if groups removed
    if (length(group_keep) < length(group_all)) {
        warning("Not all groups passed threshold min_n=", min_n, ".", 
                " Excluded: ", paste(setdiff(group_all, group_keep), collapse=", "))
    }
    
    # Generate abundance bootstrap
    if (progress) { 
        pb <- progressBar(length(group_keep))
    }
    boot_list <- list()
    abund_list <- list()
    for (g in group_keep) {
        n <- nsam[g]
        
        # Extract abundance vector
        if (!is.null(group)) {
            abund_obs <- clone_tab$clone_count[clone_tab[[group]] == g]
            names(abund_obs) <- clone_tab[[clone]][clone_tab[[group]] == g]
        } else {
            # Extract abundance vector
            abund_obs <- clone_tab$clone_count
            names(abund_obs) <- clone_tab[[clone]]
        } 
        
        # Infer complete abundance distribution
        boot_mat <- bootstrapAbundance(abund_obs, n, nboot=nboot, method="before")
        
        # Assign confidence intervals based on variance of bootstrap realizations
        p_mean <- apply(boot_mat, 1, mean)
        p_sd <- apply(boot_mat, 1, sd)
        p_err <- ci_x * p_sd
        p_lower <- pmax(p_mean - p_err, 0)
        p_upper <- p_mean + p_err
        
        # Assemble and sort abundance data.frame
	    abund_df <- tibble::tibble(!!clone := rownames(boot_mat), p=p_mean, p_sd=p_sd,
	                           lower=p_lower, upper=p_upper) %>%
	        dplyr::arrange(desc(!!rlang::sym("p"))) %>%
	        dplyr::mutate(rank=1:n())
			
        # Save summary
        abund_list[[g]] <- abund_df
        
        # Save bootstrap
        boot_list[[g]] <- as.data.frame(boot_mat) %>%
          tibble::rownames_to_column(clone)
        
        if (progress) { pb$tick() }
    }
    id_col <- "group"
    if (!is.null(group)) { id_col <- group }
    abundance_df <- as.data.frame(bind_rows(abund_list, .id=id_col))
    bootstrap_df <- as.data.frame(bind_rows(boot_list, .id=id_col))
    
    # Create a new diversity object with bootstrap
    abund_obj <- new("AbundanceCurve",
                     bootstrap=bootstrap_df, 
                     abundance=abundance_df,
                     clone_by=clone,
                     group_by=id_col,
                     #groups=if_else(is.null(group), as.character(NA), group_keep),
                     groups=group_keep,
                     n=nsam, 
                     nboot=nboot, 
                     ci=ci)
    
    return(abund_obj)
}

#### Diversity functions ####

#' Calculate the diversity index
#' 
#' \code{calcDiversity} calculates the clonal diversity index for a vector of diversity 
#' orders. 
#'
#' @param    p  numeric vector of clone (species) counts or proportions.
#' @param    q  numeric vector of diversity orders.
#' 
#' @return   A vector of diversity scores \eqn{D} for each \eqn{q}.
#' 
#' @details
#' This method, proposed by Hill (Hill, 1973), quantifies diversity as a smooth function 
#' (\eqn{D}) of a single parameter \eqn{q}. Special cases of the generalized diversity 
#' index correspond to the most popular diversity measures in ecology: species richness 
#' (\eqn{q = 0}), the exponential of the Shannon-Weiner index (\eqn{q} approaches \eqn{1}), the 
#' inverse of the Simpson index (\eqn{q = 2}), and the reciprocal abundance of the largest 
#' clone (\eqn{q} approaches \eqn{+\infty}). At \eqn{q = 0} different clones weight equally, 
#' regardless of their size. As the parameter \eqn{q} increase from \eqn{0} to \eqn{+\infty} 
#' the diversity index (\eqn{D}) depends less on rare clones and more on common (abundant) 
#' ones, thus encompassing a range of definitions that can be visualized as a single curve. 
#' 
#' Values of \eqn{q < 0} are valid, but are generally not meaningful. The value of \eqn{D} 
#' at \eqn{q=1} is estimated by \eqn{D} at \eqn{q=0.9999}. 
#'
#' @references
#' \enumerate{
#'   \item  Hill M. Diversity and evenness: a unifying notation and its consequences. 
#'            Ecology. 1973 54(2):427-32.
#' }
#' 
#' @seealso  Used by \link{alphaDiversity}.
#' 
#' @examples
#' # May define p as clonal member counts
#' p <- c(1, 1, 3, 10)
#' q <- c(0, 1, 2)
#' calcDiversity(p, q)
#'
#' # Or proportional abundance
#' p <- c(1/15, 1/15, 1/5, 2/3)
#' calcDiversity(p, q)
#' 
#' @export
calcDiversity <- function(p, q) {
    # Add jitter to q=1
    q[q == 1] <- 0.9999
    # Remove zeros
    p <- p[p > 0]
    # Convert p to proportional abundance
    p <- p / sum(p)
    # Calculate D for each q
    D <- sapply(q, function(x) sum(p^x)^(1 / (1 - x)))
    
    return(D)
}

# Calculate the inferred diversity index
# 
# \code{calcInferredDiversity} calculates the clonal diversity index for a vector of diversity 
# orders with a correction for the presence of unseen species. Does not take proportional abundance.
#
# @param    p  numeric vector of clone (species) counts.
# @param    q  numeric vector of diversity orders.
# 
# @return   A vector of diversity scores \eqn{D} for each \eqn{q}.
# 
# @details
# This method, proposed by Hill (Hill, 1973), quantifies diversity as a smooth function 
# (\eqn{D}) of a single parameter \eqn{q}. Special cases of the generalized diversity 
# index correspond to the most popular diversity measures in ecology: species richness 
# (\eqn{q = 0}), the exponential of the Shannon-Weiner index (\eqn{q} approaches \eqn{1}), the 
# inverse of the Simpson index (\eqn{q = 2}), and the reciprocal abundance of the largest 
# clone (\eqn{q} approaches \eqn{+\infty}). At \eqn{q = 0} different clones weight equally, 
# regardless of their size. As the parameter \eqn{q} increase from \eqn{0} to \eqn{+\infty} 
# the diversity index (\eqn{D}) depends less on rare clones and more on common (abundant) 
# ones, thus encompassing a range of definitions that can be visualized as a single curve. 
# 
# Values of \eqn{q < 0} are valid, but are generally not meaningful. The value of \eqn{D} 
# at \eqn{q=1} is estimated by \eqn{D} at \eqn{q=0.9999}. 
# 
# An adjusted detected species relative abundance distribution is applied before calculating diversity.
#
# @references
# \enumerate{
#   \item  Hill M. Diversity and evenness: a unifying notation and its consequences. 
#            Ecology. 1973 54(2):427-32.
# }
# 
# @seealso  Used by \link{alphaDiversity}
# 
# @examples
# # May define p as clonal member counts
# p <- c(1, 1, 3, 10)
# q <- c(0, 1, 2)
# calcInferredDiversity(p, q)
#
# 
# @export
# calcInferredDiversity <- function(p, q) {
#    # Correct abundance
#     .infer <- function(y) {
#         # Infer complete abundance distribution
#         p1 <- adjustObservedAbundance(y)
#         p2 <- inferUnseenAbundance(y)
#         names(p2) <- if (length(p2) > 0) { paste0("U", 1:length(p2)) } else { NULL }
#         return(c(p1, p2))
#     }
#    
#    # Correct abundance
#    p <- .infer(p)
#     # Add jitter to q=1
#     q[q == 1] <- 0.9999
#     # Remove zeros
#     p <- p[p > 0]
#     # Convert p to proportional abundance
#     p <- p / sum(p)
#     # Calculate D for each q
#     D <- sapply(q, function(x) sum(p^x)^(1 / (1 - x)))
#     
#     return(D)
# }


# Calculates diversity under rarefaction
# 
# Calculates Hill numbers under rarefaction
#
# @param    x  vector of observed abundance counts.
# @param    q  numeric vector of diversity orders.
# @param    m  the sequence count to rarefy to.
#
# @return   A vector of diversity scores \eqn{D} for each \eqn{q}.
inferRarefiedDiversity <- function(x, q, m) {
    x <- x[x >= 1]
    n <- sum(x)
    if (m > n) {
        stop("m must be <= the total count of observed sequences.")
    }
    q[q == 1] <- 0.9999
    
    # Tabulate frequency counts from 1:n
    fk_n <- tabulate(x, nbins=n)
    
    # Calculate estimated fk(m)
    fk_m <- sapply(1:m, function(k) sum(exp(lchoose(k:m, k) + 
                                            lchoose(n - k:m, m - k) - 
                                            lchoose(n, m))*fk_n[k:m]))
    
    # Calculate diversity
    D <- sapply(q, function(r) sum((1:m / m)^r * fk_m)^(1 / (1 - r)))
    
    return(D)
}


# Helper function for computing alpha diversity from bootrstrap outputs
#
# \code{helperAlpha} divides a set of bootstrapped clones by group annotation,
# and computes the diversity of each set. 
#
# @param    boot_output  data.frame from\link{AbundanceCurve} object containing bootstrapped clonal 
#                        abundance curves.
# @param    q            vector of Hill Diversity indices to test for diversity calculations.
# @param    clone        name of the \code{boot_output} column containing clone identifiers.
# @param    group        name of the \code{boot_output} column containing grouping information for 
#                        diversity calculation.
#
# @return   data.frame containing diversity calculations for each bootstrap iteration.
helperAlpha <- function(boot_output, q, clone="clone_id", group=NULL) {
    ## DEBUG
    # abundance <- estimateAbundance(ExampleDb, group="sample_id", nboot=100)
    # clone <- abundance@clone_by
    # group <- abundance@group_by
  
    # Compute diversity from a column of each bootstrap
    output <- boot_output %>% 
        dplyr::ungroup() %>%
        dplyr::select(-one_of(c(clone, group))) %>%
        as.matrix() %>% 
        apply(2, calcDiversity, q=q) %>%
        data.frame() %>% 
        mutate(q=q)

    return(output)
}


# Helper function for computing beta diversity from bootrstrap outputs
#
# \code{helperBeta} divides a set of bootstrapped clones by group annotation,
# and computes the alpha diversity. Group annotations are then ignored and 
# gamma diversity is computed. A multiplicative beta diversity is used corresponding
# to the gamma diversity divided by the average alpha diversity of each group.
#
# @param    boot_output   data.frame from\link{AbundanceCurve} object containing bootstrapped clonal abundance curves.
# @param    q             vector of Hill Diversity indices to test for diversity calculations.
# @param    ci_z          numeric value corresponding to confidence interval for calculating beta diversity.
# @param    clone         name of the \code{boot_output} column containing clone identifiers.
# @param    group         name of the \code{boot_output} column containing grouping information for diversity 
#                         calculation.
#
# @return   data.frame containing diversity calculations for each bootstrap iteration.
helperBeta <- function(boot_output, q, ci_x, clone="clone_id", group="group") { 
    # Hack for visibility of dplyr variables
    . <- NULL
        
    # Compute gamma diversity metrics
    gamma <- boot_output %>%
        dplyr::group_by(!!rlang::sym(clone)) %>%
        dplyr::select(-one_of(c(group))) %>%
        dplyr::summarize_all(sum) %>%
        dplyr::do(helperAlpha(., q=q, clone=clone)) %>%
        tidyr::gather(key="n", value="gamma", -!!rlang::sym("q")) %>%
        dplyr::mutate(gamma=as.numeric(!!rlang::sym("gamma")))

    # Compute alpha diversity metrics
    alpha <- boot_output %>%
        dplyr::group_by(!!rlang::sym(group)) %>%
        dplyr::do(helperAlpha(., q=q, clone=clone, group=group)) %>%
        dplyr::group_by(!!rlang::sym("q")) %>%
        dplyr::select(-one_of(c(group))) %>%
        dplyr::summarize_all(mean) %>%
        tidyr::gather(key="n", value="alpha", -!!rlang::sym("q")) %>%
        dplyr::mutate(alpha=as.numeric(!!rlang::sym("alpha")))

    # Perform comparisons of alpha and gamma to extract beta
    beta <- bind_cols(gamma, alpha) %>%
        dplyr::group_by(!!rlang::sym("q")) %>%
        dplyr::mutate(X=!!rlang::sym("gamma") / !!rlang::sym("alpha")) %>%
        dplyr::summarize(d=mean(!!rlang::sym("X"), na.rm=TRUE),
                         d_sd=sd(!!rlang::sym("X"), na.rm=TRUE)) %>%
        dplyr::mutate(d_lower=pmax(!!rlang::sym("d") - !!rlang::sym("d_sd") * ci_x, 0), 
                      d_upper=!!rlang::sym("d") + !!rlang::sym("d_sd") * ci_x)

    return(beta)
}

# Helper function for computing statistical significance
#
# \code{helperTest} computes the pairwise statistical significance of differences
# in bootstrapped diversity values between two sets defined by the group column. 
# A p-value is computed using the ECDF distribution as the frequency of bootstrap iterations
# for which no difference is observed. 
#
# @param    div_df  data.frame from\link{DiversityCurve} object containing bootstrapped 
#                   diversity curves.
# @param    group   name of the \code{boot_output} column containing grouping information 
#                   for diversity calculation.
# @param    q       vector of Hill Diversity indices to test for diversity calculations.
#
# @return   data.frame containing test results for each value of q.
helperTest <- function(div_df, q, group="group") {
    # Hack for visibility of dplyr variables
    #. <- NULL
    
    # Pairwise test
    group_pairs <- combn(unique(div_df[[group]]), 2, simplify=F)
    pvalue_list <- list()
    for (group_pair in group_pairs) {
        pair_list <- list()
        for(q_i in q) {
            
            # Currently just testing for one diversity order
            mat1 <- div_df %>%
                dplyr::filter(!!rlang::sym(group) == group_pair[1], !!rlang::sym("q") == q_i) %>%
                dplyr::select(-one_of(c(group, "q"))) %>% unlist()
            mat2 <- div_df %>%
                dplyr::filter(!!rlang::sym(group) == group_pair[2], !!rlang::sym("q") == q_i) %>%
                dplyr::select(-one_of(c(group, "q"))) %>% unlist()

            if (mean(mat1) >= mean(mat2)) { 
                g_delta <- mat1 - mat2 
            } else { 
                g_delta <- mat2 - mat1 
            }  

            # Compute p-value from ecdf
            p <- ecdf(g_delta)(0)
            p <- ifelse(p <= 0.5, p * 2, (1 - p) * 2)
            
            pair_list[[as.character(q_i)]] <- list(delta_mean=mean(g_delta), 
                                                   delta_sd=sd(g_delta), 
                                                   pvalue=p)
        }
        pvalue_list[[paste(group_pair, collapse=" != ")]] <- bind_rows(pair_list, .id="q")

    }
    test_df <- bind_rows(pvalue_list, .id="test")

    return(test_df)
}


#' Calculate clonal alpha diversity
#'
#' \code{alphaDiversity} takes in a data.frame or \link{AbundanceCurve} and computes
#' diversity scores (\eqn{D}) over an interval of diversity orders (\eqn{q}).
#' 
#' @param    data      data.frame with Change-O style columns containing clonal assignments or
#'                     a \link{AbundanceCurve} generate by \link{estimateAbundance} object 
#'                     containing a previously calculated bootstrap distributions of clonal abundance.
#' @param    min_q     minimum value of \eqn{q}.
#' @param    max_q     maximum value of \eqn{q}.
#' @param    step_q    value by which to increment \eqn{q}.
#' @param    ci        confidence interval to calculate; the value must be between 0 and 1.
#' @param    ...       additional arguments to pass to \link{estimateAbundance}. Additional arguments
#'                     are ignored if a \link{AbundanceCurve} is provided as input.
#' 
#' @return   A \link{DiversityCurve} object summarizing the diversity scores.
#' 
#' @references
#' \enumerate{
#'   \item  Hill M. Diversity and evenness: a unifying notation and its consequences. 
#'            Ecology. 1973 54(2):427-32.
#'   \item  Chao A. Nonparametric Estimation of the Number of Classes in a Population. 
#'            Scand J Stat. 1984 11, 265270.
#'   \item  Chao A, et al. Rarefaction and extrapolation with Hill numbers: 
#'            A framework for sampling and estimation in species diversity studies. 
#'            Ecol Monogr. 2014 84:45-67.
#'   \item  Chao A, et al. Unveiling the species-rank abundance distribution by 
#'            generalizing the Good-Turing sample coverage theory. 
#'            Ecology. 2015 96, 11891201.
#' }
#'  
#' @seealso  See \link{calcDiversity} for the basic calculation and 
#'           \link{DiversityCurve} for the return object. 
#'           See \link{plotDiversityCurve} for plotting the return object.
#' 
#' @details
#' Clonal diversity is calculated using the generalized diversity index (Hill numbers) 
#' proposed by Hill (Hill, 1973). See \link{calcDiversity} for further details.
#'
#' To generate a smooth curve, \eqn{D} is calculated for each value of \eqn{q} from
#' \code{min_q} to \code{max_q} incremented by \code{step_q}.  When \code{uniform=TRUE}
#' variability in total sequence counts across unique values in the \code{group} column 
#' is corrected by repeated resampling from the estimated complete clonal distribution to a 
#' common number of sequences. The complete clonal abundance distribution that is resampled 
#' from is inferred by using the Chao1 estimator to infer the number of unseen clones, 
#' followed by applying the relative abundance correction and unseen clone frequencies 
#' described in Chao et al, 2015.
#' 
#' The diversity index (\eqn{D}) for each group is the mean value of over all resampling 
#' realizations. Confidence intervals are derived using the standard deviation of the 
#' resampling realizations, as described in Chao et al, 2015.
#'
#' Significance of the difference in diversity index (\code{D}) between groups is tested by 
#' constructing a bootstrap delta distribution for each pair of unique values in the 
#' \code{group} column. The bootstrap delta distribution is built by subtracting the diversity 
#' index \code{Da} in group \code{a} from the corresponding value \eqn{Db} in group \code{b}, 
#' for all bootstrap realizations, yielding a distribution of \code{nboot} total deltas; where 
#' group \code{a} is the group with the greater mean \code{D}. The p-value for hypothesis 
#' \code{Da  !=  Db} is the value of \code{P(0)} from the empirical cumulative distribution 
#' function of the bootstrap delta distribution, multiplied by 2 for the two-tailed correction.
#' 
#' Note, this method may inflate statistical significance when clone sizes are uniformly small,
#' such as when most clones sizes are 1, sample size is small, and \code{max_n} is near
#' the total count of the smallest data group. Use caution when interpreting the results 
#' in such cases.
#'
#' @examples
#' # Group by sample identifier in two steps
#' set.seed(123)
#' abund <- estimateAbundance(ExampleDb, group="sample_id", nboot=100)
#' div <- alphaDiversity(abund, step_q=1, max_q=10)
#' plotDiversityCurve(div, legend_title="Sample")
#'                    
#' # Grouping by isotype rather than sample identifier in one step
#' div <- alphaDiversity(ExampleDb, group="c_call", min_n=40, step_q=1, max_q=10, 
#'                       nboot=100)
#' plotDiversityCurve(div, legend_title="Isotype")
#'
#' @export
alphaDiversity <- function(data, min_q=0, max_q=4, step_q=0.1, ci=0.95, ...) {
    # Hack for visibility of dplyr variables
    . <- NULL
    
    # Check input object and call estimateAbundance if required
    if (is(data, "AbundanceCurve")) {
        abundance <- data
    } else if (is(data, "data.frame")) {
        abundance <- estimateAbundance(data, ci=0.95, ...)
    } else {
        stop("Input must be either a data.frame or AbundanceCurve object.")
    }
    
    # Set diversity orders and confidence interval
    ci_z <- ci + (1 - ci) / 2
    ci_x <- qnorm(ci_z)
    q <- seq(min_q, max_q, step_q)
    if (!(0 %in% q)) { q <- c(0, q) }
    
    # Set grouping variables
    clone <- abundance@clone_by
    group <- abundance@group_by
        
    # Compute diversity metric for bootstrap instances
    boot_df <- abundance@bootstrap %>%
        dplyr::group_by(!!rlang::sym(group)) %>%
        dplyr::do(helperAlpha(., q=q, clone=clone, group=group)) %>%
        dplyr::ungroup()
    
    # Summarize diversity
    div_df <- boot_df %>%
        tidyr::gather(key="n", value="X", -one_of(c(group, "q"))) %>%
        dplyr::mutate(X=as.numeric(!!rlang::sym("X"))) %>%
        dplyr::group_by(!!!rlang::syms(c(group, "q"))) %>%
        dplyr::summarize(d=mean(!!rlang::sym("X"), na.rm=TRUE),
                         d_sd=sd(!!rlang::sym("X"), na.rm=TRUE)) %>%
        dplyr::mutate(d_lower=pmax(!!rlang::sym("d") - !!rlang::sym("d_sd") * ci_x, 0), 
                      d_upper=!!rlang::sym("d") + !!rlang::sym("d_sd") * ci_x)
    
    # Compute evenness
    div_qi <- div_df %>%
        filter(!!rlang::sym("q") == 0) %>%
        select(one_of(c(group, "d")))
    div_df <- div_df %>%
        dplyr::right_join(div_qi, by=group, suffix=c("", "_0")) %>%
        mutate(e=!!rlang::sym("d")/!!rlang::sym("d_0"), 
               e_lower=!!rlang::sym("d_lower")/!!rlang::sym("d_0"), 
               e_upper=!!rlang::sym("d_upper")/!!rlang::sym("d_0")) %>%
        select(-!!rlang::sym("d_0"))
    
    # Test
    if (length(abundance@groups) > 1) {
        test_df <- helperTest(boot_df, q=q, group=group)
    } else {
        test_df <- NULL
    }

    # Build return object
    group_set <- unique(div_df[[group]])
    div_obj <- new("DiversityCurve",
                   diversity=div_df, 
                   tests=test_df,
                   method="alpha",
                   group_by=group,
                   groups=group_set,
                   q=q,  
                   n=abundance@n,
                   ci=ci)
        
   return(div_obj) 
}

# Calculates the pairwise beta diversity
# 
# \code{betaDiversity} takes in a data.frame or \link{AbundanceCurve} and computes
# the multiplicative beta diversity across a range of Hill diversity indices.
# 
# @param    data         data.frame with Change-O style columns containing clonal assignments or
#                        an \link{AbundanceCurve} object generate by \link{estimateAbundance}.
#                        containing a previously calculated bootstrap distributions of clonal abundance.
# @param    comparisons  named list of comparisons between group members for computing beta diversity.
# @param    min_q        minimum value of \eqn{q}.
# @param    max_q        maximum value of \eqn{q}.
# @param    step_q       value by which to increment \eqn{q}.
# @param    ci           confidence interval to calculate; the value must be between 0 and 1.
# @param    ...          additional arguments to pass to \link{estimateAbundance}. Additional arguments
#                        are ignored if a \link{AbundanceCurve} is provided as input.
# 
# @return   A \link{DiversityCurve} object summarizing the diversity scores.
# 
# @details
# Beta diversity or the comparative difference between two samples as quantified using Hill
# diversity indices proposed by Jost (Jost, 2007).
# 
# Briefly, the alpha and gamma diversity components are calculated for each comparison.
# Alpha diversity is calculated as the average hill diversity across each independent sample
# while Gamma diversity is calculated as the total diversity without distinguishing between
# samples. Beta diversity is computed as Gamma/Alpha.
# 
# Diversity is calculated on the estimated clonal abundance distribution with a correction
# for unseen species much like the calculation for alpha diversity \link{alphaDiversity}.
# A smooth curve is generated in the same manner as in \link{alphaDiversity}.
# Confidence intervals are derived using the standard deviation of the resampling realizations.
# 
# \enumerate{
#   \item  Hill M. Diversity and evenness: a unifying notation and its consequences.
#            Ecology. 1973 54(2):427-32.
#   \item  Jost L. Partitioning Diversity Into Independent Alpha and Beta Components.
#            Ecology. 2007 88(10):2427–2439.
#   \item  Jost L, et al. Partitioning diversity for conservation analyses.
#            Diversity Distrib. 2010 16(1):65–76
# }
# 
# @examples
# div <- betaDiversity(ExampleDb, comparisons=list("TIME"=c("-1h", "+7d")), group="sample_id",
#                      min_n=40, step_q=1, max_q=10, nboot=100)
# 
# plotDiversityCurve(div, legend_title="Isotype")
# 
# @export
betaDiversity <- function(data, comparisons, min_q=0, max_q=4, step_q=0.1, ci=0.95, ...) {
    # Hack for visibility of dplyr variables
    . <- NULL
    
    if (!is.list(comparisons) || is.null(names(comparisons))) {
      stop("'comparisons' must be a named list")
    }
    
    # Check input object and call estimateAbundance if required
    if (is(data, "AbundanceCurve")) {
        abundance <- data
    } else if (is(data, "data.frame")) {
        abundance <- estimateAbundance(data, ci=0.95, ...)
    } else {
        stop("Input must be either a data.frame or AbundanceCurve object.")
    }
        
    # Set diversity orders and confidence interval
    ci_z <- ci + (1 - ci) / 2
    ci_x <- qnorm(ci_z)
    q <- seq(min_q, max_q, step_q)
    if (!(0 %in% q)) { q <- c(0, q) }
        
    # Compute pairwise beta diversity for bootstrap instances
    beta_diversity_list <- list()

    for (comparison in names(comparisons)) {
        beta_diversity_list[[comparison]] <- abundance@bootstrap %>%
            dplyr::ungroup() %>%
            dplyr::filter(.[[abundance@group_by]] %in% comparisons[[comparison]]) %>%
            dplyr::do(helperBeta(., q=q, clone=abundance@clone_by, group=abundance@group_by, ci_x=ci_x))
    }

    # Generate summary diversity output
    div_df <- bind_rows(beta_diversity_list, .id = "comparison")
    
    # Beta groups
    group_set <- unique(div_df[["comparison"]])
    
    # Compute evenness
    div_qi <- div_df %>%
        filter(!!rlang::sym("q") == 0) %>%
        select(one_of(c("comparison", "D")))

    div <- div_df %>%
        right_join(div_qi, by = "comparison", suffix = c("", "_0")) %>%
        mutate(d = !!rlang::sym("d")/!!rlang::sym("d_0"), 
            e_lower = !!rlang::sym("d_lower")/!!rlang::sym("d_0"), 
            e_upper = !!rlang::sym("d_upper")/!!rlang::sym("d_0")) %>%
        select(-!!rlang::sym("d_0"))
        
    # Test
    if (length(group_set) > 1) {
       test_df <- helperTest(div_df, q=q, group="comparison")
    } else {
       test_df <- NULL
    }
    
    # Build return object
    div_obj <- new("DiversityCurve",
                   diversity=div, 
                   tests=test_df,
                   method="beta",
                   group_by="comparison",
                   groups=group_set,
                   n=abundance@n,
                   q=q,  
                   ci=ci)
        
    return(div_obj)
}


#### Plotting functions ####

#' Plot a clonal abundance distribution
#' 
#' \code{plotAbundanceCurve} plots the results from estimating the complete clonal 
#' relative abundance distribution. The distribution is plotted as a log rank abundance 
#' distribution.
#' 
#' @param    data          \link{AbundanceCurve} object returned by \link{estimateAbundance}.
#' @param    colors        named character vector whose names are values in the 
#'                         \code{group} column of \code{data} and whose values are 
#'                         colors to assign to those group names.
#' @param    main_title    string specifying the plot title.
#' @param    legend_title  string specifying the legend title.
#' @param    xlim          numeric vector of two values specifying the 
#'                         \code{c(lower, upper)} x-axis limits. The lower x-axis 
#'                         value must be >=1.
#' @param    ylim          numeric vector of two values specifying the 
#'                         \code{c(lower, upper)} y-axis limits. The limits on the 
#'                         abundance values are expressed as fractions of 1: use
#'                         c(0,1) to set the lower and upper limits to 0\% and 100\%.
#' @param    annotate      string defining whether to added values to the group labels 
#'                         of the legend. When \code{"none"} (default) is specified no
#'                         annotations are added. Specifying (\code{"depth"}) adds 
#'                         sequence counts to the labels.
#' @param    silent        if \code{TRUE} do not draw the plot and just return the ggplot2 
#'                         object; if \code{FALSE} draw the plot.
#' @param    ...           additional arguments to pass to ggplot2::theme.
#' 
#' @return   A \code{ggplot} object defining the plot.
#' 
#' @seealso  
#' See \link{AbundanceCurve} for the input object and \link{estimateAbundance} for
#' generating the input abundance distribution. Plotting is performed with \link[ggplot2]{ggplot}.
#'           
#' @examples
#' # Estimate abundance by sample and plot
#' abund <- estimateAbundance(ExampleDb, group="sample_id", nboot=100)
#' plotAbundanceCurve(abund, legend_title="Sample")
#' 
#' @export
plotAbundanceCurve <- function(data, colors=NULL, main_title="Rank Abundance", 
                               legend_title=NULL, xlim=NULL, ylim=NULL, 
                               annotate=c("none", "depth"),
                               silent=FALSE, ...) {
    
    # Check if abundance is in data
    if (is.null(data@abundance)) { stop("Missing abundance data.") }
    
    # Validate abundance limits
    if (!is.null(xlim)) {
        if (xlim[1]<1) {
            stop("The lower x-axis xlim value must be >=1.")
        }
        max_xlim <- max(data@abundance$rank, na.rm = T)
        if (xlim[2]>max_xlim) {
            message("The largest x-axis value is ",max_xlim,".")
        }
    }    

    # Check arguments
    annotate <- match.arg(annotate)
    
    # Define group label annotations
    if (all(is.na(data@groups)) || length(data@groups) == 1) {
        group_labels <- NA  
    } else if (annotate == "none") {
        group_labels <- setNames(data@groups, data@groups)
    } else if (annotate == "depth") {
        group_labels <- setNames(paste0(data@groups, " (N=", data@n, ")"),
                                 data@groups)
    }

    # Stupid hack for check NOTE about `.x` in math_format
    .x <- NULL
    
    if (!all(is.na(group_labels))) {
        # Define grouped plot
        p1 <- ggplot(data@abundance, aes(x=!!rlang::sym("rank"), 
                                         y=!!rlang::sym("p"), 
                                         group=!!rlang::sym(data@group_by))) + 
            ggtitle(main_title) + 
            baseTheme() + 
            xlab("Rank") +
            ylab("Abundance") +
            scale_x_log10(
                          breaks=scales::trans_breaks("log10", function(x) 10^x),
                          labels=scales::trans_format("log10", scales::math_format(10^.x))) +
            scale_y_continuous(labels=scales::percent) +
            geom_ribbon(aes(ymin=!!rlang::sym("lower"), 
                            ymax=!!rlang::sym("upper"), 
                            fill=!!rlang::sym(data@group_by)), alpha=0.4) +
            geom_line(aes(color=!!rlang::sym(data@group_by)))
        
        # Set colors and legend
        if (!is.null(colors)) {
            p1 <- p1 + scale_color_manual(name=legend_title, labels=group_labels, values=colors) +
                scale_fill_manual(name=legend_title, labels=group_labels, values=colors)
        } else {
            p1 <- p1 + scale_color_discrete(name=legend_title, labels=group_labels) +
                scale_fill_discrete(name=legend_title, labels=group_labels)
        }
    } else {
        # Set color
        if (!is.null(colors) & length(colors) == 1) {
            line_color <- colors
        } else {
            line_color <- "black"
        }
        # Define plot
        p1 <- ggplot(data@abundance, aes(x=!!rlang::sym("rank"), 
                                         y=!!rlang::sym("p"))) + 
            ggtitle(main_title) + 
            baseTheme() + 
            xlab("Rank") +
            ylab("Abundance") +
            scale_x_log10(
                          breaks=scales::trans_breaks("log10", function(x) 10^x),
                          labels=scales::trans_format("log10", scales::math_format(10^.x))) +
            scale_y_continuous(labels=scales::percent) +
            geom_ribbon(aes(ymin=!!rlang::sym("lower"), 
                            ymax=!!rlang::sym("upper")), 
                        fill=line_color, alpha=0.4) +
            geom_line(color=line_color)
    }
    # Add additional theme elements
    p1 <- p1 + 
        coord_cartesian(xlim=xlim,ylim=ylim) +
        do.call(theme, list(...))
    
    # Plot
    if (!silent) { plot(p1) }
    
    invisible(p1)
}


#' Plot the results of alphaDiversity
#' 
#' \code{plotDiversityCurve} plots a \code{DiversityCurve} object.
#'
#' @param    data            \link{DiversityCurve} object returned by 
#'                           \link{alphaDiversity}.
#' @param    colors          named character vector whose names are values in the 
#'                           \code{group} column of the \code{data} slot of \code{data},
#'                           and whose values are colors to assign to those group names.
#' @param    main_title      string specifying the plot title.
#' @param    legend_title    string specifying the legend title.
#' @param    log_x           if \code{TRUE} then plot \eqn{q} on a log scale;
#'                           if \code{FALSE} plot on a linear scale.
#' @param    log_y           if \code{TRUE} then plot the diversity/evenness scores 
#'                           on a log scale; if \code{FALSE} plot on a linear scale.
#' @param    xlim            numeric vector of two values specifying the 
#'                           \code{c(lower, upper)} x-axis limits.
#' @param    ylim            numeric vector of two values specifying the 
#'                           \code{c(lower, upper)} y-axis limits.
#' @param    annotate        string defining whether to added values to the group labels 
#'                           of the legend. When \code{"none"} (default) is specified no
#'                           annotations are added. Specifying (\code{"depth"}) adds 
#'                           sequence counts to the labels.
#' @param    score           one of \code{"diversity"} or \code{"evenness"} specifying which
#'                           score to plot on the y-asis.
#' @param    silent          if \code{TRUE} do not draw the plot and just return the ggplot2 
#'                           object; if \code{FALSE} draw the plot.
#' @param    ...             additional arguments to pass to ggplot2::theme.
#'
#' @return   A \code{ggplot} object defining the plot.
#' 
#' @seealso  See \link{alphaDiversity} and \link{alphaDiversity} for generating 
#'           \link{DiversityCurve} objects for input. Plotting is performed with \link[ggplot2]{ggplot}.
#' 
#' @examples
#' # Calculate diversity
#' div <- alphaDiversity(ExampleDb, group="sample_id", nboot=100)
#' 
#' # Plot diversity
#' plotDiversityCurve(div, legend_title="Sample")
#'
#' # Plot diversity
#' plotDiversityCurve(div, legend_title="Sample", score="evenness")
#' 
#' @export
plotDiversityCurve <- function(data, colors=NULL, main_title="Diversity", 
                               legend_title="Group", log_x=FALSE, log_y=FALSE,
                               xlim=NULL, ylim=NULL, annotate=c("none", "depth"), 
                               score=c("diversity", "evenness"),
                               silent=FALSE, ...) {
    # Check arguments
    annotate <- match.arg(annotate)
    score <- match.arg(score)
    
    # Define group label annotations
    if (all(is.na(data@groups)) || length(data@groups) == 1) {
        group_labels <- NA  
    } else if (annotate == "none") {
        group_labels <- setNames(data@groups, data@groups)
    } else if (annotate == "depth") {
        group_labels <- setNames(paste0(data@groups, " (n=", data@n, ")"),
                                 data@groups)
    }
    
    # Define y-axis scores
    if (score == "diversity") {
        y_value <- "d"
        y_min <- "d_lower"
        y_max <- "d_upper"
        y_label <- expression(''^q * D)
    } else if (score == "evenness") {
        y_value <- "e"
        y_min <- "e_lower"
        y_max <- "e_upper"
        y_label <- expression(''^q * e)
    }
    
    # Stupid hack for check NOTE about `.x` in math_format
    .x <- NULL
    
    if (!all(is.na(group_labels))) {
        # Define grouped plot
        p1 <- ggplot(data@diversity, aes(x=q, y=!!rlang::sym(y_value), 
                                         group=!!rlang::sym(data@group_by))) + 
            ggtitle(main_title) + 
            baseTheme() + 
            xlab('q') +
            ylab(y_label) +
            geom_ribbon(aes(ymin=!!rlang::sym(y_min), ymax=!!rlang::sym(y_max), 
                            fill=!!rlang::sym(data@group_by)), alpha=0.4) +
            geom_line(aes(color=!!rlang::sym(data@group_by)))
    
        # Set colors and legend
        if (!is.null(colors)) {
            p1 <- p1 + scale_color_manual(name=legend_title, labels=group_labels, values=colors) +
                scale_fill_manual(name=legend_title, labels=group_labels, values=colors)
        } else {
            p1 <- p1 + scale_color_discrete(name=legend_title, labels=group_labels) +
                scale_fill_discrete(name=legend_title, labels=group_labels)
        }
    } else {
        # Set color
        if (!is.null(colors) & length(colors) == 1) {
          line_color <- colors
        } else {
          line_color <- "black"
        }
      
        # Define ungrouped plot
        p1 <- ggplot(data@diversity, aes(x=q, y=!!rlang::sym(y_value))) + 
            ggtitle(main_title) + 
            baseTheme() + 
            xlab('q') +
            ylab(y_label) +
            geom_ribbon(aes(ymin=!!rlang::sym(y_min), ymax=!!rlang::sym(y_max)), fill=line_color, alpha=0.4) +
            geom_line(color=line_color)
    }
    
    # Set x-axis style
    if (log_x) {
        p1 <- p1 + scale_x_continuous(trans=scales::log2_trans(),
                                      breaks=scales::trans_breaks('log2', function(x) 2^x),
                                      labels=scales::trans_format('log2', scales::math_format(2^.x)))
    } else {
        p1 <- p1 + scale_x_continuous()
    }
    
    # Set y-axis style
    if (log_y) {
        p1 <- p1 + scale_y_continuous(trans=scales::log2_trans(),
                                      breaks=scales::trans_breaks('log2', function(x) 2^x),
                                      labels=scales::trans_format('log2', scales::math_format(2^.x)))
    } else {
        p1 <- p1 + scale_y_continuous()
    }
    
    # Add additional theme elements
    p1 <- p1 + 
        coord_cartesian(xlim=xlim,ylim=ylim) +
        do.call(theme, list(...))

    # Plot
    if (!silent) { plot(p1) }
    
    invisible(p1)
}


#' Plot the results of diversity testing
#' 
#' \code{plotDiversityTest} plots summary data for a \code{DiversityCurve} object 
#' with mean and a line range indicating plus/minus one standard deviation.
#'
#' @param    data            \link{DiversityCurve} object returned by 
#'                           \link{alphaDiversity}.
#' @param    q               diversity order to plot the test for.
#' @param    colors          named character vector whose names are values in the 
#'                           \code{group} column of the \code{data} slot of \code{data},
#'                           and whose values are colors to assign to those group names.
#' @param    main_title      string specifying the plot title.
#' @param    legend_title    string specifying the legend title.
#' @param    log_d           if \code{TRUE} then plot the diversity scores \eqn{D} 
#'                           on a log scale; if \code{FALSE} plot on a linear scale.
#' @param    annotate        string defining whether to added values to the group labels 
#'                           of the legend. When \code{"none"} (default) is specified no
#'                           annotations are added. Specifying (\code{"depth"}) adds 
#'                           sequence counts to the labels.
#' @param    silent          if \code{TRUE} do not draw the plot and just return the ggplot2 
#'                           object; if \code{FALSE} draw the plot.
#' @param    ...             additional arguments to pass to ggplot2::theme.
#'
#' @return   A \code{ggplot} object defining the plot.
#' 
#' @seealso  See \link{alphaDiversity} for generating input.
#'           Plotting is performed with \link[ggplot2]{ggplot}.
#' 
#' @examples
#' # Calculate diversity
#' div <- alphaDiversity(ExampleDb, group="sample_id", min_q=0, max_q=2, step_q=1, nboot=100)
#' 
#' # Plot results at q=0 (equivalent to species richness)
#' plotDiversityTest(div, 0, legend_title="Sample")
#' 
#' # Plot results at q=2 (equivalent to Simpson's index)
#' plotDiversityTest(div, q=2, legend_title="Sample")
#' 
#' @export
plotDiversityTest <- function(data, q, colors=NULL, main_title="Diversity", legend_title="Group", 
                              log_d=FALSE, annotate=c("none", "depth"), silent=FALSE, ...) {
    # Stupid hack for check NOTE about `.x` in math_format
    .x <- NULL
    
    # Check arguments
    annotate <- match.arg(annotate)
    
    # Check if abundance is in data
    if (is.null(data@tests)) { 
        stop("Test data missing from input object.")
    }

    # Check if q is in data
    if (!(q %in% data@q)) { 
        stop("Test for order q=", q, " not found in input object.") 
    }

    # Define group label annotations
    if (annotate == "none") {
        group_labels <- setNames(data@groups, data@groups)
    } else if (annotate == "depth") {
        group_labels <- setNames(paste0(data@groups, " (N=", data@n, ")"),
                                 data@groups)
    }
    # Define plot values
    df <- data@diversity %>%
        dplyr::filter(!!rlang::sym("q") == !!rlang::enquo(q)) %>%
        dplyr::mutate(lower=!!rlang::sym("d") - !!rlang::sym("d_sd"), 
                      upper=!!rlang::sym("d") + !!rlang::sym("d_sd"))
    
    # Define base plot elements
    p1 <- ggplot(df, aes(x=!!rlang::sym(data@group_by))) + 
        ggtitle(main_title) + 
        baseTheme() + 
        xlab("") +
        ylab(bquote("Mean " ^ .(q) * D %+-% "SD")) +
        geom_linerange(aes(ymin=!!rlang::sym("lower"), 
                           ymax=!!rlang::sym("upper"), 
                           color=!!rlang::sym(data@group_by)), alpha=0.8) +
        geom_point(aes(y=!!rlang::sym("d"), color=!!rlang::sym(data@group_by)))
    
    # Set colors and legend
    if (!is.null(colors)) {
        p1 <- p1 + scale_color_manual(name=legend_title, labels=group_labels, values=colors)
    } else {
        p1 <- p1 + scale_color_discrete(name=legend_title, labels=group_labels)
    }

    # Set x-axis style
    if (log_d) {
        p1 <- p1 + scale_y_continuous(trans=scales::log2_trans(),
                                      breaks=scales::trans_breaks('log2', function(x) 2^x),
                                      labels=scales::trans_format('log2', scales::math_format(2^.x)))
    } else {
        p1 <- p1 + scale_y_continuous()
    }

    # Add additional theme elements
    p1 <- p1 + do.call(theme, list(...))
    
    # Plot
    if (!silent) { plot(p1) }
    
    invisible(p1)
}

# example_file <- file.path("tests", "data-tests", "ExampleDb.gz")
example_file <- file.path("..", "data-tests", "ExampleDb.gz")
db <- readChangeoDb(example_file)

#### calcCoverage ####

test_that("calcCoverage", {
    # Calculate clone sizes
    clones <- countClones(db, groups="SAMPLE", clone="CLONE")
    # Calculate 1st order coverage for a single sample
    obs <- calcCoverage(clones$seq_count[clones$SAMPLE == "RL01"])
    expect_equal(obs, 0.1608073, tolerance=0.001)
})


#### countClones ####

test_that("countClones", {
	# Calculate clone sizes
	clones <- countClones(db, groups="SAMPLE", clone="CLONE")
	expect_equal(clones$seq_count[1:6], c(31, 15, 5, 4, 4, 4))
	expect_equal(clones$seq_freq[1:6], 
	             c(0.15, 0.07, 0.02, 0.04, 0.04, 0.02), 
	             tolerance=0.01)

	# With copy numbers and multiple groups
    clones <- countClones(db, groups=c("SAMPLE", "ISOTYPE"), copy="DUPCOUNT", clone="CLONE")

	expect_equal(clones$seq_count[1:6], c(23, 15, 5, 3, 4, 1))
	expect_equal(clones$copy_count[1:6], c(53, 43, 24, 11, 11, 10))
	expect_equal(clones$copy_freq[6:11], 
	             c(0.71, 0.05, 0.47, 0.42, 0.04, 0.04),
	             tolerance=0.01)
    
	# Toy database
	db_toy <- tibble::tibble(SEQUENCE_ID=1:10, 
	                          GROUP=c(rep("A", 3), rep("B", 7)),
	                          CLONE=as.character(c(rep(1, 5), 2, 2, 3, 4, 5)),
                              LOCUS=rep("IGH", 10),
	                          COPY=10:1)
	ungrouped_toy <- tibble::tibble(CLONE=as.character(1:5), 
	                                 seq_count=as.integer(c(5, 2, 1, 1, 1)),
	                                 copy_count=as.integer(c(sum(10:6), sum(5:4), 3, 2, 1)),
	                                 seq_freq=c(5, 2, 1, 1, 1)/10,
	                                 copy_freq=c(sum(10:6), sum(5:4), 3, 2, 1)/sum(10:1))
	grouped_toy <- tibble::tibble(GROUP=c("A", rep("B", 5)),
	                               CLONE=as.character(c(1, 1:5)), 
	                               seq_count=as.integer(c(3, 2, 2, 1, 1, 1)),
	                               copy_count=as.integer(c(sum(10:8), sum(7:6), sum(5:4), 3, 2, 1)),
	                               seq_freq=c(3/3, 2/7, 2/7, 1/7, 1/7, 1/7),
	                               copy_freq=c(sum(10:8)/sum(10:8), 
	                                           sum(7:6)/sum(7:1), sum(5:4)/sum(7:1), 3/sum(7:1), 2/sum(7:1), 1/sum(7:1)))
	# Check toy ungrouped
	expect_equal(countClones(db_toy, clone="CLONE", copy="COPY"), 
	             ungrouped_toy, 
	             tolerance=0.01)

	# Check toy grouped
	expect_equal(countClones(db_toy, groups="GROUP", clone="CLONE", copy="COPY") %>% ungroup(), 
	             grouped_toy, 
	             tolerance=0.01)
	
	# Test how NAs are handled
	db_some_na <- data.frame(sequence_id = c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12),
	                         sample = c("S1", "S2", "S4", "S3", "S2", "S3", "S4", "S4", "S3", "S4", "S0", "S0"),
	                         dupcount = c(1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1),
                             locus = c(rep("IGH", 12)),
	                         clone_id = c(1, 2, 2, 3, 3, 3, 4, 4, 4, 4, NA, NA),
	                         stringsAsFactors=F)
	
	expect_warning(clones <- countClones(db_some_na), NULL)
	expect_equal(clones$seq_freq, 
	             c(0.4, 0.3, 0.2, 0.1),
	             tolerance=0.001)
	
	expect_warning(clones <- countClones(db_some_na, groups = "sample"), NULL)
	expect_equal(clones$seq_count, 
	             c(3, 2, 1, 1, 1, 1, 1),
	             tolerance=0.001)
	
	clones <- countClones(db_some_na, remove_na = FALSE)
	expect_equal(clones$seq_count, 
	             c(4, 3, 2, 2, 1),
	             tolerance=0.001)
	
	clones <- countClones(db_some_na, groups = "sample", remove_na = FALSE)
	expect_equal(clones$seq_count, 
	             c(3, 2, 2, 1, 1, 1, 1, 1),
	             tolerance=0.001)
	
	db_all_na <- data.frame(sequence_id = c(1, 2, 3),
	                        sample = c("S1", "S2", "S2"),
	                        dupcount = c(1, 1, 1),
                            locus = c(rep("IGH", 3)),
	                        clone_id = c(NA, NA, NA), 
	                        stringsAsFactors=F)
	
	expect_warning(clones <- countClones(db_all_na), "The column clone_id contains no data")
	expect_equal(nrow(clones), 
	             0,
	             tolerance=0.001)
	
	expect_warning(clones <- countClones(db_all_na, groups = "sample"), "The column clone_id contains no data")
	expect_equal(nrow(clones), 
	             0,
	             tolerance=0.001)
	
	expect_warning(clones <- countClones(db_all_na, remove_na = FALSE), "The column clone_id contains no data")
	expect_equal(clones$seq_count, 
	             3,
	             tolerance=0.001)
	
	expect_warning(clones <- countClones(db_all_na, groups = "sample", remove_na = FALSE), "The column clone_id contains no data")
	expect_equal(clones$seq_count, 
	             c(2, 1),
	             tolerance=0.001)
})

#### calcInferredDiversity ####

test_that("calcDiversity", {
	# May define p as clonal member counts
	p <- c(1, 1, 3, 10)
	q <- c(0, 1, 2)
	obs <- calcDiversity(p, q)
	exp <- c(4.000000, 2.594272, 2.027027)
	expect_equal(obs, exp, tolerance=0.001)

	# Or proportional abundance
	p <- c(1/15, 1/15, 1/5, 2/3)
	obs <- calcDiversity(p, q)
	expect_equal(obs, exp, tolerance=0.001)
})


#### estimateAbundance ####

test_that("estimateAbundance-current", {
	set.seed(90)
	abund <- estimateAbundance(db, group="SAMPLE", nboot=100, clone="CLONE")
	expect_equal(abund@abundance$p[1:5], 
	             c(0.0372, 0.0370, 0.0139, 0.0133, 0.0126),
	             tolerance=0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$lower[c(1:3,8:10)],
	             c(0.0041, 0.0004, 0, 0, 0, 0),
	             tolerance = 0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$upper[45:50],
	             c(0.0085, 0.0091, 0.0085, 0.0085, 0.0085, 0.0085),
	             tolerance = 0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$rank[200:203], c(200, 201, 202, 203), check.attributes = FALSE)
	
	# Grouping by isotype rather than sample identifier should raise warning
	set.seed(90)
	expect_warning(abund <- estimateAbundance(db, group="ISOTYPE", nboot=100, clone="CLONE"),
	               "Not all groups passed threshold")
})

	
#### alphaDiversity ####
#change

test_that("alphaDiversity", {
    # Test diversity
    set.seed(5)
	abund <- estimateAbundance(db, group="SAMPLE", nboot=100, clone="CLONE")
	div <- alphaDiversity(abund, step_q=1, max_q=10)
	obs <- data.frame(div@diversity[c(1,3,9,20), ])
	exp <- data.frame("SAMPLE" = c("RL01", "RL01", "RL01", "RL02"),
                      "q" = c(0, 2, 8, 8),
        	          "d" = c(88.4000, 72.0492, 33.8868, 8.9179),
	                  "d_sd" = c(3.4902, 9.3973, 11.9207, 2.1954),
        	          "d_lower" = c(81.5592, 53.6307, 10.5224, 4.6149),
        	          "d_upper" = c(95.2407, 90.4677, 57.2511, 13.2210),
        	          "e" = c(1.0000, 0.8150, 0.3833, 0.1397),
        	          "e_lower" = c(0.9226, 0.6066, 0.1190, 0.0723),
        	          "e_upper" = c(1.0773, 1.0233, 0.6476, 0.2071),
        	          stringsAsFactors = F)
	
	expect_equal(colnames(obs), colnames(exp))
	expect_equal(obs, exp, tolerance=0.001, check.attributes=F)

	# Test diversity p-values
	set.seed(3)
	abund <- estimateAbundance(db, group="SAMPLE", nboot=100, clone="CLONE")
	div <- alphaDiversity(abund, step_q=1, max_q=4)
	expect_equal(div@tests$pvalue[1:5], 
	             c(0, 0, 0, 0, 0), tolerance=0.001)
	expect_equal(div@diversity$d[1:5], 
	             c(88.5600, 82.4607, 72.2387, 60.0029, 50.0501), tolerance=0.001)
	
	# Verify two-steps == one-step
	set.seed(3)
	div_single <- alphaDiversity(db, step_q=1, max_q=4, group="SAMPLE", nboot=100, clone="CLONE")
	expect_equal(div, div_single)
})


#### betaDiversity ####

test_that("betaDiversity", {	
    skip("In development")
	set.seed(3)
	beta_db <- db %>% 
	    dplyr::rowwise() %>% 
	    dplyr::mutate(RANDOM = sample(1:4, 1, replace=TRUE)) %>% 
	    ungroup()
	diversity_obj  <- betaDiversity(beta_db, 
		comparisons=list("1-2"=c("1", "2"), "1-3"=c("1", "3")), group="RANDOM", clone="CLONE")

	obs <- data.frame(diversity_obj@diversity[c(1, 5, 12, 72), ])

	exp <- data.frame(
	        "comparison" = c("1-2", "1-2", "1-2", "1-3"),
	        "q" = c(0, 0.4, 1.1, 3.0),
	        "d" = c(1.790, 1.752, 1.634, 1.275), 
	        "d_sd" = c(0.2257, 0.1997, 0.1563, 0.3084),
	        "d_lower" = c(1.5643, 1.5524, 1.4775, 0.9664),
	        "d_upper" = c(2.0157, 1.9517, 1.79017, 1.5831),
	        "e" = c(1.00000000, 0.97879, 0.91273, 0.72458),
	        "e_lower" = c(0.87392, 0.867258, 0.82540, 0.54931),
	        "e_upper" = c(1.126074, 1.090331, 1.000072, 0.899849),
	        stringsAsFactors = F
	    )

	expect_equal(colnames(obs), colnames(exp))
	expect_equal(obs, exp, tolerance=0.001, check.attributes=F)

	# TODO: fix me. This test isn't right, SAMPLE doen'st exist is exp or obs
	# expect_equal(obs$SAMPLE, exp$SAMPLE, tolerance=0.001, check.attributes=F)

	set.seed(3)
	beta_db <- db %>% 
	    dplyr::rowwise() %>% 
	    dplyr::mutate(RANDOM = sample(1:4, 1, replace=TRUE)) %>% 
	    ungroup()
	diversity_obj  <- betaDiversity(beta_db, comparisons=list("1-2"=c("1", "2"), "1-3"=c("1", "3")), group="RANDOM")
	expect_equal(diversity_obj@tests$pvalue[1:5], 
	        c(0.5,0.5,0.5,0.5,0.5), tolerance=0.001)
	expect_equal(diversity_obj@diversity$d[1:5], 
	        c(1.40, 1.39, 1.38, 1.37, 1.36), tolerance=0.01)
})


#### Reproducibility tests ####

test_that("alphaDiversity reproduces rarefyDiversity and testDiversity", {
    # Default params for test in now deprecated rearefyDiversity
    #
    # estimateAbundance(data, group = NULL, clone = "CLONE", copy = NULL,
    #                   ci = 0.95, nboot = 2000, progress = FALSE)
    # rarefyDiversity(data, group, clone = "CLONE", copy = NULL, min_q = 0,
    #                 max_q = 4, step_q = 0.05, min_n = 30, max_n = NULL, ci = 0.95,
    #                 nboot = 2000, uniform = TRUE, progress = FALSE)
    #
    # The old test
    # set.seed(5)
    # div <- rarefyDiversity(db, "SAMPLE", step_q=1, max_q=10, nboot=100)
    # obs <- div@data[c(1, 3, 9, 20),]
    
    # Reproduce old test with alphaDiversity, and expect same results
    set.seed(5)
    diversity_obj <- alphaDiversity(as.data.frame(db), 
                                    group="SAMPLE", clone="CLONE", copy=NULL, 
                                    step_q=1, min_q=0, max_q=10, min_n=30, max_n=NULL,
                                    ci=0.95, nboot=100)
    obs <- diversity_obj@diversity[c(1, 3, 9, 20), ]
    
    # test the deprecated function
    set.seed(5)
    expect_warning(rarefy_obj <- rarefyDiversity(as.data.frame(db), 
                                                 group="SAMPLE", clone="CLONE", copy=NULL, 
                                                 step_q=1, min_q=0, max_q=10, min_n=30, max_n=NULL,
                                                 ci=0.95, nboot=100),
                  "is deprecated")
    # should match same code run outside the function
    expect_equal(diversity_obj, rarefy_obj)
    
    # Check reprod. old testDiversity
    # Defaults:
    # testDiversity(data, q, group, clone = "CLONE", copy = NULL,
    #               min_n = 30, max_n = NULL, nboot = 2000, progress = FALSE)
    # 
    # The old test:
    # set.seed(3)
    # div <- testDiversity(db, "SAMPLE", q=0, min_n=30, nboot=100)
    # expect_equal(div@tests$PVALUE, 0)
    # expect_equal(div@summary$MEAN, c(88.10, 63.11), tolerance=0.001)

    set.seed(3)
    diversity_obj <- alphaDiversity(as.data.frame(db), 
                                    group="SAMPLE", clone="CLONE", copy=NULL, 
                                    step_q=1, min_q=0, max_q=0, min_n=30, max_n=NULL,
                                    ci=0.95, nboot=100)
    
    # test the deprecated function
    set.seed(3)
    expect_warning(testdiv_obj <- testDiversity(as.data.frame(db), 
                                     group="SAMPLE", clone="CLONE", copy=NULL, 
                                     q=0, min_n=30, max_n=NULL,
                                     ci=0.95, nboot=100),
        "is deprecated")
    
    # should be the same as the code run outside the function
    expect_equal(diversity_obj, testdiv_obj)
})

test_that("estimateAbundance reproduces v0.2.11 results", {
    set.seed(90)
    abund <- estimateAbundance(db, group="SAMPLE", min_n=1, uniform=FALSE, nboot=1000, clone="CLONE")
    # Exact v0.2.11 test results with nboot=100
    # expect_equal(abund@abundance$P[1:6], 
    #              c(0.038086, 0.038086, 0.012930, 0.012930, 0.012930, 0.012930),
    #              tolerance=0.001)
    # expect_equal(abund@abundance$LOWER[c(1:3, 8:10)],
    #              c(0.001102, 0.000786, 0, 0, 0, 0),
    #              tolerance = 0.001)
    # expect_equal(abund@abundance$UPPER[45:50],
    #              c(0.00758, 0.00598, 0.00932, 0.00630, 0.00659, 0.00834),
    #              tolerance = 0.001)
    # v0.2.11 test results with nboot=1000
    expect_equal(abund@abundance$p[1:6],
                 c(0.03808, 0.03808, 0.01293, 0.01293, 0.01293, 0.01293),
                 tolerance=0.005, check.attributes = FALSE)
    expect_equal(abund@abundance$lower[c(1:3, 8:10)],
                 c(0.00049, 0, 0, 0, 0, 0),
                 tolerance = 0.005, check.attributes = FALSE)
    expect_equal(abund@abundance$upper[45:50],
                 c(0.00739, 0.00757, 0.00703, 0.00711,0.00662, 0.00709),
                 tolerance = 0.005, check.attributes = FALSE)
    expect_equal(abund@abundance$rank[1000:1005], c(36, 37, 38, 39, 40, 41))
    
    set.seed(90)
    abund <- estimateAbundance(db[c(1, 289), ], group="SAMPLE", min_n=1, uniform=FALSE, nboot=100, clone="CLONE")
    expect_equal(abund@abundance$lower, c(1, 1), check.attributes = FALSE)
    expect_equal(abund@abundance$upper, c(1, 1), check.attributes = FALSE)
    expect_equal(abund@abundance$rank, c(1, 1), check.attributes = FALSE)
})

test_that("rarefyDiversity reproduces v0.2.11 results", {
    set.seed(5)
    # Group by sample identifier
    expect_warning(div <- rarefyDiversity(db, "SAMPLE", step_q=1, max_q=10, nboot=1000),
                   "is deprecated")
    obs <- data.frame(div@diversity[c(1, 3, 9, 20), ])
    # Exact v0.2.11 test results with nboot=100
    # exp <- data.frame("SAMPLE" = c("RL01", "RL01", "RL01", "RL02"),
    #                   "Q" = c(0, 2, 8, 8),
    #                   "D" = c(88.22000, 71.51465, 32.51328, 8.94750),
    #                   "D_SD" = c(3.160936, 8.132310, 10.110378, 2.197165),
    #                   "D_LOWER" = c(82.024680, 55.575615, 12.697307, 4.641136),
    #                   "D_UPPER" = c(94.41532, 87.45369, 52.32926, 13.25386),
    #                   "E" = c(1.0000000, 0.8106399, 0.3685478, 0.1404852),
    #                   "E_LOWER" = c(0.92977420, 0.62996616, 0.14392776, 0.07287072),
    #                   "E_UPPER" = c(1.0702258, 0.9913136, 0.5931678, 0.2080996),
    #                   stringsAsFactors = FALSE)
    # v0.2.11 test results with nboot=10000
    exp <- data.frame("SAMPLE" = c("RL01", "RL01", "RL01", "RL02"),
                      "q" = c(0, 2, 8, 8),
                      "d" = c(88.16, 71.07, 33.07, 8.83),
                      "d_sd" = c(3.34, 9.78, 11.61, 2.39),
                      "d_lower" = c(81.61, 51.90, 10.32, 4.15),
                      "d_upper" = c(94.72, 90.23, 55.82, 13.53),
                      "e" = c(1.00, 0.81, 0.38, 0.14),
                      "e_lower" = c(0.93, 0.59, 0.12, 0.063),
                      "e_upper" = c(1.07, 1.02, 0.63, 0.21),
                      stringsAsFactors = FALSE)
    
    expect_equal(colnames(obs), colnames(exp))
    expect_equal(obs, exp, tolerance=0.05, check.attributes=F)
    
    # Grouping by isotype rather than sample identifier
    set.seed(25)
    expect_warning(div <- rarefyDiversity(db, "ISOTYPE", min_n=40, step_q=1, max_q=10, nboot=1000),
                   "Not all groups passed threshold")
    obs <- data.frame(div@diversity[c(5, 13, 19, 30), ])
    # Exact v0.2.11 test results with nboot=100
    # exp <- data.frame("ISOTYPE" = c("IgA", "IgG", "IgG", "IgM"),
    #                   "Q" = c(4, 1, 7, 7),
    #                   "D" = c(10.377177, 7.071540, 2.497792, 40.500567),
    #                   "D_SD" = c(3.1856002, 1.3429885, 0.5335591, 7.1375712),
    #                   "D_LOWER" = c(4.133516, 4.439331, 1.452036, 26.511184),
    #                   "D_UPPER" = c(16.620839, 9.703749, 3.543549, 54.489949),
    #                   "E" = c(0.4008180, 0.5087439, 0.1796973, 0.8343751),
    #                   "E_LOWER" = c(0.1596568, 0.3193763, 0.1044630, 0.5461719),
    #                   "E_UPPER" = c(0.6419791, 0.6981115, 0.2549316, 1.1225783),
    #                   stringsAsFactors = FALSE)
    # v0.2.11 test results with nboot=10000
    exp <- data.frame("ISOTYPE" = c("IgA", "IgG", "IgG", "IgM"),
                      "q" = c(4, 1, 7, 7),
                      "d" = c(10.02, 7.26, 2.56, 39.96),
                      "d_sd" = c(2.83, 1.45, 0.49, 6.87),
                      "d_lower" = c(4.47, 4.41, 1.59, 26.49),
                      "d_upper" = c(15.57, 10.10, 3.52, 53.43),
                      "e" = c(0.38, 0.52, 0.18, 0.82),
                      "e_lower" = c(0.17, 0.31, 0.11, 0.55),
                      "e_upper" = c(0.60, 0.72, 0.25, 1.10),
                      stringsAsFactors = FALSE)
    expect_equal(colnames(obs), colnames(exp))
    expect_equal(obs, exp, tolerance=0.05, check.attributes=F)
})

test_that("testDiversity reproduces v0.2.11 results", {
    set.seed(3)
    # Groups under the size threshold are excluded and a warning message is issued.
    expect_warning(div <- testDiversity(db, "SAMPLE", q=0, min_n=30, nboot=1000),
                   "is deprecated")
    # Exact v0.2.11 test results with nboot=100
    # expect_equal(div@tests$PVALUE, 0)
    # expect_equal(div@diversity$D, c(88.10, 63.11), tolerance=0.001)
    # v0.2.11 test results with nboot=10000
    expect_equal(div@tests$pvalue, 0, tolerance=0.05)
    expect_equal(div@diversity$d, c(88.13, 63.57), tolerance=0.05)
    
    set.seed(3)
    expect_warning(div <- testDiversity(rbind(db, db), "SAMPLE", q=0, min_n=30, nboot=1000),
                   "is deprecated")
    # Exact v0.2.11 test results with nboot=100
    # expect_equal(div@tests$PVALUE, 0.88)
    # expect_equal(div@diversity$D, c(78.63, 79.58), tolerance=0.001)
    # v0.2.11 test results with nboot=10000
    expect_equal(div@tests$pvalue, 0.88, tolerance=0.05)
    expect_equal(div@diversity$d, c(78.63, 79.80), tolerance=0.05)
})


#### Duplicated Testing for AIRR FORMAT








# example_file <- file.path("tests", "data-tests", "ExampleDb_airr.gz")
example_file <- file.path("..", "data-tests", "ExampleDb_airr.gz")
expect_warning(
    db <- readChangeoDb(example_file),
regexp="airr::read_rearrangement")

expect_warning(db_mixed_cloned <- airr::read_rearrangement(file.path("..", "data-tests", "db_test_cloned.tsv")))
db_sc_cloned <- db_mixed_cloned %>% dplyr::filter(!is.na(cell_id)) %>% dplyr::mutate("umi_count" = 2)
db_bulk_cloned <- db_mixed_cloned %>% dplyr::select(-cell_id)

#### calcCoverage ####

test_that("calcCoverage", {
    # Calculate clone sizes
    clones <- countClones(db, groups="sample")
    # Calculate 1st order coverage for a single sample
    obs <- calcCoverage(clones$seq_count[clones$sample == "RL01"])
    expect_equal(obs, 0.1608073, tolerance=0.001)
})


#### countClones ####

test_that("countClones", {
	# Calculate clone sizes
	clones <- countClones(db, groups="sample")
	expect_equal(clones$seq_count[1:6], c(31, 15, 5, 4, 4, 4))
	expect_equal(clones$seq_freq[1:6], 
	             c(0.15, 0.07, 0.02, 0.04, 0.04, 0.02), 
	             tolerance=0.01)

	# With copy numbers and multiple groups
	clones <- countClones(db, groups=c("sample", "isotype"), copy="duplicate_count")

	expect_equal(clones$seq_count[1:6], c(23, 15, 5, 3, 4, 1))
	expect_equal(clones$copy_count[1:6], c(53, 43, 24, 11, 11, 10))
	expect_equal(clones$copy_freq[6:11], 
	             c(0.71, 0.05, 0.47, 0.42, 0.04, 0.04),
	             tolerance=0.01)
    
	# Toy database
	db_toy <- tibble::tibble(SEQUENCE_ID=1:10, 
	                          GROUP=c(rep("A", 3), rep("B", 7)),
	                          CLONE=as.character(c(rep(1, 5), 2, 2, 3, 4, 5)),
                              LOCUS=rep("IGH", 10),
	                          COPY=10:1)
	ungrouped_toy <- tibble::tibble(CLONE=as.character(1:5), 
	                                 seq_count=as.integer(c(5, 2, 1, 1, 1)),
	                                 copy_count=as.integer(c(sum(10:6), sum(5:4), 3, 2, 1)),
	                                 seq_freq=c(5, 2, 1, 1, 1)/10,
	                                 copy_freq=c(sum(10:6), sum(5:4), 3, 2, 1)/sum(10:1))
	grouped_toy <- tibble::tibble(GROUP=c("A", rep("B", 5)),
	                               CLONE=as.character(c(1, 1:5)), 
	                               seq_count=as.integer(c(3, 2, 2, 1, 1, 1)),
	                               copy_count=as.integer(c(sum(10:8), sum(7:6), sum(5:4), 3, 2, 1)),
	                               seq_freq=c(3/3, 2/7, 2/7, 1/7, 1/7, 1/7),
	                               copy_freq=c(sum(10:8)/sum(10:8), 
	                                           sum(7:6)/sum(7:1), sum(5:4)/sum(7:1), 3/sum(7:1), 2/sum(7:1), 1/sum(7:1)))
	# Check toy ungrouped
	expect_equal(countClones(db_toy, clone="CLONE", copy="COPY"), 
	             ungrouped_toy,
	             tolerance=0.01)

	# Check toy grouped
	expect_equal(countClones(db_toy, groups="GROUP", clone="CLONE", copy="COPY") %>% ungroup(), 
	             grouped_toy, 
	             tolerance=0.01)
})

test_that("countClones with mixed data", {
    expect_warning(clones <- countClones(db_mixed_cloned))
    clones_expect <- tibble::tibble(
        clone_id = c("1","2","4","3"),
        seq_count = c(11,2,2,1),
        seq_freq = c(11/16,2/16,2/16,1/16)
    )
    expect_equal(clones,
                    clones_expect,
                    tolerance=0.01)
})

test_that("countClones with sc data", {
    expect_warning(clones <- countClones(db_sc_cloned))
    clones_expect <- tibble::tibble(
        clone_id = c("1","2","4","3"),
        seq_count = c(10,2,2,1),
        seq_freq = c(10/15,2/15,2/15,1/15)
    )
})

test_that("countClones with bulk data", {
    expect_warning(clones <- countClones(db_bulk_cloned))
    clones_expect <- tibble::tibble(
        clone_id = c("1","2","4","3"),
        seq_count = c(11,2,2,1),
        seq_freq = c(11/16,2/16,2/16,1/16)
    )

})

test_that("countClones with sc data and specified copy fails", {
    expect_error(clones <- countClones(db_sc_cloned, copy="umi_count"))
})

test_that("countClones with sc data and factor grouping variables", {
    # Test for issue: countClones throws error with single cell data and factors in grouping variables
    db_sc_factor <- db_sc_cloned %>% 
        dplyr::mutate(sample_id = factor("a", levels = "a"))
    
    expect_warning(clones <- countClones(db_sc_factor, groups = "sample_id"))
    expect_true(is.data.frame(clones))
    expect_true(nrow(clones) > 0)
    expect_true("sample_id" %in% names(clones))
    expect_equal(sum(clones$seq_count), 15)
})

#### calcInferredDiversity ####

test_that("calcDiversity", {
	# May define p as clonal member counts
	p <- c(1, 1, 3, 10)
	q <- c(0, 1, 2)
	obs <- calcDiversity(p, q)
	exp <- c(4.000000, 2.594272, 2.027027)
	expect_equal(obs, exp, tolerance=0.001)

	# Or proportional abundance
	p <- c(1/15, 1/15, 1/5, 2/3)
	obs <- calcDiversity(p, q)
	expect_equal(obs, exp, tolerance=0.001)
})


#### estimateAbundance ####

test_that("estimateAbundance-current", {
	set.seed(90)
	abund <- estimateAbundance(db, group="sample", nboot=100)
	expect_equal(abund@abundance$p[1:5], 
	             c(0.0372, 0.0370, 0.0139, 0.0133, 0.0126),
	             tolerance=0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$lower[c(1:3,8:10)],
	             c(0.0041, 0.0004, 0, 0, 0, 0),
	             tolerance = 0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$upper[45:50],
	             c(0.0085, 0.0091, 0.0085, 0.0085, 0.0085, 0.0085),
	             tolerance = 0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$rank[200:203], c(200, 201, 202, 203), 
	             check.attributes = FALSE)
	
	# Grouping by isotype rather than sample identifier should raise warning
	set.seed(90)
	expect_warning(abund <- estimateAbundance(db, group="isotype", nboot=100),
	               "Not all groups passed threshold")
})

	
#### alphaDiversity ####

test_that("alphaDiversity", {
    # Test diversity
    set.seed(5)
	abund <- estimateAbundance(db, group="sample", nboot=100)
	div <- alphaDiversity(abund, step_q=1, max_q=10)
	obs <- data.frame(div@diversity[c(1,3,9,20), ])
	exp <- data.frame("sample" = c("RL01", "RL01", "RL01", "RL02"),
                      "q" = c(0, 2, 8, 8),
        	          "d" = c(88.4000, 72.0492, 33.8868, 8.9179),
	                  "d_sd" = c(3.4902, 9.3973, 11.9207, 2.1954),
        	          "d_lower" = c(81.5592, 53.6307, 10.5224, 4.6149),
        	          "d_upper" = c(95.2407, 90.4677, 57.2511, 13.2210),
        	          "e" = c(1.0000, 0.8150, 0.3833, 0.1397),
        	          "e_lower" = c(0.9226, 0.6066, 0.1190, 0.0723),
        	          "e_upper" = c(1.0773, 1.0233, 0.6476, 0.2071),
        	          stringsAsFactors = F)
	
	expect_equal(colnames(obs), colnames(exp))
	expect_equal(obs, exp, tolerance=0.001, check.attributes=F)

	# Test diversity p-values
	set.seed(3)
	abund <- estimateAbundance(db, group="sample", nboot=100)
	div <- alphaDiversity(abund, step_q=1, max_q=4)
	expect_equal(div@tests$pvalue[1:5], 
	             c(0, 0, 0, 0, 0), tolerance=0.001)
	expect_equal(div@diversity$d[1:5], 
	             c(88.5600, 82.4607, 72.2387, 60.0029, 50.0501), tolerance=0.001)
	
	# Verify two-steps == one-step
	set.seed(3)
	div_single <- alphaDiversity(db, step_q=1, max_q=4, group="sample", nboot=100)
	expect_equal(div, div_single)
})


#### betaDiversity ####

test_that("betaDiversity", {	
    skip("In development")
	set.seed(3)
	beta_db <- db %>% 
	    dplyr::rowwise() %>% 
	    dplyr::mutate(RANDOM = sample(1:4, 1, replace=TRUE)) %>% 
	    ungroup()
	diversity_obj  <- betaDiversity(beta_db, 
		comparisons=list("1-2"=c("1", "2"), "1-3"=c("1", "3")), group="RANDOM")

	obs <- data.frame(diversity_obj@diversity[c(1, 5, 12, 72), ])

	exp <- data.frame(
	        "comparison" = c("1-2", "1-2", "1-2", "1-3"),
	        "q" = c(0, 0.4, 1.1, 3.0),
	        "d" = c(1.790, 1.752, 1.634, 1.275), 
	        "d_sd" = c(0.2257, 0.1997, 0.1563, 0.3084),
	        "d_lower" = c(1.5643, 1.5524, 1.4775, 0.9664),
	        "d_upper" = c(2.0157, 1.9517, 1.79017, 1.5831),
	        "e" = c(1.00000000, 0.97879, 0.91273, 0.72458),
	        "e_lower" = c(0.87392, 0.867258, 0.82540, 0.54931),
	        "e_upper" = c(1.126074, 1.090331, 1.000072, 0.899849),
	        stringsAsFactors = F
	    )

	expect_equal(colnames(obs), colnames(exp))
	expect_equal(obs, exp, tolerance=0.001, check.attributes=F)

	# TODO: fix me. This test isn't right, sample doen'st exist is exp or obs
	# expect_equal(obs$sample, exp$sample, tolerance=0.001, check.attributes=F)

	set.seed(3)
	beta_db <- db %>% 
	    dplyr::rowwise() %>% 
	    dplyr::mutate(RANDOM = sample(1:4, 1, replace=TRUE)) %>% 
	    ungroup()
	diversity_obj  <- betaDiversity(beta_db, comparisons=list("1-2"=c("1", "2"), "1-3"=c("1", "3")), group="RANDOM")
	expect_equal(diversity_obj@tests$pvalue[1:5], 
	        c(0.5,0.5,0.5,0.5,0.5), tolerance=0.001)
	expect_equal(diversity_obj@diversity$d[1:5], 
	        c(1.40, 1.39, 1.38, 1.37, 1.36), tolerance=0.01)
})


#### Reproducibility tests ####

test_that("alphaDiversity reproduces rarefyDiversity and testDiversity", {
    # Default params for test in now deprecated rearefyDiversity
    #
    # estimateAbundance(data, group = NULL, clone = "CLONE", copy = NULL,
    #                   ci = 0.95, nboot = 2000, progress = FALSE)
    # rarefyDiversity(data, group, clone = "CLONE", copy = NULL, min_q = 0,
    #                 max_q = 4, step_q = 0.05, min_n = 30, max_n = NULL, ci = 0.95,
    #                 nboot = 2000, uniform = TRUE, progress = FALSE)
    #
    # The old test
    # set.seed(5)
    # div <- rarefyDiversity(db, "sample", step_q=1, max_q=10, nboot=100)
    # obs <- div@data[c(1, 3, 9, 20),]
    
    # Reproduce old test with alphaDiversity, and expect same results
    set.seed(5)
    diversity_obj <- alphaDiversity(as.data.frame(db), 
                                    group="sample", clone="clone_id", copy=NULL, 
                                    step_q=1, min_q=0, max_q=10, min_n=30, max_n=NULL,
                                    ci=0.95, nboot=100)
    obs <- diversity_obj@diversity[c(1, 3, 9, 20), ]
    
    # test the deprecated function
    set.seed(5)
    expect_warning(rarefy_obj <- rarefyDiversity(as.data.frame(db), 
                                                 group="sample", clone="clone_id", copy=NULL, 
                                                 step_q=1, min_q=0, max_q=10, min_n=30, max_n=NULL,
                                                 ci=0.95, nboot=100),
                  "is deprecated")
    # should match same code run outside the function
    expect_equal(diversity_obj, rarefy_obj)
    
    # Check reprod. old testDiversity
    # Defaults:
    # testDiversity(data, q, group, clone = "CLONE", copy = NULL,
    #               min_n = 30, max_n = NULL, nboot = 2000, progress = FALSE)
    # 
    # The old test:
    # set.seed(3)
    # div <- testDiversity(db, "sample", q=0, min_n=30, nboot=100)
    # expect_equal(div@tests$PVALUE, 0)
    # expect_equal(div@summary$MEAN, c(88.10, 63.11), tolerance=0.001)

    set.seed(3)
    diversity_obj <- alphaDiversity(as.data.frame(db), 
                                    group="sample", clone="clone_id", copy=NULL, 
                                    step_q=1, min_q=0, max_q=0, min_n=30, max_n=NULL,
                                    ci=0.95, nboot=100)
    
    # test the deprecated function
    set.seed(3)
    expect_warning(testdiv_obj <- testDiversity(as.data.frame(db), 
                                     group="sample", clone="clone_id", copy=NULL, 
                                     q=0, min_n=30, max_n=NULL,
                                     ci=0.95, nboot=100),
        "is deprecated")
    
    # should be the same as the code run outside the function
    expect_equal(diversity_obj, testdiv_obj)
})

test_that("estimateAbundance reproduces v0.2.11 results", {
    set.seed(90)
    abund <- estimateAbundance(db, group="sample", min_n=1, uniform=FALSE, nboot=1000)
    # Exact v0.2.11 test results with nboot=100
    # expect_equal(abund@abundance$P[1:6], 
    #              c(0.038086, 0.038086, 0.012930, 0.012930, 0.012930, 0.012930),
    #              tolerance=0.001)
    # expect_equal(abund@abundance$LOWER[c(1:3, 8:10)],
    #              c(0.001102, 0.000786, 0, 0, 0, 0),
    #              tolerance = 0.001)
    # expect_equal(abund@abundance$UPPER[45:50],
    #              c(0.00758, 0.00598, 0.00932, 0.00630, 0.00659, 0.00834),
    #              tolerance = 0.001)
    # v0.2.11 test results with nboot=1000
    expect_equal(abund@abundance$p[1:6],
                 c(0.03808, 0.03808, 0.01293, 0.01293, 0.01293, 0.01293),
                 tolerance=0.005, check.attributes = FALSE)
    expect_equal(abund@abundance$lower[c(1:3, 8:10)],
                 c(0.00049, 0, 0, 0, 0, 0),
                 tolerance = 0.005, check.attributes = FALSE)
    expect_equal(abund@abundance$upper[45:50],
                 c(0.00739, 0.00757, 0.00703, 0.00711,0.00662, 0.00709),
                 tolerance = 0.005, check.attributes = FALSE)
    expect_equal(abund@abundance$rank[1000:1005], c(36, 37, 38, 39, 40, 41), 
                 check.attributes = FALSE)
    
    set.seed(90)
    abund <- estimateAbundance(db[c(1, 289), ], group="sample", min_n=1, uniform=FALSE, nboot=100)
    expect_equal(abund@abundance$lower, c(1, 1), check.attributes = FALSE)
    expect_equal(abund@abundance$upper, c(1, 1), check.attributes = FALSE)
    expect_equal(abund@abundance$rank, c(1, 1), check.attributes = FALSE)
})

test_that("rarefyDiversity reproduces v0.2.11 results", {
    set.seed(5)
    # Group by sample identifier
    expect_warning(div <- rarefyDiversity(db, "sample", step_q=1, max_q=10, nboot=1000, clone="clone_id"),
                   "is deprecated")
    obs <- data.frame(div@diversity[c(1, 3, 9, 20), ])
    # Exact v0.2.11 test results with nboot=100
    # exp <- data.frame("sample" = c("RL01", "RL01", "RL01", "RL02"),
    #                   "Q" = c(0, 2, 8, 8),
    #                   "D" = c(88.22000, 71.51465, 32.51328, 8.94750),
    #                   "D_SD" = c(3.160936, 8.132310, 10.110378, 2.197165),
    #                   "D_LOWER" = c(82.024680, 55.575615, 12.697307, 4.641136),
    #                   "D_UPPER" = c(94.41532, 87.45369, 52.32926, 13.25386),
    #                   "E" = c(1.0000000, 0.8106399, 0.3685478, 0.1404852),
    #                   "E_LOWER" = c(0.92977420, 0.62996616, 0.14392776, 0.07287072),
    #                   "E_UPPER" = c(1.0702258, 0.9913136, 0.5931678, 0.2080996),
    #                   stringsAsFactors = FALSE)
    # v0.2.11 test results with nboot=10000
    exp <- data.frame("sample" = c("RL01", "RL01", "RL01", "RL02"),
                      "q" = c(0, 2, 8, 8),
                      "d" = c(88.16, 71.07, 33.07, 8.83),
                      "d_sd" = c(3.34, 9.78, 11.61, 2.39),
                      "d_lower" = c(81.61, 51.90, 10.32, 4.15),
                      "d_upper" = c(94.72, 90.23, 55.82, 13.53),
                      "e" = c(1.00, 0.81, 0.38, 0.14),
                      "e_lower" = c(0.93, 0.59, 0.12, 0.063),
                      "e_upper" = c(1.07, 1.02, 0.63, 0.21),
                      stringsAsFactors = FALSE)
    
    expect_equal(colnames(obs), colnames(exp))
    expect_equal(obs, exp, tolerance=0.05, check.attributes=F)
    
    # Grouping by isotype rather than sample identifier
    set.seed(25)
    expect_warning(div <- rarefyDiversity(db, "isotype", min_n=40, step_q=1, max_q=10, nboot=1000, clone="clone_id"),
                   "Not all groups passed threshold")
    obs <- data.frame(div@diversity[c(5, 13, 19, 30), ])
    # Exact v0.2.11 test results with nboot=100
    # exp <- data.frame("isotype" = c("IgA", "IgG", "IgG", "IgM"),
    #                   "Q" = c(4, 1, 7, 7),
    #                   "D" = c(10.377177, 7.071540, 2.497792, 40.500567),
    #                   "D_SD" = c(3.1856002, 1.3429885, 0.5335591, 7.1375712),
    #                   "D_LOWER" = c(4.133516, 4.439331, 1.452036, 26.511184),
    #                   "D_UPPER" = c(16.620839, 9.703749, 3.543549, 54.489949),
    #                   "E" = c(0.4008180, 0.5087439, 0.1796973, 0.8343751),
    #                   "E_LOWER" = c(0.1596568, 0.3193763, 0.1044630, 0.5461719),
    #                   "E_UPPER" = c(0.6419791, 0.6981115, 0.2549316, 1.1225783),
    #                   stringsAsFactors = FALSE)
    # v0.2.11 test results with nboot=10000
    exp <- data.frame("isotype" = c("IgA", "IgG", "IgG", "IgM"),
                      "q" = c(4, 1, 7, 7),
                      "d" = c(10.02, 7.26, 2.56, 39.96),
                      "d_sd" = c(2.83, 1.45, 0.49, 6.87),
                      "d_lower" = c(4.47, 4.41, 1.59, 26.49),
                      "d_upper" = c(15.57, 10.10, 3.52, 53.43),
                      "e" = c(0.38, 0.52, 0.18, 0.82),
                      "e_lower" = c(0.17, 0.31, 0.11, 0.55),
                      "e_upper" = c(0.60, 0.72, 0.25, 1.10),
                      stringsAsFactors = FALSE)
    expect_equal(colnames(obs), colnames(exp))
    expect_equal(obs, exp, tolerance=0.05, check.attributes=F)
})

test_that("testDiversity reproduces v0.2.11 results", {
    set.seed(3)
    # Groups under the size threshold are excluded and a warning message is issued.
    expect_warning(div <- testDiversity(db, "sample", q=0, min_n=30, nboot=1000, clone="clone_id"),
                   "is deprecated")
    # Exact v0.2.11 test results with nboot=100
    # expect_equal(div@tests$PVALUE, 0)
    # expect_equal(div@diversity$D, c(88.10, 63.11), tolerance=0.001)
    # v0.2.11 test results with nboot=10000
    expect_equal(div@tests$pvalue, 0, tolerance=0.05)
    expect_equal(div@diversity$d, c(88.13, 63.57), tolerance=0.05)
    
    set.seed(3)
    expect_warning(div <- testDiversity(rbind(db, db), "sample", q=0, min_n=30, nboot=1000, clone="clone_id"),
                   "is deprecated")
    # Exact v0.2.11 test results with nboot=100
    # expect_equal(div@tests$PVALUE, 0.88)
    # expect_equal(div@diversity$D, c(78.63, 79.58), tolerance=0.001)
    # v0.2.11 test results with nboot=10000
    expect_equal(div@tests$pvalue, 0.88, tolerance=0.05)
    expect_equal(div@diversity$d, c(78.63, 79.80), tolerance=0.05)
})


test_that("estimateAbundance-mixed", {
	set.seed(90)
    expect_warning(
        abund <- estimateAbundance(db_mixed_cloned, ci=0.95, nboot=100, clone="clone_id", min_n = 2),
        regexp="NA(s) found in 2 row(s) of the clone_id column and excluded from tabulation",
        fixed=TRUE
    )
	expect_equal(abund@abundance$p[1:5], 
	             c(0.69500, 0.11750, 0.10375, 0.04375, 0.04000),
	             tolerance=0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$lower[c(1:3)],
	             c(0.4671344, 0, 0),
	             tolerance = 0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$upper[2:5],
	             c(0.2735016, 0.2399798, 0.1503706, 0.1329823),
	             tolerance = 0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$rank[1:5], c(1,  2,  3,  4,  5), check.attributes = FALSE)
	
})

test_that("estimateAbundance-sc", {

	set.seed(90)
    # seq18 has clone_id=NA
    expect_warning(
        abund <- estimateAbundance(db_sc_cloned, ci=0.95, nboot=100, clone="clone_id", min_n = 2),
        regexp="NA(s) found in 1 row(s) of the clone_id column and excluded from tabulation",
        fixed=TRUE
    )
	expect_equal(abund@abundance$p[1:5], 
	             c(0.68000000, 0.11933333, 0.11000000, 0.04733333, 0.04333333),
	             tolerance=0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$lower[c(1:3)],
	             c(0.430833, 0.0000000, 0.0000000),
	             tolerance = 0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$upper[2:5],
	             c(0.2836650, 0.2613061, 0.1607507, 0.1439901),
	             tolerance = 0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$rank[1:5], c(1,  2,  3,  4,  5), check.attributes = FALSE)
	
})
test_that("estimateAbundance-bulk", {
	set.seed(90)
	expect_warning(
	    abund <- estimateAbundance(db_bulk_cloned, ci=0.95, nboot=100, clone="clone_id", min_n = 2),
	    regexp="NA(s) found in 2 row(s) of the clone_id column and excluded from tabulation",
	    fixed=TRUE
	)
	expect_equal(abund@abundance$p[1:4], 
	             c(0.67357143, 0.14035714, 0.11250000, 0.07357143),
	             tolerance=0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$lower[c(1:3)],
	             c(0.501816121, 0.012458845, 0.005633074),
	             tolerance = 0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$upper[2:4],
	             c( 0.2682554, 0.2193669, 0.1607731),
	             tolerance = 0.001, check.attributes = FALSE)
	expect_equal(abund@abundance$rank[1:4], c(1,  2,  3,  4), check.attributes = FALSE)
	
})

test_that("alphaDiversity-mixed", {
    # Test diversity
    set.seed(5)
    expect_warning(
	    abund <- estimateAbundance(db_mixed_cloned, ci=0.95, nboot=100, clone="clone_id", min_n = 2),
	    "NA(s) found in 2 row(s) of the clone_id column and excluded from tabulation",
	    fixed=TRUE
    )
	div <- alphaDiversity(abund, step_q=1, max_q=10)
	obs <- data.frame(div@diversity[c(1,3,9), ])
	exp <- data.frame("group" = c("All", "All", "All"),
                      "q" = c(0, 2, 8),
        	          "d" = c(3.570000, 1.943491, 1.567165),
	                  "d_sd" = c(0.7946157, 0.5334659, 0.3656269),
        	          "d_lower" = c(2.0125818, 0.8979167, 0.8505499),
        	          "d_upper" = c(5.127418, 2.989065, 2.283781),
        	          "e" = c(1.0000000, 0.5443952, 0.4389819),
        	          "e_lower" = c(0.5637484, 0.2515173, 0.2382493),
        	          "e_upper" = c(1.4362516, 0.8372730, 0.6397146),
        	          stringsAsFactors = F)
	
	expect_equal(colnames(obs), colnames(exp))
	expect_equal(obs, exp, tolerance=0.001, check.attributes=F)

	# Test diversity p-values
	# Verify two-steps == one-step
    # Not performed above tested since there all the sample belong to the same group in the testing mixed data
})

test_that("alphaDiversity-sc", {
    # Test diversity
    set.seed(5)
    expect_warning(
        abund <- estimateAbundance(db_sc_cloned, ci=0.95, nboot=100, clone="clone_id", min_n = 2),
        regexp="NA(s) found in 1 row(s) of the clone_id column and excluded from tabulation",
        fixed=TRUE
    )
	div <- alphaDiversity(abund, step_q=1, max_q=10)
	obs <- data.frame(div@diversity[c(1,3,9), ])
	exp <- data.frame("group" = c("All", "All", "All"),
                      "q" = c(0, 2, 8),
        	          "d" = c(3.580000, 2.022895, 1.629327),
	                  "d_sd" = c(0.7936618, 0.5732313, 0.4164018),
        	          "d_lower" = c(2.0244515, 0.8993822, 0.8131947),
        	          "d_upper" = c(5.135548, 3.146408, 2.445460),
        	          "e" = c(1.0000000, 0.5650544, 0.4551193),
        	          "e_lower" = c(0.5654893, 0.2512241, 0.2271494),
        	          "e_upper" = c(1.4345107, 0.8788848, 0.6830893),
        	          stringsAsFactors = F)
	
	expect_equal(colnames(obs), colnames(exp))
	expect_equal(obs, exp, tolerance=0.001, check.attributes=F)

	# Test diversity p-values
	# Verify two-steps == one-step
    # Not performed above tested since there all the sample belong to the same group in the testing mixed data
})

test_that("alphaDiversity-bulk", {
    # Test diversity
    set.seed(5)
    expect_warning(
        abund <- estimateAbundance(db_bulk_cloned, ci=0.95, nboot=100, clone="clone_id", min_n = 2),
        regexp="NA(s) found in 2 row(s) of the clone_id column and excluded from tabulation",
        fixed=TRUE
    )
	div <- alphaDiversity(abund, step_q=1, max_q=10)
	obs <- data.frame(div@diversity[c(1,3,9), ])
	exp <- data.frame("group" = c("All", "All", "All"),
                      "q" = c(0, 2, 8),
        	          "d" = c(3.860000, 1.976337, 1.568243),
	                  "d_sd" = c(0.3487351, 0.3777932, 0.2492464),
        	          "d_lower" = c(3.176492, 1.235876, 1.079729),
        	          "d_upper" = c(4.543508, 2.716798, 2.056757),
        	          "e" = c(1.0000000, 0.5120044, 0.4062806),
        	          "e_lower" = c(0.8229253, 0.3201751, 0.2797226),
        	          "e_upper" = c(1.1770747, 0.7038337, 0.5328386),
        	          stringsAsFactors = F)
	
	expect_equal(colnames(obs), colnames(exp))
	expect_equal(obs, exp, tolerance=0.001, check.attributes=F)

	# Test diversity p-values
	# Verify two-steps == one-step
    # Not performed above tested since there all the sample belong to the same group in the testing mixed data
})

# Classes, Generics and Methods

#### Generics ####

# @exportMethod print
#setGeneric("print")

# @exportMethod plot
#setGeneric("plot")

setClassUnion("DfNULL", members=c("data.frame", "NULL"))
setClassUnion("CharNULL", members=c("character", "NULL"))

#### Diversity classes ####

#' S4 class defining a clonal abundance curve
#'
#' \code{AbundanceCurve} defines clonal abundance values.
#' 
#' @slot  abundance  data.frame with relative clonal abundance data and confidence intervals,
#'                   containing the following columns:
#'                   \itemize{
#'                     \item  \code{group}:  group identifier.
#'                     \item  \code{clone_id} or \code{CLONE}:  clone identifier. 
#'                     \item  \code{p}:      relative abundance of the clone.
#'                     \item  \code{lower}:  lower confidence interval bound.
#'                     \item  \code{upper}:  upper confidence interval bound.
#'                     \item  \code{rank}:   the rank of the clone abundance.
#'                   }
#' @slot  bootstrap  data.frame of bootstrapped clonal distributions. 
#' @slot  clone_by   string specifying the name of the clone column. 
#' @slot  group_by   string specifying the name of the grouping column. 
#' @slot  groups     vector specifying the names of unique groups in group column. 
#' @slot  n          numeric vector indication the number of sequences sampled in each group.
#' @slot  nboot      numeric specifying the number of bootstrap iterations to use.  
#' @slot  ci         confidence interval defining the upper and lower bounds 
#'                   (a value between 0 and 1).
#' 
#' @name         AbundanceCurve-class
#' @rdname       AbundanceCurve-class
#' @aliases      AbundanceCurve
#' @exportClass  AbundanceCurve
setClass("AbundanceCurve", 
         slots=c(abundance="data.frame",
                 bootstrap="data.frame",
                 clone_by="character",
                 group_by="character",
                 groups="character",
                 n="numeric", 
                 ci="numeric",
                 nboot="numeric"))

#' S4 class defining a diversity curve 
#'
#' \code{DiversityCurve} defines diversity (\eqn{D}) scores over multiple diversity 
#' orders (\eqn{Q}).
#' 
#' @slot  diversity  data.frame defining the diversity curve with the following columns:
#'                   \itemize{
#'                     \item  \code{group}:    group label.
#'                     \item  \code{q}:        diversity order.
#'                     \item  \code{d}:        mean diversity index over all bootstrap 
#'                                             realizations.
#'                     \item  \code{d_sd}:     standard deviation of the diversity index 
#'                                             over all bootstrap realizations.
#'                     \item  \code{d_lower}:  diversity lower confidence interval bound.
#'                     \item  \code{d_upper}:  diversity upper confidence interval bound.
#'                     \item  \code{e}:        evenness index calculated as \code{D} 
#'                                             divided by \code{D} at \code{Q=0}.
#'                     \item  \code{e_lower}:  evenness lower confidence interval bound.
#'                     \item  \code{e_upper}:  evenness upper confidence interval bound.
#'                   }
#' @slot  tests    data.frame describing the significance test results with columns:
#'                 \itemize{
#'                   \item  \code{test}:        string listing the two groups tested.
#'                   \item  \code{delta_mean}:  mean of the \eqn{D} bootstrap delta 
#'                                              distribution for the test.
#'                   \item  \code{delta_sd}:    standard deviation of the \eqn{D} 
#'                                              bootstrap delta distribution for the test.
#'                   \item  \code{pvalue}:      p-value for the test.
#'                 }
#' @slot  group_by   string specifying the name of the grouping column in diversity calculation.
#' @slot  groups     vector specifying the names of unique groups in group column in diversity calculation.
#' @slot  method     string specifying the type of diversity calculated. 
#' @slot  q          vector of diversity hill diversity indices used for computing diversity.  
#' @slot  n          numeric vector indication the number of sequences sampled in each group.
#' @slot  ci         confidence interval defining the upper and lower bounds 
#'                   (a value between 0 and 1).
#' 
#' @name         DiversityCurve-class
#' @rdname       DiversityCurve-class
#' @aliases      DiversityCurve
#' @exportClass  DiversityCurve
setClass("DiversityCurve", 
         slots=c(diversity="data.frame", 
                 tests="DfNULL",
                 method="character",
                 group_by="character",
                 groups="character",
                 q="numeric",
                 n="numeric", 
                 ci="numeric"))
              
#### Diversity methods ####

#' @param    x    AbundanceCurve object
#' 
#' @rdname   AbundanceCurve-class
#' @aliases  AbundanceCurve-method
#' @export
setMethod("print", c(x="AbundanceCurve"), function(x) { print(x@abundance) })

#' @param    y    ignored.
#' @param    ...  arguments to pass to \link{plotDiversityCurve}.
#' 
#' @rdname   AbundanceCurve-class
#' @aliases  AbundanceCurve-method
#' @export
setMethod("plot", c(x="AbundanceCurve", y="missing"),
          function(x, y, ...) { plotAbundanceCurve(x, ...) })

#' @param    x    DiversityCurve object
#' 
#' @rdname   DiversityCurve-class
#' @aliases  DiversityCurve-method
#' @export
setMethod("print", c(x="DiversityCurve"), function(x) { print(x@diversity) })

#' @param    y    diversity order to plot (q).
#' @param    ...  arguments to pass to \link{plotDiversityCurve} or \link{plotDiversityTest}.
#' 
#' @rdname   DiversityCurve-class
#' @aliases  DiversityCurve-method
#' @export
setMethod("plot", c(x="DiversityCurve", y="missing"),
          function(x, y, ...) { plotDiversityCurve(x, ...) })

#' @rdname   DiversityCurve-class
#' @aliases  DiversityCurve-method
#' @export
setMethod("plot", c(x="DiversityCurve", y="numeric"),
          function(x, y, ...) { plotDiversityTest(x, y, ...) })

#### Lineage classes ####

#' S4 class defining a clone
#' 
#' \code{ChangeoClone} defines a common data structure for perform lineage reconstruction
#' from Change-O data.
#' 
#' @slot     data      data.frame containing sequences and annotations. Contains the
#'                     columns \code{SEQUENCE_ID} and \code{SEQUENCE}, as well as any additional 
#'                     sequence-specific annotation columns.
#' @slot     clone     string defining the clone identifier.
#' @slot     germline  string containing the germline sequence for the clone.
#' @slot     v_gene    string defining the V segment gene call.
#' @slot     j_gene    string defining the J segment gene call.
#' @slot     junc_len  numeric junction length (nucleotide count).
#' 
#' @seealso  See \link{makeChangeoClone} and \link{buildPhylipLineage} for use.
#'           
#' @name         ChangeoClone-class
#' @rdname       ChangeoClone-class
#' @aliases      ChangeoClone
#' @exportClass  ChangeoClone
setClass("ChangeoClone", 
         slots=c(data="data.frame",
                 clone="character",
                 germline="character", 
                 v_gene="character", 
                 j_gene="character", 
                 junc_len="numeric"))


#### Topology classes ####

#' S4 class defining edge significance
#'
#' \code{MRCATest} defines the significance of enrichment for annotations appearing at
#' the MRCA of the tree.
#' 
#' @slot  tests         data.frame describing the significance test results with columns:
#'                      \itemize{
#'                        \item  \code{annotation}:  annotation value.
#'                        \item  \code{count}:       observed count of MRCA positions 
#'                                                   with the given annotation.
#'                        \item  \code{expected}:    expected mean count of MRCA occurrence
#'                                                   for the annotation.
#'                        \item  \code{pvalue}:      one-sided p-value for the hypothesis that 
#'                                                   the observed annotation abundance is greater 
#'                                                   than expected.
#'                      }
#' @slot  permutations  data.frame containing the raw permutation test data with columns:
#'                      \itemize{
#'                        \item  \code{annotation}:  annotation value.
#'                        \item  \code{count}:       count of MRCA positions with the 
#'                                                   given annotation.
#'                        \item  \code{iter}:        numerical index define which 
#'                                                   permutation realization each 
#'                                                   observation corresponds to.
#'                      }
#' @slot  nperm         number of permutation realizations.
#' 
#' @name         MRCATest-class
#' @rdname       MRCATest-class
#' @aliases      MRCATest
#' @exportClass  MRCATest
setClass("MRCATest", 
         slots=c(tests="data.frame",
                 permutations="data.frame",
                 nperm="numeric"))


#' S4 class defining edge significance
#'
#' \code{EdgeTest} defines the significance of parent-child annotation enrichment.
#' 
#' @slot  tests         data.frame describing the significance test results with columns:
#'                      \itemize{
#'                        \item  \code{parent}:    parent node annotation.
#'                        \item  \code{child}:     child node annotation
#'                        \item  \code{count}:     count of observed edges with the given 
#'                                                 parent-child annotation set.
#'                        \item  \code{expected}:  mean count of expected edges for the 
#'                                                 given parent-child relationship.
#'                        \item  \code{pvalue}:    one-sided p-value for the hypothesis that 
#'                                                  the observed edge abundance is greater 
#'                                                  than expected.
#'                      }
#' @slot  permutations  data.frame containing the raw permutation test data with columns:
#'                      \itemize{
#'                        \item  \code{parent}:  parent node annotation.
#'                        \item  \code{child}:   child node annotation
#'                        \item  \code{count}:   count of edges with the given parent-child 
#'                                               annotation set.
#'                        \item  \code{iter}:    numerical index define which permutation
#'                                               realization each observation corresponds 
#'                                               to.
#'                      }
#' @slot  nperm         number of permutation realizations.
#' 
#' @name         EdgeTest-class
#' @rdname       EdgeTest-class
#' @aliases      EdgeTest
#' @exportClass  EdgeTest
setClass("EdgeTest", 
         slots=c(tests="data.frame",
                 permutations="data.frame",
                 nperm="numeric"))


#### Topology methods ####

#' @param    x    MRCATest object.
#' 
#' @rdname   MRCATest-class
#' @aliases  MRCATest-method
#' @export
setMethod("print", c(x="MRCATest"), function(x) { print(x@tests) })

#' @param    y    ignored.
#' @param    ...  arguments to pass to \link{plotMRCATest}.
#' 
#' @rdname   MRCATest-class
#' @aliases  MRCATest-method
#' @export
setMethod("plot", c(x="MRCATest", y="missing"),
          function(x, y, ...) { plotMRCATest(x, ...) })

#' @param    x    EdgeTest object.
#' 
#' @rdname   EdgeTest-class
#' @aliases  EdgeTest-method
#' @export
setMethod("print", c(x="EdgeTest"), function(x) { print(x@tests) })

#' @param    y    ignored.
#' @param    ...  arguments to pass to \link{plotEdgeTest}.
#' 
#' @rdname   EdgeTest-class
#' @aliases  EdgeTest-method
#' @export
setMethod("plot", c(x="EdgeTest", y="missing"),
          function(x, y, ...) { plotEdgeTest(x, ...) })

---
title: 'Alakazam: Analysis of clonal abundance and diversity'
author: "Jason Anthony Vander Heiden"
date: '`r Sys.Date()`'
output:
  pdf_document:
    dev: pdf
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    toc: yes
  html_document:
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    theme: readable
    toc: yes
  md_document:
    fig_height: 4
    fig_width: 7.5
    preserve_yaml: no
    toc: yes
geometry: margin=1in
fontsize: 11pt
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Diversity analysis}
  %\usepackage[utf8]{inputenc}
---

The clonal diversity of the repertoire can be analyzed using the general form
of the diversity index, as proposed by Hill in:

    Hill, M. Diversity and evenness: a unifying notation and its consequences. 
        Ecology 54, 427-432 (1973).

Coupled with resampling strategies to correct for variations in sequencing 
depth, as well as inference of complete clonal abundance distributions as 
described in:

    Chao A, et al. Rarefaction and extrapolation with Hill numbers: 
        A framework for sampling and estimation in species diversity studies. 
        Ecol Monogr. 2014 84:45-67.
    Chao A, et al. Unveiling the species-rank abundance distribution by 
        generalizing the Good-Turing sample coverage theory. 
        Ecology. 2015 96, 11891201.

This package provides methods for the inference of a complete clonal 
abundance distribution (using the `estimateAbundance` function) along with 
two approaches to assess the diversity of these distributions: 

1. Generation of a smooth diversity (D) curve over a range of diversity orders (q) 
using `alphaDiversity`, and
2. A significance test of the diversity (D) at a fixed diversity order (q).


## Example data

A small example AIRR database, `ExampleDb`, is included in the `alakazam` package. 
Diversity calculation requires the `clone` field (column) to be present in the 
AIRR file, as well as an additional grouping column. In this example we 
will use the grouping columns `sample_id` and `c_call`.

```{r, eval=TRUE, warning=FALSE, message=FALSE}
# Load required packages
library(alakazam)

# Load example data
data(ExampleDb)
```

For details about the AIRR format, visit the [AIRR Community documentation site](https://docs.airr-community.org/en/stable/datarep/rearrangements.html).

## Generate a clonal abundance curve

A simple table of the observed clonal abundance counts and frequencies may be
generated using the `countClones` function either with or without copy numbers, where
the size of each clone is determined by the number of sequence members:

```{r, eval=TRUE, warning=FALSE}
# Partitions the data based on the sample column
clones <- countClones(ExampleDb, group="sample_id")
head(clones, 5)
```

You may also specify a column containing the abundance count of each sequence 
(usually copy numbers), that will include weighting of each clone size by the 
corresponding abundance count. Furthermore, multiple grouping columns may be
specified such that `seq_freq` (unweighted clone size as a fraction
of total sequences in the group) and `copy_freq` (weighted fraction) are 
normalized to within multiple group data partitions.

```{r, eval=TRUE, warning=FALSE}
# Partitions the data based on both the sample_id and c_call columns
# Weights the clone sizes by the duplicate_count column
clones <- countClones(ExampleDb, group=c("sample_id", "c_call"), copy="duplicate_count", clone="clone_id")
head(clones, 5)
```

While `countClones` will report observed abundances, it will not provide confidence 
intervals. A complete clonal abundance distribution may be inferred using the 
`estimateAbundance` function with confidence intervals derived via bootstrapping.  
This output may be visualized using the `plotAbundanceCurve` function.

```{r, eval=TRUE, results='hide', warning=FALSE, fig.width=6, fig.height=4}
# Partitions the data on the sample column
# Calculates a 95% confidence interval via 100 bootstrap realizations
set.seed(123) # For reproducibility of example bootstrap results
curve <- estimateAbundance(ExampleDb, group="sample_id", ci=0.95, nboot=100, clone="clone_id")
```

```{r, eval=TRUE, warning=FALSE, fig.width=6, fig.height=4}
# Plots a rank abundance curve of the relative clonal abundances
sample_colors <- c("-1h"="seagreen", "+7d"="steelblue")
plot(curve, colors = sample_colors, legend_title="Sample")
```

## Generate a diversity curve

The function `alphaDiversity` performs uniform resampling of the input 
sequences and recalculates the clone size distribution, and diversity, with each 
resampling realization. Diversity (D) is calculated over a range of diversity 
orders (q) to generate a smooth curve.

```{r, eval=TRUE, results='hide'}
# Compare diversity curve across values in the "sample" column
# q ranges from 0 (min_q=0) to 4 (max_q=4) in 0.05 increments (step_q=0.05)
# A 95% confidence interval will be calculated (ci=0.95)
# 100 resampling realizations are performed (nboot=100)
set.seed(123) # For reproducibility of example alphaDiversity results
sample_curve <- alphaDiversity(ExampleDb, group="sample_id", clone="clone_id",
                               min_q=0, max_q=4, step_q=0.1,
                               ci=0.95, nboot=100)

# Compare diversity curve across values in the c_call column
# Analyse is restricted to c_call values with at least 30 sequences by min_n=30
# Excluded groups are indicated by a warning message
isotype_curve <- alphaDiversity(ExampleDb, group="c_call", clone="clone_id",
                                min_q=0, max_q=4, step_q=0.1,
                                ci=0.95, nboot=100)
```

```{r, eval=TRUE, fig.width=6, fig.height=4}
# Plot a log-log (log_q=TRUE, log_d=TRUE) plot of sample diversity
# Indicate number of sequences resampled from each group in the title
sample_main <- paste0("Sample diversity")
sample_colors <- c("-1h"="seagreen", "+7d"="steelblue")
plot(sample_curve, colors=sample_colors, main_title=sample_main, 
     legend_title="Sample")

# Plot isotype diversity using default set of Ig isotype colors
isotype_main <- paste0("Isotype diversity")
plot(isotype_curve, colors=IG_COLORS, main_title=isotype_main, 
     legend_title="Isotype")
```


## View diversity tests at a fixed diversity order

Significance testing across groups is performed using the delta of the bootstrap
distributions between groups when running `alphaDiversity` for all values of `q` 
specified.

```{r, eval=TRUE, fig.width=6, fig.height=3}
# Test diversity at q=0, q=1 and q=2 (equivalent to species richness, Shannon entropy, 
# Simpson's index) across values in the sample_id column
# 100 bootstrap realizations are performed (nboot=100)
set.seed(123) # For reproducibility of example alphaDiversity results
isotype_test <- alphaDiversity(ExampleDb, group="c_call", min_q=0, max_q=2, step_q=1, nboot=100, clone="clone_id")

# Print P-value table
print(isotype_test@tests)

# Plot results at q=0 and q=2
# Plot the mean and standard deviations at q=0 and q=2
plot(isotype_test, 0, colors=IG_COLORS, main_title=isotype_main, 
     legend_title="Isotype")
plot(isotype_test, 2, colors=IG_COLORS, main_title=isotype_main, 
     legend_title="Isotype")
```

---
title: 'Alakazam: Using sequencing quality scores'
author: "Susanna Marquez"
date: '`r Sys.Date()`'
output:
  pdf_document:
    dev: pdf
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    toc: yes  
    toc_depth: 3
  md_document:
    fig_height: 4
    fig_width: 7.5
    preserve_yaml: no
    toc: yes
    toc_depth: 3
  html_document:
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    theme: readable
    toc: yes
    toc_depth: 3
geometry: margin=1in
fontsize: 11pt
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Fastq}
  %\usepackage[utf8]{inputenc}
---

The `alakazam` package includes a set of functions to inspect the sequencing quality.

## Example data

Load example data:

```{r, eval=TRUE, warning=FALSE, message=FALSE}
library(alakazam)
library(dplyr)
library(airr)

db <- read_rearrangement(system.file("extdata", "example_quality.tsv", package="alakazam"))
fastq_file <- system.file("extdata", "example_quality.fastq", package="alakazam")
```

## Load quality scores

This method allows to add the quality scores to the repertoire `data.frame` as strings.

```{r}
original_cols <- colnames(db)
db <- readFastqDb(db, fastq_file, style="both", quality_sequence=TRUE)
new_cols <- setdiff(colnames(db), original_cols)
db[,new_cols] %>% head()
```

The function `readFastq` takes as main inputs a repertoire `data.frame` (`db`) and 
a path to the corresponding `.fastq` file (`fastq_file`). The sequencing quality scores will
be merged into the `data.frame` by `sequence_id`. The newly added columns are:
`r paste(new_cols, collapse=", ")`. The other fields, contain the ASCII quality scores in the 
form of a vector, where values are comma separated, and `-` or `.` positions 
have value `" "` (blank).

After loading the quality scores with `readFastqDb`,  `getPositionQuality`
can be used to generate a `data.frame` of sequencing quality values 
per position.

```{r}
quality <- getPositionQuality(db, sequence_id="sequence_id", 
                              sequence="sequence_alignment",
                              quality_num="quality_alignment_num")
head(quality)
```

```{r, fig.cap="Sequence quality per IMGT position for one sequence.", fig.asp=0.25}
min_pos <- min(quality$position)
max_pos <- max(quality$position)

ggplot(quality, aes(x=position,
                    y=quality_alignment_num,
                    color=nt)) +
  geom_point() +
  coord_cartesian(xlim=c(110,120)) +
  xlab("IMGT position") +
  ylab("Sequencing quality") +
  scale_fill_gradient(low = "light blue",  high = "dark red") +
  scale_x_continuous(breaks=c(min_pos:max_pos)) +
  alakazam::baseTheme()
```

You can add use the quality `data.frame` to complement analysis performed
with other tools from the Immcantation framework. For example, you could inspect
the sequencing quality of novel polymorphisms identified with `tigger`, or
the sequencing quality in mutated/unmutated regions.

## Mask low quality positions

Use `maskPositionsByQuality` to mask low quality positions. Positions with
a sequencing quality < `min_quality` will be replaced with an 'N'. A message
will show the number of sequences in `db` that had at least one position
masked.

```{r}
db <- maskPositionsByQuality(db, min_quality=70,
                             sequence="sequence_alignment",
                             quality="quality_alignment_num")
```

---
title: 'Alakazam: Amino acid physicochemical property analysis'
author: "Susanna Marquez"
date: '`r Sys.Date()`'
output:
  pdf_document:
    dev: pdf
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    toc: yes  
  md_document:
    fig_height: 4
    fig_width: 7.5
    preserve_yaml: no
    toc: yes
  html_document:
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    theme: readable
    toc: yes
geometry: margin=1in
fontsize: 11pt
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Amino acid property analysis}
  %\usepackage[utf8]{inputenc}
---

The `alakazam` package includes a set of functions to analyze the physicochemical 
properties of Ig and TCR amino acid sequences. Of particular interest is the analysis of 
CDR3 properties, which this vignette will demonstrate. The same process 
can be applied to other regions simply by altering the sequence data column used.

    Wu YC, et al. High-throughput immunoglobulin repertoire analysis distinguishes 
        between human IgM memory and switched memory B-cell populations. 
        Blood 116, 1070-8 (2010).
    Wu YC, et al. The relationship between CD27 negative and positive B cell 
        populations in human peripheral blood. 
        Front Immunol 2, 1-12 (2011).


## Example data

A small example AIRR database, `ExampleDb`, is included in the `alakazam` package.

```{r, eval=TRUE, warning=FALSE, message=FALSE}
# Load required packages
library(alakazam)
library(dplyr)

# Subset example data
data(ExampleDb)
db <- ExampleDb[ExampleDb$sample_id == "+7d", ]
```

For details about the AIRR format, visit the [AIRR Community documentation site](https://docs.airr-community.org/en/stable/datarep/rearrangements.html).

## Calculate the properties of amino acid sequences

Multiple amino acid physicochemical properties can be obtained with the function 
`aminoAcidProperties`. The available properties are:

* `length`: total amino acid count
* `gravy`: grand average of hydrophobicity
* `bulkiness`: average bulkiness
* `polarity`: average polarity
* `aliphatic`: normalized aliphatic index
* `charge`: normalized net charge
* `acidic`: acidic side chain residue content
* `basic`: basic side chain residue content
* `aromatic`: aromatic side chain content

This example demonstrates how to calculate all of the available amino acid 
properties from DNA sequences found in the `junction` column of the previously loaded AIRR file.

Translation of the DNA sequences to amino acid sequences is accomplished by 
default with the `nt=TRUE` argument. To reduce the junction sequence to the CDR3 
sequence we specify the argument `trim=TRUE` which will strip the first and last
codon (the conserved residues) prior to analysis. The prefix `cdr3` is added
to the output column names using the `label="cdr3"` argument.

```{r, eval=TRUE, warning=FALSE, fig.width=7.5, fig.height=6}
db_props <- aminoAcidProperties(db, seq="junction", trim=TRUE, 
                                label="cdr3")

# The full set of properties are calculated by default
dplyr::select(db_props[1:3, ], starts_with("cdr3"))

# Define a ggplot theme for all plots
tmp_theme <- theme_bw() + theme(legend.position="bottom")

# Generate plots for all four of the properties
g1 <- ggplot(db_props, aes(x=c_call, y=cdr3_aa_length)) + tmp_theme +
    ggtitle("CDR3 length") + 
    xlab("Isotype") + ylab("Amino acids") +
    scale_fill_manual(name="Isotype", values=IG_COLORS) +
    geom_boxplot(aes(fill=c_call))
g2 <- ggplot(db_props, aes(x=c_call, y=cdr3_aa_gravy)) + tmp_theme + 
    ggtitle("CDR3 hydrophobicity") + 
    xlab("Isotype") + ylab("GRAVY") +
    scale_fill_manual(name="Isotype", values=IG_COLORS) +
    geom_boxplot(aes(fill=c_call))
g3 <- ggplot(db_props, aes(x=c_call, y=cdr3_aa_basic)) + tmp_theme +
    ggtitle("CDR3 basic residues") + 
    xlab("Isotype") + ylab("Basic residues") +
    scale_y_continuous(labels=scales::percent) +
    scale_fill_manual(name="Isotype", values=IG_COLORS) +
    geom_boxplot(aes(fill=c_call))
g4 <- ggplot(db_props, aes(x=c_call, y=cdr3_aa_acidic)) + tmp_theme +
    ggtitle("CDR3 acidic residues") + 
    xlab("Isotype") + ylab("Acidic residues") +
    scale_y_continuous(labels=scales::percent) +
    scale_fill_manual(name="Isotype", values=IG_COLORS) +
    geom_boxplot(aes(fill=c_call))

# Plot in a 2x2 grid
gridPlot(g1, g2, g3, g4, ncol=2)
```

### Obtaining properties individually

A subset of the properties may be calculated using the `property` argument of
`aminoAcidProperties`. For example, calculations may be restricted to only the 
grand average of hydrophobicity (`gravy`) index and normalized net charge 
(`charge`) by specifying `property=c("gravy", "charge")`. 

```{r, eval=TRUE, warning=FALSE}
db_props <- aminoAcidProperties(db, seq="junction", property=c("gravy", "charge"),
                                trim=TRUE, label="cdr3")
dplyr::select(db_props[1:3, ], starts_with("cdr3"))
```

### Using user defined scales

Each property has a default scale setting, but users may specify alternate scales 
if they wish. The following example shows how to import and use the
Kidera et al, 1985 hydrophobicity scale and the Murrary et al, 2006 pK values from 
the `seqinr` package instead of the defaults for calculating the GRAVY index and 
net charge.

```{r, eval=TRUE, warning=FALSE, message=FALSE}
# Load the relevant data objects from the seqinr package
library(seqinr)
data(aaindex)
data(pK)
h <- aaindex[["KIDA850101"]]$I
p <- setNames(pK[["Murray"]], rownames(pK))
# Rename the hydrophobicity vector to use single-letter codes
names(h) <- translateStrings(names(h), ABBREV_AA)
db_props <- aminoAcidProperties(db, seq="junction", property=c("gravy", "charge"), 
                                trim=TRUE, label="cdr3", 
                                hydropathy=h, pK=p)
dplyr::select(db_props[1:3, ], starts_with("cdr3"))
```

### Getting vectors of individual properties

The `aminoAcidProperties` function provides a convenient wrapper for calculating
multiple properties at once from a `data.frame`. If a vector of a specific property is
required this may be accomplished using one of the worker functions:

* `gravy`: grand average of hydrophobicity
* `bulk`: average bulkiness
* `polar`: average polarity
* `aliphatic`: aliphatic index
* `charge`: net charge
* `countPatterns`: counts the occurrence of patterns in amino acid sequences

The input to each function must be a vector of amino acid sequences.

```{r, eval=TRUE, warning=FALSE, message=FALSE}
# Translate junction DNA sequences to amino acids and trim first and last codons
cdr3 <- translateDNA(db$junction[1:3], trim=TRUE)

# Grand average of hydrophobicity
gravy(cdr3)

# Average bulkiness
bulk(cdr3)

# Average polarity
polar(cdr3)

# Normalized aliphatic index
aliphatic(cdr3)
# Unnormalized aliphatic index
aliphatic(cdr3, normalize=FALSE)

# Normalized net charge
charge(cdr3)
# Unnormalized net charge
charge(cdr3, normalize=FALSE)

# Count of acidic amino acids
# Takes a named list of regular expressions
countPatterns(cdr3, nt=FALSE, c(ACIDIC="[DE]"), label="cdr3")
```

## Default scales

The following references were used for the default physicochemical scales:

* Aliphatic index:  
  Ikai AJ. Thermostability and aliphatic index of globular proteins. 
        J Biochem 88, 1895-1898 (1980).
* Bulkiness scale:  
  Zimmerman JM, Eliezer N, Simha R. The characterization of amino acid sequences 
    in proteins by statistical methods. 
    J Theor Biol 21, 170-201 (1968).
* Hydrophobicity scale:  
  Kyte J, Doolittle RF. A simple method for displaying the hydropathic character 
    of a protein.
    J Mol Biol 157, 105-32 (1982).
* pK values:  
  \url{https://emboss.sourceforge.net/apps/cvs/emboss/apps/iep.html}
* Polarity scale:  
  Grantham R. Amino acid difference formula to help explain protein evolution. 
    Science 185, 862-864 (1974).

---
title: 'Alakazam: Gene usage analysis'
author: "Susanna Marquez"
date: '`r Sys.Date()`'
output:
  pdf_document:
    dev: pdf
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    toc: yes
  md_document:
    fig_height: 4
    fig_width: 7.5
    preserve_yaml: no
    toc: yes
  html_document:
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    theme: readable
    toc: yes
geometry: margin=1in
fontsize: 11pt
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Gene usage analysis}
  %\usepackage[utf8]{inputenc}
---

The 'alakazam' package provides basic gene usage quantification by either sequence count 
or clonal grouping; with or without consideration of duplicate reads/mRNA. Additionally, 
a set of accessory functions for sorting and parsing V(D)J gene names 
are also provided.

## Example data

A small example AIRR database, `ExampleDb`, is included in the `alakazam` package. For details 
about the AIRR format, visit the [AIRR Community documentation site](https://docs.airr-community.org/en/stable/datarep/rearrangements.html).

Gene usage analysis requires only the following columns:

* `v_call`
* `d_call`
* `j_call`

However, the optional clonal clustering (`clone_id`) and duplicate count (`duplicate_count`) 
columns may be used to quantify usage by different abundance criteria.

```{r, eval=TRUE, warning=FALSE, message=FALSE}
# Load required packages
library(alakazam)
library(dplyr)
library(scales)

# Subset example data
data(ExampleDb)
```

## Tabulate V(D)J allele, gene or family usage by sample

The relative abundance of V(D)J alleles, genes or families within groups can be obtained 
with the function `countGenes`. To analyze differences in the V gene usage across 
different samples we will set `gene="v_call"` (the column containing gene data) and 
`groups="sample_id"` (the columns containing grouping variables). To quantify abundance at
the gene level we set `mode="gene"`:

```{r, eval=TRUE, warning=FALSE}
# Quantify usage at the gene level
gene <- countGenes(ExampleDb, gene="v_call", groups="sample_id", mode="gene")
head(gene, n=4)
```

In the resultant `data.frame`, the `seq_count` column is the number of raw sequences within each `sample_id` 
group for the given `gene`. `seq_freq` is the frequency of each `gene` within the given `sample_id`.

Below we plot only the IGHV1 abundance by filtering on the `gene` column to only rows 
containing IGHV1 family genes. We extract the family portion of the gene name using the 
`getFamily` function. Also, we take advantage of the `sortGenes` function to convert the 
`gene` column to a factor with gene name lexicographically ordered in the factor levels 
(`method="name"`) for axis ordering using the `ggplot2` package. Alternatively, we could have 
ordered the genes by genomic position by passing `method="position"` to `sortGenes`.

```{r, eval=TRUE, warning=FALSE}
# Assign sorted levels and subset to IGHV1
ighv1 <- gene %>%
    mutate(gene=factor(gene, levels=sortGenes(unique(gene), method="name"))) %>%
    filter(getFamily(gene) == "IGHV1")

# Plot V gene usage in the IGHV1 family by sample
g1 <- ggplot(ighv1, aes(x=gene, y=seq_freq)) +
    theme_bw() +
    ggtitle("IGHV1 Usage") +
    theme(axis.text.x=element_text(angle=45, hjust=1, vjust=1)) +
    ylab("Percent of repertoire") +
    xlab("") +
    scale_y_continuous(labels=percent) +
    scale_color_brewer(palette="Set1") +
    geom_point(aes(color=sample_id), size=5, alpha=0.8, position=position_dodge(width = 0.1))
plot(g1)
```

Alternatively, usage can be quantified at the allele (`mode="allele"`) or 
family level (`mode="family"`):

```{r, eval=TRUE, warning=FALSE}
# Quantify V family usage by sample
family <- countGenes(ExampleDb, gene="v_call", groups="sample_id", mode="family")

# Plot V family usage by sample
g2 <- ggplot(family, aes(x=gene, y=seq_freq)) +
    theme_bw() +
    ggtitle("Family Usage") +
    theme(axis.text.x=element_text(angle=45, hjust=1, vjust=1)) +
    ylab("Percent of repertoire") +
    xlab("") +
    scale_y_continuous(labels=percent) +
    scale_color_brewer(palette="Set1") +
    geom_point(aes(color=sample_id), size=5, alpha=0.8, position=position_dodge(width = 0.1))
plot(g2)
```


## Tabulating gene abundance using additional groupings

The `groups` argument to `countGenes` can accept multiple grouping columns and
will calculate abundance within each unique combination. In the examples below,
groupings will be perform by unique sample and isotype pairs 
(`groups=c("sample_id", "c_call")`). Furthermore, instead of quantifying abundance
by sequence count, we will quantify it by clone count (each clone will
be counted only once regardless of how many sequences the clone represents).  
Clonal criteria are added by passing a value to the `clone` argument of `countGenes`
(`clone="clone_id"`). For each clonal group, only the most common allele/gene/family will
be considered for counting.

```{r, eval=TRUE, warning=FALSE}
# Quantify V family clonal usage by sample and isotype
family <- countGenes(ExampleDb, gene="v_call", groups=c("sample_id", "c_call"), 
                     clone="clone_id", mode="family")
head(family, n=4)
```

The output `data.frame` contains the additional grouping column (`c_call`) along with the 
`clone_count` and `clone_freq` columns that represent the count of clones for each V family 
and the frequencies within the given `sample_id` and `c_call` pair, respectively.

```{r, eval=TRUE, warning=FALSE}
# Subset to IGHM and IGHG for plotting
family <- filter(family, c_call %in% c("IGHM", "IGHG"))
# Plot V family clonal usage by sample and isotype
g3 <- ggplot(family, aes(x=gene, y=clone_freq)) +
    theme_bw() +
    ggtitle("Clonal Usage") +
    theme(axis.text.x=element_text(angle=45, hjust=1, vjust=1)) +
    ylab("Percent of repertoire") +
    xlab("") +
    scale_y_continuous(labels=percent) +
    scale_color_brewer(palette="Set1") +
    geom_point(aes(color=sample_id), size=5, alpha=0.8, position=position_dodge(width = 0.1)) +
    facet_grid(. ~ c_call)
plot(g3)
```

Instead of calculating abundance by sequence or clone count, abundance can be calculated
using copy numbers for the individual sequences. This is accomplished by passing
a copy number column to the `copy` argument (`copy="duplicate_count"`). Specifying both
`clone` and `copy` arguments is not meaningful and will result in the `clone` argument
being ignored.

```{r, eval=TRUE, warning=FALSE}
# Calculate V family copy numbers by sample and isotype
family <- countGenes(ExampleDb, gene="v_call", groups=c("sample_id", "c_call"), 
                     mode="family", copy="duplicate_count")
head(family, n=4)
```

The output `data.frame` includes the `seq_count` and `seq_freq` columns as previously defined, 
as well as the additional copy number columns `copy_count` and `copy_freq` reflected the summed 
copy number (`duplicate_count`) for each sequence within the given `gene`, `sample_id` and `c_call`.

```{r, eval=TRUE, warning=FALSE}
# Subset to IGHM and IGHG for plotting
family <- filter(family, c_call %in% c("IGHM", "IGHG"))
# Plot V family copy abundance by sample and isotype
g4 <- ggplot(family, aes(x=gene, y=copy_freq)) +
    theme_bw() +
    ggtitle("Copy Number") +
    theme(axis.text.x=element_text(angle=45, hjust=1, vjust=1)) +
    ylab("Percent of repertoire") +
    xlab("") +
    scale_y_continuous(labels=percent) +
    scale_color_brewer(palette="Set1") +
    geom_point(aes(color=sample_id), size=5, alpha=0.8, position=position_dodge(width = 0.1)) +
    facet_grid(. ~ c_call)
plot(g4)
```

---
title: "Alakazam: How to read and write files"
author: "Edel Aron"
date: '`r Sys.Date()`'
output:
  pdf_document:
    dev: pdf
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    toc: yes
  html_document:
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    theme: readable
    toc: yes
  md_document:
    fig_height: 4
    fig_width: 7.5
    preserve_yaml: no
    toc: yes
geometry: margin=1in
fontsize: 11pt
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{File input and output}
  %\usepackage[utf8]{inputenc}
---

As part of the Immcantation suite of tools, the `alakazam` package includes a set of 
built-in functions capable of reading and writing tab-delimited database files created by 
[Change-O](https://changeo.readthedocs.io/en/stable/) into R data.frames. However, due to 
differences in how certain values and sequences are handled, `alakazam::readChangeoDb` and 
`alakazam::writeChangeoDb` will not properly read in [AIRR](https://docs.airr-community.org) 
formatted files. These files should instead be loaded using the functions included 
in the `airr` package (`airr::read_rearrangement` and `airr::write_rearrangement`).

You can read more about how we use both data standards
[here](https://immcantation.readthedocs.io/en/stable/datastandards.html) and 
[here](https://changeo.readthedocs.io/en/stable/standard.html). *Please note that the default 
file format for all functions in Immcantation is the AIRR-C format as of Immcantation 
v4.0.0, which corresponds to alakazam v1.0.0.*

## Reading data

Small example databases for both the Change-O format (`ExampleDbChangeo`) and the AIRR format (`ExampleDb`) 
are included in the `alakazam` package. For specific details about the latter, visit the 
[AIRR Community documentation site](https://docs.airr-community.org/en/stable/datarep/rearrangements.html).

```{r, eval=TRUE, warning=FALSE, message=FALSE}
# Set the file paths from inside the package directory
# These files are smaller versions of the example databases previously mentioned
changeo_file <- system.file("extdata", "example_changeo.tab.gz", package="alakazam")
airr_file <- system.file("extdata", "example_airr.tsv.gz", package="alakazam")

# Read in the data
db_changeo <- alakazam::readChangeoDb(changeo_file)
db_airr <- airr::read_rearrangement(airr_file)
```

## Writing data

```{r, eval=FALSE, warning=FALSE, message=FALSE}
# Write the data to a tab-delimited file
alakazam::writeChangeoDb(db_changeo, "changeo.tsv")
airr::write_rearrangement(db_airr, "airr.tsv")
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Lineage.R
\name{makeChangeoClone}
\alias{makeChangeoClone}
\title{Generate a ChangeoClone object for lineage construction}
\usage{
makeChangeoClone(
  data,
  id = "sequence_id",
  seq = "sequence_alignment",
  germ = "germline_alignment",
  v_call = "v_call",
  j_call = "j_call",
  junc_len = "junction_length",
  clone = "clone_id",
  mask_char = "N",
  locus = "locus",
  max_mask = 0,
  pad_end = FALSE,
  text_fields = NULL,
  num_fields = NULL,
  seq_fields = NULL,
  add_count = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{data}{data.frame containing the AIRR or Change-O data for a clone. See Details
for the list of required columns and their default values.}

\item{id}{name of the column containing sequence identifiers.}

\item{seq}{name of the column containing observed DNA sequences. All 
sequences in this column must be multiple aligned.}

\item{germ}{name of the column containing germline DNA sequences. All entries 
in this column should be identical for any given clone, and they
must be multiple aligned with the data in the \code{seq} column.}

\item{v_call}{name of the column containing V-segment allele assignments. All 
entries in this column should be identical to the gene level.}

\item{j_call}{name of the column containing J-segment allele assignments. All 
entries in this column should be identical to the gene level.}

\item{junc_len}{name of the column containing the length of the junction as a 
numeric value. All entries in this column should be identical 
for any given clone.}

\item{clone}{name of the column containing the identifier for the clone. All 
entries in this column should be identical.}

\item{mask_char}{character to use for masking and padding.}

\item{locus}{name of the column containing locus specification. Must be present
and only contain the value "IGH", representing heavy chains.}

\item{max_mask}{maximum number of characters to mask at the leading and trailing
sequence ends. If \code{NULL} then the upper masking bound will 
be automatically determined from the maximum number of observed 
leading or trailing Ns amongst all sequences. If set to \code{0} 
(default) then masking will not be performed.}

\item{pad_end}{if \code{TRUE} pad the end of each sequence with \code{mask_char}
to make every sequence the same length.}

\item{text_fields}{text annotation columns to retain and merge during duplicate removal.}

\item{num_fields}{numeric annotation columns to retain and sum during duplicate removal.}

\item{seq_fields}{sequence annotation columns to retain and collapse during duplicate 
removal. Note, this is distinct from the \code{seq} and \code{germ} 
arguments, which contain the primary sequence data for the clone
and should not be repeated in this argument.}

\item{add_count}{if \code{TRUE} add an additional annotation column called 
\code{collapse_count} during duplicate removal that indicates the 
number of sequences that were collapsed.}

\item{verbose}{passed on to \code{collapseDuplicates}. If \code{TRUE}, report the 
numbers of input, discarded and output sequences; otherwise, process
sequences silently.}
}
\value{
A \link{ChangeoClone} object containing the modified clone.
}
\description{
\code{makeChangeoClone} takes a data.frame with AIRR or Change-O style columns as input and 
masks gap positions, masks ragged ends, removes duplicate sequences, and merges 
annotations associated with duplicate sequences. It returns a \code{ChangeoClone} 
object which serves as input for lineage reconstruction. \strong{Note}: To use the 
most recent methods for building, visualizing and analyzing 
trees, use the R package [Dowser](https://dowser.readthedocs.io).
}
\details{
The input data.frame (\code{data}) must columns for each of the required column name 
arguments: \code{id}, \code{seq}, \code{germ}, \code{v_call}, \code{j_call}, 
\code{junc_len}, and \code{clone}.  The default values are as follows:
\itemize{
  \item  \code{id       = "sequence_id"}:         unique sequence identifier.
  \item  \code{seq      = "sequence_alignment"}:  IMGT-gapped sample sequence.
  \item  \code{germ     = "germline_alignment"}:  IMGT-gapped germline sequence.
  \item  \code{v_call   = "v_call"}:              V segment allele call.
  \item  \code{j_call   = "j_call"}:              J segment allele call.
  \item  \code{junc_len = "junction_length"}:     junction sequence length.
  \item  \code{clone    = "clone_id"}:            clone identifier.
}
Additional annotation columns specified in the \code{text_fields}, \code{num_fields} 
or \code{seq_fields} arguments will be retained in the \code{data} slot of the return 
object, but are not required. If the input data.frame \code{data} already contains a 
column named \code{sequence}, which is not used as the \code{seq} argument, then that 
column will not be retained.

The default columns are IMGT-gapped sequence columns, but this is not a requirement. 
However, all sequences (both observed and germline) must be multiple aligned using
some scheme for both proper duplicate removal and lineage reconstruction. 

The value for the germline sequence, V-segment gene call, J-segment gene call, 
junction length, and clone identifier are determined from the first entry in the 
\code{germ}, \code{v_call}, \code{j_call}, \code{junc_len} and \code{clone} columns, 
respectively. For any given clone, each value in these columns should be identical.
}
\examples{
# Example data
db <- data.frame(sequence_id=LETTERS[1:4],
                 sequence_alignment=c("CCCCTGGG", "CCCCTGGN", "NAACTGGN", "NNNCTGNN"),
                 germline_alignment="CCCCAGGG",
                 v_call="Homsap IGKV1-39*01 F",
                 j_call="Homsap IGKJ5*01 F",
                 junction_length=2,
                 clone_id=1,
                 locus=rep("IGH", length=4),
                 c_call=c("IGHM", "IGHG", "IGHG", "IGHA"),
                 duplicate_count=1:4,
                 stringsAsFactors=FALSE)


 # Without end masking
 makeChangeoClone(db, text_fields="c_call", num_fields="duplicate_count")

 # With end masking
 makeChangeoClone(db, max_mask=3, text_fields="c_call", num_fields="duplicate_count")

}
\seealso{
Executes in order \link{maskSeqGaps}, \link{maskSeqEnds}, 
          \link{padSeqEnds}, and \link{collapseDuplicates}. 
          Returns a \link{ChangeoClone} object which serves as input to
          \link{buildPhylipLineage}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Topology.R
\name{testEdges}
\alias{testEdges}
\title{Tests for parent-child annotation enrichment in lineage trees}
\usage{
testEdges(
  graphs,
  field,
  indirect = FALSE,
  exclude = c("Germline", NA),
  nperm = 200,
  progress = FALSE
)
}
\arguments{
\item{graphs}{list of igraph objects with vertex annotations.}

\item{field}{string defining the annotation field to permute.}

\item{indirect}{if \code{FALSE} count direct connections (edges) only. If 
\code{TRUE} walk through any nodes with annotations specified in 
the \code{argument} to count indirect connections. Specifying
\code{indirect=TRUE} with \code{exclude=NULL} will have no effect.}

\item{exclude}{vector of strings defining \code{field} values to exclude from 
permutation.}

\item{nperm}{number of permutations to perform.}

\item{progress}{if \code{TRUE} show a progress bar.}
}
\value{
An \link{EdgeTest} object containing the test results and permutation
          realizations.
}
\description{
\code{testEdges} performs a permutation test on a set of lineage trees to determine
the significance of an annotation's association with parent-child relationships.
}
\examples{
\donttest{
# Define example tree set
graphs <- ExampleTrees[1:10]

# Perform edge test on isotypes
x <- testEdges(graphs, "c_call", nperm=10)
print(x)
}

}
\seealso{
Uses \link{tableEdges} and \link{permuteLabels}. 
          See \link{plotEdgeTest} for plotting the permutation distributions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Lineage.R
\name{phyloToGraph}
\alias{phyloToGraph}
\title{Convert a tree in ape \code{phylo} format to igraph \code{graph} format.}
\usage{
phyloToGraph(phylo, germline = "Germline")
}
\arguments{
\item{phylo}{An ape \code{phylo} object.}

\item{germline}{If specified, places specified tip sequence as the direct 
ancestor of the tree}
}
\value{
A \code{graph} object representing the input tree.
}
\description{
\code{phyloToGraph} converts a tree in \code{phylo} format to and 
\code{graph} format.
}
\details{
Convert from phylo to graph object. Uses the node.label vector to label internal nodes. Nodes 
may rotate but overall topology will remain constant.
}
\examples{
\dontrun{
   library(igraph)
   library(ape)

   #convert to phylo
   phylo = graphToPhylo(graph)
   
   #plot tree using ape
   plot(phylo,show.node.label=TRUE)
   
   #store as newick tree
   write.tree(phylo,file="tree.newick")
   
   #read in tree from newick file
   phylo_r = read.tree("tree.newick")
   
   #convert to igraph
   graph_r = phyloToGraph(phylo_r,germline="Germline")
   
   #plot graph - same as before, possibly rotated
   plot(graph_r,layout=layout_as_tree)
}

}
\references{
\enumerate{
  \item  Hoehn KB, Lunter G, Pybus OG - A Phylogenetic Codon Substitution Model for Antibody 
             Lineages. Genetics 2017 206(1):417-427
             https://doi.org/10.1534/genetics.116.196303 
 \item  Hoehn KB, Vander Heiden JA, Zhou JQ, Lunter G, Pybus OG, Kleinstein SHK - 
             Repertoire-wide phylogenetic models of B cell molecular evolution reveal 
             evolutionary signatures of aging and vaccination. bioRxiv 2019  
             https://doi.org/10.1101/558825 
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Lineage.R
\name{buildPhylipLineage}
\alias{buildPhylipLineage}
\title{Infer an Ig lineage using PHYLIP}
\usage{
buildPhylipLineage(
  clone,
  phylip_exec,
  dist_mat = getDNAMatrix(gap = 0),
  rm_temp = FALSE,
  verbose = FALSE,
  temp_path = NULL,
  onetree = FALSE,
  branch_length = c("mutations", "distance")
)
}
\arguments{
\item{clone}{\link{ChangeoClone} object containing clone data.}

\item{phylip_exec}{absolute path to the PHYLIP dnapars executable.}

\item{dist_mat}{character distance matrix to use for reassigning edge weights. 
Defaults to a Hamming distance matrix returned by \link{getDNAMatrix} 
with \code{gap=0}. If gap characters, \code{c("-", ".")}, are assigned 
a value of -1 in \code{dist_mat} then contiguous gaps of any run length,
which are not present in both sequences, will be counted as a 
distance of 1. Meaning, indels of any length will increase
the sequence distance by 1. Gap values other than -1 will 
return a distance that does not consider indels as a special case.}

\item{rm_temp}{if \code{TRUE} delete the temporary directory after running dnapars;
if \code{FALSE} keep the temporary directory.}

\item{verbose}{if \code{FALSE} suppress the output of dnapars; 
if \code{TRUE} STDOUT and STDERR of dnapars will be passed to 
the console.}

\item{temp_path}{specific path to temp directory if desired.}

\item{onetree}{if \code{TRUE} save only one tree.}

\item{branch_length}{specifies how to define branch lengths; one of \code{"mutations"} 
or \code{"distance"}. If set to \code{"mutations"} (default), then branch 
lengths represent the number of mutations between nodes. 
If set to \code{"distance"}, then branch lengths represent
the expected number of mutations per site, unaltered from PHYLIP output.}
}
\value{
An igraph \code{graph} object defining the Ig lineage tree. Each unique input 
          sequence in \code{clone} is a vertex of the tree, with additional vertices being
          either the germline (root) sequences or inferred intermediates. The \code{graph} 
          object has the following attributes.
          
          Vertex attributes:
          \itemize{
            \item  \code{name}:      value in the \code{sequence_id} column of the \code{data} 
                                     slot of the input \code{clone} for observed sequences. 
                                     The germline (root) vertex is assigned the name 
                                     "Germline" and inferred intermediates are assigned
                                     names with the format \{"Inferred1", "Inferred2", ...\}.
            \item  \code{sequence}:  value in the \code{sequence} column of the \code{data} 
                                     slot of the input \code{clone} for observed sequences.
                                     The germline (root) vertex is assigned the sequence
                                     in the \code{germline} slot of the input \code{clone}.
                                     The sequence of inferred intermediates are extracted
                                     from the dnapars output.
            \item  \code{label}:     same as the \code{name} attribute.
          }
          Additionally, each other column in the \code{data} slot of the input 
          \code{clone} is added as a vertex attribute with the attribute name set to 
          the source column name. For the germline and inferred intermediate vertices,
          these additional vertex attributes are all assigned a value of \code{NA}.
          
          Edge attributes:
          \itemize{
            \item  \code{weight}:    Hamming distance between the \code{sequence} attributes
                                     of the two vertices.
            \item  \code{label}:     same as the \code{weight} attribute.
          }
          Graph attributes:
          \itemize{
            \item  \code{clone}:     clone identifier from the \code{clone} slot of the
                                     input \code{ChangeoClone}.
            \item  \code{v_gene}:    V-segment gene call from the \code{v_gene} slot of 
                                     the input \code{ChangeoClone}.
            \item  \code{j_gene}:    J-segment gene call from the \code{j_gene} slot of 
                                     the input \code{ChangeoClone}.
            \item  \code{junc_len}:  junction length (nucleotide count) from the 
                                     \code{junc_len} slot of the input \code{ChangeoClone}.

             Alternatively, this function will return an \code{phylo} object, which is compatible
             with the ape package. This object will contain reconstructed ancestral sequences in
             \code{nodes} attribute.
          }
}
\description{
\code{buildPhylipLineage} reconstructs an Ig lineage via maximum parsimony using the 
dnapars application, or maximum likelihood using the dnaml application of the PHYLIP package.
\strong{Note}: To use the most recent methods for building, visualizing and analyzing 
trees, use the R package [Dowser](https://dowser.readthedocs.io).
}
\details{
\code{buildPhylipLineage} builds the lineage tree of a set of unique Ig sequences via
maximum parsimony through an external call to the dnapars application of the PHYLIP
package. dnapars is called with default algorithm options, except for the search option, 
which is set to "Rearrange on one best tree". The germline sequence of the clone is used 
for the outgroup. 

Following tree construction using dnapars, the dnapars output is modified to allow
input sequences to appear as internal nodes of the tree. Intermediate sequences 
inferred by dnapars are replaced by children within the tree having a Hamming distance 
of zero from their parent node. With the default \code{dist_mat}, the distance calculation 
allows IUPAC ambiguous character matches, where an ambiguous character has distance zero 
to any character in the set of characters it represents. Distance calculation and movement of 
child nodes up the tree is repeated until all parent-child pairs have a distance greater than zero 
between them. The germline sequence (outgroup) is moved to the root of the tree and
excluded from the node replacement processes, which permits the trunk of the tree to be
the only edge with a distance of zero. Edge weights of the resultant tree are assigned 
as the distance between each sequence.
}
\examples{
\dontrun{
# Preprocess clone
db <- subset(ExampleDb, clone_id == 3138)
clone <- makeChangeoClone(db, text_fields=c("sample_id", "c_call"), 
                          num_fields="duplicate_count")

# Run PHYLIP and process output
phylip_exec <- "~/apps/phylip-3.695/bin/dnapars"
graph <- buildPhylipLineage(clone, phylip_exec, rm_temp=TRUE)

# Plot graph with a tree layout
library(igraph)
plot(graph, layout=layout_as_tree, vertex.label=V(graph)$c_call, 
     vertex.size=50, edge.arrow.mode=0, vertex.color="grey80")

# To consider each indel event as a mutation, change the masking character 
# and distance matrix
clone <- makeChangeoClone(db, text_fields=c("sample_id", "c_call"), 
                          num_fields="duplicate_count", mask_char="-")
graph <- buildPhylipLineage(clone, phylip_exec, dist_mat=getDNAMatrix(gap=-1), 
                            rm_temp=TRUE)
}

}
\references{
\enumerate{
  \item  Felsenstein J. PHYLIP - Phylogeny Inference Package (Version 3.2). 
           Cladistics. 1989 5:164-166.
  \item  Stern JNH, Yaari G, Vander Heiden JA, et al. B cells populating the multiple 
           sclerosis brain mature in the draining cervical lymph nodes. 
           Sci Transl Med. 2014 6(248):248ra107.
}
}
\seealso{
Takes as input a \link{ChangeoClone}. 
          Temporary directories are created with \link{makeTempDir}.
          Distance is calculated using \link{seqDist}. 
          See [igraph](http://www.rdocumentation.org/packages/igraph/topics/aaa-igraph-package) 
          and [igraph.plotting](http://www.rdocumentation.org/packages/igraph/topics/plot.common) 
          for working  with igraph \code{graph} objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Sequence.R
\name{junctionAlignment}
\alias{junctionAlignment}
\title{Calculate junction region alignment properties}
\usage{
junctionAlignment(
  data,
  germline_db,
  v_call = "v_call",
  d_call = "d_call",
  j_call = "j_call",
  v_germline_start = "v_germline_start",
  v_germline_end = "v_germline_end",
  d_germline_start = "d_germline_start",
  d_germline_end = "d_germline_end",
  j_germline_start = "j_germline_start",
  j_germline_end = "j_germline_end",
  np1_length = "np1_length",
  np2_length = "np2_length",
  junction = "junction",
  junction_length = "junction_length",
  sequence_alignment = "sequence_alignment"
)
}
\arguments{
\item{data}{\code{data.frame} containing sequence data.}

\item{germline_db}{reference germline database for the V, D and J genes.
in \code{data}}

\item{v_call}{V gene assignment column.}

\item{d_call}{D gene assignment column.}

\item{j_call}{J gene assignment column.}

\item{v_germline_start}{column containing the start position of the alignment 
in the V reference germline.}

\item{v_germline_end}{column containing the end position of the alignment in the 
V reference germline.}

\item{d_germline_start}{column containing the start position of the alignment 
in the D reference germline.}

\item{d_germline_end}{column containing the start position of the alignment 
in the D reference germline.}

\item{j_germline_start}{column containing the start position of the alignment 
in the J reference germline.}

\item{j_germline_end}{column containing the start position of the alignment 
in the J reference germline.}

\item{np1_length}{combined length of the N and P regions between the 
V and D regions (heavy chain) or V and J regions (light chain).}

\item{np2_length}{combined length of the N and P regions between the 
D and J regions (heavy chain).}

\item{junction}{column containing the junction sequence.}

\item{junction_length}{column containing the length of the junction region in nucleotides.}

\item{sequence_alignment}{column containing the aligned sequence.}
}
\value{
A modified input \code{data.frame} with the following additional columns storing 
         junction alignment information:
         \enumerate{
             \item  \code{e3v_length}:     number of 3' V germline nucleotides deleted.
             \item  \code{e5d_length}:     number of 5' D germline nucleotides deleted.
             \item  \code{e3d_length}:     number of 3' D germline nucleotides deleted.
             \item  \code{e5j_length}:     number of 5' J germline nucleotides deleted.
             \item  \code{v_cdr3_length}:  number of sequence_alignment V nucleotides in the CDR3.
             \item  \code{j_cdr3_length}:  number of sequence_alignment J nucleotides in the CDR3.
         }
}
\description{
\code{junctionAlignment} determines the number of deleted germline nucleotides in the 
junction region and the number of V gene and J gene nucleotides in the CDR3.
}
\examples{
germline_db <- list(
"IGHV3-11*05"="CAGGTGCAGCTGGTGGAGTCTGGGGGA...GGCTTGGTCAAGCCTGGAGGGTCCCTGAGACT
CTCCTGTGCAGCCTCTGGATTCACCTTC............AGTGACTACTACATGAGCTGGATCCGCCAGGCTCCAG
GGAAGGGGCTGGAGTGGGTTTCATACATTAGTAGTAGT......AGTAGTTACACAAACTACGCAGACTCTGTGAAG
...GGCCGATTCACCATCTCCAGAGACAACGCCAAGAACTCACTGTATCTGCAAATGAACAGCCTGAGAGCCGAGGA
CACGGCCGTGTATTACTGTGCGAGAGA",
"IGHD3-10*01"="GTATTACTATGGTTCGGGGAGTTATTATAAC",
"IGHJ5*02"="ACAACTGGTTCGACCCCTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAG"
)

db <- junctionAlignment(SingleDb, germline_db)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Core.R
\name{baseTheme}
\alias{baseTheme}
\title{Standard ggplot settings}
\usage{
baseTheme(sizing = c("figure", "window"))
}
\arguments{
\item{sizing}{defines the style and sizing of the theme. One of 
\code{c("figure", "window")} where \code{sizing="figure"} is appropriately
sized for pdf export at 7 to 7.5 inch width, and \code{sizing="window"}
is sized for an interactive session.}
}
\value{
A ggplot2 object.
}
\description{
\code{baseTheme} defines common ggplot theme settings for plotting.
}
\seealso{
\link[ggplot2]{theme}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Core.R
\name{readChangeoDb}
\alias{readChangeoDb}
\title{Read a Change-O tab-delimited database file}
\usage{
readChangeoDb(file, select = NULL, drop = NULL, seq_upper = TRUE)
}
\arguments{
\item{file}{tab-delimited database file output by a Change-O tool.}

\item{select}{columns to select from database file.}

\item{drop}{columns to drop from database file.}

\item{seq_upper}{if \code{TRUE} convert sequence columns to upper case;
if \code{FALSE} do not alter sequence columns. See Value 
for a list of which columns are effected.}
}
\value{
A data.frame of the database file. Columns will be imported as is, except for 
          the following columns which will be explicitly converted into character 
          values:
          \itemize{
            \item  SEQUENCE_ID
            \item  CLONE
            \item  SAMPLE
          }
          And the following sequence columns which will converted to upper case if
          \code{seq_upper=TRUE} (default).
          \itemize{
            \item  SEQUENCE_INPUT
            \item  SEQUENCE_VDJ
            \item  SEQUENCE_IMGT
            \item  JUNCTION
            \item  GERMLINE_IMGT
            \item  GERMLINE_IMGT_D_MASK
          }
}
\description{
\code{readChangeoDb} reads a tab-delimited database file created by a Change-O tool 
into a data.frame.
}
\examples{
\dontrun{
    # Read all columns in and convert sequence fields to upper case
    db <- readChangeoDb("changeo.tsv")

    # Subset columns and convert sequence fields to upper case
    db <- readChangeoDb("changeo.tsv", select=c("SEQUENCE_ID", "SEQUENCE_IMGT"))

    # Drop columns and do not alter sequence field case
    db <- readChangeoDb("changeo.tsv", drop=c("D_CALL", "DUPCOUNT"), 
                        seq_upper=FALSE)
}

}
\seealso{
Wraps \link[readr]{read_delim}. 
          See \link{writeChangeoDb} for writing to Change-O files.
          See \link[airr:read_tabular]{read_rearrangement} and \link[airr:write_tabular]{write_rearrangement}
          to read and write AIRR-C Standard formatted repertoires.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Data.R
\docType{data}
\name{IMGT_REGIONS}
\alias{IMGT_REGIONS}
\title{IMGT V-segment regions}
\format{
A list with regions named one of \code{c("fwr1", "cdr1", "fwr2", "cdr2", "fwr3")} 
         with values containing a numeric vector of length two defining the 
         \code{c(start, end)} positions of the named region.
}
\usage{
IMGT_REGIONS
}
\description{
A list defining the boundaries of V-segment framework regions (FWRs) and complementarity 
determining regions (CDRs) for IMGT-gapped immunoglobulin (Ig) nucleotide sequences 
according to the IMGT numbering scheme.
}
\references{
\url{https://www.imgt.org/}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AminoAcids.R
\name{gravy}
\alias{gravy}
\title{Calculates the hydrophobicity of amino acid sequences}
\usage{
gravy(seq, hydropathy = NULL)
}
\arguments{
\item{seq}{vector of strings containing amino acid sequences.}

\item{hydropathy}{named numerical vector defining hydropathy index values for 
each amino acid, where names are single-letter amino acid 
character codes. If \code{NULL}, then the Kyte & Doolittle
scale is used.}
}
\value{
A vector of gravy scores for the sequence(s).
}
\description{
\code{gravy} calculates the Grand Average of Hydrophobicity (gravy) index 
of amino acid sequences using the method of Kyte & Doolittle. Non-informative
positions are excluded, where non-informative is defined as any character in 
\code{c("X", "-", ".", "*")}.
}
\examples{
# Default scale
seq <- c("CARDRSTPWRRGIASTTVRTSW", "XXTQMYVRT")
gravy(seq)

# Use the Kidera et al, 1985 scores from the seqinr package
library(seqinr)
data(aaindex)
x <- aaindex[["KIDA850101"]]$I
# Rename the score vector to use single-letter codes
names(x) <- translateStrings(names(x), ABBREV_AA)
# Calculate hydrophobicity
gravy(seq, hydropathy=x)

}
\references{
\enumerate{
  \item  Kyte J, Doolittle RF. A simple method for displaying the hydropathic character 
           of a protein. J Mol Biol. 157, 105-32 (1982).
}
}
\seealso{
For additional hydrophobicity indices see \code{\link[seqinr]{aaindex}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AminoAcids.R
\name{countPatterns}
\alias{countPatterns}
\title{Count sequence patterns}
\usage{
countPatterns(seq, patterns, nt = TRUE, trim = FALSE, label = "region")
}
\arguments{
\item{seq}{character vector of either DNA or amino acid sequences.}

\item{patterns}{list of sequence patterns to count in each sequence. If the 
list is named, then names will be assigned as the column names of 
output data.frame.}

\item{nt}{if \code{TRUE} then \code{seq} are DNA sequences and and will be 
translated before performing the pattern search.}

\item{trim}{if \code{TRUE} remove the first and last codon or amino acid from 
each sequence before the pattern search. If \code{FALSE} do
not modify the input sequences.}

\item{label}{string defining a label to add as a prefix to the output 
column names.}
}
\value{
A data.frame containing the fraction of times each sequence pattern was 
         found.
}
\description{
\code{countPatterns} counts the fraction of times a set of character patterns occur 
in a set of sequences.
}
\examples{
seq <- c("TGTCAACAGGCTAACAGTTTCCGGACGTTC",
         "TGTCAGCAATATTATATTGCTCCCTTCACTTTC",
         "TGTCAAAAGTATAACAGTGCCCCCTGGACGTTC")
patterns <- c("A", "V", "[LI]")
names(patterns) <- c("arg", "val", "iso_leu")
countPatterns(seq, patterns, nt=TRUE, trim=TRUE, label="cdr3")
            
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Sequence.R
\name{getDNAMatrix}
\alias{getDNAMatrix}
\title{Build a DNA distance matrix}
\usage{
getDNAMatrix(gap = -1)
}
\arguments{
\item{gap}{value to assign to characters in the set \code{c("-", ".")}.}
}
\value{
A \code{matrix} of DNA character distances with row and column names 
          indicating the character pair. By default, distances will be either 0 
          (equivalent), 1 (non-equivalent or missing), or -1 (gap).
}
\description{
\code{getDNAMatrix} returns a Hamming distance matrix for IUPAC ambiguous
DNA characters with modifications for gap, \code{c("-", ".")}, and missing, 
\code{c("?")}, character values.
}
\examples{
# Set gap characters to Inf distance
# Distinguishes gaps from Ns
getDNAMatrix()

# Set gap characters to 0 distance
# Makes gap characters equivalent to Ns
getDNAMatrix(gap=0)

}
\seealso{
Creates DNA distance matrix for \link{seqDist}.
          See \link{getAAMatrix} for amino acid distances.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Classes.R
\docType{class}
\name{EdgeTest-class}
\alias{EdgeTest-class}
\alias{EdgeTest}
\alias{print,EdgeTest-method}
\alias{EdgeTest-method}
\alias{plot,EdgeTest,missing-method}
\title{S4 class defining edge significance}
\usage{
\S4method{print}{EdgeTest}(x)

\S4method{plot}{EdgeTest,missing}(x, y, ...)
}
\arguments{
\item{x}{EdgeTest object.}

\item{y}{ignored.}

\item{...}{arguments to pass to \link{plotEdgeTest}.}
}
\description{
\code{EdgeTest} defines the significance of parent-child annotation enrichment.
}
\section{Slots}{

\describe{
\item{\code{tests}}{data.frame describing the significance test results with columns:
\itemize{
  \item  \code{parent}:    parent node annotation.
  \item  \code{child}:     child node annotation
  \item  \code{count}:     count of observed edges with the given 
                           parent-child annotation set.
  \item  \code{expected}:  mean count of expected edges for the 
                           given parent-child relationship.
  \item  \code{pvalue}:    one-sided p-value for the hypothesis that 
                            the observed edge abundance is greater 
                            than expected.
}}

\item{\code{permutations}}{data.frame containing the raw permutation test data with columns:
\itemize{
  \item  \code{parent}:  parent node annotation.
  \item  \code{child}:   child node annotation
  \item  \code{count}:   count of edges with the given parent-child 
                         annotation set.
  \item  \code{iter}:    numerical index define which permutation
                         realization each observation corresponds 
                         to.
}}

\item{\code{nperm}}{number of permutation realizations.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Alakazam.R
\docType{package}
\name{alakazam-package}
\alias{alakazam-package}
\title{alakazam: Immunoglobulin Clonal Lineage and Diversity Analysis}
\description{
Provides methods for high-throughput adaptive immune receptor repertoire sequencing (AIRR-Seq; Rep-Seq) analysis. In particular, immunoglobulin (Ig) sequence lineage reconstruction, lineage topology analysis, diversity profiling, amino acid property analysis and gene usage. Citations: Gupta and Vander Heiden, et al (2017) \doi{10.1093/bioinformatics/btv359}, Stern, Yaari and Vander Heiden, et al (2014) \doi{10.1126/scitranslmed.3008879}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://alakazam.readthedocs.io/}
  \item Report bugs at \url{https://github.com/immcantation/alakazam/issues}
}

}
\author{
\strong{Maintainer}: Susanna Marquez \email{susanna.marquez@yale.edu}

Authors:
\itemize{
  \item Namita Gupta \email{namita.gupta@yale.edu}
  \item Nima Nouri \email{nima.nouri@yale.edu}
  \item Ruoyi Jiang \email{ruoyi.jiang@yale.edu}
  \item Julian Zhou \email{julian.zhou@bulldogs.yale.edu}
  \item Kenneth Hoehn \email{kenneth.hoehn@yale.edu}
  \item Cole Jensen \email{cole.jensen@yale.edu}
  \item Jason Vander Heiden \email{jason.vanderheiden@gmail.com}
  \item Steven Kleinstein \email{steven.kleinstein@yale.edu} [copyright holder]
}

Other contributors:
\itemize{
  \item Daniel Gadala-Maria \email{daniel.gadala-maria@yale.edu} [contributor]
  \item Edel Aron \email{edel.aron@yale.edu} [contributor]
  \item Gisela Gabernet \email{gisela.gabernet@yale.edu} [contributor]
  \item Caroline Sullivan \email{caroline.sullivan@yale.edu} [contributor]
  \item Hailong Meng \email{hailong.meng@yale.edu} [contributor]
  \item Huimin Lyu \email{huimin.lyu@yale.edu} [contributor]
  \item Burhan Sabuwala \email{burhan.sabuwala@yale.edu} [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Lineage.R
\name{combineIgphyml}
\alias{combineIgphyml}
\title{Combine IgPhyML object parameters into a dataframe}
\usage{
combineIgphyml(iglist, format = c("wide", "long"))
}
\arguments{
\item{iglist}{list of igphyml objects returned by \link{readIgphyml}. 
Each must have an \code{id} column in its \code{param} attribute, 
which can be added automatically using the \code{id} option of 
\code{readIgphyml}.}

\item{format}{string specifying whether each column of the resulting data.frame
should represent a parameter (\code{wide}) or if 
there should only be three columns; i.e. id, variable, and value
(\code{long}).}
}
\value{
A data.frame containing HLP model parameter estimates for all igphyml objects.
          Only parameters shared among all objects will be returned.
}
\description{
\code{combineIgphyml} combines IgPhyML object parameters into a data.frame.
}
\details{
\code{combineIgphyml} combines repertoire-wide parameter estimates from multiple igphyml
objects produced by readIgphyml into a dataframe that can be easily used for plotting and 
other hypothesis testing analyses.

All igphyml objects used must have an "id" column in their \code{param} attribute, which
can be added automatically from the \code{id} flag of \code{readIgphyml}.
}
\examples{
\dontrun{
   # Read in and combine two igphyml runs
   s1 <- readIgphyml("IB+7d_lineages_gy.tsv_igphyml_stats_hlp.tab", id="+7d")
   s2 <- readIgphyml("IB+7d_lineages_gy.tsv_igphyml_stats_hlp.tab", id="s2")
   combineIgphyml(list(s1, s2))
}

}
\references{
\enumerate{
  \item  Hoehn KB, Lunter G, Pybus OG - A Phylogenetic Codon Substitution Model for Antibody 
             Lineages. Genetics 2017 206(1):417-427
             https://doi.org/10.1534/genetics.116.196303 
 \item  Hoehn KB, Vander Heiden JA, Zhou JQ, Lunter G, Pybus OG, Kleinstein SHK - 
             Repertoire-wide phylogenetic models of B cell molecular evolution reveal 
             evolutionary signatures of aging and vaccination. bioRxiv 2019  
             https://doi.org/10.1101/558825 
}
}
\seealso{
\link{readIgphyml}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AminoAcids.R
\name{charge}
\alias{charge}
\title{Calculates the net charge of amino acid sequences.}
\usage{
charge(seq, pH = 7.4, pK = NULL, normalize = FALSE)
}
\arguments{
\item{seq}{vector strings defining of amino acid sequences.}

\item{pH}{environmental pH.}

\item{pK}{named vector defining pK values for each charged amino acid,
where names are the single-letter amino acid character codes
\code{c("R", "H", "K", "D", "E", "C", "Y")}). If \code{NULL}, 
then the EMBOSS scale is used.}

\item{normalize}{if \code{TRUE} then divide the net charge of each amino acid 
sequence by the number of informative positions. Non-informative 
position are defined by the presence any character in 
\code{c("X", "-", ".", "*")}. If \code{FALSE} then return the raw
net charge.}
}
\value{
A vector of net charges for the sequence(s).
}
\description{
\code{charge} calculates the net charge of amino acid sequences using 
the method of Moore, 1985, with exclusion of the C-terminus and N-terminus charges.
}
\examples{
seq <- c("CARDRSTPWRRGIASTTVRTSW", "XXTQMYVRT") 
# Unnormalized charge
charge(seq)
# Normalized charge
charge(seq, normalize=TRUE)

# Use the Murray et al, 2006 scores from the seqinr package
library(seqinr)
data(pK)
x <- setNames(pK[["Murray"]], rownames(pK))
# Calculate charge
charge(seq, pK=x)

}
\references{
\enumerate{
  \item  Moore DS. Amino acid and peptide net charges: A simple calculational procedure. 
           Biochem Educ. 13, 10-11 (1985).
  \item  \url{https://emboss.sourceforge.net/apps/cvs/emboss/apps/iep.html}
  }
}
\seealso{
For additional pK scales see \code{\link[seqinr]{pK}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Classes.R
\docType{class}
\name{MRCATest-class}
\alias{MRCATest-class}
\alias{MRCATest}
\alias{print,MRCATest-method}
\alias{MRCATest-method}
\alias{plot,MRCATest,missing-method}
\title{S4 class defining edge significance}
\usage{
\S4method{print}{MRCATest}(x)

\S4method{plot}{MRCATest,missing}(x, y, ...)
}
\arguments{
\item{x}{MRCATest object.}

\item{y}{ignored.}

\item{...}{arguments to pass to \link{plotMRCATest}.}
}
\description{
\code{MRCATest} defines the significance of enrichment for annotations appearing at
the MRCA of the tree.
}
\section{Slots}{

\describe{
\item{\code{tests}}{data.frame describing the significance test results with columns:
\itemize{
  \item  \code{annotation}:  annotation value.
  \item  \code{count}:       observed count of MRCA positions 
                             with the given annotation.
  \item  \code{expected}:    expected mean count of MRCA occurrence
                             for the annotation.
  \item  \code{pvalue}:      one-sided p-value for the hypothesis that 
                             the observed annotation abundance is greater 
                             than expected.
}}

\item{\code{permutations}}{data.frame containing the raw permutation test data with columns:
\itemize{
  \item  \code{annotation}:  annotation value.
  \item  \code{count}:       count of MRCA positions with the 
                             given annotation.
  \item  \code{iter}:        numerical index define which 
                             permutation realization each 
                             observation corresponds to.
}}

\item{\code{nperm}}{number of permutation realizations.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Data.R
\docType{data}
\name{ExampleDbChangeo}
\alias{ExampleDbChangeo}
\title{Example Change-O database}
\format{
A data.frame with the following Change-O style columns:
  \itemize{
    \item  \code{SEQUENCE_ID}:           Sequence identifier
    \item  \code{SEQUENCE_IMGT}:         IMGT-gapped observed sequence.
    \item  \code{GERMLINE_IMGT_D_MASK}:  IMGT-gapped germline sequence with N, P and 
                                         D regions masked.
    \item  \code{V_CALL}:                V region allele assignments.
    \item  \code{V_CALL_GENOTYPED}:      TIgGER corrected V region allele assignment.
    \item  \code{D_CALL}:                D region allele assignments.
    \item  \code{J_CALL}:                J region allele assignments.
    \item  \code{JUNCTION}:              Junction region sequence.
    \item  \code{JUNCTION_LENGTH}:       Length of the junction region in nucleotides.
    \item  \code{NP1_LENGTH}:            Combined length of the N and P regions proximal
                                         to the V region.
    \item  \code{NP2_LENGTH}:            Combined length of the N and P regions proximal
                                         to the J region.
    \item  \code{SAMPLE}:                Sample identifier. Time in relation to vaccination.
    \item  \code{ISOTYPE}:               Isotype assignment.
    \item  \code{DUPCOUNT}:              Copy count (number of duplicates) of the sequence.
    \item  \code{CLONE}:                 Change-O assignment clonal group identifier.
}
}
\usage{
ExampleDbChangeo
}
\description{
A small example database subset from Laserson and Vigneault et al, 2014.
}
\references{
\enumerate{
  \item  Laserson U and Vigneault F, et al. High-resolution antibody dynamics of 
           vaccine-induced immune responses. 
           Proc Natl Acad Sci USA. 2014 111:4928-33.
}
}
\seealso{
\link{ExampleDb} \link{ExampleTrees}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Sequence.R
\name{maskSeqGaps}
\alias{maskSeqGaps}
\title{Masks gap characters in DNA sequences}
\usage{
maskSeqGaps(seq, mask_char = "N", outer_only = FALSE)
}
\arguments{
\item{seq}{character vector of DNA sequence strings.}

\item{mask_char}{character to use for masking.}

\item{outer_only}{if \code{TRUE} replace only contiguous leading and trailing gaps;
if \code{FALSE} replace all gap characters.}
}
\value{
A modified \code{seq} vector with \code{"N"} in place of \code{c("-", ".")} 
          characters.
}
\description{
\code{maskSeqGaps} substitutes gap characters, \code{c("-", ".")}, with \code{"N"} 
in a vector of DNA sequences.
}
\examples{
# Mask with Ns
maskSeqGaps(c("ATG-C", "CC..C"))
maskSeqGaps("--ATG-C-")
maskSeqGaps("--ATG-C-", outer_only=TRUE)

# Mask with dashes
maskSeqGaps(c("ATG-C", "CC..C"), mask_char="-")

}
\seealso{
See \link{maskSeqEnds} for masking ragged edges.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Classes.R
\docType{class}
\name{DiversityCurve-class}
\alias{DiversityCurve-class}
\alias{DiversityCurve}
\alias{print,DiversityCurve-method}
\alias{DiversityCurve-method}
\alias{plot,DiversityCurve,missing-method}
\alias{plot,DiversityCurve,numeric-method}
\title{S4 class defining a diversity curve}
\usage{
\S4method{print}{DiversityCurve}(x)

\S4method{plot}{DiversityCurve,missing}(x, y, ...)

\S4method{plot}{DiversityCurve,numeric}(x, y, ...)
}
\arguments{
\item{x}{DiversityCurve object}

\item{y}{diversity order to plot (q).}

\item{...}{arguments to pass to \link{plotDiversityCurve} or \link{plotDiversityTest}.}
}
\description{
\code{DiversityCurve} defines diversity (\eqn{D}) scores over multiple diversity 
orders (\eqn{Q}).
}
\section{Slots}{

\describe{
\item{\code{diversity}}{data.frame defining the diversity curve with the following columns:
\itemize{
  \item  \code{group}:    group label.
  \item  \code{q}:        diversity order.
  \item  \code{d}:        mean diversity index over all bootstrap 
                          realizations.
  \item  \code{d_sd}:     standard deviation of the diversity index 
                          over all bootstrap realizations.
  \item  \code{d_lower}:  diversity lower confidence interval bound.
  \item  \code{d_upper}:  diversity upper confidence interval bound.
  \item  \code{e}:        evenness index calculated as \code{D} 
                          divided by \code{D} at \code{Q=0}.
  \item  \code{e_lower}:  evenness lower confidence interval bound.
  \item  \code{e_upper}:  evenness upper confidence interval bound.
}}

\item{\code{tests}}{data.frame describing the significance test results with columns:
\itemize{
  \item  \code{test}:        string listing the two groups tested.
  \item  \code{delta_mean}:  mean of the \eqn{D} bootstrap delta 
                             distribution for the test.
  \item  \code{delta_sd}:    standard deviation of the \eqn{D} 
                             bootstrap delta distribution for the test.
  \item  \code{pvalue}:      p-value for the test.
}}

\item{\code{group_by}}{string specifying the name of the grouping column in diversity calculation.}

\item{\code{groups}}{vector specifying the names of unique groups in group column in diversity calculation.}

\item{\code{method}}{string specifying the type of diversity calculated.}

\item{\code{q}}{vector of diversity hill diversity indices used for computing diversity.}

\item{\code{n}}{numeric vector indication the number of sequences sampled in each group.}

\item{\code{ci}}{confidence interval defining the upper and lower bounds 
(a value between 0 and 1).}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Diversity.R
\name{estimateAbundance}
\alias{estimateAbundance}
\title{Estimates the complete clonal relative abundance distribution}
\usage{
estimateAbundance(
  data,
  clone = "clone_id",
  copy = NULL,
  group = NULL,
  min_n = 30,
  max_n = NULL,
  uniform = TRUE,
  ci = 0.95,
  nboot = 200,
  cell_id = "cell_id",
  progress = FALSE
)
}
\arguments{
\item{data}{data.frame with Change-O style columns containing clonal assignments.}

\item{clone}{name of the \code{data} column containing clone identifiers.}

\item{copy}{name of the \code{data} column containing copy numbers for each 
sequence. If \code{copy=NULL} (the default), then clone abundance
is determined by the number of sequences. If a \code{copy} column
is specified, then clone abundances is determined by the sum of 
copy numbers within each clonal group.}

\item{group}{name of the \code{data} column containing group identifiers. 
If \code{NULL} then no grouping is performed and the \code{group} 
column of the output will contain the value \code{NA} for each row.}

\item{min_n}{minimum number of observations to sample.
A group with less observations than the minimum is excluded.}

\item{max_n}{maximum number of observations to sample. If \code{NULL} then no 
maximum is set.}

\item{uniform}{if \code{TRUE} then uniformly resample each group to the same 
number of observations. If \code{FALSE} then allow each group to
be resampled to its original size or, if specified, \code{max_size}.}

\item{ci}{confidence interval to calculate; the value must be between 0 and 1.}

\item{nboot}{number of bootstrap realizations to generate.}

\item{cell_id}{name of the \code{data} column containing cell identifiers. If
\code{cell_id=NULL} then the function will assume bulk data.}

\item{progress}{if \code{TRUE} show a progress bar.}
}
\value{
A \link{AbundanceCurve} object summarizing the abundances.
}
\description{
\code{estimateAbundance} estimates the complete clonal relative abundance distribution 
and confidence intervals on clone sizes using bootstrapping.
}
\examples{
abund <- estimateAbundance(ExampleDb, group="sample_id", nboot=100)

}
\references{
\enumerate{
  \item  Chao A. Nonparametric Estimation of the Number of Classes in a Population. 
           Scand J Stat. 1984 11, 265270.
  \item  Chao A, et al. Rarefaction and extrapolation with Hill numbers: 
           A framework for sampling and estimation in species diversity studies. 
           Ecol Monogr. 2014 84:45-67.
  \item  Chao A, et al. Unveiling the species-rank abundance distribution by 
           generalizing the Good-Turing sample coverage theory. 
           Ecology. 2015 96, 11891201.
}
}
\seealso{
See \link{plotAbundanceCurve} for plotting of the abundance distribution.
See \link{alphaDiversity} for a similar application to clonal diversity.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AminoAcids.R
\name{aliphatic}
\alias{aliphatic}
\title{Calculates the aliphatic index of amino acid sequences}
\usage{
aliphatic(seq, normalize = TRUE)
}
\arguments{
\item{seq}{vector of strings containing amino acid sequences.}

\item{normalize}{if \code{TRUE} then divide the aliphatic index of each amino acid 
sequence by the number of informative positions. Non-informative 
position are defined by the presence any character in 
\code{c("X", "-", ".", "*")}. If \code{FALSE} then return the raw
aliphatic index.}
}
\value{
A vector of the aliphatic indices for the sequence(s).
}
\description{
\code{aliphatic} calculates the aliphatic index of amino acid sequences using 
the method of Ikai. Non-informative positions are excluded, where non-informative 
is defined as any character in \code{c("X", "-", ".", "*")}.
}
\examples{
seq <- c("CARDRSTPWRRGIASTTVRTSW", NA, "XXTQMYVRT")
aliphatic(seq)

}
\references{
\enumerate{
  \item  Ikai AJ. Thermostability and aliphatic index of globular proteins. 
           J Biochem. 88, 1895-1898 (1980).
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AminoAcids.R
\name{isValidAASeq}
\alias{isValidAASeq}
\title{Validate amino acid sequences}
\usage{
isValidAASeq(seq)
}
\arguments{
\item{seq}{character vector of sequences to check.}
}
\value{
A logical vector with \code{TRUE} for each valid amino acid sequences 
          and \code{FALSE} for each invalid sequence.
}
\description{
\code{isValidAASeq} checks that a set of sequences are valid non-ambiguous 
amino acid sequences. A sequence is considered valid if it contains only 
characters in the the non-ambiguous IUPAC character set or any characters in 
\code{c("X", ".", "-", "*")}.
}
\examples{
seq <- c("CARDRSTPWRRGIASTTVRTSW", "XXTQMYVR--XX", "CARJ", "10") 
isValidAASeq(seq)

}
\seealso{
See \link{ABBREV_AA} for the set of non-ambiguous amino acid characters.
See \link{IUPAC_AA} for the full set of ambiguous amino acid characters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Topology.R
\name{permuteLabels}
\alias{permuteLabels}
\title{Permute the node labels of a tree}
\usage{
permuteLabels(graph, field, exclude = c("Germline", NA))
}
\arguments{
\item{graph}{igraph object containing an annotated lineage tree.}

\item{field}{string defining the annotation field to permute.}

\item{exclude}{vector of strings defining \code{field} values to exclude 
from permutation.}
}
\value{
A modified igraph object with vertex annotations permuted.
}
\description{
\code{permuteLabels} permutes the node annotations of a lineage tree.
}
\examples{
# Define and plot example graph
library(igraph)
graph <- ExampleTrees[[23]]
plot(graph, layout=layout_as_tree, vertex.label=V(graph)$c_call, 
     vertex.size=50, edge.arrow.mode=0, vertex.color="grey80")

# Permute annotations and plot new tree
set.seed(123)
g <- permuteLabels(graph, "c_call")
plot(g, layout=layout_as_tree, vertex.label=V(g)$c_call,
     vertex.size=50, edge.arrow.mode=0, vertex.color="grey80")

}
\seealso{
\link{testEdges}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Gene.R
\name{sortGenes}
\alias{sortGenes}
\title{Sort V(D)J genes}
\usage{
sortGenes(genes, method = c("name", "position"))
}
\arguments{
\item{genes}{vector of strings representing V(D)J gene names.}

\item{method}{string defining the method to use for sorting genes. One of:
\itemize{
  \item \code{"name"}:      sort in lexicographic order. Order is by
                            family first, then gene, and then allele.
  \item \code{"position"}:  sort by position in the locus, as
                            determined by the final two numbers
                            in the gene name. Non-localized genes
                            are assigned to the highest positions.
}}
}
\value{
A sorted character vector of gene names.
}
\description{
\code{sortGenes} sorts a vector of V(D)J gene names by either lexicographic ordering
or locus position.
}
\examples{
# Create a list of allele names
genes <- c(
    "IGHV1-69D*01", "IGHV1-69*01", "IGHV4-38-2*01", "IGHV1-69-2*01",
    "IGHV2-5*01", "IGHV1-NL1*01", "IGHV1-2*01,IGHV1-2*05",
    "IGHV1-2", "IGHV1-2*02", "IGHV1-69*02"
)

# Sort genes by name
sortGenes(genes)

# Sort genes by position in the locus
sortGenes(genes, method = "pos")

}
\seealso{
See \code{getAllele}, \code{getGene} and \code{getFamily} for parsing
          gene names.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Data.R
\docType{data}
\name{ABBREV_AA}
\alias{ABBREV_AA}
\title{Amino acid abbreviation translations}
\format{
Named character vector defining single-letter character codes to 
         three-letter abbreviation mappings.
}
\usage{
ABBREV_AA
}
\description{
Mappings of amino acid abbreviations.
}
\examples{
aa <- c("Ala", "Ile", "Trp")
translateStrings(aa, ABBREV_AA)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{pairwiseEqual}
\alias{pairwiseEqual}
\title{Calculate pairwise equivalence between sequences}
\usage{
pairwiseEqual(seq)
}
\arguments{
\item{seq}{character vector containing a DNA sequences.}
}
\value{
A logical matrix of equivalence between each entry in \code{seq}. 
          Values are \code{TRUE} when sequences are equivalent and \code{FALSE}
          when they are not.
}
\description{
\code{pairwiseEqual} determined pairwise equivalence between a pairs in a 
set of sequences, excluding ambiguous positions (Ns and gaps).
}
\examples{
# Gaps and Ns will match any character
seq <- c(A="ATGGC", B="ATGGG", C="ATGGG", D="AT--C", E="NTGGG")
d <- pairwiseEqual(seq)
rownames(d) <- colnames(d) <- seq
d

}
\seealso{
Uses \link{seqEqual} for testing equivalence between pairs.
          See \link{pairwiseDist} for generating a sequence distance matrix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Alakazam.R
\name{alakazam}
\alias{alakazam}
\title{The Alakazam package}
\description{
\code{alakazam} in a member of the Immcantation framework of tools and serves five main 
purposes:
\itemize{
  \item  Providing core functionality for other R packages in Immcantation. This
         includes common tasks such as file I/O, basic DNA sequence manipulation, and
         interacting with V(D)J segment and gene annotations.
  \item  Providing an R interface for interacting with the output of the pRESTO and 
         Change-O tool suites.
  \item  Performing clonal abundance and diversity analysis on lymphocyte repertoires.
  \item  Performing lineage reconstruction on clonal populations of immunoglobulin 
         (Ig) sequences.
  \item  Performing physicochemical property analyses of lymphocyte receptor sequences.
}
For additional details regarding the use of the \code{alakazam} package see the 
vignettes:\cr
\code{browseVignettes("alakazam")}
}
\section{File I/O}{

\itemize{
  \item  \link{readChangeoDb}:        Input Change-O style files.
  \item  \link{writeChangeoDb}:       Output Change-O style files.
}
}

\section{Sequence cleaning}{

\itemize{
  \item  \link{maskSeqEnds}:          Mask ragged ends.
  \item  \link{maskSeqGaps}:          Mask gap characters.
  \item  \link{collapseDuplicates}:   Remove duplicate sequences.
}
}

\section{Lineage reconstruction}{

\itemize{
  \item  \link{makeChangeoClone}:     Clean sequences for lineage reconstruction.
  \item  \link{buildPhylipLineage}:   Perform lineage reconstruction of Ig sequences.
}
}

\section{Lineage topology analysis}{

\itemize{
  \item  \link{tableEdges}:           Tabulate annotation relationships over edges.
  \item  \link{testEdges}:            Significance testing of annotation edges.
  \item  \link{testMRCA}:             Significance testing of MRCA annotations.
  \item  \link{summarizeSubtrees}:    Various summary statistics for subtrees.
  \item  \link{plotSubtrees}:         Plot distributions of summary statistics 
                                      for a population of trees.
}
}

\section{Diversity analysis}{

\itemize{
  \item  \link{countClones}:          Calculate clonal abundance.
  \item  \link{estimateAbundance}:  	 Bootstrap clonal abundance curves.
  \item  \link{alphaDiversity}:  	 Generate clonal alpha diversity curves.
  \item  \link{plotAbundanceCurve}:   Plot clone size distribution as a rank-abundance 
  \item  \link{plotDiversityCurve}:   Plot clonal diversity curves.
  \item  \link{plotDiversityTest}:    Plot testing at given diversity hill indices. 
}
}

\section{Ig and TCR sequence annotation}{

\itemize{
  \item  \link{countGenes}:           Calculate Ig and TCR allele, gene and family usage.
  \item  \link{extractVRegion}:       Extract CDRs and FWRs sub-sequences.
  \item  \link{getAllele}:            Get V(D)J allele names.
  \item  \link{getGene}:              Get V(D)J gene names.
  \item  \link{getFamily}:            Get V(D)J family names.
  \item  \link{junctionAlignment}: Junction alignment properties 
}
}

\section{Sequence distance calculation}{

\itemize{
  \item  \link{seqDist}:        Calculate Hamming distance between two sequences.
  \item  \link{seqEqual}:       Test two sequences for equivalence.
  \item  \link{pairwiseDist}:   Calculate a matrix of pairwise Hamming distances for a 
                                set of sequences.
  \item  \link{pairwiseEqual}:  Calculate a logical matrix of pairwise equivalence for a 
                                set of sequences.
}
}

\section{Amino acid properties}{

\itemize{
  \item  \link{translateDNA}:         Translate DNA sequences to amino acid sequences.
  \item  \link{aminoAcidProperties}:  Calculate various physicochemical properties of amino acid 
                                      sequences.
  \item  \link{countPatterns}:        Count patterns in sequences.
                                             
}
}

\references{
\enumerate{
  \item  Vander Heiden JA, Yaari G, et al. pRESTO: a toolkit for processing 
           high-throughput sequencing raw reads of lymphocyte receptor repertoires. 
           Bioinformatics. 2014 30(13):1930-2.
  \item  Stern JNH, Yaari G, Vander Heiden JA, et al. B cells populating the multiple 
           sclerosis brain mature in the draining cervical lymph nodes. 
           Sci Transl Med. 2014 6(248):248ra107.
  \item  Wu Y-CB, et al. Influence of seasonal exposure to grass pollen on local and 
           peripheral blood IgE repertoires in patients with allergic rhinitis. 
           J Allergy Clin Immunol. 2014 134(3):604-12.
  \item  Gupta NT, Vander Heiden JA, et al. Change-O: a toolkit for analyzing 
           large-scale B cell immunoglobulin repertoire sequencing data.
           Bioinformatics. 2015 Oct 15;31(20):3356-8.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Lineage.R
\name{readIgphyml}
\alias{readIgphyml}
\title{Read in output from IgPhyML}
\usage{
readIgphyml(
  file,
  id = NULL,
  format = c("graph", "phylo"),
  collapse = FALSE,
  branches = c("mutations", "distance")
)
}
\arguments{
\item{file}{IgPhyML output file (.tab).}

\item{id}{ID to assign to output object.}

\item{format}{if \code{"graph"} return trees as igraph \code{graph} objects. 
if \code{"phylo"} return trees as ape \code{phylo} objects.}

\item{collapse}{if \code{TRUE} transform branch lengths to units of substitutions, 
rather than substitutions per site, and collapse internal nodes
separated by branches < 0.1 substitutions. Will also remove all
internal node labels, as it makes them inconsistent.}

\item{branches}{if \code{"distance"} branch lengths are in expected mutations per
site. If \code{"mutations"} branches are in expected mutations.}
}
\value{
A list containing IgPhyML model parameters and estimated lineage trees. 
          
          Object attributes:
          \itemize{
            \item  \code{param}:     Data.frame of parameter estimates for each clonal 
                                     lineage. Columns include: \code{CLONE}, which is the 
                                     clone id; \code{NSEQ}, the total number of sequences in 
                                     the lineage; \code{NSITE}, the number of codon sites;
                                     \code{TREE_LENGTH}, the sum of all branch lengths in 
                                     the estimated lineage tree; and \code{LHOOD}, the log 
                                     likelihood of the clone's sequences given the tree and
                                     parameters. Subsequent columns are parameter estimates 
                                     from IgPhyML, which will depend on the model used. 
                                     Parameter columns ending with \code{_MLE} are maximum 
                                     likelihood estimates; those ending with \code{_LCI} are 
                                     the lower 95%% confidence interval estimate; those ending 
                                     with \code{_UCI} are the upper 95%% confidence interval 
                                     estimate. The first line of \code{param} is for clone 
                                     \code{REPERTOIRE}, 
                                     which is a summary of all lineages within the repertoire.
                                     For this row, \code{NSEQ} is the total number of sequences, 
                                     \code{NSITE} is the average number of sites, and
                                     \code{TREE_LENGTH} is the mean tree length. For most 
                                     applications, parameter values will be the same for all 
                                     lineages within the repertoire, so access them simply by:
                                     \code{<object>$param$OMEGA_CDR_MLE[1]} to, for instance,
                                     get the estimate of dN/dS on the CDRs at the repertoire level.
            \item  \code{trees}:     List of tree objects estimated by IgPhyML. If 
                                     \code{format="graph"} these are igraph \code{graph} objects. 
                                     If \code{format="phylo"}, these are ape \code{phylo} objects.
            \item  \code{command}:   Command used to run IgPhyML.
          }
}
\description{
\code{readIgphyml} reads output from the IgPhyML phylogenetics inference package for 
B cell repertoires
}
\details{
\code{readIgphyml} reads output from the IgPhyML repertoire phylogenetics inference package. 
The resulting object is divided between parameter estimates (usually under the HLP19 model),
which provide information about mutation and selection pressure operating on the sequences.

Trees returned from this function are either igraph objects or phylo objects, and each may be 
visualized accordingly. Further, branch lengths in tree may represent either the expected number of
substitutions per site (codon, if estimated under HLP or GY94 models), or the total number of 
expected substitutions per site. If the latter, internal nodes - but not tips - separated by branch
lengths less than 0.1 are collapsed to simplify viewing.
}
\examples{
\dontrun{
   # Read in and plot a tree from an igphyml run
   library(igraph)
   s1 <- readIgphyml("IB+7d_lineages_gy.tsv_igphyml_stats_hlp.tab", id="+7d")
   print(s1$param$OMEGA_CDR_MLE[1])
   plot(s1$trees[[1]], layout=layout_as_tree, edge.label=E(s1$trees[[1]])$weight)
}

}
\references{
\enumerate{
  \item  Hoehn KB, Lunter G, Pybus OG - A Phylogenetic Codon Substitution Model for Antibody 
             Lineages. Genetics 2017 206(1):417-427
             https://doi.org/10.1534/genetics.116.196303 
 \item  Hoehn KB, Vander Heiden JA, Zhou JQ, Lunter G, Pybus OG, Kleinstein SHK - 
             Repertoire-wide phylogenetic models of B cell molecular evolution reveal 
             evolutionary signatures of aging and vaccination. bioRxiv 2019  
             https://doi.org/10.1101/558825 
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Diversity.R
\name{plotAbundanceCurve}
\alias{plotAbundanceCurve}
\title{Plot a clonal abundance distribution}
\usage{
plotAbundanceCurve(
  data,
  colors = NULL,
  main_title = "Rank Abundance",
  legend_title = NULL,
  xlim = NULL,
  ylim = NULL,
  annotate = c("none", "depth"),
  silent = FALSE,
  ...
)
}
\arguments{
\item{data}{\link{AbundanceCurve} object returned by \link{estimateAbundance}.}

\item{colors}{named character vector whose names are values in the 
\code{group} column of \code{data} and whose values are 
colors to assign to those group names.}

\item{main_title}{string specifying the plot title.}

\item{legend_title}{string specifying the legend title.}

\item{xlim}{numeric vector of two values specifying the 
\code{c(lower, upper)} x-axis limits. The lower x-axis 
value must be >=1.}

\item{ylim}{numeric vector of two values specifying the 
\code{c(lower, upper)} y-axis limits. The limits on the 
abundance values are expressed as fractions of 1: use
c(0,1) to set the lower and upper limits to 0\% and 100\%.}

\item{annotate}{string defining whether to added values to the group labels 
of the legend. When \code{"none"} (default) is specified no
annotations are added. Specifying (\code{"depth"}) adds 
sequence counts to the labels.}

\item{silent}{if \code{TRUE} do not draw the plot and just return the ggplot2 
object; if \code{FALSE} draw the plot.}

\item{...}{additional arguments to pass to ggplot2::theme.}
}
\value{
A \code{ggplot} object defining the plot.
}
\description{
\code{plotAbundanceCurve} plots the results from estimating the complete clonal 
relative abundance distribution. The distribution is plotted as a log rank abundance 
distribution.
}
\examples{
# Estimate abundance by sample and plot
abund <- estimateAbundance(ExampleDb, group="sample_id", nboot=100)
plotAbundanceCurve(abund, legend_title="Sample")

}
\seealso{
See \link{AbundanceCurve} for the input object and \link{estimateAbundance} for
generating the input abundance distribution. Plotting is performed with \link[ggplot2]{ggplot}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Gene.R
\name{countGenes}
\alias{countGenes}
\title{Tabulates V(D)J allele, gene or family usage within each locus.}
\usage{
countGenes(
  data,
  gene,
  groups = NULL,
  copy = NULL,
  clone = NULL,
  fill = FALSE,
  first = TRUE,
  collapse = TRUE,
  mode = c("gene", "allele", "family", "asis"),
  cell_id = "cell_id",
  remove_na = TRUE
)
}
\arguments{
\item{data}{data.frame with AIRR-format or Change-O style columns.}

\item{gene}{column containing allele assignments. Only the first allele in the
column will be considered when \code{mode} is "gene", "family" or
"allele". The value will be used as it is with \code{mode="asis"}.}

\item{groups}{columns containing grouping variables. If \code{NULL} do not group.}

\item{copy}{name of the \code{data} column containing copy numbers for each
sequence. If this value is specified, then total copy abundance
is determined by the sum of copy numbers within each gene.
This argument is ignored if \code{clone} is specified.}

\item{clone}{name of the \code{data} column containing clone identifiers for each
sequence. If this value is specified, then one gene will be considered
for each clone. Note, this is accomplished by using the most
common gene within each \code{clone} identifier. As such,
ambiguous alleles within a clone will not be accurately represented.}

\item{fill}{logical of \code{c(TRUE, FALSE)} specifying when if groups (when specified)
lacking a particular gene should be counted as 0 if TRUE or not (omitted).}

\item{first}{if TRUE return only the first allele/gene/family call for computing the frequency; if FALSE return all calls delimited by commas.}

\item{collapse}{if TRUE check for duplicates and return only unique allele/gene/family assignments per sequence; if
FALSE return all assignments (faster). Has no effect if first=TRUE.}

\item{mode}{one of \code{c("gene", "family", "allele", "asis")} defining
the degree of specificity regarding allele calls. Determines whether
to return counts for genes (calling \code{getGene}),
families (calling \code{getFamily}), alleles (calling
\code{getAllele}) or using the value as it is in the column
\code{gene}, without any processing.}

\item{cell_id}{name of the \code{data} column containing the cell identifiers for each sequence.}

\item{remove_na}{removes rows with \code{NA} values in the gene column if \code{TRUE} and issues a warning.
Otherwise, keeps those rows and considers \code{NA} as a gene in the final counts
and relative abundances.}
}
\value{
A data.frame summarizing family, gene or allele counts and frequencies
          with columns:
          \itemize{
            \item \code{locus}:        locus of the gene (IGH, IGK, IGL, TRA, TRB, TRD, TRG). Note that frequencies are calculated within each locus.
            \item \code{gene}:         name of the family, gene or allele.
            \item \code{seq_count}:    total number of sequences for the gene in the locus.
            \item \code{locus_count}:  total number of sequences in the locus.
            \item \code{seq_freq}:     frequency of the gene as a fraction of the total
                                       number of sequences within each grouping.
            \item \code{copy_count}:   sum of the copy counts in the \code{copy} column.
                                       for each gene. Only present if the \code{copy}
                                       argument is specified.
            \item \code{locus_copy_count}: sum of the copy counts in the \code{copy} column.
                                       for all gene in the locus. Only present if the 
                                       \code{copy} argument is specified.
            \item \code{copy_freq}:    frequency of the gene as a fraction of the total
                                       copy number within each group. Only present if
                                       the \code{copy} argument is specified.
            \item \code{clone_count}:  total number of clones for the gene. Only present if
                                       the \code{clone} argument is specified.
            \item \code{clone_freq}:   frequency of the gene as a fraction of the total
                                       number of clones within each grouping. Only present if
                                       the \code{clone} argument is specified.
          }
          Additional columns defined by the \code{groups} argument will also be present.
}
\description{
Determines the count and relative abundance of V(D)J alleles, genes or families within
groups. If sequences from multiple loci are present, the frequency is calculated within 
each locus.
}
\examples{
# Without copy numbers
genes <- countGenes(ExampleDb, gene = "v_call", groups = "sample_id", mode = "family")
genes <- countGenes(ExampleDb, gene = "v_call", groups = "sample_id", mode = "gene")
genes <- countGenes(ExampleDb, gene = "v_call", groups = "sample_id", mode = "allele")

# With copy numbers and multiple groups
genes <- countGenes(ExampleDb,
    gene = "v_call", groups = c("sample_id", "c_call"),
    copy = "duplicate_count", mode = "family"
)

# Count by clone
genes <- countGenes(ExampleDb,
    gene = "v_call", groups = c("sample_id", "c_call"),
    clone = "clone_id", mode = "family"
)

# Count absent genes
genes <- countGenes(ExampleDb,
    gene = "v_call", groups = "sample_id",
    mode = "allele", fill = TRUE
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Topology.R
\name{getPathLengths}
\alias{getPathLengths}
\title{Calculate path lengths from the tree root}
\usage{
getPathLengths(graph, root = "Germline", field = NULL, exclude = NULL)
}
\arguments{
\item{graph}{igraph object containing an annotated lineage tree.}

\item{root}{name of the root (germline) node.}

\item{field}{annotation field to use for exclusion of nodes from step count.}

\item{exclude}{annotation values specifying which nodes to exclude from step count. 
If \code{NULL} consider all nodes. This does not affect the weighted
(distance) path length calculation.}
}
\value{
A data.frame with columns:
          \itemize{
            \item  \code{name}:      node name
            \item  \code{steps}:     path length as the number of nodes traversed
            \item  \code{distance}:  path length as the sum of edge weights
          }
}
\description{
\code{getPathLengths} calculates the unweighted (number of steps) and weighted (distance) 
path lengths from the root of a lineage tree.
}
\examples{
# Define example graph
graph <- ExampleTrees[[24]]

# Consider all nodes
getPathLengths(graph, root="Germline")

# Exclude nodes without an isotype annotation from step count
getPathLengths(graph, root="Germline", field="c_call", exclude=NA)

}
\seealso{
See \link{buildPhylipLineage} for generating input trees.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Topology.R
\name{plotMRCATest}
\alias{plotMRCATest}
\title{Plot the results of a founder permutation test}
\usage{
plotMRCATest(
  data,
  color = "black",
  main_title = "MRCA Test",
  style = c("histogram", "cdf"),
  silent = FALSE,
  ...
)
}
\arguments{
\item{data}{\link{MRCATest} object returned by \link{testMRCA}.}

\item{color}{color of the histogram or lines.}

\item{main_title}{string specifying the plot title.}

\item{style}{type of plot to draw. One of:
\itemize{
  \item \code{"histogram"}:  histogram of the annotation count 
                             distribution with a red dotted line
                             denoting the observed value.
  \item \code{"cdf"}:        cumulative distribution function 
                             of annotation counts with a red dotted 
                             line denoting the observed value and
                             a blue dotted line indicating the 
                             p-value.
}}

\item{silent}{if \code{TRUE} do not draw the plot and just return the ggplot2 
object; if \code{FALSE} draw the plot.}

\item{...}{additional arguments to pass to ggplot2::theme.}
}
\value{
A \code{ggplot} object defining the plot.
}
\description{
\code{plotMRCATest} plots the results of a founder permutation test performed with 
\code{testMRCA}.
}
\examples{
\donttest{
# Define example tree set
graphs <- ExampleTrees[1:10]

# Perform MRCA test on isotypes
set.seed(123)
x <- testMRCA(graphs, "c_call", nperm=10)

# Plot
plotMRCATest(x, color="steelblue", style="hist")
plotMRCATest(x, style="cdf")
}

}
\seealso{
See \link{testEdges} for performing the test.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Data.R
\docType{data}
\name{Example10x}
\alias{Example10x}
\title{Small example 10x Genomics Ig V(D)J sequences from CD19+ B cells isolated from PBMCs of a healthy 
human donor. Down-sampled from data provided by 10x Genomics under a Creative Commons Attribute license,
and processed with their Cell Ranger pipeline.}
\format{
A data.frame with the following AIRR style columns:
  \itemize{
    \item  \code{sequence_id}:                Sequence identifier
    \item  \code{sequence_alignment}:         IMGT-gapped observed sequence.
    \item  \code{germline_alignment}:         IMGT-gapped germline sequence.
    \item  \code{v_call}:                     V region allele assignments.
    \item  \code{d_call}:                     D region allele assignments.
    \item  \code{j_call}:                     J region allele assignments.
    \item  \code{c_call}:                     Isotype (C region) assignment.
    \item  \code{junction}:                   Junction region sequence.
    \item  \code{junction_length}:            Length of the junction region in nucleotides.
    \item  \code{np1_length}:                 Combined length of the N and P regions proximal
                                              to the V region.
    \item  \code{np2_length}:                 Combined length of the N and P regions proximal
                                              to the J region.
    \item  \code{umi_count}:                  Number of unique molecular identifies atttributed to sequence.
    \item  \code{cell_id}:                    Cell identifier.
    \item  \code{locus}:                      Genomic locus of sequence.
}
}
\usage{
Example10x
}
\description{
Small example 10x Genomics Ig V(D)J sequences from CD19+ B cells isolated from PBMCs of a healthy 
human donor. Down-sampled from data provided by 10x Genomics under a Creative Commons Attribute license,
and processed with their Cell Ranger pipeline.
}
\references{
\enumerate{
  \item  Data source: https://support.10xgenomics.com/single-cell-vdj/datasets/2.2.0/vdj_v1_hs_cd19_b
  \item  License: https://creativecommons.org/licenses/by/4.0/
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Classes.R
\docType{class}
\name{ChangeoClone-class}
\alias{ChangeoClone-class}
\alias{ChangeoClone}
\title{S4 class defining a clone}
\description{
\code{ChangeoClone} defines a common data structure for perform lineage reconstruction
from Change-O data.
}
\section{Slots}{

\describe{
\item{\code{data}}{data.frame containing sequences and annotations. Contains the
columns \code{SEQUENCE_ID} and \code{SEQUENCE}, as well as any additional 
sequence-specific annotation columns.}

\item{\code{clone}}{string defining the clone identifier.}

\item{\code{germline}}{string containing the germline sequence for the clone.}

\item{\code{v_gene}}{string defining the V segment gene call.}

\item{\code{j_gene}}{string defining the J segment gene call.}

\item{\code{junc_len}}{numeric junction length (nucleotide count).}
}}

\seealso{
See \link{makeChangeoClone} and \link{buildPhylipLineage} for use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Topology.R
\name{getMRCA}
\alias{getMRCA}
\title{Retrieve the first non-root node of a lineage tree}
\usage{
getMRCA(
  graph,
  path = c("distance", "steps"),
  root = "Germline",
  field = NULL,
  exclude = NULL
)
}
\arguments{
\item{graph}{igraph object containing an annotated lineage tree.}

\item{path}{string defining whether to use unweighted (steps) or weighted (distance) 
measures for determining the founder node set..}

\item{root}{name of the root (germline) node.}

\item{field}{annotation field to use for both unweighted path length exclusion and
consideration as an MRCA node. If \code{NULL} do not exclude any nodes.}

\item{exclude}{vector of annotation values in \code{field} to exclude from the potential 
MRCA set. If \code{NULL} do not exclude any nodes. Has no effect if 
\code{field=NULL}.}
}
\value{
A data.frame of the MRCA node(s) containing the columns:
          \itemize{
            \item  \code{name}:      node name
            \item  \code{steps}:     path length as the number of nodes traversed
            \item  \code{distance}:  path length as the sum of edge weights
          }
          Along with additional columns corresponding to the 
          annotations of the input graph.
}
\description{
\code{getMRCA} returns the set of lineage tree nodes with the minimum weighted or 
unweighted path length from the root (germline) of the lineage tree, allowing for 
exclusion of specific groups of nodes.
}
\examples{
# Define example graph
graph <- ExampleTrees[[23]]

# Use unweighted path length and do not exclude any nodes
getMRCA(graph, path="steps", root="Germline")

# Exclude nodes without an isotype annotation and use weighted path length
getMRCA(graph, path="distance", root="Germline", field="c_call", exclude=NA)

}
\seealso{
Path lengths are determined with \link{getPathLengths}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Core.R
\name{checkColumns}
\alias{checkColumns}
\title{Check data.frame for valid columns and issue message if invalid}
\usage{
checkColumns(data, columns, logic = c("all", "any"))
}
\arguments{
\item{data}{data.frame to check.}

\item{columns}{vector of column names to check.}

\item{logic}{one of \code{"all"} or \code{"any"} controlling whether all,
or at least one, of the columns must be valid, respectively.}
}
\value{
\code{TRUE} if columns are valid and a string message if not.
}
\description{
Check data.frame for valid columns and issue message if invalid
}
\examples{
df <- data.frame(A=1:3, B=4:6, C=rep(NA, 3))
checkColumns(df, c("A", "B"), logic="all")
checkColumns(df, c("A", "B"), logic="any")
checkColumns(df, c("A", "C"), logic="all")
checkColumns(df, c("A", "C"), logic="any")
checkColumns(df, c("A", "D"), logic="all")
checkColumns(df, c("A", "D"), logic="any")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Core.R
\name{gridPlot}
\alias{gridPlot}
\title{Plot multiple ggplot objects}
\usage{
gridPlot(..., ncol = 1)
}
\arguments{
\item{...}{ggplot objects to plot.}

\item{ncol}{number of columns in the plot.}
}
\description{
Plots multiple ggplot objects in an equally sized grid.
}
\references{
Modified from:
http://www.cookbook-r.com/Graphs/Multiple_graphs_on_one_page_(ggplot2)
}
\seealso{
\link[ggplot2]{ggplot}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Sequence.R
\name{collapseDuplicates}
\alias{collapseDuplicates}
\title{Remove duplicate DNA sequences and combine annotations}
\usage{
collapseDuplicates(
  data,
  id = "sequence_id",
  seq = "sequence_alignment",
  text_fields = NULL,
  num_fields = NULL,
  seq_fields = NULL,
  add_count = FALSE,
  ignore = c("N", "-", ".", "?"),
  sep = ",",
  dry = FALSE,
  verbose = FALSE
)
}
\arguments{
\item{data}{data.frame containing Change-O columns. The data.frame 
must contain, at a minimum, a unique identifier column 
and a column containing a character vector of DNA sequences.}

\item{id}{name of the column containing sequence identifiers.}

\item{seq}{name of the column containing DNA sequences.}

\item{text_fields}{character vector of textual columns to collapse. The textual 
annotations of duplicate sequences will be merged into a single 
string with each unique value alphabetized and delimited by 
\code{sep}.}

\item{num_fields}{vector of numeric columns to collapse. The numeric annotations
of duplicate sequences will be summed.}

\item{seq_fields}{vector of nucleotide sequence columns to collapse. The sequence 
with the fewest number of non-informative characters will be 
retained. Where a non-informative character is one of 
\code{c("N", "-", ".", "?")}. Note, this is distinct from the 
\code{seq} parameter which is used to determine duplicates.}

\item{add_count}{if \code{TRUE} add the column \code{collpase_count} that 
indicates the number of sequences that were collapsed to build 
each unique entry.}

\item{ignore}{vector of characters to ignore when testing for equality.}

\item{sep}{character to use for delimiting collapsed annotations in the 
\code{text_fields} columns. Defines both the input and output 
delimiter.}

\item{dry}{if \code{TRUE} perform dry run. Only labels the sequences without 
collapsing them.}

\item{verbose}{if \code{TRUE} report the number input, discarded and output 
sequences; if \code{FALSE} process sequences silently.}
}
\value{
A modified \code{data} data.frame with duplicate sequences removed and 
          annotation fields collapsed if \code{dry=FALSE}. If \code{dry=TRUE}, 
          sequences will be labeled with the collapse action, but the input will be
          otherwise unmodified (see Details).
}
\description{
\code{collapseDuplicates} identifies duplicate DNA sequences, allowing for ambiguous 
characters, removes the duplicate entries, and combines any associated annotations.
}
\details{
\code{collapseDuplicates} identifies duplicate sequences in the \code{seq} column by
testing for character identity, with consideration of IUPAC ambiguous nucleotide codes. 
A cluster of sequences are considered duplicates if they are all equivalent, and no 
member of the cluster is equivalent to a sequence in a different cluster. 

Textual annotations, specified by \code{text_fields}, are collapsed by taking the unique
set of values within in each duplicate cluster and delimiting those values by \code{sep}.
Numeric annotations, specified by \code{num_fields}, are collapsed by summing all values 
in the duplicate cluster. Sequence annotations, specified by \code{seq_fields}, are 
collapsed by retaining the first sequence with the fewest number of N characters.

Columns that are not specified in either \code{text_fields}, \code{num_fields}, or 
\code{seq_fields} will be retained, but the value will be chosen from a random entry 
amongst all sequences in a cluster of duplicates.

An ambiguous sequence is one that can be assigned to two different clusters, wherein
the ambiguous sequence is equivalent to two sequences which are themselves 
non-equivalent. Ambiguous sequences arise due to ambiguous characters at positions that
vary across sequences, and are discarded along with their annotations when \code{dry=FALSE}. 
Thus, ambiguous sequences are removed as duplicates of some sequence, but do not create a potential
false-positive annotation merger. Ambiguous sequences are not included in the 
\code{collapse_count} annotation that is added when \code{add_count=TRUE}.

If \code{dry=TRUE} sequences will not be removed from the input. Instead, the following columns
will be appended to the input defining the collapse action that would have been performed in the
\code{dry=FALSE} case.

\itemize{
  \item  \code{collapse_id}:     an identifier for the group of identical sequences.
  \item  \code{collapse_class}:  string defining how the sequence matches to the other in the set.
                                 one of \code{"duplicated"} (has duplicates),
                                 \code{"unique"} (no duplicates), \code{"ambiguous_duplicate"} 
                                 (no duplicates after ambiguous sequences are removed), 
                                 or \code{"ambiguous"} (matches multiple non-duplicate sequences).
  \item  \code{collapse_pass}:   \code{TRUE} for the sequences that would be retained.
}
}
\examples{
# Example data.frame
db <- data.frame(sequence_id=LETTERS[1:4],
                 sequence_alignment=c("CCCCTGGG", "CCCCTGGN", "NAACTGGN", "NNNCTGNN"),
                 c_call=c("IGHM", "IGHG", "IGHG", "IGHA"),
                 sample_id=c("S1", "S1", "S2", "S2"),
                 duplicate_count=1:4,
                 stringsAsFactors=FALSE)

# Annotations are not parsed if neither text_fields nor num_fields is specified
# The retained sequence annotations will be random
collapseDuplicates(db, verbose=TRUE)

# Unique text_fields annotations are combined into a single string with ","
# num_fields annotations are summed
# Ambiguous duplicates are discarded
collapseDuplicates(db, text_fields=c("c_call", "sample_id"), num_fields="duplicate_count", 
                   verbose=TRUE)

# Use alternate delimiter for collapsing textual annotations
collapseDuplicates(db, text_fields=c("c_call", "sample_id"), num_fields="duplicate_count", 
                   sep="/", verbose=TRUE)

# Add count of duplicates
collapseDuplicates(db, text_fields=c("c_call", "sample_id"), num_fields="duplicate_count", 
                   add_count=TRUE, verbose=TRUE)

# Masking ragged ends may impact duplicate removal
db$sequence_alignment <- maskSeqEnds(db$sequence_alignment)
collapseDuplicates(db, text_fields=c("c_call", "sample_id"), num_fields="duplicate_count", 
                   add_count=TRUE, verbose=TRUE)

}
\seealso{
Equality is tested with \link{seqEqual} and \link{pairwiseEqual}. 
          For IUPAC ambiguous character codes see \link{IUPAC_DNA}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Diversity.R
\name{plotDiversityTest}
\alias{plotDiversityTest}
\title{Plot the results of diversity testing}
\usage{
plotDiversityTest(
  data,
  q,
  colors = NULL,
  main_title = "Diversity",
  legend_title = "Group",
  log_d = FALSE,
  annotate = c("none", "depth"),
  silent = FALSE,
  ...
)
}
\arguments{
\item{data}{\link{DiversityCurve} object returned by 
\link{alphaDiversity}.}

\item{q}{diversity order to plot the test for.}

\item{colors}{named character vector whose names are values in the 
\code{group} column of the \code{data} slot of \code{data},
and whose values are colors to assign to those group names.}

\item{main_title}{string specifying the plot title.}

\item{legend_title}{string specifying the legend title.}

\item{log_d}{if \code{TRUE} then plot the diversity scores \eqn{D} 
on a log scale; if \code{FALSE} plot on a linear scale.}

\item{annotate}{string defining whether to added values to the group labels 
of the legend. When \code{"none"} (default) is specified no
annotations are added. Specifying (\code{"depth"}) adds 
sequence counts to the labels.}

\item{silent}{if \code{TRUE} do not draw the plot and just return the ggplot2 
object; if \code{FALSE} draw the plot.}

\item{...}{additional arguments to pass to ggplot2::theme.}
}
\value{
A \code{ggplot} object defining the plot.
}
\description{
\code{plotDiversityTest} plots summary data for a \code{DiversityCurve} object 
with mean and a line range indicating plus/minus one standard deviation.
}
\examples{
# Calculate diversity
div <- alphaDiversity(ExampleDb, group="sample_id", min_q=0, max_q=2, step_q=1, nboot=100)

# Plot results at q=0 (equivalent to species richness)
plotDiversityTest(div, 0, legend_title="Sample")

# Plot results at q=2 (equivalent to Simpson's index)
plotDiversityTest(div, q=2, legend_title="Sample")

}
\seealso{
See \link{alphaDiversity} for generating input.
          Plotting is performed with \link[ggplot2]{ggplot}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Diversity.R
\name{calcDiversity}
\alias{calcDiversity}
\title{Calculate the diversity index}
\usage{
calcDiversity(p, q)
}
\arguments{
\item{p}{numeric vector of clone (species) counts or proportions.}

\item{q}{numeric vector of diversity orders.}
}
\value{
A vector of diversity scores \eqn{D} for each \eqn{q}.
}
\description{
\code{calcDiversity} calculates the clonal diversity index for a vector of diversity 
orders.
}
\details{
This method, proposed by Hill (Hill, 1973), quantifies diversity as a smooth function 
(\eqn{D}) of a single parameter \eqn{q}. Special cases of the generalized diversity 
index correspond to the most popular diversity measures in ecology: species richness 
(\eqn{q = 0}), the exponential of the Shannon-Weiner index (\eqn{q} approaches \eqn{1}), the 
inverse of the Simpson index (\eqn{q = 2}), and the reciprocal abundance of the largest 
clone (\eqn{q} approaches \eqn{+\infty}). At \eqn{q = 0} different clones weight equally, 
regardless of their size. As the parameter \eqn{q} increase from \eqn{0} to \eqn{+\infty} 
the diversity index (\eqn{D}) depends less on rare clones and more on common (abundant) 
ones, thus encompassing a range of definitions that can be visualized as a single curve. 

Values of \eqn{q < 0} are valid, but are generally not meaningful. The value of \eqn{D} 
at \eqn{q=1} is estimated by \eqn{D} at \eqn{q=0.9999}.
}
\examples{
# May define p as clonal member counts
p <- c(1, 1, 3, 10)
q <- c(0, 1, 2)
calcDiversity(p, q)

# Or proportional abundance
p <- c(1/15, 1/15, 1/5, 2/3)
calcDiversity(p, q)

}
\references{
\enumerate{
  \item  Hill M. Diversity and evenness: a unifying notation and its consequences. 
           Ecology. 1973 54(2):427-32.
}
}
\seealso{
Used by \link{alphaDiversity}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AminoAcids.R
\name{polar}
\alias{polar}
\title{Calculates the average polarity of amino acid sequences}
\usage{
polar(seq, polarity = NULL)
}
\arguments{
\item{seq}{vector of strings containing amino acid sequences.}

\item{polarity}{named numerical vector defining polarity scores for 
each amino acid, where names are single-letter amino acid 
character codes. If \code{NULL}, then the Grantham, 1974
scale is used.}
}
\value{
A vector of bulkiness scores for the sequence(s).
}
\description{
\code{polar} calculates the average polarity score of amino acid sequences. 
Non-informative positions are excluded, where non-informative is defined as any 
character in \code{c("X", "-", ".", "*")}.
}
\examples{
# Default scale
seq <- c("CARDRSTPWRRGIASTTVRTSW", "XXTQMYVRT")
polar(seq)

# Use the Zimmerman et al, 1968 polarity scale from the seqinr package
library(seqinr)
data(aaindex)
x <- aaindex[["ZIMJ680103"]]$I
# Rename the score vector to use single-letter codes
names(x) <- translateStrings(names(x), ABBREV_AA)
# Calculate polarity
polar(seq, polarity=x)

}
\references{
\enumerate{
  \item  Grantham R. Amino acid difference formula to help explain protein evolution. 
           Science 185, 862-864 (1974).
}
}
\seealso{
For additional size related indices see \code{\link[seqinr]{aaindex}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{seqEqual}
\alias{seqEqual}
\title{Test DNA sequences for equality.}
\usage{
seqEqual(seq1, seq2, ignore = as.character(c("N", "-", ".", "?")))
}
\arguments{
\item{seq1}{character string containing a DNA sequence.}

\item{seq2}{character string containing a DNA sequence.}

\item{ignore}{vector of characters to ignore when testing for equality.
Default is to ignore c("N",".","-","?")}
}
\value{
Returns \code{TRUE} if sequences are equal and \code{FALSE} if they are not.
          Sequences of unequal length will always return \code{FALSE} regardless of
          their character values.
}
\description{
\code{seqEqual} checks if two DNA sequences are identical.
}
\examples{
# Ignore gaps
seqEqual("ATG-C", "AT--C")
seqEqual("ATGGC", "ATGGN")
seqEqual("AT--T", "ATGGC")

# Ignore only Ns
seqEqual("ATG-C", "AT--C", ignore="N")
seqEqual("ATGGC", "ATGGN", ignore="N")
seqEqual("AT--T", "ATGGC", ignore="N")

}
\seealso{
Used by \link{pairwiseEqual} within \link{collapseDuplicates}.
          See \link{seqDist} for calculation Hamming distances between sequences.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AminoAcids.R
\name{bulk}
\alias{bulk}
\title{Calculates the average bulkiness of amino acid sequences}
\usage{
bulk(seq, bulkiness = NULL)
}
\arguments{
\item{seq}{vector of strings containing amino acid sequences.}

\item{bulkiness}{named numerical vector defining bulkiness scores for 
each amino acid, where names are single-letter amino acid 
character codes. If \code{NULL}, then the Zimmerman et al, 1968
scale is used.}
}
\value{
A vector of bulkiness scores for the sequence(s).
}
\description{
\code{bulk} calculates the average bulkiness score of amino acid sequences. 
Non-informative positions are excluded, where non-informative is defined as any 
character in \code{c("X", "-", ".", "*")}.
}
\examples{
# Default bulkiness scale
seq <- c("CARDRSTPWRRGIASTTVRTSW", "XXTQMYVRT")
bulk(seq)

# Use the Grantham, 1974 side chain volumn scores from the seqinr package
library(seqinr)
data(aaindex)
x <- aaindex[["GRAR740103"]]$I
# Rename the score vector to use single-letter codes
names(x) <- translateStrings(names(x), ABBREV_AA)
# Calculate average volume
bulk(seq, bulkiness=x)

}
\references{
\enumerate{
  \item  Zimmerman JM, Eliezer N, Simha R. The characterization of amino acid sequences 
           in proteins by statistical methods. J Theor Biol 21, 170-201 (1968).
}
}
\seealso{
For additional size related indices see \link[seqinr]{aaindex}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Diversity.R
\name{calcCoverage}
\alias{calcCoverage}
\title{Calculate sample coverage}
\usage{
calcCoverage(x, r = 1)
}
\arguments{
\item{x}{numeric vector of abundance counts.}

\item{r}{coverage order to calculate.}
}
\value{
The sample coverage of the given order \code{r}.
}
\description{
\code{calcCoverage} calculates the sample coverage estimate, a measure of sample 
completeness, for varying orders using the method of Chao et al, 2015, falling back 
to the Chao1 method in the first order case.
}
\examples{
# Calculate clone sizes
clones <- countClones(ExampleDb, groups="sample_id")

# Calculate 1first order coverage for a single sample
calcCoverage(clones$seq_count[clones$sample_id == "+7d"])

}
\references{
\enumerate{
  \item  Chao A. Nonparametric Estimation of the Number of Classes in a Population. 
           Scand J Stat. 1984 11, 265270.
  \item  Chao A, et al. Unveiling the species-rank abundance distribution by 
           generalizing the Good-Turing sample coverage theory. 
           Ecology. 2015 96, 11891201.
}
}
\seealso{
Used by \link{alphaDiversity}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Lineage.R
\name{graphToPhylo}
\alias{graphToPhylo}
\title{Convert a tree in igraph \code{graph} format to ape \code{phylo} format.}
\usage{
graphToPhylo(graph)
}
\arguments{
\item{graph}{An igraph \code{graph} object.}
}
\value{
A \code{phylo} object representing the input tree. Tip and internal node names are 
          stored in the \code{tip.label} and \code{node.label} vectors, respectively.
}
\description{
\code{graphToPhylo} a tree in igraph \code{graph} format to ape \code{phylo} 
format.
}
\details{
Convert from igraph \code{graph} object to ape \code{phylo} object. If \code{graph} object
was previously rooted with the germline as the direct ancestor, this will re-attach the 
germline as a descendant node with a zero branch length to a new universal common ancestor (UCA) 
node and store the germline node ID in the \code{germid} attribute and UCA node number in 
the \code{uca} attribute. Otherwise these attributes will not be specified in the \code{phylo} object. 
Using \code{phyloToGraph(phylo, germline=phylo$germid)} creates a \code{graph} object with the germline 
back as the direct ancestor. Tip and internal node names are 
stored in the \code{tip.label} and \code{node.label} vectors, respectively.
}
\examples{
\dontrun{
   library(igraph)
   library(ape)

   #convert to phylo
   phylo = graphToPhylo(graph)
   
   #plot tree using ape
   plot(phylo,show.node.label=TRUE)
   
   #store as newick tree
   write.tree(phylo,file="tree.newick")
   
   #read in tree from newick file
   phylo_r = read.tree("tree.newick")
   
   #convert to igraph
   graph_r = phyloToGraph(phylo_r,germline="Germline")
   
   #plot graph - same as before, possibly rotated
   plot(graph_r,layout=layout_as_tree)
}

}
\references{
\enumerate{
  \item  Hoehn KB, Lunter G, Pybus OG - A Phylogenetic Codon Substitution Model for Antibody 
             Lineages. Genetics 2017 206(1):417-427
             https://doi.org/10.1534/genetics.116.196303 
 \item  Hoehn KB, Vander Heiden JA, Zhou JQ, Lunter G, Pybus OG, Kleinstein SHK - 
             Repertoire-wide phylogenetic models of B cell molecular evolution reveal 
             evolutionary signatures of aging and vaccination. bioRxiv 2019  
             https://doi.org/10.1101/558825 
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Fastq.R
\name{getPositionQuality}
\alias{getPositionQuality}
\title{Get a data.frame with sequencing qualities per position}
\usage{
getPositionQuality(
  data,
  sequence_id = "sequence_id",
  sequence = "sequence_alignment",
  quality_num = "quality_alignment_num"
)
}
\arguments{
\item{data}{\code{data.frame} containing sequence data.}

\item{sequence_id}{column in \code{data} with sequence identifiers.}

\item{sequence}{column in \code{data} with sequence data.}

\item{quality_num}{column in \code{data} with quality scores (as
strings of numeric values, comma separated) for \code{sequence}.}
}
\value{
\code{data} with one additional field with masked sequences. The 
          name of this field is created concatenating \code{sequence} 
          and '_masked'.
}
\description{
\code{getPositionQuality} takes a data.frame with sequence quality scores 
in the form of a strings of comma separated numeric values, split the quality 
scores values by \code{","},  and returns a data.frame with the values
for each position.
}
\examples{
db <- airr::read_rearrangement(system.file("extdata", "example_quality.tsv", package="alakazam"))
fastq_file <- system.file("extdata", "example_quality.fastq", package="alakazam")
db <- readFastqDb(db, fastq_file, quality_offset=-33)
head(getPositionQuality(db))
}
\seealso{
\link{readFastqDb} and \link{maskPositionsByQuality}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Core.R
\name{writeChangeoDb}
\alias{writeChangeoDb}
\title{Write a Change-O tab-delimited database file}
\usage{
writeChangeoDb(data, file)
}
\arguments{
\item{data}{data.frame of Change-O data.}

\item{file}{output file name.}
}
\description{
\code{writeChangeoDb} is a simple wrapper around \link[readr]{write_delim} with defaults 
appropriate for writing a Change-O tab-delimited database file from a data.frame.
}
\examples{
\dontrun{
  # Write a database
  writeChangeoDb(data, "changeo.tsv")
}

}
\seealso{
Wraps \link[readr]{write_delim}. See \link{readChangeoDb} for reading to Change-O files.
          See \link[airr:read_tabular]{read_rearrangement} and \link[airr:write_tabular]{write_rearrangement}
          to read and write AIRR-C Standard formatted repertoires.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Gene.R
\name{getSegment}
\alias{getSegment}
\alias{getAllele}
\alias{getGene}
\alias{getFamily}
\alias{getLocus}
\alias{getChain}
\title{Get Ig segment allele, gene and family names}
\usage{
getSegment(
  segment_call,
  segment_regex,
  first = TRUE,
  collapse = TRUE,
  strip_d = TRUE,
  omit_nl = FALSE,
  sep = ","
)

getAllele(
  segment_call,
  first = TRUE,
  collapse = TRUE,
  strip_d = TRUE,
  omit_nl = FALSE,
  sep = ","
)

getGene(
  segment_call,
  first = TRUE,
  collapse = TRUE,
  strip_d = TRUE,
  omit_nl = FALSE,
  sep = ","
)

getFamily(
  segment_call,
  first = TRUE,
  collapse = TRUE,
  strip_d = TRUE,
  omit_nl = FALSE,
  sep = ","
)

getLocus(
  segment_call,
  first = TRUE,
  collapse = TRUE,
  strip_d = TRUE,
  omit_nl = FALSE,
  sep = ","
)

getChain(
  segment_call,
  first = TRUE,
  collapse = TRUE,
  strip_d = TRUE,
  omit_nl = FALSE,
  sep = ","
)
}
\arguments{
\item{segment_call}{character vector containing segment calls delimited by commas.}

\item{segment_regex}{string defining the segment match regular expression.}

\item{first}{if \code{TRUE} return only the first call in
\code{segment_call}; if \code{FALSE} return all calls
delimited by commas.}

\item{collapse}{if \code{TRUE} check for duplicates and return only unique
segment assignments; if \code{FALSE} return all assignments
(faster). Has no effect if \code{first=TRUE}.}

\item{strip_d}{if \code{TRUE} remove the "D" from the end of gene annotations
(denoting a duplicate gene in the locus);
if \code{FALSE} do not alter gene names.}

\item{omit_nl}{if \code{TRUE} remove non-localized (NL) genes from the result.
Only applies at the gene or allele level.}

\item{sep}{character defining both the input and output segment call
delimiter.}
}
\value{
A character vector containing allele, gene or family names.
}
\description{
\code{getSegment} performs generic matching of delimited segment calls with a custom
regular expression. \link{getAllele}, \link{getGene} and \link{getFamily} extract
the allele, gene and family names, respectively, from a character vector of
immunoglobulin (Ig) or TCR segment allele calls in IMGT format.
}
\examples{
# Light chain examples
kappa_call <- c(
    "Homsap IGKV1D-39*01 F,Homsap IGKV1-39*02 F,Homsap IGKV1-39*01",
    "Homsap IGKJ5*01 F"
)

getAllele(kappa_call)
getAllele(kappa_call, first = FALSE)
getAllele(kappa_call, first = FALSE, strip_d = FALSE)

getGene(kappa_call)
getGene(kappa_call, first = FALSE)
getGene(kappa_call, first = FALSE, strip_d = FALSE)

getFamily(kappa_call)
getFamily(kappa_call, first = FALSE)
getFamily(kappa_call, first = FALSE, collapse = FALSE)
getFamily(kappa_call, first = FALSE, strip_d = FALSE)

getLocus(kappa_call)
getChain(kappa_call)

# Heavy chain examples
heavy_call <- c(
    "Homsap IGHV1-69*01 F,Homsap IGHV1-69D*01 F",
    "Homsap IGHD1-1*01 F",
    "Homsap IGHJ1*01 F"
)

getAllele(heavy_call, first = FALSE)
getAllele(heavy_call, first = FALSE, strip_d = FALSE)

getGene(heavy_call, first = FALSE)
getGene(heavy_call, first = FALSE, strip_d = FALSE)

getFamily(heavy_call)
getLocus(heavy_call)
getChain(heavy_call)

# Filtering non-localized genes
nl_call <- c(
    "IGHV3-NL1*01,IGHV3-30-3*01,IGHV3-30*01",
    "Homosap IGHV3-30*01 F,Homsap IGHV3-NL1*01 F",
    "IGHV1-NL1*01"
)

getAllele(nl_call, first = FALSE, omit_nl = TRUE)
getGene(nl_call, first = FALSE, omit_nl = TRUE)
getFamily(nl_call, first = FALSE, omit_nl = TRUE)

# Temporary designation examples
tmp_call <- c("IGHV9S3*01", "IGKV10S12*01")

getAllele(tmp_call)
getGene(tmp_call)
getFamily(tmp_call)

}
\references{
\url{https://www.imgt.org/}
}
\seealso{
\link{countGenes}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Data.R
\docType{data}
\name{SingleDb}
\alias{SingleDb}
\title{Single sequence AIRR database}
\format{
An object of class \code{spec_tbl_df} (inherits from \code{tbl_df}, \code{tbl}, \code{data.frame}) with 1 rows and 32 columns.
}
\usage{
SingleDb
}
\description{
A database with just one sequence from \code{ExampleDb} and additional AIRR Rearrangement fields 
containing alignment information. The sequence was reanalyzed with a recent versions of 
alignment software (IgBLAST 1.16.0) and reference germlines (IMGT 2020-08-12).
}
\seealso{
\link{ExampleDb}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Sequence.R
\name{seqDist}
\alias{seqDist}
\title{Calculate distance between two sequences}
\usage{
seqDist(seq1, seq2, dist_mat = getDNAMatrix())
}
\arguments{
\item{seq1}{character string containing a DNA sequence.}

\item{seq2}{character string containing a DNA sequence.}

\item{dist_mat}{Character distance matrix. Defaults to a Hamming distance 
matrix returned by \link{getDNAMatrix}. If gap 
characters, \code{c("-", ".")}, are assigned a value of -1 
in \code{dist_mat} then contiguous gaps of any run length,
which are not present in both sequences, will be counted as a 
distance of 1. Meaning, indels of any length will increase
the sequence distance by 1. Gap values other than -1 will 
return a distance that does not consider indels as a special case.}
}
\value{
Numerical distance between \code{seq1} and \code{seq2}.
}
\description{
\code{seqDist} calculates the distance between two DNA sequences.
}
\examples{
# Ungapped examples
seqDist("ATGGC", "ATGGG")
seqDist("ATGGC", "ATG??")

# Gaps will be treated as Ns with a gap=0 distance matrix
seqDist("ATGGC", "AT--C", dist_mat=getDNAMatrix(gap=0))

# Gaps will be treated as universally non-matching characters with gap=1
seqDist("ATGGC", "AT--C", dist_mat=getDNAMatrix(gap=1))

# Gaps of any length will be treated as single mismatches with a gap=-1 distance matrix
seqDist("ATGGC", "AT--C", dist_mat=getDNAMatrix(gap=-1))

# Gaps of equivalent run lengths are not counted as gaps
seqDist("ATG-C", "ATG-C", dist_mat=getDNAMatrix(gap=-1))

# Overlapping runs of gap characters are counted as a single gap
seqDist("ATG-C", "AT--C", dist_mat=getDNAMatrix(gap=-1))
seqDist("A-GGC", "AT--C", dist_mat=getDNAMatrix(gap=-1))
seqDist("AT--C", "AT--C", dist_mat=getDNAMatrix(gap=-1))

# Discontiguous runs of gap characters each count as separate gaps
seqDist("-TGGC", "AT--C", dist_mat=getDNAMatrix(gap=-1))

}
\seealso{
Nucleotide distance matrix may be built with 
          \link{getDNAMatrix}. Amino acid distance matrix may be built
          with \link{getAAMatrix}. Used by \link{pairwiseDist} for generating
          distance matrices. See \link{seqEqual} for testing sequence equivalence.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Sequence.R
\name{getAAMatrix}
\alias{getAAMatrix}
\title{Build an AA distance matrix}
\usage{
getAAMatrix(gap = 0)
}
\arguments{
\item{gap}{value to assign to characters in the set \code{c("-", ".")}.}
}
\value{
A \code{matrix} of amino acid character distances with row and column names 
          indicating the character pair.
}
\description{
\code{getAAMatrix} returns a Hamming distance matrix for IUPAC ambiguous
amino acid characters.
}
\examples{
getAAMatrix()

}
\seealso{
Creates an amino acid distance matrix for \link{seqDist}.
          See \link{getDNAMatrix} for nucleotide distances.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Data.R
\docType{data}
\name{ExampleTrees}
\alias{ExampleTrees}
\title{Example Ig lineage trees}
\format{
A list of igraph objects output by \link{buildPhylipLineage}.
          Each node of each tree has the following annotations (vertex attributes):
  \itemize{
    \item  \code{sample_id}:          Sample identifier(s). Time in relation to vaccination.
    \item  \code{c_call}:             Isotype assignment(s). 
    \item  \code{duplication_count}:  Copy count (number of duplicates) of the sequence.
  }
}
\usage{
ExampleTrees
}
\description{
A set of Ig lineage trees generated from the \code{ExampleDb} file, subset to
only those trees with at least four nodes.
}
\seealso{
\link{ExampleTrees}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Sequence.R
\name{translateDNA}
\alias{translateDNA}
\title{Translate nucleotide sequences to amino acids}
\usage{
translateDNA(seq, trim = FALSE)
}
\arguments{
\item{seq}{vector of strings defining DNA sequence(s) to be converted to translated.}

\item{trim}{boolean flag to remove 3 nts from both ends of seq
(converts IMGT junction to CDR3 region).}
}
\value{
A vector of translated sequence strings.
}
\description{
\code{translateDNA} translates nucleotide sequences to amino acid sequences.
}
\examples{
# Translate a single sequence
translateDNA("ACTGACTCGA")

# Translate a vector of sequences
translateDNA(ExampleDb$junction[1:3])

# Remove the first and last codon from the translation
translateDNA(ExampleDb$junction[1:3], trim=TRUE)

}
\seealso{
\code{\link[seqinr]{translate}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Sequence.R
\name{maskSeqEnds}
\alias{maskSeqEnds}
\title{Masks ragged leading and trailing edges of aligned DNA sequences}
\usage{
maskSeqEnds(seq, mask_char = "N", max_mask = NULL, trim = FALSE)
}
\arguments{
\item{seq}{character vector of DNA sequence strings.}

\item{mask_char}{character to use for masking.}

\item{max_mask}{the maximum number of characters to mask. If set to 0 then
no masking will be performed. If set to \code{NULL} then the upper 
masking bound will be automatically determined from the maximum 
number of observed leading or trailing \code{"N"} characters amongst 
all strings in \code{seq}.}

\item{trim}{if \code{TRUE} leading and trailing characters will be cut rather 
than masked with \code{"N"} characters.}
}
\value{
A modified \code{seq} vector with masked (or optionally trimmed) sequences.
}
\description{
\code{maskSeqEnds} takes a vector of DNA sequences, as character strings,
and replaces the leading and trailing characters with \code{"N"} characters to create 
a sequence vector with uniformly masked outer sequence segments.
}
\examples{
# Default behavior uniformly masks ragged ends
seq <- c("CCCCTGGG", "NAACTGGN", "NNNCTGNN")
maskSeqEnds(seq)

# Does nothing
maskSeqEnds(seq, max_mask=0)

# Cut ragged sequence ends
maskSeqEnds(seq, trim=TRUE)

# Set max_mask to limit extent of masking and trimming
maskSeqEnds(seq, max_mask=1)
maskSeqEnds(seq, max_mask=1, trim=TRUE)

# Mask dashes instead of Ns
seq <- c("CCCCTGGG", "-AACTGG-", "---CTG--")
maskSeqEnds(seq, mask_char="-")

}
\seealso{
See \link{maskSeqGaps} for masking internal gaps.
          See \link{padSeqEnds} for padding sequence of unequal length.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Core.R
\name{stoufferMeta}
\alias{stoufferMeta}
\title{Weighted meta-analysis of p-values via Stouffer's method}
\usage{
stoufferMeta(p, w = NULL)
}
\arguments{
\item{p}{numeric vector of p-values.}

\item{w}{numeric vector of weights.}
}
\value{
A named numeric vector with the combined Z-score and p-value in the form
          \code{c(Z, pvalue)}.
}
\description{
\code{stoufferMeta} combines multiple weighted p-values into a meta-analysis p-value
using Stouffer's Z-score method.
}
\examples{
# Define p-value and weight vectors
p <- c(0.1, 0.05, 0.3)
w <- c(5, 10, 1)

# Unweighted
stoufferMeta(p)

# Weighted
stoufferMeta(p, w)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Topology.R
\name{summarizeSubtrees}
\alias{summarizeSubtrees}
\title{Generate subtree summary statistics for a tree}
\usage{
summarizeSubtrees(graph, fields = NULL, root = "Germline")
}
\arguments{
\item{graph}{igraph object containing an annotated lineage tree.}

\item{fields}{annotation fields to add to the output.}

\item{root}{name of the root (germline) node.}
}
\value{
A data.frame with columns: 
          \itemize{
            \item  \code{name}:             node name.
            \item  \code{parent}:           name of the parent node.
            \item  \code{outdegree}:        number of edges leading from the node.
            \item  \code{size}:             total number of nodes within the subtree rooted 
                                            at the node.
            \item  \code{depth}:            the depth of the subtree that is rooted at 
                                            the node.
            \item  \code{pathlength}:       the maximum pathlength beneath the node.
            \item  \code{outdegree_norm}:   \code{outdegree} normalized by the total 
                                            number of edges.
            \item  \code{size_norm}:        \code{size} normalized by the largest
                                            subtree size (the germline).
            \item  \code{depth_norm}:       \code{depth} normalized by the largest
                                            subtree depth (the germline).
            \item  \code{pathlength_norm}:  \code{pathlength} normalized by the largest
                                            subtree pathlength (the germline).
          }
          An additional column corresponding to the value of \code{field} is added when
          specified.
}
\description{
\code{summarizeSubtrees} calculates summary statistics for each node of a tree. Includes
both node properties and subtree properties.
}
\examples{
# Summarize a tree
graph <- ExampleTrees[[23]]
summarizeSubtrees(graph, fields="c_call", root="Germline")

}
\seealso{
See \link{buildPhylipLineage} for generating input trees. 
          See \link{getPathLengths} for calculating path length to nodes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Diversity.R
\name{alphaDiversity}
\alias{alphaDiversity}
\title{Calculate clonal alpha diversity}
\usage{
alphaDiversity(data, min_q = 0, max_q = 4, step_q = 0.1, ci = 0.95, ...)
}
\arguments{
\item{data}{data.frame with Change-O style columns containing clonal assignments or
a \link{AbundanceCurve} generate by \link{estimateAbundance} object 
containing a previously calculated bootstrap distributions of clonal abundance.}

\item{min_q}{minimum value of \eqn{q}.}

\item{max_q}{maximum value of \eqn{q}.}

\item{step_q}{value by which to increment \eqn{q}.}

\item{ci}{confidence interval to calculate; the value must be between 0 and 1.}

\item{...}{additional arguments to pass to \link{estimateAbundance}. Additional arguments
are ignored if a \link{AbundanceCurve} is provided as input.}
}
\value{
A \link{DiversityCurve} object summarizing the diversity scores.
}
\description{
\code{alphaDiversity} takes in a data.frame or \link{AbundanceCurve} and computes
diversity scores (\eqn{D}) over an interval of diversity orders (\eqn{q}).
}
\details{
Clonal diversity is calculated using the generalized diversity index (Hill numbers) 
proposed by Hill (Hill, 1973). See \link{calcDiversity} for further details.

To generate a smooth curve, \eqn{D} is calculated for each value of \eqn{q} from
\code{min_q} to \code{max_q} incremented by \code{step_q}.  When \code{uniform=TRUE}
variability in total sequence counts across unique values in the \code{group} column 
is corrected by repeated resampling from the estimated complete clonal distribution to a 
common number of sequences. The complete clonal abundance distribution that is resampled 
from is inferred by using the Chao1 estimator to infer the number of unseen clones, 
followed by applying the relative abundance correction and unseen clone frequencies 
described in Chao et al, 2015.

The diversity index (\eqn{D}) for each group is the mean value of over all resampling 
realizations. Confidence intervals are derived using the standard deviation of the 
resampling realizations, as described in Chao et al, 2015.

Significance of the difference in diversity index (\code{D}) between groups is tested by 
constructing a bootstrap delta distribution for each pair of unique values in the 
\code{group} column. The bootstrap delta distribution is built by subtracting the diversity 
index \code{Da} in group \code{a} from the corresponding value \eqn{Db} in group \code{b}, 
for all bootstrap realizations, yielding a distribution of \code{nboot} total deltas; where 
group \code{a} is the group with the greater mean \code{D}. The p-value for hypothesis 
\code{Da  !=  Db} is the value of \code{P(0)} from the empirical cumulative distribution 
function of the bootstrap delta distribution, multiplied by 2 for the two-tailed correction.

Note, this method may inflate statistical significance when clone sizes are uniformly small,
such as when most clones sizes are 1, sample size is small, and \code{max_n} is near
the total count of the smallest data group. Use caution when interpreting the results 
in such cases.
}
\examples{
# Group by sample identifier in two steps
set.seed(123)
abund <- estimateAbundance(ExampleDb, group="sample_id", nboot=100)
div <- alphaDiversity(abund, step_q=1, max_q=10)
plotDiversityCurve(div, legend_title="Sample")
                   
# Grouping by isotype rather than sample identifier in one step
div <- alphaDiversity(ExampleDb, group="c_call", min_n=40, step_q=1, max_q=10, 
                      nboot=100)
plotDiversityCurve(div, legend_title="Isotype")

}
\references{
\enumerate{
  \item  Hill M. Diversity and evenness: a unifying notation and its consequences. 
           Ecology. 1973 54(2):427-32.
  \item  Chao A. Nonparametric Estimation of the Number of Classes in a Population. 
           Scand J Stat. 1984 11, 265270.
  \item  Chao A, et al. Rarefaction and extrapolation with Hill numbers: 
           A framework for sampling and estimation in species diversity studies. 
           Ecol Monogr. 2014 84:45-67.
  \item  Chao A, et al. Unveiling the species-rank abundance distribution by 
           generalizing the Good-Turing sample coverage theory. 
           Ecology. 2015 96, 11891201.
}
}
\seealso{
See \link{calcDiversity} for the basic calculation and 
          \link{DiversityCurve} for the return object. 
          See \link{plotDiversityCurve} for plotting the return object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Sequence.R
\name{padSeqEnds}
\alias{padSeqEnds}
\title{Pads ragged ends of aligned DNA sequences}
\usage{
padSeqEnds(seq, len = NULL, start = FALSE, pad_char = "N", mod3 = TRUE)
}
\arguments{
\item{seq}{character vector of DNA sequence strings.}

\item{len}{length to pad to. Only applies if longer than the maximum length of
the data in \code{seq}.}

\item{start}{if \code{TRUE} pad the beginning of each sequence instead of the end.}

\item{pad_char}{character to use for padding.}

\item{mod3}{if \code{TRUE} pad sequences to be of length multiple three.}
}
\value{
A modified \code{seq} vector with padded sequences.
}
\description{
\code{padSeqEnds} takes a vector of DNA sequences, as character strings,
and appends the ends of each sequence with an appropriate number of \code{"N"} 
characters to create a sequence vector with uniform lengths.
}
\examples{
# Default behavior uniformly pads ragged ends
seq <- c("CCCCTGGG", "ACCCTG", "CCCC")
padSeqEnds(seq)

# Pad to fixed length
padSeqEnds(seq, len=15)

# Add padding to the beginning of the sequences instead of the ends
padSeqEnds(seq, start=TRUE)
padSeqEnds(seq, len=15, start=TRUE)

}
\seealso{
See \link{maskSeqEnds} for creating uniform masking from existing masking.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Core.R
\name{cpuCount}
\alias{cpuCount}
\title{Available CPU cores}
\usage{
cpuCount()
}
\value{
Count of available cores. Returns 1 if undeterminable.
}
\description{
\code{cpuCount} determines the number of CPU cores available.
}
\examples{
cpuCount()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AminoAcids.R
\name{aminoAcidProperties}
\alias{aminoAcidProperties}
\title{Calculates amino acid chemical properties for sequence data}
\usage{
aminoAcidProperties(
  data,
  property = c("length", "gravy", "bulk", "aliphatic", "polarity", "charge", "basic",
    "acidic", "aromatic"),
  seq = "junction",
  nt = TRUE,
  trim = FALSE,
  label = NULL,
  ...
)
}
\arguments{
\item{data}{\code{data.frame} containing sequence data.}

\item{property}{vector strings specifying the properties to be calculated. Defaults
to calculating all defined properties.}

\item{seq}{\code{character} name of the column containing input 
sequences.}

\item{nt}{boolean, TRUE if the sequences (or sequence) are DNA and will be translated.}

\item{trim}{if \code{TRUE} remove the first and last codon/amino acids from each
sequence before calculating properties. If \code{FALSE} do
not modify input sequences.}

\item{label}{name of sequence region to add as prefix to output column names.}

\item{...}{additional named arguments to pass to the functions 
\link{gravy}, \link{bulk}, \link{aliphatic}, \link{polar} or \link{charge}.}
}
\value{
A modified \code{data} data.frame with the following columns:
         \itemize{
           \item  \code{*_aa_length}:     number of amino acids.
           \item  \code{*_aa_gravy}:      grand average of hydrophobicity (gravy) index.
           \item  \code{*_aa_bulk}:       average bulkiness of amino acids.
           \item  \code{*_aa_aliphatic}:  aliphatic index.
           \item  \code{*_aa_polarity}:   average polarity of amino acids.
           \item  \code{*_aa_charge}:     net charge.
           \item  \code{*_aa_basic}:      fraction of informative positions that are 
                                          Arg, His or Lys.
           \item  \code{*_aa_acidic}:     fraction of informative positions that are 
                                          Asp or Glu.
           \item  \code{*_aa_aromatic}:   fraction of informative positions that are 
                                          His, Phe, Trp or Tyr.
           
         }
         
         Where \code{*} is the value from \code{label} or the name specified for 
         \code{seq} if \code{label=NULL}.
}
\description{
\code{aminoAcidProperties} calculates amino acid sequence physicochemical properties, including
length, hydrophobicity, bulkiness, polarity, aliphatic index, net charge, acidic residue
content, basic residue content, and aromatic residue content.
}
\details{
For all properties except for length, non-informative positions are excluded, 
where non-informative is defined as any character in \code{c("X", "-", ".", "*")}.

The scores for gravy, bulkiness and polarity are calculated as simple averages of the 
scores for each informative positions. The basic, acid and aromatic indices are 
calculated as the fraction of informative positions falling into the given category.

The aliphatic index is calculated using the Ikai, 1980 method.

The net charge is calculated using the method of Moore, 1985, excluding the N-terminus and
C-terminus charges, and normalizing by the number of informative positions.  The default 
pH for the calculation is 7.4.

The following data sources were used for the default property scores:
\itemize{
  \item  hydropathy:  Kyte & Doolittle, 1982.  
  \item  bulkiness:   Zimmerman et al, 1968. 
  \item  polarity:    Grantham, 1974.
  \item  pK:          EMBOSS.
}
}
\examples{
# Subset example data
db <- ExampleDb[c(1,10,100), c("sequence_id", "junction")]

# Calculate default amino acid properties from DNA sequences
aminoAcidProperties(db, seq="junction")
# Calculate default amino acid properties from amino acid sequences
# Use a custom output column prefix
db$junction_aa <- translateDNA(db$junction)
aminoAcidProperties(db, seq="junction_aa", label="junction", nt=FALSE)

# Use the Grantham, 1974 side chain volume scores from the seqinr package
# Set pH=7.0 for the charge calculation
# Calculate only average volume and charge
# Remove the head and tail amino acids from the junction, thus making it the CDR3
library(seqinr)
data(aaindex)
x <- aaindex[["GRAR740103"]]$I
# Rename the score vector to use single-letter codes
names(x) <- translateStrings(names(x), ABBREV_AA)
# Calculate properties
aminoAcidProperties(db, property=c("bulk", "charge"), seq="junction", 
                    trim=TRUE, label="cdr3", bulkiness=x, pH=7.0)

}
\references{
\enumerate{
  \item  Zimmerman JM, Eliezer N, Simha R. The characterization of amino acid sequences 
           in proteins by statistical methods. J Theor Biol 21, 170-201 (1968).
  \item  Grantham R. Amino acid difference formula to help explain protein evolution. 
           Science 185, 862-864 (1974).
  \item  Ikai AJ. Thermostability and aliphatic index of globular proteins. 
           J Biochem 88, 1895-1898 (1980).
  \item  Kyte J, Doolittle RF. A simple method for displaying the hydropathic character 
           of a protein. J Mol Biol 157, 105-32 (1982).
  \item  Moore DS. Amino acid and peptide net charges: A simple calculational procedure. 
           Biochem Educ 13, 10-11 (1985).
  \item  Wu YC, et al. High-throughput immunoglobulin repertoire analysis distinguishes 
           between human IgM memory and switched memory B-cell populations. 
           Blood 116, 1070-8 (2010).
  \item  Wu YC, et al. The relationship between CD27 negative and positive B cell 
           populations in human peripheral blood. 
           Front Immunol 2, 1-12 (2011).
  \item  \url{https://emboss.sourceforge.net/apps/cvs/emboss/apps/iep.html}
}
}
\seealso{
See \link{countPatterns} for counting the occurrence of specific amino acid subsequences.
See \link{gravy}, \link{bulk}, \link{aliphatic}, \link{polar} and \link{charge} for functions 
that calculate the included properties individually.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Deprecated.R
\name{testDiversity}
\alias{testDiversity}
\title{Pairwise test of the diversity index}
\usage{
testDiversity(
  data,
  q,
  group,
  clone = "CLONE",
  copy = NULL,
  min_n = 30,
  max_n = NULL,
  nboot = 2000,
  progress = FALSE,
  ci = 0.95,
  cell_id = "cell_id"
)
}
\arguments{
\item{data}{data.frame with Change-O style columns containing clonal assignments.}

\item{q}{diversity order to test.}

\item{group}{name of the \code{data} column containing group identifiers.}

\item{clone}{name of the \code{data} column containing clone identifiers.}

\item{copy}{name of the \code{data} column containing copy numbers for each 
sequence. If \code{copy=NULL} (the default), then clone abundance
is determined by the number of sequences. If a \code{copy} column
is specified, then clone abundances is determined by the sum of 
copy numbers within each clonal group.}

\item{min_n}{minimum number of observations to sample.
A group with less observations than the minimum is excluded.}

\item{max_n}{maximum number of observations to sample. If \code{NULL} the maximum
if automatically determined from the size of the largest group.}

\item{nboot}{number of bootstrap realizations to perform.}

\item{progress}{if \code{TRUE} show a progress bar.}

\item{ci}{confidence interval to calculate; the value must be between 0 and 1.}

\item{cell_id}{the name of the \code{data} column containing cell identifiers.}
}
\value{
A \link{DiversityCurve} object containing slot test with p-values and summary 
            statistics.
}
\description{
\code{testDiversity} performs pairwise significance tests of the diversity index 
(\eqn{D}) at a given diversity order (\eqn{q}) for a set of annotation groups via
rarefaction and bootstrapping.
}
\details{
Clonal diversity is calculated using the generalized diversity index proposed by 
Hill (Hill, 1973). See \link{calcDiversity} for further details.

Diversity is calculated on the estimated complete clonal abundance distribution.
This distribution is inferred by using the Chao1 estimator to estimate the number
of seen clones, and applying the relative abundance correction and unseen clone
frequency described in Chao et al, 2014.

Variability in total sequence counts across unique values in the \code{group} column is 
corrected by repeated resampling from the estimated complete clonal distribution to 
a common number of sequences. The diversity index estimate (\eqn{D}) for each group is 
the mean value of over all bootstrap realizations. 

Significance of the difference in diversity index (\eqn{D}) between groups is tested by 
constructing a bootstrap delta distribution for each pair of unique values in the 
\code{group} column. The bootstrap delta distribution is built by subtracting the diversity 
index \eqn{Da} in \eqn{group-a} from the corresponding value \eqn{Db} in \eqn{group-b}, 
for all bootstrap realizations, yielding a distribution of \code{nboot} total deltas; where 
\eqn{group-a} is the group with the greater mean \eqn{D}. The p-value for hypothesis 
\eqn{Da  !=  Db} is the value of \eqn{P(0)} from the empirical cumulative distribution 
function of the bootstrap delta distribution, multiplied by 2 for the two-tailed correction.
}
\note{
This method may inflate statistical significance when clone sizes are uniformly small,
such as when most clones sizes are 1, sample size is small, and \code{max_n} is near
the total count of the smallest data group. Use caution when interpreting the results 
in such cases. We are currently investigating this potential problem.
}
\examples{
 
\dontrun{
# Groups under the size threshold are excluded and a warning message is issued.
testDiversity(ExampleDb, "sample_id", q=0, min_n=30, nboot=100)
}

}
\references{
\enumerate{
  \item  Hill M. Diversity and evenness: a unifying notation and its consequences. 
           Ecology. 1973 54(2):427-32.
  \item  Chao A. Nonparametric Estimation of the Number of Classes in a Population. 
           Scand J Stat. 1984 11, 265270.
  \item  Wu Y-CB, et al. Influence of seasonal exposure to grass pollen on local and 
           peripheral blood IgE repertoires in patients with allergic rhinitis. 
           J Allergy Clin Immunol. 2014 134(3):604-12.
  \item  Chao A, et al. Rarefaction and extrapolation with Hill numbers: 
           A framework for sampling and estimation in species diversity studies. 
           Ecol Monogr. 2014 84:45-67.
  \item  Chao A, et al. Unveiling the species-rank abundance distribution by 
           generalizing the Good-Turing sample coverage theory. 
           Ecology. 2015 96, 11891201.
}
}
\seealso{
\link{alphaDiversity}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Core.R
\name{makeTempDir}
\alias{makeTempDir}
\title{Create a temporary folder}
\usage{
makeTempDir(prefix)
}
\arguments{
\item{prefix}{prefix name for the folder.}
}
\value{
The path to the temporary folder.
}
\description{
\code{makeTempDir} creates a randomly named temporary folder in the 
system temp location.
}
\examples{
\dontrun{
makeTempDir("Clone50")
}

}
\seealso{
This is just a wrapper for \link{tempfile} and 
          \link{dir.create}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Topology.R
\name{tableEdges}
\alias{tableEdges}
\title{Tabulate the number of edges between annotations within a lineage tree}
\usage{
tableEdges(graph, field, indirect = FALSE, exclude = NULL)
}
\arguments{
\item{graph}{igraph object containing an annotated lineage tree.}

\item{field}{string defining the annotation field to count.}

\item{indirect}{if \code{FALSE} count direct connections (edges) only. If 
\code{TRUE} walk through any nodes with annotations specified in 
the \code{argument} to count indirect connections. Specifying
\code{indirect=TRUE} with \code{exclude=NULL} will have no effect.}

\item{exclude}{vector of strings defining \code{field} values to exclude from counts.
Edges that either start or end with the specified annotations will not
be counted. If \code{NULL} count all edges.}
}
\value{
A data.frame defining total annotation connections in the tree with columns:
          \itemize{
            \item  \code{parent}:  parent annotation
            \item  \code{child}:   child annotation
            \item  \code{count}:   count of edges for the parent-child relationship
          }
}
\description{
\code{tableEdges} creates a table of the total number of connections (edges) for each 
unique pair of annotations within a tree over all nodes.
}
\examples{
# Define example graph
graph <- ExampleTrees[[23]]

# Count direct edges between isotypes including inferred nodes
tableEdges(graph, "c_call")

# Count direct edges excluding edges to and from germline and inferred nodes
tableEdges(graph, "c_call", exclude=c("Germline", NA))

# Count indirect edges walking through germline and inferred nodes
tableEdges(graph, "c_call", indirect=TRUE, exclude=c("Germline", NA))

}
\seealso{
See \link{testEdges} for performed a permutation test on edge relationships.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Data.R
\docType{data}
\name{IUPAC_CODES}
\alias{IUPAC_CODES}
\alias{IUPAC_DNA}
\alias{IUPAC_AA}
\alias{DNA_IUPAC}
\title{IUPAC ambiguous characters}
\format{
A list with single character codes as names and values containing character 
         vectors that define the set of standard characters that match to each each 
         ambiguous character.
\itemize{
  \item  \code{IUPAC_DNA}:  DNA ambiguous character translations.
  \item  \code{IUPAC_AA}:   Amino acid ambiguous character translations.
  \item  \code{DNA_IUPAC}:  Ordered DNA to ambiguous characters
}

An object of class \code{list} of length 15.

An object of class \code{list} of length 25.

An object of class \code{list} of length 15.
}
\usage{
IUPAC_DNA

IUPAC_AA

DNA_IUPAC
}
\description{
A translation list mapping IUPAC ambiguous characters code to corresponding nucleotide
amino acid characters.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Core.R
\name{translateStrings}
\alias{translateStrings}
\title{Translate a vector of strings}
\usage{
translateStrings(strings, translation)
}
\arguments{
\item{strings}{vector of character strings to modify.}

\item{translation}{named character vector or a list of character vectors specifying 
the strings to replace (values) and their replacements (names).}
}
\value{
A modified \code{strings} vector.
}
\description{
\code{translateStrings} modifies a character vector by substituting one or more 
strings with a replacement string.
}
\details{
Does not perform partial replacements. Each translation value must match a complete 
\code{strings} value or it will not be replaced.  Values that do not have a replacement
named in the \code{translation} parameter will not be modified.

Replacement is accomplished using \link{gsub}.
}
\examples{
# Using a vector translation
strings <- LETTERS[1:5]
translation <- c("POSITION1"="A", "POSITION5"="E")
translateStrings(strings, translation)

# Using a list translation
strings <- LETTERS[1:5]
translation <- list("1-3"=c("A","B","C"), "4-5"=c("D","E"))
translateStrings(strings, translation)

}
\seealso{
See \link{gsub} for single value replacement in the base package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Topology.R
\name{testMRCA}
\alias{testMRCA}
\title{Tests for MRCA annotation enrichment in lineage trees}
\usage{
testMRCA(
  graphs,
  field,
  root = "Germline",
  exclude = c("Germline", NA),
  nperm = 200,
  progress = FALSE
)
}
\arguments{
\item{graphs}{list of igraph object containing annotated lineage trees.}

\item{field}{string defining the annotation field to test.}

\item{root}{name of the root (germline) node.}

\item{exclude}{vector of strings defining \code{field} values to exclude from the
set of potential founder annotations.}

\item{nperm}{number of permutations to perform.}

\item{progress}{if \code{TRUE} show a progress bar.}
}
\value{
An \link{MRCATest} object containing the test results and permutation
          realizations.
}
\description{
\code{testMRCA} performs a permutation test on a set of lineage trees to determine
the significance of an annotation's association with the MRCA position of the lineage
trees.
}
\examples{
\donttest{
# Define example tree set
graphs <- ExampleTrees[1:10]

# Perform MRCA test on isotypes
x <- testMRCA(graphs, "c_call", nperm=10)
print(x)
}

}
\seealso{
Uses \link{getMRCA} and \link{getPathLengths}. 
          See \link{plotMRCATest} for plotting the permutation distributions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Classes.R
\docType{class}
\name{AbundanceCurve-class}
\alias{AbundanceCurve-class}
\alias{AbundanceCurve}
\alias{print,AbundanceCurve-method}
\alias{AbundanceCurve-method}
\alias{plot,AbundanceCurve,missing-method}
\title{S4 class defining a clonal abundance curve}
\usage{
\S4method{print}{AbundanceCurve}(x)

\S4method{plot}{AbundanceCurve,missing}(x, y, ...)
}
\arguments{
\item{x}{AbundanceCurve object}

\item{y}{ignored.}

\item{...}{arguments to pass to \link{plotDiversityCurve}.}
}
\description{
\code{AbundanceCurve} defines clonal abundance values.
}
\section{Slots}{

\describe{
\item{\code{abundance}}{data.frame with relative clonal abundance data and confidence intervals,
containing the following columns:
\itemize{
  \item  \code{group}:  group identifier.
  \item  \code{clone_id} or \code{CLONE}:  clone identifier. 
  \item  \code{p}:      relative abundance of the clone.
  \item  \code{lower}:  lower confidence interval bound.
  \item  \code{upper}:  upper confidence interval bound.
  \item  \code{rank}:   the rank of the clone abundance.
}}

\item{\code{bootstrap}}{data.frame of bootstrapped clonal distributions.}

\item{\code{clone_by}}{string specifying the name of the clone column.}

\item{\code{group_by}}{string specifying the name of the grouping column.}

\item{\code{groups}}{vector specifying the names of unique groups in group column.}

\item{\code{n}}{numeric vector indication the number of sequences sampled in each group.}

\item{\code{nboot}}{numeric specifying the number of bootstrap iterations to use.}

\item{\code{ci}}{confidence interval defining the upper and lower bounds 
(a value between 0 and 1).}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Topology.R
\name{plotSubtrees}
\alias{plotSubtrees}
\title{Plots subtree statistics for multiple trees}
\usage{
plotSubtrees(
  graphs,
  field,
  stat,
  root = "Germline",
  exclude = c("Germline", NA),
  colors = NULL,
  main_title = "Subtrees",
  legend_title = "Annotation",
  style = c("box", "violin"),
  silent = FALSE,
  ...
)
}
\arguments{
\item{graphs}{list of igraph objects containing annotated lineage trees.}

\item{field}{string defining the annotation field.}

\item{stat}{string defining the subtree statistic to plot. One of:
\itemize{
  \item  \code{outdegree}:   distribution of normalized node 
                             outdegrees.
  \item  \code{size}:        distribution of normalized subtree sizes.
  \item  \code{depth}:       distribution of subtree depths.
  \item  \code{pathlength}:  distribution of maximum pathlength 
                             beneath nodes.
}}

\item{root}{name of the root (germline) node.}

\item{exclude}{vector of strings defining \code{field} values to exclude from
plotting.}

\item{colors}{named vector of colors for values in \code{field}, with 
names defining annotation names \code{field} column and values
being colors. Also controls the order in which values appear on the
plot. If \code{NULL} alphabetical ordering and a default color palette 
will be used.}

\item{main_title}{string specifying the plot title.}

\item{legend_title}{string specifying the legend title.}

\item{style}{string specifying the style of plot to draw. One of:
\itemize{
  \item \code{"histogram"}:  histogram of the annotation count 
                             distribution with a red dotted line
                             denoting the observed value.
  \item \code{"cdf"}:        cumulative distribution function 
                             of annotation counts with a red 
                             dotted line denoting the observed 
                             value and a blue dotted line 
                             indicating the p-value.
}}

\item{silent}{if \code{TRUE} do not draw the plot and just return the ggplot2 
object; if \code{FALSE} draw the plot.}

\item{...}{additional arguments to pass to ggplot2::theme.}
}
\value{
A \code{ggplot} object defining the plot.
}
\description{
\code{plotSubtree} plots distributions of normalized subtree statistics for a 
set of lineage trees, broken down by annotation value.
}
\examples{
# Define example tree set
graphs <- ExampleTrees[1:10]

# Violin plots of node outdegree by sample
plotSubtrees(graphs, "sample_id", "out", style="v")

# Violin plots of subtree size by sample
plotSubtrees(graphs, "sample_id", "size", style="v")

# Boxplot of node depth by isotype
plotSubtrees(graphs, "c_call", "depth", style="b")

}
\seealso{
Subtree statistics are calculated with \link{summarizeSubtrees}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Data.R
\docType{data}
\name{DEFAULT_COLORS}
\alias{DEFAULT_COLORS}
\alias{DNA_COLORS}
\alias{IG_COLORS}
\alias{TR_COLORS}
\title{Default colors}
\format{
Named character vectors with hexcode colors as values.
\itemize{
  \item  \code{DNA_COLORS}:  DNA character colors 
                             \code{c("A", "C", "G", "T")}.
  \item  \code{IG_COLORS}:   Ig isotype colors 
                             \code{c("IGHA", "IGHD", "IGHE", "IGHG", "IGHM", "IGHK", "IGHL")}.
  \item  \code{TR_COLORS}:   TCR chain colors 
                             \code{c("TRA", "TRB", "TRD", "TRG")}.
}

An object of class \code{character} of length 4.

An object of class \code{character} of length 7.

An object of class \code{character} of length 4.
}
\usage{
DNA_COLORS

IG_COLORS

TR_COLORS
}
\description{
Default color palettes for DNA characters, Ig isotypes, and TCR chains.
}
\examples{
# IG_COLORS as an isotype color set for ggplot
isotype <- c("IGHG", "IGHM", "IGHM", "IGHA")
db <- data.frame(x=1:4, y=1:4, iso=isotype)
g1 <- ggplot(db, aes(x=x, y=y, color=iso)) + 
    scale_color_manual(name="Isotype", values=IG_COLORS) +
    geom_point(size=10)
plot(g1)

# DNA_COLORS to translate nucleotide values to a vector of colors 
# for use in base graphics plots
seq <- c("A", "T", "T", "C")
colors <- translateStrings(seq, setNames(names(DNA_COLORS), DNA_COLORS))
plot(1:4, 1:4, col=colors, pch=16, cex=6)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Diversity.R
\name{countClones}
\alias{countClones}
\title{Tabulates clones sizes}
\usage{
countClones(
  data,
  groups = NULL,
  copy = NULL,
  clone = "clone_id",
  cell_id = "cell_id",
  remove_na = TRUE
)
}
\arguments{
\item{data}{data.frame with columns containing clonal assignments.}

\item{groups}{character vector defining \code{data} columns containing grouping 
variables. If \code{groups=NULL}, then do not group data.}

\item{copy}{name of the \code{data} column containing copy numbers for each 
sequence. If this value is specified, then total copy abundance
is determined by the sum of copy numbers within each clonal group.}

\item{clone}{name of the \code{data} column containing clone identifiers.}

\item{cell_id}{name of the \code{data} column containing cell identifiers. If
\code{cell_id} column is not present the function will assume bulk data.}

\item{remove_na}{removes rows with \code{NA} values in the clone column if \code{TRUE} and issues a warning. 
Otherwise, keeps those rows and considers \code{NA} as a clone in the final counts 
and relative abundances.}
}
\value{
A data.frame summarizing clone counts and frequencies with columns:
          \itemize{
            \item \code{clone_id}:    clone identifier. This is the default column
                                      name, specified with \code{clone='clone_id'}.
                                      If the function call uses Change-O 
                                      formatted data and \code{clone='CLONE'}, this
                                      column will have name \code{CLONE}.
            \item \code{seq_count}:   total number of sequences for the clone.
            \item \code{seq_freq}:    frequency of the clone as a fraction of the total
                                      number of sequences within each group.
            \item \code{copy_count}:  sum of the copy counts in the \code{copy} column.
                                      Only present if the \code{copy} argument is 
                                      specified.
            \item \code{copy_freq}:   frequency of the clone as a fraction of the total
                                      copy number within each group. Only present if 
                                      the \code{copy} argument is specified.
          }
          Also includes additional columns specified in the \code{groups} argument.
}
\description{
\code{countClones} determines the number of sequences and total copy number of 
clonal groups.
}
\examples{
# Without copy numbers
clones <- countClones(ExampleDb, groups="sample_id")

# With copy numbers and multiple groups
clones <- countClones(ExampleDb, groups=c("sample_id", "c_call"), copy="duplicate_count")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Fastq.R
\name{readFastqDb}
\alias{readFastqDb}
\title{Load sequencing quality scores from a FASTQ file}
\usage{
readFastqDb(
  data,
  fastq_file,
  quality_offset = -33,
  header = c("presto", "asis"),
  sequence_id = "sequence_id",
  sequence = "sequence",
  sequence_alignment = "sequence_alignment",
  v_cigar = "v_cigar",
  d_cigar = "d_cigar",
  j_cigar = "j_cigar",
  np1_length = "np1_length",
  np2_length = "np2_length",
  v_sequence_end = "v_sequence_end",
  d_sequence_end = "d_sequence_end",
  style = c("num", "ascii", "both"),
  quality_sequence = FALSE
)
}
\arguments{
\item{data}{\code{data.frame} containing sequence data.}

\item{fastq_file}{path to the fastq file}

\item{quality_offset}{offset value to be used by ape::read.fastq. It is 
the value to be added to the quality scores 
(the default -33 applies to the Sanger format and 
should work for most recent FASTQ files).}

\item{header}{FASTQ file header format; one of \code{"presto"} or 
\code{"asis"}. Use \code{"presto"} to specify 
that the fastq file headers are using the pRESTO
format and can be parsed to extract 
the \code{sequence_id}. Use \code{"asis"} to skip 
any processing and use the sequence names as they are.}

\item{sequence_id}{column in \code{data} that contains sequence 
identifiers to be matched to sequence identifiers in 
\code{fastq_file}.}

\item{sequence}{column in \code{data} that contains sequence data.}

\item{sequence_alignment}{column in \code{data} that contains IMGT aligned sequence data.}

\item{v_cigar}{column in \code{data} that contains CIGAR 
strings for the V gene alignments.}

\item{d_cigar}{column in \code{data} that contains CIGAR 
strings for the D gene alignments.}

\item{j_cigar}{column in \code{data} that contains CIGAR 
strings for the J gene alignments.}

\item{np1_length}{column in \code{data} that contains the number
of nucleotides between the V gene and first D gene 
alignments or between the V gene and J gene alignments.}

\item{np2_length}{column in \code{data} that contains the number
of nucleotides between either the first D gene and J 
gene alignments or the first D gene and second D gene
alignments.}

\item{v_sequence_end}{column in \code{data} that contains the 
end position of the V gene in \code{sequence}.}

\item{d_sequence_end}{column in \code{data} that contains the 
end position of the D gene in \code{sequence}.}

\item{style}{how the sequencing quality should be returned;
one of \code{"num"}, \code{"phred"}, or \code{"both"}.
Specify \code{"num"} to store the quality scores as strings of 
comma separated numeric values. Use \code{"phred"} to have
the function return the scores as Phred (ASCII) scores. 
Use \code{"both"} to retrieve both.}

\item{quality_sequence}{specify \code{TRUE} to keep the quality scores for 
\code{sequence}. If false, only the quality score
for \code{sequence_alignment} will be added to \code{data}.}
}
\value{
Modified \code{data} with additional fields:
          \enumerate{
                \item \code{quality_alignment}:     A character vector with ASCII Phred 
                                                    scores for \code{sequence_alignment}.
                \item \code{quality_alignment_num}: A character vector, with comma separated 
                                                    numerical quality values for each 
                                                    position in \code{sequence_alignment}.
                \item \code{quality}:      A character vector with ASCII Phred 
                                                    scores for \code{sequence}.
                \item \code{quality_num}:  A character vector, with comma separated 
                                                    numerical quality values for each 
                                                    position in \code{sequence}.
          }
}
\description{
\code{readFastqDb} adds the sequencing quality scores to a data.frame
from a FASTQ file. Matching is done by `sequence_id`.
}
\examples{
db <- airr::read_rearrangement(system.file("extdata", "example_quality.tsv", package="alakazam"))
fastq_file <- system.file("extdata", "example_quality.fastq", package="alakazam")
db <- readFastqDb(db, fastq_file, quality_offset=-33)

}
\seealso{
\link{maskPositionsByQuality} and \link{getPositionQuality}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Gene.R
\name{groupGenes}
\alias{groupGenes}
\title{Group sequences by gene assignment}
\usage{
groupGenes(
  data,
  v_call = "v_call",
  j_call = "j_call",
  junc_len = NULL,
  sequence_alignment = NULL,
  cell_id = NULL,
  split_light = FALSE,
  locus = "locus",
  only_heavy = TRUE,
  first = FALSE
)
}
\arguments{
\item{data}{data.frame containing sequence data.}

\item{v_call}{name of the column containing the heavy/long chain
V-segment allele calls.}

\item{j_call}{name of the column containing the heavy/long chain
J-segment allele calls.}

\item{junc_len}{name of column containing the junction length.
If \code{NULL} then 1-stage partitioning is perform
considering only the V and J genes is performed.
See Details for further clarification.}

\item{sequence_alignment}{name of the column containing the sequence alignment.}

\item{cell_id}{name of the column containing cell identifiers or barcodes.
If specified, grouping will be performed in single-cell mode
with the behavior governed by the \code{locus} and
\code{only_heavy} arguments. If set to \code{NULL} then the
bulk sequencing data is assumed.}

\item{split_light}{A deprecated parameter. This would split clones by the light chain.
For similar function use dowser::resolveLightChains}

\item{locus}{name of the column containing locus information.
Only applicable to single-cell data.
Ignored if \code{cell_id=NULL}.}

\item{only_heavy}{This is deprecated. Only heavy chains will be used in clustering.
Use only the IGH (BCR) or TRB/TRD (TCR) sequences
for grouping. Only applicable to single-cell data.
Ignored if \code{cell_id=NULL}.}

\item{first}{if \code{TRUE} only the first call of the gene assignments
is used. if \code{FALSE} the union of ambiguous gene
assignments is used to group all sequences with any
overlapping gene calls.}
}
\value{
Returns a modified data.frame with disjoint union indices
          in a new \code{vj_group} column.

          If \code{junc_len} is supplied, the grouping this \code{vj_group}
          will have been based on V, J, and junction length simultaneously. However,
          the output column name will remain \code{vj_group}.

          The output \code{v_call}, \code{j_call}, \code{cell_id}, and \code{locus}
          columns will be converted to type \code{character} if they were of type
          \code{factor} in the input \code{data}.
}
\description{
\code{groupGenes} will group rows by shared V and J gene assignments,
and optionally also by junction lengths. IGH:IGK/IGL, TRB:TRA, and TRD:TRG
paired single-cell BCR/TCR sequencing and unpaired bulk sequencing
(IGH, TRB, TRD chain only) are supported. In the case of ambiguous (multiple)
gene assignments, the grouping may be specified to be a union across all
ambiguous V and J gene pairs, analogous to single-linkage clustering
(i.e., allowing for chaining).
}
\details{
To invoke single-cell mode the \code{cell_id} argument must be specified and the \code{locus}
column must be correct. Otherwise, \code{groupGenes} will be run with bulk sequencing assumptions,
using all input sequences regardless of the values in the \code{locus} column.

Values in the \code{locus} column must be one of \code{c("IGH", "IGI", "IGK", "IGL")} for BCR
or \code{c("TRA", "TRB", "TRD", "TRG")} for TCR sequences. Otherwise, the function returns an
error message and stops.

Under single-cell mode with paired chained sequences, there was a choice of whether
grouping should be done by (a) using IGH (BCR) or TRB/TRD (TCR) sequences only or
(b) using IGH plus IGK/IGL (BCR) or TRB/TRD plus TRA/TRG (TCR).
This was governed by the \code{only_heavy} argument, now deprecated.

Specifying \code{junc_len} will force \code{groupGenes} to perform a 1-stage partitioning of the
sequences/cells based on V gene, J gene, and junction length simultaneously.
If \code{junc_len=NULL} (no column specified), then \code{groupGenes} performs only the first
stage of a 2-stage partitioning in which sequences/cells are partitioned in the first stage
based on V gene and J gene, and then in the second stage further splits the groups based on
junction length (the second stage must be performed independently, as this only returns the
first stage results).

In the input \code{data}, the \code{v_call}, \code{j_call}, \code{cell_id}, and \code{locus}
columns, if present, must be of type \code{character} (as opposed to \code{factor}).

It is assumed that ambiguous gene assignments are separated by commas.

All rows containing \code{NA} values in any of the \code{v_call}, \code{j_call}, and \code{junc_len}
(if \code{junc_len != NULL}) columns will be removed. A warning will be issued when a row
containing an \code{NA} is removed.
}
\section{Expectations for single-cell data}{


Single-cell paired chain data assumptions:
  \itemize{
     \item every row represents a sequence (chain).
     \item heavy/long and light/short chains of the same cell are linked by \code{cell_id}.
     \item the value in \code{locus} column indicates whether the chain is the heavy/long or light/short chain.
     \item each cell possibly contains multiple heavy/long and/or light/short chains.
     \item every chain has its own V(D)J annotation, in which ambiguous V(D)J
           annotations, if any, are separated by a comma.
  }

Single-cell example:
  \itemize{
     \item A cell has 1 heavy chain and 2 light chains.
     \item There should be 3 rows corresponding to this cell.
     \item One of the light chains may have an ambiguous V annotation which looks like \code{"Homsap IGKV1-39*01 F,Homsap IGKV1D-39*01 F"}.
  }
}

\examples{
# Group by genes
db <- groupGenes(ExampleDb)
head(db$vj_group)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Deprecated.R
\name{rarefyDiversity}
\alias{rarefyDiversity}
\title{Generate a clonal diversity index curve}
\usage{
rarefyDiversity(
  data,
  group,
  clone = "CLONE",
  copy = NULL,
  min_q = 0,
  max_q = 4,
  step_q = 0.05,
  min_n = 30,
  max_n = NULL,
  ci = 0.95,
  nboot = 2000,
  uniform = TRUE,
  cell_id = "cell_id",
  progress = FALSE
)
}
\arguments{
\item{data}{data.frame with Change-O style columns containing clonal assignments.}

\item{group}{name of the \code{data} column containing group identifiers.}

\item{clone}{name of the \code{data} column containing clone identifiers.}

\item{copy}{name of the \code{data} column containing copy numbers for each 
sequence. If \code{copy=NULL} (the default), then clone abundance
is determined by the number of sequences. If a \code{copy} column
is specified, then clone abundances is determined by the sum of 
copy numbers within each clonal group.}

\item{min_q}{minimum value of \eqn{q}.}

\item{max_q}{maximum value of \eqn{q}.}

\item{step_q}{value by which to increment \eqn{q}.}

\item{min_n}{minimum number of observations to sample.
A group with less observations than the minimum is excluded.}

\item{max_n}{maximum number of observations to sample. If \code{NULL} then no 
maximum is set.}

\item{ci}{confidence interval to calculate; the value must be between 0 and 1.}

\item{nboot}{number of bootstrap realizations to generate.}

\item{uniform}{if \code{TRUE} then uniformly resample each group to the same 
number of observations. If \code{FALSE} then allow each group to
be resampled to its original size or, if specified, \code{max_size}.}

\item{cell_id}{name of the \code{data} column containing cell identifiers.}

\item{progress}{if \code{TRUE} show a progress bar.}
}
\value{
A \link{DiversityCurve} object summarizing the diversity scores.
}
\description{
\code{rarefyDiversity} divides a set of clones by a group annotation,
resamples the sequences from each group, and calculates diversity
scores (\eqn{D}) over an interval of diversity orders (\eqn{q}).
}
\details{
Clonal diversity is calculated using the generalized diversity index (Hill numbers) 
proposed by Hill (Hill, 1973). See \link{calcDiversity} for further details.

Diversity is calculated on the estimated complete clonal abundance distribution.
This distribution is inferred by using the Chao1 estimator to estimate the number
of seen clones, and applying the relative abundance correction and unseen clone
frequency described in Chao et al, 2015.

To generate a smooth curve, \eqn{D} is calculated for each value of \eqn{q} from
\code{min_q} to \code{max_q} incremented by \code{step_q}.  When \code{uniform=TRUE}
variability in total sequence counts across unique values in the \code{group} column 
is corrected by repeated resampling from the estimated complete clonal distribution to a 
common number of sequences.

The diversity index (\eqn{D}) for each group is the mean value of over all resampling 
realizations. Confidence intervals are derived using the standard deviation of the 
resampling realizations, as described in Chao et al, 2015.
}
\examples{
\dontrun{
# Group by sample identifier
div <- rarefyDiversity(ExampleDb, "sample_id", step_q=1, max_q=10, nboot=100)
plotDiversityCurve(div, legend_title="Sample")
                   
# Grouping by isotype rather than sample identifier
div <- rarefyDiversity(ExampleDb, "c_call", min_n=40, step_q=1, max_q=10, 
                       nboot=100)
plotDiversityCurve(div, legend_title="Isotype")
}
}
\references{
\enumerate{
  \item  Hill M. Diversity and evenness: a unifying notation and its consequences. 
           Ecology. 1973 54(2):427-32.
  \item  Chao A. Nonparametric Estimation of the Number of Classes in a Population. 
           Scand J Stat. 1984 11, 265270.
  \item  Chao A, et al. Rarefaction and extrapolation with Hill numbers: 
           A framework for sampling and estimation in species diversity studies. 
           Ecol Monogr. 2014 84:45-67.
  \item  Chao A, et al. Unveiling the species-rank abundance distribution by 
           generalizing the Good-Turing sample coverage theory. 
           Ecology. 2015 96, 11891201.
}
}
\seealso{
\link{alphaDiversity}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Sequence.R
\name{extractVRegion}
\alias{extractVRegion}
\title{Extracts FWRs and CDRs from IMGT-gapped sequences}
\usage{
extractVRegion(sequences, region = c("fwr1", "cdr1", "fwr2", "cdr2", "fwr3"))
}
\arguments{
\item{sequences}{character vector of IMGT-gapped nucleotide sequences.}

\item{region}{string defining the region(s) of the V segment to extract. 
May be a single region or multiple regions (as a vector) from
\code{c("fwr1", "cdr1", "fwr2", "cdr2" ,"fwr3")}.  By default, all
regions will be returned.}
}
\value{
If only one region is specified in the \code{region} argument, a character 
           vector of the extracted sub-sequences will be returned. If multiple regions 
           are specified, then a character matrix will be returned with columns 
           corresponding to the specified regions and a row for each entry in 
           \code{sequences}.
}
\description{
\code{extractVRegion} extracts the framework and complementarity determining regions of 
the V segment for IMGT-gapped immunoglobulin (Ig) nucleotide sequences according to the 
IMGT numbering scheme.
}
\examples{
# Assign example clone
clone <- subset(ExampleDb, clone_id == 3138)

# Get all regions
extractVRegion(clone$sequence_alignment)

# Get single region
extractVRegion(clone$sequence_alignment, "fwr1")

# Get all CDRs
extractVRegion(clone$sequence_alignment, c("cdr1", "cdr2"))

# Get all FWRs
extractVRegion(clone$sequence_alignment, c("fwr1", "fwr2", "fwr3"))

}
\references{
\enumerate{
  \item  Lefranc M-P, et al. IMGT unique numbering for immunoglobulin and T cell 
           receptor variable domains and Ig superfamily V-like domains.
           Dev Comp Immunol. 2003 27(1):55-77.
}
}
\seealso{
IMGT-gapped region boundaries are defined in \link{IMGT_REGIONS}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Fastq.R
\name{maskPositionsByQuality}
\alias{maskPositionsByQuality}
\title{Mask sequence positions with low quality}
\usage{
maskPositionsByQuality(
  data,
  min_quality = 70,
  sequence = "sequence_alignment",
  quality_num = "quality_alignment_num"
)
}
\arguments{
\item{data}{\code{data.frame} containing sequence data.}

\item{min_quality}{minimum quality score. Positions with sequencing quality 
less than \code{min_qual} will be masked.}

\item{sequence}{column in \code{data} with sequence data to be masked.}

\item{quality_num}{column in \code{data} with quality scores (a
string of numeric values, comma separated) that can
be used to mask \code{sequence}.}
}
\value{
Modified \code{data} data.frame with an additional field containing 
          quality masked sequences. The  name of this field is created 
          concatenating the \code{sequence} name and \code{"_masked"}.
}
\description{
\code{maskPositionsByQuality} will replace positions that 
have a sequencing quality score lower that \code{min_quality} with an
\code{"N"} character.
}
\examples{
db <- airr::read_rearrangement(system.file("extdata", "example_quality.tsv", package="alakazam"))
fastq_file <- system.file("extdata", "example_quality.fastq", package="alakazam")
db <- readFastqDb(db, fastq_file, quality_offset=-33)
maskPositionsByQuality(db, min_quality=90, quality_num="quality_alignment_num")

}
\seealso{
\link{readFastqDb} and \link{getPositionQuality}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Sequence.R
\name{pairwiseDist}
\alias{pairwiseDist}
\title{Calculate pairwise distances between sequences}
\usage{
pairwiseDist(seq, dist_mat = getDNAMatrix())
}
\arguments{
\item{seq}{character vector containing a DNA sequences.}

\item{dist_mat}{Character distance matrix. Defaults to a Hamming distance 
matrix returned by \link{getDNAMatrix}. If gap 
characters, \code{c("-", ".")}, are assigned a value of -1 
in \code{dist_mat} then contiguous gaps of any run length,
which are not present in both sequences, will be counted as a 
distance of 1. Meaning, indels of any length will increase
the sequence distance by 1. Gap values other than -1 will 
return a distance that does not consider indels as a special case.}
}
\value{
A matrix of numerical distance between each entry in \code{seq}. 
          If \code{seq} is a named vector, row and columns names will be added 
          accordingly.

          Amino acid distance matrix may be built with \link{getAAMatrix}. 
          Uses \link{seqDist} for calculating distances between pairs.
          See \link{pairwiseEqual} for generating an equivalence matrix.
}
\description{
\code{pairwiseDist} calculates all pairwise distance between a set of sequences.
}
\examples{
# Gaps will be treated as Ns with a gap=0 distance matrix
pairwiseDist(c(A="ATGGC", B="ATGGG", C="ATGGG", D="AT--C"), 
             dist_mat=getDNAMatrix(gap=0))

# Gaps will be treated as universally non-matching characters with gap=1
pairwiseDist(c(A="ATGGC", B="ATGGG", C="ATGGG", D="AT--C"), 
             dist_mat=getDNAMatrix(gap=1))

# Gaps of any length will be treated as single mismatches with a gap=-1 distance matrix
pairwiseDist(c(A="ATGGC", B="ATGGG", C="ATGGG", D="AT--C"), 
             dist_mat=getDNAMatrix(gap=-1))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Core.R
\name{progressBar}
\alias{progressBar}
\title{Standard progress bar}
\usage{
progressBar(n)
}
\arguments{
\item{n}{maximum number of ticks}
}
\value{
A \link[progress]{progress_bar} object.
}
\description{
\code{progressBar} defines a common progress bar format.
}
